Type: Package
Package: magclass
Title: Data Class and Tools for Handling Spatial-Temporal Data
Version: 7.5.0
Date: 2026-06-25
Authors@R: c(
    person("Jan Philipp", "Dietrich", , "dietrich@pik-potsdam.de", role = c("aut", "cre"),
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research", ORCID = "0000-0002-4309-6431")),
    person("Benjamin Leon", "Bodirsky", , "bodirsky@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Markus", "Bonsch", role = "aut"),
    person("Florian", "Humpenoeder", , "humpenoeder@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Stephen", "Bi", role = "aut"),
    person("Kristine", "Karstens", , "karstens@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Debbora", "Leip", , "leip@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Pascal", "Sauer", , "pascal.sauer@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research", ORCID = "0000-0002-6856-8239")),
    person("Lavinia", "Baumstark", , "lavinia@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Christoph", "Bertram", , "bertram@pik-potsdam.de", role = "ctb"),
    person("Anastasis", "Giannousakis", , "giannou@pik-potsdam.de", role = "ctb"),
    person("David", "Klein", , "dklein@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Ina", "Neher", role = "ctb"),
    person("Michaja", "Pehl", , "pehl@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Anselm", "Schultes", role = "ctb"),
    person("Miodrag", "Stevanovic", , "miodrag@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Xiaoxi", "Wang", , "wang@pik-potsdam.de", role = "ctb"),
    person("Felicitas", "Beier", , "beier@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Mika", "Pflüger", role = "ctb"),
    person("Oliver", "Richters", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Patrick", "Rein", , "patrick.rein@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research", ORCID = "0000-0001-9454-8381"))
  )
Description: Data class for increased interoperability working with
    spatial-temporal data together with corresponding functions and
    methods (conversions, basic calculations and basic data manipulation).
    The class distinguishes between spatial, temporal and other dimensions
    to facilitate the development and interoperability of tools build for
    it. Additional features are name-based addressing of data and internal
    consistency checks (e.g. checking for the right data order in
    calculations).
License: LGPL-3 | file LICENSE
URL: https://github.com/pik-piam/magclass,
        https://doi.org/10.5281/zenodo.1158580
BugReports: https://github.com/pik-piam/magclass/issues
Depends: methods, R (>= 2.10.0)
Imports: abind, data.table, rlang, stats
Suggests: covr, ggplot2, knitr, lpjmlkit, ncdf4, pkgconfig, quitte,
        raster, rmarkdown, terra, testthat (>= 3.1.5), tibble, withr
VignetteBuilder: knitr
Additional_repositories: https://pik-piam.r-universe.dev
Config/Keywords: tool
Encoding: UTF-8
LazyData: true
Config/Needs/website: tidyverse/tidytemplate
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-29 07:56:03 UTC; patrickr
Author: Jan Philipp Dietrich [aut, cre] (affiliation: Potsdam Institute for
    Climate Impact Research, ORCID:
    <https://orcid.org/0000-0002-4309-6431>),
  Benjamin Leon Bodirsky [aut] (affiliation: Potsdam Institute for
    Climate Impact Research),
  Markus Bonsch [aut],
  Florian Humpenoeder [aut] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Stephen Bi [aut],
  Kristine Karstens [aut] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Debbora Leip [aut] (affiliation: Potsdam Institute for Climate Impact
    Research),
  Pascal Sauer [aut] (affiliation: Potsdam Institute for Climate Impact
    Research, ORCID: <https://orcid.org/0000-0002-6856-8239>),
  Lavinia Baumstark [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Christoph Bertram [ctb],
  Anastasis Giannousakis [ctb],
  David Klein [ctb] (affiliation: Potsdam Institute for Climate Impact
    Research),
  Ina Neher [ctb],
  Michaja Pehl [ctb] (affiliation: Potsdam Institute for Climate Impact
    Research),
  Anselm Schultes [ctb],
  Miodrag Stevanovic [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Xiaoxi Wang [ctb],
  Felicitas Beier [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Mika Pflüger [ctb],
  Oliver Richters [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Patrick Rein [aut] (affiliation: Potsdam Institute for Climate Impact
    Research, ORCID: <https://orcid.org/0000-0001-9454-8381>)
Maintainer: Jan Philipp Dietrich <dietrich@pik-potsdam.de>
Repository: CRAN
Date/Publication: 2026-06-29 09:50:03 UTC
Built: R 4.5.2; ; 2026-06-29 12:07:52 UTC; unix

Type: Package
Package: magclass
Title: Data Class and Tools for Handling Spatial-Temporal Data
Version: 7.4.3
Date: 2026-03-25
Authors@R: c(
    person("Jan Philipp", "Dietrich", , "dietrich@pik-potsdam.de", role = c("aut", "cre"),
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research", ORCID = "0000-0002-4309-6431")),
    person("Benjamin Leon", "Bodirsky", , "bodirsky@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Markus", "Bonsch", role = "aut"),
    person("Florian", "Humpenoeder", , "humpenoeder@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Stephen", "Bi", role = "aut"),
    person("Kristine", "Karstens", , "karstens@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Debbora", "Leip", , "leip@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Pascal", "Sauer", , "pascal.sauer@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research", ORCID = "0000-0002-6856-8239")),
    person("Lavinia", "Baumstark", , "lavinia@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Christoph", "Bertram", , "bertram@pik-potsdam.de", role = "ctb"),
    person("Anastasis", "Giannousakis", , "giannou@pik-potsdam.de", role = "ctb"),
    person("David", "Klein", , "dklein@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Ina", "Neher", role = "ctb"),
    person("Michaja", "Pehl", , "pehl@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Anselm", "Schultes", role = "ctb"),
    person("Miodrag", "Stevanovic", , "miodrag@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Xiaoxi", "Wang", , "wang@pik-potsdam.de", role = "ctb"),
    person("Felicitas", "Beier", , "beier@pik-potsdam.de", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Mika", "Pflüger", role = "ctb"),
    person("Oliver", "Richters", role = "ctb",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research")),
    person("Patrick", "Rein", , "patrick.rein@pik-potsdam.de", role = "aut",
           comment = c(affiliation = "Potsdam Institute for Climate Impact Research", ORCID = "0000-0001-9454-8381"))
  )
Description: Data class for increased interoperability working with
    spatial-temporal data together with corresponding functions and
    methods (conversions, basic calculations and basic data manipulation).
    The class distinguishes between spatial, temporal and other dimensions
    to facilitate the development and interoperability of tools build for
    it. Additional features are name-based addressing of data and internal
    consistency checks (e.g. checking for the right data order in
    calculations).
License: LGPL-3 | file LICENSE
URL: https://github.com/pik-piam/magclass,
        https://doi.org/10.5281/zenodo.1158580
BugReports: https://github.com/pik-piam/magclass/issues
Depends: methods, R (>= 2.10.0)
Imports: abind, data.table, rlang, stats
Suggests: covr, ggplot2, knitr, lpjmlkit, ncdf4, pkgconfig, quitte,
        raster, rmarkdown, terra, testthat (>= 3.1.5), tibble, withr
VignetteBuilder: knitr
Additional_repositories: https://pik-piam.r-universe.dev
Config/Keywords: tool
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Config/Needs/website: tidyverse/tidytemplate
NeedsCompilation: no
Packaged: 2026-04-17 11:49:59 UTC; patrickr
Author: Jan Philipp Dietrich [aut, cre] (affiliation: Potsdam Institute for
    Climate Impact Research, ORCID:
    <https://orcid.org/0000-0002-4309-6431>),
  Benjamin Leon Bodirsky [aut] (affiliation: Potsdam Institute for
    Climate Impact Research),
  Markus Bonsch [aut],
  Florian Humpenoeder [aut] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Stephen Bi [aut],
  Kristine Karstens [aut] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Debbora Leip [aut] (affiliation: Potsdam Institute for Climate Impact
    Research),
  Pascal Sauer [aut] (affiliation: Potsdam Institute for Climate Impact
    Research, ORCID: <https://orcid.org/0000-0002-6856-8239>),
  Lavinia Baumstark [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Christoph Bertram [ctb],
  Anastasis Giannousakis [ctb],
  David Klein [ctb] (affiliation: Potsdam Institute for Climate Impact
    Research),
  Ina Neher [ctb],
  Michaja Pehl [ctb] (affiliation: Potsdam Institute for Climate Impact
    Research),
  Anselm Schultes [ctb],
  Miodrag Stevanovic [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Xiaoxi Wang [ctb],
  Felicitas Beier [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Mika Pflüger [ctb],
  Oliver Richters [ctb] (affiliation: Potsdam Institute for Climate
    Impact Research),
  Patrick Rein [aut] (affiliation: Potsdam Institute for Climate Impact
    Research, ORCID: <https://orcid.org/0000-0001-9454-8381>)
Maintainer: Jan Philipp Dietrich <dietrich@pik-potsdam.de>
Repository: CRAN
Date/Publication: 2026-04-17 13:00:02 UTC
Built: R 4.6.0; ; 2026-04-25 20:09:53 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(aperm,magpie)
S3method(cbind,magpie)
S3method(head,magpie)
S3method(print,magpie)
S3method(str,magpie)
S3method(tail,magpie)
S3method(tibble::as_tibble,magpie)
export("copy.attributes<-")
export("getCells<-")
export("getComment<-")
export("getCoords<-")
export("getItems<-")
export("getNames<-")
export("getRegionList<-")
export("getRegions<-")
export("getSets<-")
export("getYears<-")
export("mselect<-")
export(addDim)
export(add_columns)
export(add_dimension)
export(as.RasterBrick)
export(as.SpatRaster)
export(as.SpatRasterDataset)
export(as.SpatVector)
export(clean_magpie)
export(collapseDim)
export(collapseNames)
export(commonYears)
export(complete_magpie)
export(convergence)
export(copy.attributes)
export(copy.magpie)
export(dimCode)
export(dimExists)
export(dimOrder)
export(dimReduce)
export(dimSums)
export(extend)
export(fulldim)
export(getCPR)
export(getCells)
export(getComment)
export(getCoords)
export(getDim)
export(getItems)
export(getNames)
export(getRegionList)
export(getRegions)
export(getSets)
export(getYears)
export(guessResolution)
export(hasCoords)
export(hasSets)
export(is.magpie)
export(is.spatial)
export(is.temporal)
export(isYear)
export(lowpass)
export(magpie_expand)
export(magpiesort)
export(magpply)
export(matchDim)
export(maxample)
export(mbind)
export(mcalc)
export(mplot)
export(mselect)
export(ncells)
export(ndata)
export(ndim)
export(new.magpie)
export(nregions)
export(nyears)
export(read.magpie)
export(read.report)
export(replace_non_finite)
export(sameDims)
export(setCells)
export(setComment)
export(setItems)
export(setYears)
export(spatRasterToDataset)
export(suppressSpecificWarnings)
export(time_interpolate)
export(unitjoin)
export(unitsplit)
export(unwrap)
export(where)
export(wrap)
export(write.magpie)
export(write.report)
export(write.report2)
exportClasses(magpie)
exportMethods("[")
exportMethods("[<-")
exportMethods(Ops)
exportMethods(as.array)
exportMethods(as.data.frame)
exportMethods(as.magpie)
exportMethods(colMeans)
exportMethods(colSums)
exportMethods(ifelse)
exportMethods(is.finite)
exportMethods(is.infinite)
exportMethods(is.na)
exportMethods(is.nan)
exportMethods(log)
exportMethods(logb)
exportMethods(pmax)
exportMethods(pmin)
exportMethods(round)
exportMethods(rowMeans)
exportMethods(rowSums)
exportMethods(setNames)
exportMethods(show)
import(data.table)
import(methods)
importFrom(abind,abind)
importFrom(data.table,as.data.table)
importFrom(data.table,melt)
importFrom(data.table,tstrsplit)
importFrom(methods,Ops)
importFrom(methods,as)
importFrom(methods,callGeneric)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,signature)
importFrom(rlang,.data)
importFrom(stats,as.formula)
importFrom(utils,head)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,type.convert)
importFrom(utils,write.table)

# Data Class and Tools for Handling Spatial-Temporal Data

R package **magclass**, version **7.4.3**

[![CRAN status](https://www.r-pkg.org/badges/version/magclass)](https://cran.r-project.org/package=magclass) [![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.1158580.svg)](https://doi.org/10.5281/zenodo.1158580) [![R build status](https://github.com/pik-piam/magclass/workflows/check/badge.svg)](https://github.com/pik-piam/magclass/actions) [![codecov](https://codecov.io/gh/pik-piam/magclass/branch/master/graph/badge.svg)](https://app.codecov.io/gh/pik-piam/magclass) [![r-universe](https://pik-piam.r-universe.dev/badges/magclass)](https://pik-piam.r-universe.dev/builds)

## Purpose and Functionality

Data class for increased interoperability working with
    spatial-temporal data together with corresponding functions and
    methods (conversions, basic calculations and basic data manipulation).
    The class distinguishes between spatial, temporal and other dimensions
    to facilitate the development and interoperability of tools build for
    it. Additional features are name-based addressing of data and internal
    consistency checks (e.g. checking for the right data order in
    calculations).


## Installation

For installation of the most recent package version an additional repository has to be added in R:

```r
options(repos = c(CRAN = "@CRAN@", pik = "https://rse.pik-potsdam.de/r/packages"))
```
The additional repository can be made available permanently by adding the line above to a file called `.Rprofile` stored in the home folder of your system (`Sys.glob("~")` in R returns the home directory).

After that the most recent version of the package can be installed using `install.packages`:

```r 
install.packages("magclass")
```

Package updates can be installed using `update.packages` (make sure that the additional repository has been added before running that command):

```r 
update.packages()
```

## Tutorial

The package comes with vignettes describing the basic functionality of the package and how to use it. You can load them with the following command (the package needs to be installed):

```r
vignette("magclass-concept")   # Concept behind magclass
vignette("magclass-expansion") # MAGPIE Class Object Expansion
vignette("magclass")           # MAGPIE Class Tutorial
vignette("magclass6")          # magclass 6
```

## Questions / Problems

In case of questions / problems please contact Jan Philipp Dietrich <dietrich@pik-potsdam.de>.

## Citation

To cite package **magclass** in publications use:

Dietrich J, Bodirsky B, Bonsch M, Humpenoeder F, Bi S, Karstens K, Leip D, Sauer P, Baumstark L, Bertram C, Giannousakis A, Klein D, Neher I, Pehl M, Schultes A, Stevanovic M, Wang X, Beier F, Pflüger M, Richters O, Rein P (2026). "magclass: Data Class and Tools for Handling Spatial-Temporal Data." doi:10.5281/zenodo.1158580 <https://doi.org/10.5281/zenodo.1158580>, Version: 7.4.3, <https://github.com/pik-piam/magclass>.

A BibTeX entry for LaTeX users is

 ```latex
@Misc{,
  title = {magclass: Data Class and Tools for Handling Spatial-Temporal Data},
  author = {Jan Philipp Dietrich and Benjamin Leon Bodirsky and Markus Bonsch and Florian Humpenoeder and Stephen Bi and Kristine Karstens and Debbora Leip and Pascal Sauer and Lavinia Baumstark and Christoph Bertram and Anastasis Giannousakis and David Klein and Ina Neher and Michaja Pehl and Anselm Schultes and Miodrag Stevanovic and Xiaoxi Wang and Felicitas Beier and Mika Pflüger and Oliver Richters and Patrick Rein},
  doi = {10.5281/zenodo.1158580},
  date = {2026-03-25},
  year = {2026},
  url = {https://github.com/pik-piam/magclass},
  note = {Version: 7.4.3},
}
```

#' @exportMethod rowMeans
#'
setMethod("rowMeans",
  signature(x = "magpie"),
  function(x, na.rm = FALSE, dims = 1, ...) { # nolint: object_name_linter.
    out <- rowMeans(as.array(x), na.rm = na.rm, dims = dims, ...)
    out <- as.magpie(as.array(x))
    return(out)
  }
)

#' @title magpply
#' @description apply command for magpieobjects. Very efficient for replacing loops.
#'
#' @param X magpie object
#' @param FUN function that shall be applied X
#' @param MARGIN dimension over which FUN shall be applied (like a loop over that dimension).
#' This dimension will be preserved in the output object (see also \code{DIM}).
#' @param DIM dimension in which FUN shall be applied. This dimension will be missing in the output. DIM and MARGIN
#' are opposite ways of expressing the dimensions to be addressed and you must only use one of them with MARGIN
#' excluding dimensions from the calculation and DIM including them.
#' @param ... further parameters passed on to FUN
#' @param INTEGRATE if TRUE, the output will be filled into an magpie object of the same dimensionality as X
#' @return magpie object
#'
#' @author Jan Philipp Dietrich, Benjamin Leon Bodirsky
#' @examples
#' pop <- maxample("pop")
#' magpply(pop, FUN = sum, MARGIN = 2)
#' fourdim <- pop * setNames(pop, c("jkk", "lk"))
#' magpply(fourdim, FUN = sum, MARGIN = c(1, 3.1))
#' @family SelectionCalculation
#' @export
magpply <- function( # nolint: cyclocomp_linter.
  X, FUN, MARGIN = NULL, DIM = NULL, ..., INTEGRATE = FALSE # nolint: object_name_linter.
) {
  x <- X
  margin <- MARGIN
  dim <- DIM
  if (!is.magpie(x)) {
    stop("Input is not a MAgPIE object!")
  }
  if (!is.null(margin) && !is.null(dim)) {
    stop("Please specify either MARGIN or DIM, not both at the same time!")
  }
  if (!is.null(margin)) {
    # converting margin to dim
    margin <- dimCode(margin, x)
    dim <- NULL
    for (i in 1:3) {
      if (!is.element(i, floor(margin))) {
        dim <- c(dim, i)
      } else if (is.element(i, floor(margin)) && !is.element(i, margin)) {
        dim <- c(dim, setdiff(i + seq_len(ndim(x, dim = i)) / 10, margin))
      }
    }
  }
  dim <- sort(dimCode(dim, x), decreasing = TRUE)
  if (any(dim == 0)) {
    stop("Invalid dimension(s) specified")
  }
  if (length(x) == 0) {
    return(NULL)
  }
  if (INTEGRATE) {
    xIn <- x
  }
  for (d in dim) {
    getItems(x, dim = d, raw = TRUE) <- NULL
  }
  noNames <- which(vapply(dimnames(x), is.null, logical(1)))
  for (i in noNames) {
    getItems(x, dim = i) <- rep("dummy", dim(x)[i])
  }
  xd <- as.data.frame.table(x)
  out <- new("magpie", tapply(xd[[4]], xd[1:3], FUN, ...))
  for (i in noNames) {
    getItems(out, dim = i) <- NULL
  }
  if (INTEGRATE) {
    out <- magpie_expand(out, xIn)
  }
  return(out)
}

a <- maxample("animal")
p <- maxample("pop")

test_that("dimSums works", {
  ref1 <- new("magpie",
              .Data = structure(c(0, 0, 1, 1, 6, 4, 7, 5, 6, 4, 7, 5, 0, 0, 0, 0),
                                .Dim = c(2L, 2L, 4L),
                                .Dimnames = list(x.y.country.cell = c("5p75.53p25.NLD.14084", "6p25.53p25.NLD.14113"),
                                                 year.month.day = c("y2000.april.20", "y2000.may.20"),
                                                 type.color = c("animal.black", "animal.white",
                                                                "animal.red", "animal.brown"))))
  expect_identical(dimSums(a[1:2, 1:2, ], dim = "species"), ref1)
  expect_identical(dimSums(a[1:2, 1:2, ], dim = 3.2), ref1)

  ref2 <- new("magpie",
              .Data = structure(c(9397, 7653, 3475, 3345, 1900),
                                .Dim = c(1L, 1L, 5L),
                                .Dimnames = list(d1 = NULL, d2 = NULL,
                                                 type.species.color = c("animal.rabbit.black", "animal.rabbit.white",
                                                                        "animal.bird.black", "animal.bird.red",
                                                                        "animal.dog.brown"))))
  expect_identical(dimSums(a, dim = 1:2), ref2)
  expect_error(dimSums(1), "Input is not a MAgPIE object")
  expect_error(dimSums(a, dim = 4), "Invalid dimension")

  ax <- a[1, 1:2, 1:4]
  getItems(ax, dim = 3, raw = TRUE) <- c("animal.rabbit.black", "animal.rabbit.white",
                                         "animal.bird.black", "animal.bird.white")
  ref4 <- new("magpie",
              .Data = structure(c(12, 15), .Dim = c(1L, 2L, 1L),
                                .Dimnames = list(x.y.country.cell = "5p75.53p25.NLD.14084",
                                                 year.month.day = c("y2000.april.20", "y2000.may.20"),
                                                 type = "animal")))
  expect_identical(dimSums(ax, dim = c(3.2, 3.3)), ref4)

  ref5 <- new("magpie",
              .Data = structure(c(0, 13, 1, 13), .Dim = c(1L, 1L, 4L),
                                .Dimnames = list(d1 = NULL, d2 = NULL,
                                                 type.species.color = c("animal.rabbit.black", "animal.rabbit.white",
                                                                        "animal.bird.black", "animal.bird.white"))))
  expect_identical(dimSums(ax, dim = 1:2), ref5)

  ref3 <- new("magpie", .Data = structure(25770, .Dim = c(1L, 1L, 1L),
                                          .Dimnames = list(d1 = NULL, d2 = NULL, d3 = NULL)))
  expect_identical(dimSums(a, dim = 1:3), ref3)

  expect_true(all(dimSums(p, dim = 1) - colSums(p) == 0))
  expect_true(all(dimSums(p, dim = 2:3) - rowSums(p) == 0))
  expect_true(all(dimSums(p, dim = 1) / dim(p)[1] - colMeans(p) == 0))
  expect_true(all(magpply(p, mean, dim = 2:3) - rowMeans(p) == 0))

  p0 <- p[, , -1:-2]
  expect_null(dimSums(p0, dim = 3))
})

a <- maxample("animal")
p <- maxample("pop")

test_that("magpply works", {
  ref1 <- new("magpie",
              .Data = structure(c(0, 0, 1, 1, 6, 4, 7, 5, 6, 4, 7, 5, 0, 0, 0, 0),
                                .Dim = c(2L, 2L, 4L),
                                .Dimnames = list(x.y.country.cell = c("5p75.53p25.NLD.14084", "6p25.53p25.NLD.14113"),
                                                 year.month.day = c("y2000.april.20", "y2000.may.20"),
                                                 type.color = c("animal.black", "animal.white",
                                                                "animal.red", "animal.brown"))))
  expect_identical(magpply(a[1:2, 1:2, ], FUN = sum, DIM = "species"), ref1)
  expect_identical(magpply(a[1:2, 1:2, ], FUN = sum, DIM = 3.2, INTEGRATE = TRUE), magpie_expand(ref1, a[1:2, 1:2, ]))
  expect_identical(magpply(a[1:2, 1:2, ], FUN = sum, MARGIN = c(1, 2, 3.1, 3.3)), ref1)

  ref2 <- new("magpie",
              .Data = structure(c(9397, 7653, 3475, 3345, 1900),
                                .Dim = c(1L, 1L, 5L),
                                .Dimnames = list(d1 = NULL, d2 = NULL,
                                                 type.species.color = c("animal.rabbit.black", "animal.rabbit.white",
                                                                        "animal.bird.black", "animal.bird.red",
                                                                        "animal.dog.brown"))))
  expect_error(magpply(a, FUN = mean, DIM = 1, MARGIN = 2:3), "specify either MARGIN or DIM")

  ref3 <- new("magpie", .Data = structure(c(2709, 1965), .Dim = c(2L, 1L, 1L),
                                          .Dimnames = list(i = c("AFR", "CPA"),
                                                           d2 = NULL, d3 = NULL)))

  expect_identical(magpply(round(p[1:2, , ]), max, MARGIN = 1), ref3)

})

p <- maxample("pop")
attr(p, "Metadata") <- NULL


test_that("time interpolate works", {
  expect_silent(p2 <- time_interpolate(p, "y2000", integrate_interpolated_years = TRUE))
  expect_identical(p2[, 2000, , invert = TRUE], p)
  expect_true(all(p2[, 2000, ] - magpply(p[, c(1995, 2005), ], mean, DIM = 2) == 0))
  expect_silent(p3 <- time_interpolate(p, c(1980, 2000), integrate_interpolated_years = TRUE,
                                       extrapolation_type = "constant"))
  expect_identical(p3[, c(1980, 2000), , invert = TRUE], p)
  expect_true(all(p3[, 1980, ] - setYears(p3[, 1995, ], NULL) == 0))
  expect_error(time_interpolate(1), "Invalid data format")
  expect_error(time_interpolate(as.magpie(1), interpolated_year = "blablub"), "not in the right format")
  expect_silent(p4 <- time_interpolate(p[, 1, ], c(1900, 2100)))
  expect_true(all(dimSums(p4, dim = 2) - 2 * p[, 1, ] == 0))
  expect_silent(p5 <- time_interpolate(p[, 1, ], c(1900, 1995, 2100)))
  expect_true(all(dimSums(p5, dim = 2) - 3 * p[, 1, ] == 0))
  expect_silent(p6 <- time_interpolate(p[, 1, ], c(1900, 1995, 2100), integrate_interpolated_years = TRUE))
  expect_identical(p5, p6)
})

test_that("time interpolate can handle faulty data", {
  a <- maxample("animal")
  a <- dimSums(a, dim = c(2.2, 2.3))
  names(dimnames(a))[3] <- "bla.blub"
  expect_silent(out <- time_interpolate(a, interpolated_year = c(2012, 2013)))
})

#' magpie_expand
#'
#' Expands a MAgPIE object based on a reference
#'
#' Expansion means here that the dimensions of x are expanded accordingly to
#' ref. Please note that this is really only about expansion. In the case that
#' one dimension of ref is smaller than of x nothing happens with this
#' dimension.
#'
#' You can influence the verbosity of this function by setting the option
#' "magclass.verbosity". By default verbosity is set to 1 which means that
#' only warnings are returned. Setting verbosity to 2 means that warnings
#' as well as additional notes are returned. This is done by
#' options(verbosity.level=2)
#'
#' With version 5 of the package magpie_expand has been updated to a newer version
#' (currently 2.1) and since version 6 this is the only currently supported
#' version. To switch to the old setup you have to install magclass  in a version < 6
#' and set \code{options(magclass_expand_version=1)}.
#'
#' By default expansion is based on the elements in a dimension ignoring the set name of
#' the dimension. To expand based on set names instead of contents (recommended) you can
#' switch \code{options(magclass_setMatching=TRUE)}. Please be careful with this setting
#' as it alters the behavior of magclass objects quite significantly! For more information
#' have a look at \code{vignette("magclass-expansion")}.
#'
#' @param x MAgPIE object that should be expanded
#' @param ref MAgPIE object that serves as a reference
#' @return An expanded version of x.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{as.magpie}}, \code{\link[base]{options}}
#' @examples
#'
#' a <- new.magpie(c("AFR", "CPA"), "y1995", c("m", "n"))
#' b <- new.magpie("GLO", "y1995", c("bla", "blub"))
#' magpie_expand(b, a)
#' options(magclass.verbosity = 2)
#' magpie_expand(b, a)
#' @family DataBinding
#' @export
magpie_expand <- function(x, ref) { # nolint: object_name_linter, cyclocomp_linter.

  version <- getOption("magclass_expand_version")
  if (is.null(version)) options("magclass_expand_version" = 2.1) # nolint: undesirable_function_linter.
  if (!is.null(version) && version != 2.1) stop("Unsupported magclass expand version (", version, ")!")

  setMatching <- isTRUE(getOption("magclass_setMatching"))

  # x: MAgPIE object which should be expanded
  # ref: Reference object defining the structure to which x should be expanded

  # 1.spatial dimension
  # 2.temporal dimension
  # 3.data dimension

  for (i in 1:3) {
    # Remove "GLO" from spatial dimension in case another spatial dimension exists
    if (i == 1) {
      if (dim(ref)[1] == 1 && !is.null(rownames(ref)) && rownames(ref) == "GLO") rownames(ref) <- NULL
      else if (dim(x)[1] == 1 && !is.null(rownames(x)) && rownames(x) == "GLO") rownames(x) <- NULL
    }

    if (is.null(dimnames(ref)[[i]]) && dim(ref)[i] > 1) {
      stop("Inconsistent MAgPIE reference file: more than 1 element in dimension ", i, " but no names given!")
    }

    if (is.null(dimnames(ref)[[i]])) next # Nothing to do if ref has no dimnames

    if (is.null(dimnames(x)[[i]])) {
      if (dim(x)[i] > 1) stop("Inconsistent MAgPIE file: more than 1 element in dimension ", i, " but no names given!")
      if (dim(ref)[i] > 1) { # Expand single element dimension to dimension of ref
        x <- x[rep(1, dim(ref)[i]), dim = i]
      }
      if (!is.null(dimnames(ref)[[i]])) dimnames(x)[[i]] <- dimnames(ref)[[i]]
    } else if (dim(x)[i] == dim(ref)[i] && all(dimnames(x)[[i]] == dimnames(ref)[[i]]) &&
                 (!setMatching || names(dimnames(x))[i] == names(dimnames(ref))[i])) {
      # dimension is identical
      next
    } else if (dim(x)[i] == dim(ref)[i] && setequal(dimnames(x)[[i]], dimnames(ref)[[i]]) &&
                 (!setMatching || names(dimnames(x))[i] == names(dimnames(ref))[i])) {
      # same length and entries, but different order
      x <- x[dimnames(ref)[[i]], dim = i]
    } else {
      x <- magpie_expand_dim(x, ref, dim = i)
    }
  }
  getSets(x) <- make.unique(getSets(clean_magpie(x, what = "sets")), sep = "")
  return(x)
}

library(testthat)

context("Expansion Test")

mev <- getOption("magclass_expand_version")
msm <- getOption("magclass_setMatching")

on.exit({
  options(magclass_expand_version = mev)
  options(magclass_setMatching = msm)
})

test_that("partial expansion", {
  a <- new.magpie(c("AFR.1", "AFR.2", "EUR.1"), fill = 1)
  b <- new.magpie(c("AFR", "EUR"), fill = 1)

  options(magclass_expand_version = NULL)
  options(magclass_setMatching = TRUE)
  expect_identical(magpie_expand(b, a), a)
  options(magclass_setMatching = msm)
  expect_identical(magpie_expand(b, a), a)

  d <- new.magpie(c("AFR.BLUB.1", "AFR.BLUB.2", "EUR.BLUB.1",
                    "AFR.BLA.1", "AFR.BLA.2", "EUR.BLA.1"), fill = 1)
  e <- new.magpie(c("BLA.AFR.A", "BLA.EUR.A", "BLUB.AFR.A", "BLUB.EUR.A",
                    "BLA.AFR.B", "BLA.EUR.B", "BLUB.AFR.B", "BLUB.EUR.B"), fill = 1)
  ee <- magpie_expand(e, d)
  expect_identical(ee, magpie_expand(d, ee))
})

test_that("expansion with unusual set names", {
  d <- new.magpie(c("AFR.BLUB.1", "AFR.BLUB.2", "EUR.BLUB.1", "EUR.BLUB.2",
                    "AFR.BLA.1", "AFR.BLA.2", "EUR.BLA.1", "EUR.BLA.2"), fill = 1)
  e <- new.magpie(c("BLA.AFR.A", "BLA.EUR.A", "BLUB.AFR.A", "BLUB.EUR.A",
                    "BLA.AFR.B", "BLA.EUR.B", "BLUB.AFR.B", "BLUB.EUR.B"), fill = 1)

  # test for cases with missing set names
  names(dimnames(e)) <- NULL
  ee <- magpie_expand(e, d)
  expect_identical(ee, magpie_expand(d, ee))

  # test case with NA set name
  f <- new.magpie("GLO", 1900, "value", fill = 1)
  g <- new.magpie("GLO", 1900, "blub", fill = 1)
  names(dimnames(g))[3] <- NA
  ff <- magpie_expand(f, g)
  expect_identical(ff, magpie_expand(g, ff))

})

test_that("subdimension order", {
  # subdimension order should remain unchanged if possible
  names <- c("bla.value1", "bla.value2")
  h <- new.magpie("GLO", 1900, names, fill = 1)
  i <- new.magpie("GLO", 1900, c("value1", "value2"), fill = 1)
  expect_identical(getItems(magpie_expand(i, h), 3), names)

  x <- new.magpie("GLO", 2000, c("a", "b", "c"))
  y <- new.magpie("GLO", 2000, c("1", "2", "3"))
  xx <- magpie_expand(x, y)
  expect_identical(getItems(xx, 3), getItems(magpie_expand(y, xx), 3))
})

test_that("magpie_expand_dim works", {
  a <- maxample("animal")
  expect_error(magclass:::magpie_expand_dim(a, a, dim = 4), "Unsupported dim setting")
  expect_error(magclass:::magpie_expand_dim(a[, , 2:5], a, dim = 3), "Identical set names but different content")
  expect_identical(magclass:::magpie_expand_dim(collapseDim(a, "day"), a, dim = 2), a)
})

#' Write MAgPIE-object to file
#'
#' Writes a magpie object to a file. The file type is determined by the filename extension.
#' The written file can be read again using \code{\link{read.magpie}}.
#'
#' This function supports writing the following file types:
#' \itemize{
#' \item "cs2" is the new standard format for cellular data with or without header
#' and the first columns (year,regiospatial) or only (regiospatial)
#' \item "cs2b" is identical to "cs2" except that it will suppress the data name
#' if it has only 1 element in the data dimension.
#' \item "csv" is the standard format for regional data with or without header and
#' the first columns (year,region,cellnumber) or only (region,cellnumber)
#' \item "cs3" is another csv format which is
#' specifically designed for multidimensional data for usage in GAMS.
#' \item "cs4" alternative multidimensional format compatible to GAMS, in contrast
#' to cs3 it can also handle sparse data
#' \item "csvr", "cs2r", "cs3r", "cs4r" which are the same formats as the ones
#' previously explained with the only difference that they have a REMIND
#' compatible format
#' \item "cs5" a more generalized version of cs4
#' \item "rds" is an R-default format for storing R objects
#' \item "m" (magpie) and "mz" (magpie zipped) are new formats developed to allow
#' a less storage intensive management of MAgPIE-data. The only difference
#' between both formats is that .mz is gzipped whereas .m is not compressed. So
#' .mz needs less memory, whereas .m might have a higher compatibility to other
#' languages
#' \item "asc" is the ASCII grid format. It can only be applied for gridded data
#' and writes one file per year per data column. In the case that more than
#' one year and data column is supplied several files are written with the
#' structure filename_year_datacolumn.asc
#' \item "tif" is the  GEOtiff format for gridded data.
#' \item "grd" is the native raster format for gridded data.
#' \item "nc" is the netCDF format for gridded data. Check ?magclass::writeNC
#' for more details on how nc files are written.
#' }
#'
#' @param x a magclass object. An exception is that formats written via the raster package
#' (currently "asc", "grd" and "tif") also accept RasterBrick objects which have
#' been previously created from a magclass object via as.RasterBrick)
#' @param file_name file name including file ending (wildcards are supported).
#' Optionally also the full path can be specified here (instead of splitting it
#' to file_name and file_folder)
#' @param file_folder folder the file should be written to (alternatively you
#' can also specify the full path in file_name - wildcards are supported)
#' @param file_type Format the data should be stored as. If file_type=NULL
#' the file ending of the file_name is used as format. See detailed
#' description for a list of available file types. Please be aware that the
#' file_name is independent of the file_type you choose here, so no additional
#' file ending will be added!
#' @param append Decides whether an existing file should be overwritten (FALSE)
#' or the data should be added to it (TRUE). Append = TRUE only works if the
#' existing data can be combined with the new data using the mbind function
#' @param comment Vector of strings: Optional comment giving additional
#' information about the data. If different to NULL this will overwrite the
#' content of attr(x,"comment"). For nc files the unit can be passed
#' with e.g. 'comment = "unit: kg"'.
#' @param comment.char character: a character vector of length one containing a
#' single character or an empty string. Use "" to turn off the interpretation
#' of comments altogether.
#' @param mode File permissions the file should be written with as 3-digit
#' number (e.g. "777" means full access for user, group and all, "750" means
#' full access for user, read access for group and no acess for anybody else).
#' Set to NULL system defaults will be used. Access codes are identical to the
#' codes used in unix function chmod.
#' @param zname name of the time variable for raster files like nc, asc, grd and tif
#' @param ... additional arguments passed to specific write functions.
#' Check ?magclass::writeNC for available arguments when writing nc files.
#' @note
#'
#' The binary MAgPIE formats .m and .mz have the following content/structure
#' (you only have to care for that if you want to implement
#' read.magpie/write.magpie functions in other languages): \cr \cr
#' [ FileFormatVersion | Current file format version number (currently 6) | integer | 2 Byte ] \cr
#' [ nchar_comment | Number of character bytes of the file comment | integer | 4 Byte ] \cr
#' [ nbyte_metadata | Number of bytes of the serialized metadata (currently = 0) | integer | 4 Byte ] \cr
#' [ nchar_sets | Number of characters bytes of all regionnames + 2 delimiter | integer | 2 Byte] \cr
#' [ nyears | Number of years | integer | 2 Byte ]\cr
#' [ yearList | All years of the dataset (0, if year is not present) | integer | 2*nyears Byte ] \cr
#' [ ncells | Number of cells | integer | 4 Byte ]\cr
#' [ nchar_cell | Number of characters bytes of all regionnames + (nreg-1) for delimiters | integer | 4 Byte ] \cr
#' [ cells | Cell names saved as cell1\\cell2 (\\n is the delimiter) | character | 1*nchar_cell Byte ] \cr
#' [ nelem | Total number of data elements | integer | 4 Byte ] \cr
#' [ nchar_data | Number of char. bytes of all datanames + (ndata - 1) for delimiters | integer | 4 Byte ] \cr
#' [ datanames | Names saved in the format data1\\ndata2 (\\n as del.) | character | 1*nchar_data Byte ] \cr
#' [ data | Data of the MAgPIE array in vectorized form | numeric | 4*nelem Byte ] \cr
#' [ comment | Comment with additional information about the data | character | 1*nchar_comment Byte ] \cr
#' [ sets | Set names with \\n as delimiter | character | 1*nchar_sets Byte] \cr
#' [ metadata | serialized metadata information (currently not in use) | bytes | 1*nbyte_metadata Byte] \cr
#'
#' @author Jan Philipp Dietrich, Stephen Bi, Florian Humpenoeder, Pascal Sauer
#' @seealso \code{"\linkS4class{magpie}"},
#' \code{\link{read.magpie}},\code{\link{mbind}}
#' @examples
#' pop <- maxample("pop")
#' path <- tempfile(fileext = ".mz")
#' write.magpie(pop, path)
#' pop2 <- read.magpie(path)
#' @family FileIO
#' @export
write.magpie <- function(x, # nolint: object_name_linter, cyclocomp_linter.
                         file_name, file_folder = "", file_type = NULL, # nolint: object_name_linter.
                         append = FALSE, comment = NULL,
                         comment.char = "*", # nolint: object_name_linter.
                         mode = NULL, zname = "time", ...) {
  umask <- Sys.umask()
  if (!is.null(mode)) {
    umaskMode <- as.character(777 - as.integer(mode))
    Sys.umask(umaskMode)
  } else {
    mode <- as.character(777 - as.integer(as.character(umask)))
  }
  if (is.null(x)) x <- as.magpie(numeric(0))
  if (is.magpie(x) || inherits(x, "RasterBrick")) {
    years <- !(is.null(dimnames(x)[[2]]))

    # if file-type is not mentioned file-ending is used as file-type
    if (is.null(file_type)) {
      file_type <- tail(strsplit(file_name, "\\.")[[1]], 1) # nolint: object_name_linter.
    }
    if (inherits(x, "RasterBrick") && !(file_type %in% c("asc", "grd", "tif"))) {
      stop("RasterBrick format is only allowed for file types: asc, grd and tif")
    }
    if (!file_folder == "") {
      filePath <- paste(file_folder, file_name, sep = "/")
    } else {
      filePath <- file_name
    }

    # look for comment/additional information
    if (is.null(comment) && !is.null(attr(x, "comment"))) comment <- attr(x, "comment")
    if (is.null(comment)) comment <- ""
    comment <- unlist(strsplit(comment, "\n"))

    # expand wildcards
    filePath <- file.path(Sys.glob(dirname(filePath)), basename(filePath))
    if (length(filePath) > 1) {
      filePath <- filePath[1]
      warning("file name is ambiguous, only first alternative is used!")
    }

    if (append && file.exists(filePath)) {
      x2 <- read.magpie(filePath)
      x <- mbind(x2, x)
    }

    if (file_type %in% c("m", "mz")) {
      fformatVersion <- "6"  # File format version (oldest data has version 0)
      comment <- paste(comment, collapse = "\n")
      ncells <- dim(x)[1]
      nyears <- dim(x)[2]
      ndata  <- dim(x)[3]
      cells <- dimnames(x)[[1]]
      cellsCollapsed <- paste(cells, collapse = "\n")
      datanames <- dimnames(x)[[3]]
      datanamesCollapsed <- paste(datanames, collapse = "\n")
      setsCollapsed <- paste(getSets(x, fulldim = FALSE), collapse = "\n")

      if (years) {
        yearList <- as.integer(substr(dimnames(x)[[2]], 2, 5))
      } else {
        yearList <- 0
      }

      if (file_type == "mz") {
        zz <- gzfile(filePath, "wb")
      } else {
        zz <- file(filePath, "wb")
      }

      writeBin(as.integer(fformatVersion), zz, size = 2)
      writeBin(as.integer(nchar(comment, type = "bytes")), zz, size = 4)
      writeBin(as.integer(0), zz, size = 4)
      writeBin(as.integer(nchar(setsCollapsed, type = "bytes")), zz, size = 2)
      writeBin(as.integer(c(nyears, yearList)), zz, size = 2)
      writeBin(as.integer(c(ncells, nchar(cellsCollapsed, type = "bytes"))), zz, size = 4)
      writeChar(cellsCollapsed, zz, eos = NULL)
      writeBin(as.integer(c(ndata * ncells * nyears, nchar(datanamesCollapsed, type = "bytes"))), zz, size = 4)
      if (datanamesCollapsed != "") writeChar(datanamesCollapsed, zz, eos = NULL)
      writeBin(as.numeric(as.vector(x)), zz, size = 4)
      if (comment != "") writeChar(comment, zz, eos = NULL)
      if (nchar(setsCollapsed, type = "bytes") > 0) writeChar(setsCollapsed, zz, eos = NULL)
      close(zz)
      Sys.chmod(filePath, mode)
    } else if (file_type %in% c("asc", "grd", "tif")) {
      if (!requireNamespace("raster", quietly = TRUE)) {
        stop("The package \"raster\" is required!")
      }
      format <- c(asc = "ascii", grd = "raster", tif = "GTiff")
      if (is.magpie(x)) {
        if (ndata(x) != 1) stop("Currently no support for multiple variables for file type \"", file_type,
                                "\". Please store each variable separately.")
        varname <- getItems(x, dim = 3)
        zunit <- ifelse(all(isYear(getYears(x))), "years", "")
        x <- as.RasterBrick(x)
      } else if (inherits(x, "RasterBrick")) {
        tmp <- names(x)
        tmp <- strsplit(tmp, "\\..")
        years <- sort(unique(unlist(lapply(tmp, function(x) x[1]))))
        varname <- sort(unique(unlist(lapply(tmp, function(x) x[2]))))
        zunit <- ifelse(all(isYear(years)), "years", "")
        if (length(varname) != 1) {
          stop("Currently no support for multiple variables for file type \"", file_type,
               "\". Please store each variable separately.")
        }
      }
      if (file_type == "asc") {
        if (dim(x)[3] != 1) stop("asc does not support multiple year layers. Please choose just one!")
        x <- x[[1]]
      }
      if (is.null(varname)) varname <- "Variable"
      raster::writeRaster(x, filename = filePath, format = format[file_type], overwrite = TRUE,
                          zname = zname, zunit = zunit, varname = varname, ...)
    } else if (file_type == "nc") {
      if (!hasCoords(x) && dimExists(1.2, x)) {
        items <- getItems(x, dim = 1.2)
        if (length(items) == 59199 && all(items == seq_len(59199))) {
          # special treatment for data with 59199 cells as this
          # is the originally used 0.5deg resolution in magclass
          # before it had been generalized
          getCoords(x) <- magclassdata$half_deg[c("lon", "lat")]
        }
      }
      if (is.null(comment)) {
        unit <- ""
      } else {
        indicators <- sub(":.*$", "", comment)
        units <- sub("^.*: ", "", comment)
        if (any(grepl("unit", indicators))) {
          unit <- units[grep("unit", indicators)]
        } else {
          unit <- ""
        }
      }
      writeNC(x = x, filename = file_name, unit = unit, zname = zname, ...)
    } else if (file_type == "rds") {
      saveRDS(object = x, file = filePath, ...)
    } else if (file_type == "cs3" || file_type == "cs3r") {
      if (file_type == "cs3r") dimnames(x)[[2]] <- sub("y", "", dimnames(x)[[2]])
      if (dim(x)[3] != prod(vapply(getItems(x, dim = 3, split = TRUE), length, integer(1)))) {
        stop("Input data seems to be sparse but ", file_type, " does not support sparse data. Please use ",
             sub("3", "4", file_type), " instead!")
      }
      x <- unwrap(x)
      if (dim(x)[1] == 1 && (is.null(dimnames(x)[[1]]) || length(grep("GLO", dimnames(x)[[1]])) == 1)) {
        dimnames(x)[[1]] <- "TODELETE"
      } else {
        if (nregions(x) == dim(x)[1]) {
          dimnames(x)[[1]] <- sub("\\..*$", "", dimnames(x)[[1]])
        } else {
          dimnames(x)[[1]] <- sub("\\.", "_", dimnames(x)[[1]])
        }
      }
      x <- wrap(x, map = list(NA, length(dim(x))))
      dimnames(x)[[1]] <- sub("^([^\\.]*)\\.([^\\.]*)", "\\2\\.\\1", dimnames(x)[[1]])

      dimnames(x)[[1]] <- gsub("TODELETE", "", dimnames(x)[[1]])
      dimnames(x)[[1]] <- gsub("\\.\\.", "\\.", dimnames(x)[[1]])
      dimnames(x)[[1]] <- gsub("^\\.", "", dimnames(x)[[1]])
      dimnames(x)[[1]] <- gsub("\\.$", "", dimnames(x)[[1]])
      dimnames(x)[[1]] <- gsub("\\.", ",", dimnames(x)[[1]])


      header <- dimnames(x)[[2]]
      x <- cbind(dimnames(x)[[1]], x)
      dimnames(x)[[2]] <- c(gsub("[^,]*(,|$)", "dummy\\1", x[1, 1]), header)
      zz <- file(filePath, open = "w")
      if (any(comment != "")) writeLines(paste(comment.char, comment, sep = ""), zz)
      utils::write.csv(x, file = zz, quote = FALSE, row.names = FALSE)
      close(zz)
      Sys.chmod(filePath, mode)
    } else if (file_type == "cs4" || file_type == "cs4r") {
      printCells <- nregions(x) < ncells(x)
      printRegions <- (!is.null(getItems(x, dim = 1.1)) && getItems(x, dim = 1.1)[1] != "GLO")
      printData <- ((ndata(x) > 1) | !is.null(getNames(x)))

      output <- as.data.frame(x)
      output <- output[c("Year", "Region", "Cell", names(output)[-c(1:3)])]

      if (!printCells) output["Cell"] <- NULL
      if (!printRegions) output["Region"] <- NULL
      if (!printData) output["Data1"] <- NULL
      if (!years) {
        output["Year"] <- NULL
      } else {
        if (file_type == "cs4") levels(output[["Year"]]) <- paste0("y", levels(output[["Year"]]))
      }
      zz <- file(filePath, open = "w")
      if (any(comment != "")) writeLines(paste(comment.char, comment, sep = ""), zz)
      utils::write.table(output, file = zz, quote = FALSE, row.names = FALSE, col.names = FALSE, sep = ",")
      close(zz)
      Sys.chmod(filePath, mode)
    } else if (file_type == "cs5") {
      output <- as.data.frame(x, rev = 3, raw = TRUE)

      .formatMeta <- function(meta, commentChar) {
        out <- NULL
        for (n in names(meta)) {
          out <- c(out, paste0(commentChar, "META ", n, ": ", paste(meta[[n]], collapse = ", ")))
        }
        return(out)
      }

      zz <- file(filePath, open = "w")
      if (any(comment != "")) writeLines(paste(comment.char, comment, sep = ""), zz)
      writeLines(.formatMeta(attributes(output)[c("names", "dimtype")], comment.char), zz)
      utils::write.table(output, file = zz, quote = FALSE, row.names = FALSE, col.names = FALSE, sep = ",")
      close(zz)

    } else {
      printCells <- nregions(x) < ncells(x)
      printRegions <- (!is.null(getItems(x, dim = 1.1)) && getItems(x, dim = 1.1)[1] != "GLO")
      printData <- ((ndata(x) > 1) | !is.null(getNames(x)))

      if (file_type == "cs2b" && ndata(x) == 1) getNames(x) <- NULL

      # non-cellular data
      if (!printCells && (!printData || !years || !printRegions)) {
        if (file_type == "csvr" || file_type == "cs2r") dimnames(x)[[2]] <- sub("y", "", dimnames(x)[[2]])
        if (!printData) {
          output <-  array(x, dim = dim(x)[1:2], dimnames = list(dimnames(x)[[1]], dimnames(x)[[2]]))
          output <- aperm(output)
          if (printRegions) {
            if (all(dimExists(c(1.2, 1.3), x) == c(TRUE, FALSE))) {
              tmp <- sub("\\.[^.]*$", "", dimnames(x)[[1]])
            } else {
              tmp <- dimnames(x)[[1]]
            }
            output <- rbind(tmp, output)
            if (years) output <- cbind(c("dummy", dimnames(x)[[2]]), output)
          } else {
            if (years) output <- cbind(dimnames(x)[[2]], output)
          }
          header <- FALSE
        } else if (!years) {
          output <-  array(x, dim = dim(x)[c(1, 3)], dimnames = list(dimnames(x)[[1]], dimnames(x)[[3]]))
          header <- !is.null(dimnames(output)[[2]])
          if (printRegions) output <- cbind(substring(dimnames(x)[[1]], 1, 3), output)
          if (header && !printRegions) {
            output <- t(output)
            header <- FALSE
            output <- cbind(dimnames(x)[[3]], output)
          }
        } else {
          output <-  array(x, dim = dim(x)[2:3], dimnames = list(dimnames(x)[[2]], dimnames(x)[[3]]))
          header <- !is.null(dimnames(output)[[2]])
          output <- cbind(dimnames(x)[[2]], output)
          dimnames(output)[[2]][1] <- "dummy"
        }
        if (header && printRegions) dimnames(output)[[2]][1] <- "dummy"
        zz <- file(filePath, open = "w")
        if (any(comment != "")) writeLines(paste(comment.char, comment, sep = ""), zz)
        utils::write.table(output, zz, sep = ",", col.names = header, row.names = FALSE, quote = FALSE)
        close(zz)
        Sys.chmod(filePath, mode)
      } else {
        if (file_type == "csvr" || file_type == "cs2r") dimnames(x)[[2]] <- sub("y", "", dimnames(x)[[2]])
        if (file_type %in% c("cs2", "cs2b", "cs2r")) printRegions <- FALSE
        output <- array(NA, c(dim(x)[1] * dim(x)[2], dim(x)[3] + printRegions + printCells + years))
        output[, (1 + printRegions + printCells + years):dim(output)[2]] <- as.vector(as.matrix(x))
        if (years) {
          yearvec <- c()
          for (year in dimnames(x)[[2]]) yearvec <- c(yearvec, rep(year, dim(x)[1]))
          output[, 1] <- yearvec
        }
        if (printRegions) output[, 1 + years] <- substring(rep(dimnames(x)[[1]], dim(x)[2]), 1, 3)
        if (printCells) {
          if (file_type %in% c("cs2", "cs2b", "cs2r")) {
            output[, 1 + printRegions + years] <- rep(gsub(".", "_", dimnames(x)[[1]], fixed = TRUE), dim(x)[2])
          } else {
            output[, 1 + printRegions + years] <- rep(seq_len(dim(x)[1]), dim(x)[2])
          }
        }
        if (!is.null(dimnames(x)[[3]])) {
          dimnames(output)[[2]] <- c(rep("dummy", printRegions + printCells + years), dimnames(x)[[3]])
          header <- TRUE
        } else {
          header <- FALSE
        }
        zz <- file(filePath, open = "w")
        if (any(comment != "")) writeLines(paste0(comment.char, comment, sep = ""), zz)
        utils::write.table(output, zz, sep = ",", col.names = header, row.names = FALSE, quote = FALSE)
        close(zz)
        Sys.chmod(filePath, mode)
      }
    }
  } else {
    stop("Input is not in MAgPIE-format!")
  }
  if (!is.null(mode)) {
    Sys.umask(umask)
  }
}

test_that("read/write does not affect content", {
  mag <- maxample("pop")
  names(dimnames(mag)) <- NULL
  getNames(mag) <- c("A2-A", "B1-A")
  getComment(mag) <- "This is a comment"
  for (ext in c(".csv", ".cs3", ".cs3r", ".cs4", ".cs4r", ".cs5")) {
    tmpfile <- withr::local_tempfile(fileext = ext)
    write.magpie(mag, tmpfile)
    mag2 <- read.magpie(tmpfile)
    names(dimnames(mag2)) <- NULL
    expect_identical(names(dimnames(mag)), names(dimnames(mag2)))
    expect_equivalent(mag, mag2)
    unlink(tmpfile)
  }

  for (ext in c(".m", ".mz", ".rds")) {
    tmpfile <- withr::local_tempfile(fileext = ext)
    write.magpie(mag, tmpfile)
    mag2 <- read.magpie(tmpfile)
    expect_equivalent(mag, mag2)
    expect_identical(read.magpie(tmpfile, as.array = TRUE), as.array(read.magpie(tmpfile))[, , ])
    unlink(tmpfile)
  }

  bi <- maxample("bilateral")
  for (ext in c(".m", ".mz", ".rds", ".cs5")) {
    tmpfile <- withr::local_tempfile(fileext = ext)
    write.magpie(bi, tmpfile)
    bi2 <- round(read.magpie(tmpfile), 4)
    expect_equivalent(bi, bi2)
    unlink(tmpfile)
  }


  m2 <- mag
  getCells(m2) <- paste0("AFR.", 1:10)
  tmpfile <- withr::local_tempfile(fileext = ".cs3")
  expect_silent(write.magpie(m2, tmpfile))
  expect_silent(in2 <- read.magpie(tmpfile))
  names(dimnames(in2)) <- NULL
  attr(m2, "Metadata") <- NULL
  expect_equal(in2, m2)
  unlink(tmpfile)

  m3 <- setCells(mag[1, , ], "GLO")
  tmpfile <- withr::local_tempfile(fileext = ".cs2r")
  write.magpie(m3, tmpfile)
  mag3 <- read.magpie(tmpfile)
  names(dimnames(mag3)) <- NULL
  expect_equivalent(m3, mag3)
  unlink(tmpfile)

  expect_error(write.magpie(1), "Input is not in MAgPIE-format")

  mag <- maxample("pop")
  tmpfile <- withr::local_tempfile(fileext = ".cs5")
  write.magpie(mag, tmpfile)
  mag2 <- read.magpie(tmpfile)
  expect_identical(names(dimnames(mag)), names(dimnames(mag2)))
  unlink(tmpfile)
})

test_that("read supports older formats", {
  f1 <- system.file("extdata", "testdata", "oldformat1.m", package = "magclass")
  expect_silent(a <- read.magpie(f1))
  ref1 <- new("magpie", .Data = structure(552.666381835938, .Dim = c(1L, 1L, 1L),
                                          .Dimnames = list(i = "AFR", t = "y1995", scenario = "A2")))
  attr(ref1, "FileFormatVersion") <- 4
  expect_equal(a, ref1)

  f2 <- system.file("extdata", "testdata", "oldformat2.mz", package = "magclass")
  expect_silent(a2 <- read.magpie(f2))
  ref2 <- new("magpie", .Data = structure(c(552.666381835938, 1280.63500976562), .Dim = c(2L, 1L, 1L),
                                          .Dimnames = list(i.j = c("AFR.1", "AFR.2"),
                                                           t = "y1995", scenario = "A2")))
  attr(ref2, "FileFormatVersion") <- 4
  expect_equal(a2, ref2)
})

test_that("handling of spatial data works", {
  skip_if_not_installed("ncdf4")
  td <- withr::local_tempdir()

  md <- magclass:::magclassdata$half_deg
  m05 <- new.magpie(paste0(md$region, ".", seq_len(dim(md)[1])), years = c(2000, 2001),
                    names = "nctest", fill = c(md$lon, md$lat))
  m10 <- mbind(m05, m05)

  expect_silent(write.magpie(m05, file.path(td, "test.nc")))
  expect_silent(m05in <- read.magpie(file.path(td, "test.nc")))

  getCoords(m05) <- md[c("lon", "lat")]
  getItems(m05, "region") <- NULL
  getItems(m05, "region1") <- NULL
  m05 <- m05[getItems(m05in, dim = 1), , ]

  expect_identical(m05, m05in)

  getNames(m10) <- c("bla", "blub")
  expect_error(write.magpie(m10, file.path(td, "test.grd")), "no support for multiple variables")
  expect_no_warning(write.magpie(m10, file.path(td, "test.nc")))

  a <- maxample("animal")
  a <- dimSums(a, dim = c(1.3, 1.4, 2, 3))
  write.magpie(a, file.path(td, "animal.asc"))
  asc <- read.magpie(file.path(td, "animal.asc"))
  attr(asc, ".internal.selfref")  <- NULL
  getItems(asc, dim = 2) <- NULL
  getItems(asc, dim = 3) <- NULL
  expect_identical(asc, a)

  expect_error({
    write.magpie(a, file.path(td, "animal.nc"), unknownArg = TRUE, blabla = 1)
  }, "Unknown argument passed to writeNC: unknownArg, blabla")

  write.magpie(a, file.path(td, "animal.nc"), verbose = FALSE)
  expect_silent(anc <- read.magpie(file.path(td, "animal.nc")))
  getItems(anc, dim = 3) <- NULL
  getSets(anc) <- c("x", "y", "d2", "d3")
  expect_identical(anc, a)

  write.magpie(a, file.path(td, "animal2.nc"), gridDefinition = c(3, 7, 49, 54, 0.25))
  animalRaster <- ncdf4::nc_open(file.path(td, "animal2.nc"))
  withr::defer(ncdf4::nc_close(animalRaster))
  expect_identical(as.vector(animalRaster$dim$lon$vals), seq(3, 7, 0.25))
  expect_identical(as.vector(animalRaster$dim$lat$vals), seq(54, 49, -0.25))
})


test_that("read/write conserves cell naming", {
  p <- new.magpie(c("AFR.2", "CPA.3", "AFR.1", "CPA.4"), fill = 0)
  tmpfile <- withr::local_tempfile(fileext = ".mz")
  write.magpie(p, tmpfile)
  p2 <- read.magpie(tmpfile)
  expect_identical(getCells(p), getCells(p2))
})

.fakeFile <- function(filePath) {
  zz <- gzfile(filePath, "wb")
  fformatVersion <- 100
  writeBin(as.integer(fformatVersion), zz, size = 2)
  close(zz)
}

test_that("read/write triggers errors and warnings correctly", {
  a <- maxample("animal")
  expect_error(write.magpie(a, file.path(tempdir(), "tmp.cs3")), "cs3 does not support sparse data")
  expect_error(read.magpie("doesnotexist.csv"), "does not exist")
  path <- file.path(tempdir(), "exists.blub")
  writeLines("exists", path)
  expect_error(read.magpie(path), "Unknown file type")
  expect_error(read.magpie(file.path(tempdir(), "*.blub")), "Unknown file type")

  path <- file.path(tempdir(), "fake.mz")
  .fakeFile(path)
  expect_error(read.magpie(path), "File format is newer")
  unlink(path)

  path1 <- file.path(tempdir(), "animal1.rds")
  write.magpie(a, path1)
  path2 <- file.path(tempdir(), "animal2.rds")
  write.magpie(a, path2)
  expect_warning(read.magpie(file.path(tempdir(), "animal*.rds")), "ambiguous")
  unlink(path1)
  unlink(path2)

  tmpfile <- withr::local_tempfile(fileext = ".rds")
  saveRDS(1, tmpfile)
  expect_error(read.magpie(tmpfile), "does not contain a magpie object")
  unlink(tmpfile)

  expect_error(write.magpie(a, "blub.grd"), "no support for multiple variables")
  expect_error(write.magpie(a[, , 1], "blub.asc"), "choose just one")

  skip_if_not_installed("ncdf4")
  expect_error(write.magpie(as.magpie(1), "bla.nc"), "No coordinates")
})

test_that("copy.magpie works", {
  p <- maxample("pop")
  td <- tempdir()
  write.magpie(p, file.path(td, "test.mz"))
  expect_silent(copy.magpie(file.path(td, "test.mz"), file.path(td, "test2.mz")))
  expect_silent(copy.magpie(file.path(td, "test.mz"), file.path(td, "test3.cs3")))
  expect_silent(copy.magpie(file.path(td, "test.mz"), file.path(td, "test4.mz"), round = 0))
  expect_identical(read.magpie(file.path(td, "test.mz")), read.magpie(file.path(td, "test2.mz")))
  expect_equivalent(read.magpie(file.path(td, "test.mz")), read.magpie(file.path(td, "test3.cs3")))
  expect_identical(round(read.magpie(file.path(td, "test.mz"))), read.magpie(file.path(td, "test4.mz")))
})

test_that("edge cases work", {
  td <- tempdir()
  a <- new.magpie("GLO", 1:10, fill = 1)
  expect_silent(write.magpie(a, file.path(td, "test.csv")))
  expect_silent(b <- read.magpie(file.path(td, "test.csv")))
  expect_equivalent(a, b)
})

#' where
#'
#' Analysis function for magpie objects
#'
#'
#' @param x A logical statement with a magpie object
#' @param plot deprecated. Use the function whereplot in package luplot.
#' @return A list of analysis parameters
#' @author Benjamin Leon Bodirsky, Jan Philipp Dietrich
#' @seealso whereplot in package luplot
#' @examples
#'
#' p <- maxample("pop")
#' where(p > 500)
#' @family SelectionCalculation
#' @export
where <- function(x, plot = NULL) {
  if (!is.null(plot)) warning("Argument plot is deprecated. Use whereplot() for plot=T and where() for plot=F.")

  analysis <- list()

  .whichX <- function(x) {
    tmp <- which(x, arr.ind = TRUE)
    for (i in 1:3) {
      items <- getItems(x, dim = i)
      if (is.null(items)) items <- "dummy"
      tmp[, i] <- items[as.numeric(tmp[, i])]
    }
    out <- list(individual = tmp)
    out$regions <- unique(out$individual[, 1])
    out$years   <- unique(out$individual[, 2])
    out$data    <- unique(out$individual[, 3])
    return(out)
  }

  analysis$true  <- .whichX(x == TRUE)
  analysis$false <- .whichX(x == FALSE)
  analysis$na    <- .whichX(is.na(x))

  # check how many T, F, NAs and NANs-

  tmp <- c(sum((x == TRUE) * 1, na.rm = TRUE), sum((x == FALSE) * 1, na.rm = TRUE), sum(is.na(x) * 1, na.rm = TRUE))
  other <- length(x) - sum(tmp)
  if (other != 0) {
    warning("function is made to analyse logical statements. Your values contain values that are not 1/0/NA")
  }
  tmp <- c(tmp, other)
  names(tmp) <- c("TRUE", "FALSE", "NA", "other")
  analysis$summary <- tmp

  return(analysis)
}

p <- maxample("pop")

test_that("where works", {
  expect_warning(where(p > 10, plot = TRUE), "Argument plot is deprecated")
  expect_warning(where(p), "function is made to analyse logical statements")
  ref <- list(true = list(individual = structure(c("CPA", "CPA", "y1995", "y1995", "A2", "B1"), .Dim = 2:3,
                                                 .Dimnames = list(c("CPA", "CPA"), c("i", "t", "scenario"))),
                          regions = "CPA", years = "y1995", data = c("A2", "B1")),
              false = list(individual = structure(c("AFR", "AFR", "y1995", "y1995", "A2", "B1"), .Dim = 2:3,
                                                  .Dimnames = list(c("AFR", "AFR"), c("i", "t", "scenario"))),
                           regions = "AFR",  years = "y1995", data = c("A2", "B1")),
              na = list(individual = structure(character(0), .Dim = c(0L, 3L),
                                               .Dimnames = list(NULL, c("i", "t", "scenario"))),
                        regions = character(0), years = character(0), data = character(0)),
              summary = c(`TRUE` = 2, `FALSE` = 2, `NA` = 0, other = 0))

  expect_identical(where(p[1:2, 1, ] > 1000), ref)
  p1 <- p2 <- p[, , 1]
  getSets(p2)[3] <- "d3"
  getItems(p2, dim = 3) <- "dummy"
  getItems(p1, dim = 3) <- NULL
  expect_identical(where(p1 > 3000), where(p2 > 3000))
})

#' Has Coordinates
#'
#' Checks, whether object contains coordinates.
#'
#'
#' @param x MAgPIE object
#' @param xlab label of x-dimension
#' @param ylab label of y-dimension
#' @return Boolean indicating whether coordinates were found or not
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getCoords}}
#' @examples
#'
#' hasCoords(maxample("pop"))
#' hasCoords(maxample("animal"))
#'
#' @family ObjectInfo
#' @export
hasCoords <- function(x, xlab = "x", ylab = "y") {
  return(all(dimExists(c(xlab, ylab), x)))
}

#' Set Items
#'
#' Set items of a given (sub-)dimension of a MAgPIE-object
#'
#'
#' @param x MAgPIE object
#' @param dim Dimension for which the items should be returned. Either number or
#' name of dimension or a vector of these. See \code{\link{dimCode}} for more details.
#' @param value a vector with the length of the main dimension the dimnames should be replaced in / added to.
#' If set to NULL the corresponding dimension will be removed.
#' @param maindim main dimension the data should be added to (does not need to be set if \code{dim} exists
#' in the data. Should be set if \code{dim} might not exist, or if \code{dim} might potentially exist
#' in a different main dimension than the one anticipated).
#' @param raw if set to FALSE inputs will be corrected (e.g. dots replaced by the letter "p") if necessary. If
#' TRUE data will be written as is (risking the creation of inconsistent objects).
#' @return the manipulated MAgPIE object
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getItems}}
#' @examples
#' x <- maxample("pop")
#' setItems(x, "i", paste0("REG", 1:ncells(x)))
#' @family GetterSetter
#' @export
setItems <- function(x, dim, value, maindim = NULL, raw = FALSE) {
  getItems(x, dim = dim, maindim = maindim, raw = raw) <- value
  return(x)
}

a <- maxample("animal")

test_that("mselect works", {
  expect_identical(getItems(mselect(a, species = "rabbit"), dim = 3), c("animal.rabbit.black", "animal.rabbit.white"))
  expect_identical(getItems(mselect(a, species = "rabbit", collapseNames = TRUE), dim = 3), c("black", "white"))
  expect_silent(a2 <- mselect(a, year = 2001))
  expect_equal(dim(a2)[2], 0)
  names(dimnames(a)) <- NULL
  expect_error(mselect(a, species = "rabbit"), "Dimnames must have names")
  a <- maxample("animal")
  attr(a, "Metadata") <- NULL # nolint: object_name_linter.
  expect_silent(mselect(a, species = "rabbit", country = "BEL") <- 99) # nolint: object_name_linter.
  expect_true(all(a["BEL", , "rabbit"] == 99))
  a0 <- setItems(setItems(a[, 1, 1], dim = 2, NULL), dim = 3, NULL)
  ref <- new("magpie", .Data = structure(10, .Dim = c(1L, 1L, 1L),
                                         .Dimnames = list(x.y.country.cell = "6p25.49p75.LUX.14106",
                                                          d2 = NULL, d3 = NULL)))
  expect_silent(a0sel <- mselect(a0, country = "LUX"))
  attr(a0sel, ".internal.selfref") <- NULL # nolint: object_name_linter.
  expect_identical(a0sel, ref)
  expect_error(mselect(a, notthere = "nix"), "not found")
  getSets(a)[1] <- "y"
  expect_error(mselect(a, y = "nix"), "found more than once")

  expect_identical(mselect(a, species = "dog"), mselect(a, species = rep("dog", 10000)))

})

context("Item Manipulation Test")

test_that("getItems even works for objects without set names", {
  x <- new.magpie("GLO", 1995, "bla")
  names(dimnames(x)) <- NULL
  expect_identical(getItems(x, 1), "GLO")
  expect_identical(getItems(x, 2), "y1995")
  expect_identical(getItems(x, 3), "bla")
})

test_that("getItems works for missing dimnames", {
  x <- new.magpie("GLO", NULL, NULL)
  expect_identical(getItems(x, 3), NULL)
  expect_identical(getItems(x, 3.1), NULL)
  expect_identical(getItems(x, 3, split = TRUE)[[1]], NULL)
})

pop <- maxample("pop")

test_that("getItems replacement works in case of incorrect set names", {
  getItems(pop, dim = 1, raw = TRUE) <- paste0(getItems(pop, dim = 1), ".", 1:10)
  value <- paste0("x", 1:10)
  getItems(pop, dim = "x", maindim = 1) <- value
  expect_identical(getItems(pop, "x"), value)
})

test_that("getItems can add and replace (sub)dimensions and separators are replaced with commas", {
  x <- pop
  expect_silent(getItems(x, "j", maindim = 1) <- seq_len(dim(x)[1]))
  expect_identical(dimnames(x)[[1]], paste0(dimnames(pop)[[1]], ".", seq_len(dim(x)[1])))
  expect_identical(names(dimnames(x))[1], "i.j")

  expect_silent(getItems(x, "i") <- paste0("A.", rev(seq_len(dim(x)[1]))))
  expect_identical(dimnames(x)[[1]], paste0("Ap", rev(seq_len(dim(x)[1])), ".", seq_len(dim(x)[1])))
  expect_identical(names(dimnames(x))[1], "i.j")

  expect_silent(getItems(x, 1) <- getItems(pop, dim = 1, split = FALSE))
  expect_identical(getItems(x, 1), getItems(pop, 1))
  expect_false(grepl(".", names(dimnames(x))[[1]], fixed = TRUE))

  expect_silent(getItems(x, 1.2) <- rep("GLO", 10))
  expect_identical(unname(getSets(x)["d1.2"]), "newdim")
  expect_true(dimExists(1.2, x))
  expect_silent(getItems(x, 1.2) <- NULL)
  expect_false(dimExists(1.2, x))
  expect_warning(getItems(x, 1.2) <- NULL, "Nothing to do here")
  expect_error(getItems(pop, 1) <- NULL, "Cannot unset dimension")
  pop1 <- pop[1, , ]
  expect_silent(getItems(pop1, 1) <- NULL)
})

test_that("getItems returns errors for unsupported inputs", {
  x <- new.magpie(c("GLO", "GLA"), 1995, "bla")
  expect_error(getItems(x, 1:2) <- "blub", "not supported")
  expect_error(getItems(x, 1) <- "blub", "Wrong number of items")
  expect_error(getItems(x, 4) <- "blub", "main dimension is not specified")
  expect_error(getItems(x, 4, maindim = 4) <- "blub", "Unsupported maindim")

  expect_error(getItems(pop, "t", maindim = 1) <- 12, "dimension different to maindim")

  expect_error(getItems(pop, dim = 1.3), "Subdimension 1.3 does not exist")

  expect_silent(tmp <- getItems(pop, 1))
  names(tmp) <- tmp
  names(tmp)[1] <- "BLA"
  expect_error(getItems(pop, 1) <- tmp, "not all names match")
})

test_that("getItems maps entries when input vector is named", {
  x <- pop
  value <- seq_len(dim(x)[1])
  names(value) <- rev(getItems(x, 1))
  expect_silent(getItems(x, 1) <- value)
  expect_identical(getItems(x, 1), as.character(rev(seq_len(dim(x)[1]))))

  # test for subdimensioin
  x <- pop
  expect_warning(getItems(x, "j", maindim = 1) <- value, "Names of input vector are being ignored")
  expect_silent(x2 <- setItems(x, "i", value))
  expect_silent(getItems(x, "i") <- value)
  expect_identical(getItems(x, "i"), as.character(rev(seq_len(dim(x)[1]))))
  expect_identical(x, x2)

  a <- maxample("animal")
  expect_error(getItems(a, 3.2) <- c("hase", "vogel"), "Wrong number of items")
  expect_error(getItems(a, 3.2) <- c(rabbit = "hase", bird = "vogel", cat = "hund"),
               "Names of input vector do not match")

  expect_silent(getItems(a, 3.2) <- c("hase", "vogel", "hund"))
  expect_identical(getItems(a, dim = 3), c("animal.hase.black", "animal.hase.white", "animal.vogel.black",
                                           "animal.vogel.red", "animal.hund.brown"))
  expect_error(getItems(a, 3.2) <- c("hase", "vogel"), "Wrong number of items")
  expect_error(getItems(a, 3.2) <- c("hase", "vogel"), "Wrong number of items")

  a2 <- maxample("animal")
  expect_silent(getItems(a2, 3.2) <- c(bird = "vogel", dog = "hund", rabbit = "hase"))
  expect_identical(a, a2)
})

a <- maxample("animal")

test_that("mcalc works", {
  a2 <- setItems(2 * a[, , "rabbit"], dim = 3.2, "rabbit2")
  expect_identical(mcalc(a, `rabbit2` ~ 2 * rabbit), a2)
  expect_true(all(mcalc(a, `GLO` ~ 0.5 * NLD - 1 / 2 * NLD) == 0))
  expect_error(mcalc(a, `BLUB` ~ 10 * NOTanElement), "no match")
  getItems(a, dim = 3.1) <- "NLD"
  expect_error(mcalc(a, `BLUB` ~ 10 * NLD), "multiple matches")
  expect_silent(mcalc(a, `dog2` ~ 2 * dog, append = TRUE))
  expect_identical(a[, , "dog2"], mcalc(a, `dog2` ~ 2 * dog))
})

test_that("calculations work", {
  p <- maxample("pop")
  p1 <- setItems(p[, , 1], dim = 3, NULL)
  p0 <- p1[, , -1]
  expect_identical(p0, p1 * p0)
  expect_identical(p0, p0 * p1)
})

a <- maxample("animal")
attr(a, "Metadata") <- NULL
p <- maxample("pop")
attr(p, "Metadata") <- NULL


test_that("fulldim works", {
  # nolint start: undesirable_function_linter.
  expect_warning(fd <- fulldim(a[1:2, 1:2, ]), "deprecated")
  ref <- list(c(2L, 2L, 1L, 3L, 4L), list(c("5p75.53p25.NLD.14084", "6p25.53p25.NLD.14113"),
                                          c("y2000.april.20", "y2000.may.20"), "animal",
                                          c("rabbit", "bird", "dog"),
                                          c("black", "white", "red", "brown")))
  expect_identical(fd, ref)
  p0 <- setItems(p[1, 1, 1], dim = 3, value = NULL)
  expect_identical(suppressWarnings(fulldim(p0)), list(c(1L, 1L, 1L), list(i = "AFR", t = "y1995", d3 = NULL)))
  p2 <- p[1, 1, ]
  getItems(p2, dim = 3, raw = TRUE) <- c("A.B", "C")
  expect_warning(fulldim(p2), "cannot be splitted")
  a2 <- dimSums(a, dim = c(1.1, 1.2, 2.2, 2.3))
  ref <- list(c(2L, 1L, 1L, 3L, 4L),
              list(country.cell = c("NLD.14084", "NLD.14113"), year = "y2000",
                   type = "animal", species = c("rabbit", "bird", "dog"), color = c("black", "white", "red", "brown")))
  expect_identical(suppressWarnings(fulldim(a2[1:2, 1, ])), ref)
  a3 <- dimSums(a2, dim = 1.2)
  ref <- list(c(3L, 3L, 1L, 3L, 4L), list(country = c("NLD", "BEL", "LUX"),
                                          year = c("y2000", "y2001", "y2002"), type = "animal",
                                          species = c("rabbit", "bird", "dog"),
                                          color = c("black", "white", "red", "brown")))
  expect_identical(suppressWarnings(fulldim(a3)), ref)
  # nolint end
})

#' Wrap
#'
#' Reshape an array or a matrix by permuting and/or joining dimensions.
#'
#'
#' @param x An array
#' @param map A list of length equal to the number of dimensions in the
#' reshaped array. Each element should be an integer vectors specifying the
#' dimensions to be joined in corresponding new dimension. One element may
#' equal NA to indicate that that dimension should be a join of all
#' non-specified (remaining) dimensions. Default is to wrap everything into a
#' vector.
#' @param sep A character separating joined dimension names
#' @note This function is extracted from the R.utils library which is licensed
#' under LGPL>=2.1 and written by Henrik Bengtsson.
#' @author Henrik Bengtsson, Jan Philipp Dietrich
#' @seealso \code{\link{unwrap}},\code{\link{fulldim}}
#' @family MAgPIE-Conversions
#' @export
wrap <- function(x, map = list(NA), sep = ".") {
  if (!is.array(x) && !is.matrix(x))
    stop("Argument 'x' is not an array or a matrix: ", class(x)[1])
  if (!is.list(map))
    stop("Argument 'map' is not a list: ", class(map)[1])
  umap <- unlist(map)
  if (any(duplicated(umap))) {
    stop("Argument 'map' contains duplicated dimension indices: ",
         paste(umap[duplicated(umap)], collapse = ", "))
  }
  dim <- dim(x)
  ndims <- length(dim)
  missingDims <- setdiff(1:ndims, umap)
  if (length(missingDims) > 0) {
    wildcard <- is.na(map)
    if (any(wildcard)) {
      map[[which(wildcard)]] <- missingDims
      umap <- unlist(map)
    } else {
      stop("Argument 'map' miss some dimensions: ", paste(missingDims,
                                                          collapse = ", "))
    }
  }
  falseDims <- setdiff(umap, 1:ndims)
  if (length(falseDims) > 0) {
    stop("Argument 'map' contains non-existing dimensions: ",
         paste(falseDims, collapse = ", "))
  }
  if (any(diff(umap) < 0)) {
    perm <- umap
    x <- aperm(x, perm = perm)
    map <- lapply(map, FUN = function(ii) match(ii, perm))
  }
  dim <- dim(x)
  dim2 <- vapply(map, FUN = function(ii) prod(dim[ii]), numeric(1))
  dimnames <- dimnames(x)

  tmpdn <- function(map, dimnames) {
    dimnames2 <- list()
    nn <- NULL
    for (dim in seq_along(map)) {
      names <- NULL
      for (ii in map[[dim]]) {
        if (is.null(names)) {
          names <- dimnames[[ii]]
          nameNames <- names(dimnames)[ii]
        } else {
          names <- paste(names, rep(dimnames[[ii]], each = length(names)),
                         sep = sep)
          nameNames <- paste(nameNames, names(dimnames)[ii], sep = sep)
        }
      }
      dimnames2[[dim]] <- names
      nn <- c(nn, nameNames)
    }
    # Trick to set names even for NULL entries
    dimnames2[[dim + 1]] <- "fake"
    names(dimnames2) <- c(nn, "fake")
    dimnames2[[dim + 1]] <- NULL
    return(dimnames2)
  }

  dim(x) <- dim2
  dimnames <- tmpdn(map, dimnames)
  if (any(dim(x) == 0)) {
    dimnames[dim(x) == 0] <- NULL
  }
  dimnames(x) <- dimnames
  return(x)
}

a <- maxample("animal")
attr(a, "Metadata") <- NULL  # nolint: object_name_linter.
p <- maxample("pop")
attr(p, "Metadata") <- NULL  # nolint: object_name_linter.


test_that("getComment works", {
  x <- as.magpie(1)
  expect_silent(getComment(x) <- "this is a comment")
  expect_identical(getComment(x), "this is a comment")
  expect_identical(getComment(setComment(x, "new comment")), "new comment")
})

test_that("is-methods work", {
  tf <- as.magpie(c(TRUE, FALSE))
  expect_identical(is.nan(as.magpie(c(NaN, 2))), tf)
  expect_identical(is.na(as.magpie(c(NA, 2))), tf)
  expect_identical(is.infinite(as.magpie(c(-Inf, 2))), tf)
  expect_identical(is.finite(as.magpie(c(2, Inf))), tf)
})

test_that("Incorrect inputs are properly detected", {
  expect_error(maxample("nothere"), "Unknown data set")
  expect_warning(magclass:::.duplicates_check(getCoords(a)[c(1, 1, 2:5), ]), "Duplicate entries")
})

test_that("magpiesort works", {
  expect_error(magpiesort(1), "not a MAgPIE object")
  expect_identical(magpiesort(a[, dim(a)[2]:1, ]), magpiesort(a)) # nolint: seq_linter.
  expect_identical(getItems(magpiesort(a), dim = 1),
                   getItems(magpiesort(a[, , NULL]), dim = 1))
  expect_identical(getItems(magpiesort(a[1, , ]), dim = 2),
                   getItems(magpiesort(a[NULL, , ]), dim = 2))
  expect_identical(magpiesort(p), magpiesort(p[dim(p)[1]:1, , ])) # nolint: seq_linter.
})

test_that("(un)wrap works", {
  x <- as.magpie(array(1:6, c(3, 2), list(c("bla", "blub", "ble"), c("up", "down"))))
  ref <- structure(1:6, .Dim = c(1L, 1L, 3L, 2L),
                   .Dimnames = list("GLO", NULL, c("bla", "blub", "ble"), c("up", "down")))
  expect_identical(unwrap(x), ref)
  expect_warning(unwrap(x, sep = ":"), "not supported anymore")
  expect_error(unwrap(1), "not a MAgPIE object")
  expect_error(unwrap(x[, , c(1, 1)]), "Duplicated names detected")
  expect_error(unwrap(a), "needs to be complete")
  p0 <- p[, , 1]
  getItems(p0, dim = 3) <- NULL
  a0 <- as.array(p0)
  names(dimnames(a0)) <- NULL
  expect_identical(unwrap(p0), a0)
  expect_error(wrap(1), "not an array")
  expect_error(wrap(p0, map = 1), "not a list")
  expect_error(wrap(p0, map = list(1, 1)), "duplicated dimension indices")
  expect_error(wrap(p0, map = list(1:4)), "non-existing dimension")
  expect_error(wrap(p0, map = list(1)), "miss some dimensions")
  ref <- new("magpie", .Data = structure(c(AFR.y1995. = 553, CPA.y1995. = 1281),
                                         .Dim = 2L, .Dimnames = list(i.t.d3 = c("AFR.y1995.", "CPA.y1995."))))
  expect_identical(wrap(round(p0[1:2, 1, 1]), map = list(1:3)), ref)
  expect_identical(wrap(p0[1:2, 1, NULL]), new("magpie", .Data = structure(numeric(0), .Dim = 0L)))
})

test_that("head and tail work", {
  expect_identical(head(a, n2 = 2), a[1:3, 1:2, 1:2])
  expect_identical(tail(a, n2 = 2), a[dim(a)[1] - 2:0, dim(a)[2] - 1:0, dim(a)[3] - 1:0])
  expect_identical(tail(a[1, 1, 1]), a[1, 1, 1])
  expect_identical(head(a[1, 1, 1]), a[1, 1, 1])
})

test_that("getCells works", {
  expect_error(getCells(p) <- 12, "Wrong number of cell names supplied")
  p0a <- p0b <- p[NULL, , ]
  expect_silent(getCells(p0a) <- NULL)
  expect_identical(p0a, p0b)
})

test_that("getRegions works", {
  a <- collapseDim(a, dim = c("x", "y"))
  expect_identical(getRegions(a), c("NLD", "BEL", "LUX")) # nolint: undesirable_function_linter.
  expect_warning(getRegions(a) <- rep("BLA", ncells(a)), "deprecated") # nolint: undesirable_function_linter.
  expect_identical(getRegions(a), "BLA") # nolint: undesirable_function_linter.
  a0 <- dimSums(a, dim = 1)
  expect_null(getRegions(a0)) # nolint: undesirable_function_linter.
  expect_equal(nregions(a0), 1) # nolint: undesirable_function_linter.
})

test_that("getRegionList works", {
  a <- collapseDim(a, dim = c("x", "y"))
  expect_warning(rl <- getRegionList(a), "deprecated") # nolint: undesirable_function_linter.
  expect_identical(rl, as.factor(getItems(a, dim = "country", full = TRUE)))
  expect_error(suppressWarnings(getRegionList(a) <- "GLO"), # nolint: undesirable_function_linter.
               "Lengths of RegionLists do not agree")
  expect_warning(getRegionList(a) <- rep("GLO", dim(a)[1]), "deprecated") # nolint: undesirable_function_linter.
  expect_identical(suppressWarnings(getRegionList(a)), # nolint: undesirable_function_linter.
                   as.factor(rep("GLO", dim(a)[1])))
})

test_that("rounding works", {
  ref <- new("magpie", .Data = structure(c(552.67, 1280.64), .Dim = c(2L, 1L, 1L),
                                         .Dimnames = list(i = c("AFR", "CPA"), t = "y1995", scenario = "A2")))
  expect_identical(round(p[1:2, 1, 1], 2), ref)
})

test_that("isYear works", {
  expect_true(all(isYear(getYears(p, as.integer = TRUE), with_y = FALSE)))
  expect_true(all(isYear(getYears(p), with_y = TRUE)))
  expect_false(isYear("yabcd"))
  expect_false(isYear("y12345"))
  expect_false(isYear("abcd", with_y = FALSE))
  expect_false(isYear("12345", with_y = FALSE))
  expect_error(isYear(p), "is no Vector")
})

test_that("getYear works", {
  expect_error(getYears(p) <- 1999, "Wrong number of years")
  expect_silent(getYears(p[, -(1:nyears(p)), ]) <- NULL)
  p1 <- p[, 1, ]
  expect_silent(getYears(p1) <- NULL)
  expect_null(getYears(p1))
  expect_error(getYears(p) <- NULL, "Setting years to NULL is not possible")
  expect_error(getYears(p1) <- "x1000", "Wrong year format")
})

test_that("sizeCheck works", {
  limit <- getOption("magclass_sizeLimit")
  withr::defer(options(magclass_sizeLimit = limit))

  # passing magpie objects instead of magpie object dimensions
  expect_error(magclass:::sizeCheck(p),
               "Numeric vector of dimension sizes expected.")

  # passing magpie object dimensions
  expect_null(magclass:::sizeCheck(dim(p)))

  # passing too large magpie object dimensions
  options(magclass_sizeLimit = 1)
  expect_error(magclass:::sizeCheck(dim(p)),
               "magclass object size .* exceeds limit")
})

test_that("log methods work", {
  expect_identical(as.vector(logb(as.magpie(10), 10)), 1)
  expect_identical(as.vector(log(as.magpie(10), 10)), 1)
})

test_that("sameDims works", {
  expect_true(sameDims(new.magpie(), new.magpie()))

  expect_true(sameDims(new.magpie(c("A", "B")), new.magpie(c("A", "B"))))
  expect_true(sameDims(new.magpie(c("A", "B")), new.magpie(c("B", "A"))))
  expect_false(sameDims(new.magpie(c("A", "C")), new.magpie(c("A", "B"))))

  expect_true(sameDims(new.magpie(years =  c(2, 3)),    new.magpie(years = c(2, 3))))
  expect_true(sameDims(new.magpie(years =  c(2, 3)),    new.magpie(years = c(3, 2))))
  expect_false(sameDims(new.magpie(years = c(1, 3)),    new.magpie(years = c(2, 3))))

  expect_false(sameDims(new.magpie(names =  c("A", "B")),    new.magpie(names = c("B", "C"))))

  expect_true(sameDims(new.magpie(c("A", "B"), c(2, 3), c("SSP1.urban", "SSP1.rural")),
                       new.magpie(c("B", "A"), c(3, 2), c("SSP1.rural", "SSP1.urban"))))

  # Variadic
  expect_true(sameDims(new.magpie(c("A", "B")), new.magpie(c("A", "B")), new.magpie(c("B", "A"))))
  expect_true(sameDims())
  expect_true(sameDims(new.magpie(c("A", "B"))))

  # Wrong input
  expect_error(sameDims(new.magpie(c("A", "B")), list("A", "B")), "cannot handle non-magpie objects")
})

#' @importFrom methods new
#' @exportMethod is.na
setMethod("is.na",
  signature(x = "magpie"),
  function(x) {
    return(as(is.na(x@.Data), "magpie"))
  }
)

#' @exportMethod is.nan
setMethod("is.nan",
  signature(x = "magpie"),
  function(x) {
    return(as(is.nan(x@.Data), "magpie"))
  }
)

#' @exportMethod is.infinite
setMethod("is.infinite",
  signature(x = "magpie"),
  function(x) {
    return(as(is.infinite(x@.Data), "magpie"))
  }
)

#' @exportMethod is.finite
setMethod("is.finite",
  signature(x = "magpie"),
  function(x) {
    return(as(is.finite(x@.Data), "magpie"))
  }
)

#' Write file in report format
#'
#' This function writes the content of a MAgPIE object into a file or returns
#' it directly using the reporting format as it is used for many model
#' inter-comparisons.
#'
#' @param x MAgPIE object or a list of lists with MAgPIE objects as created by
#' read.report. In the latter case settings for model and scenario are
#' overwritten by the information given in the list.
#' @param file file name the object should be written to. If NULL the formatted
#' content is returned
#' @param model Name of the model which calculated the results
#' @param scenario The scenario which was used to get that results.
#' @param unit Unit of the data. Only relevant if unit is not already supplied
#' in Dimnames (format "name (unit)"). Can be either a single string or a
#' vector of strings with a length equal to the number of different data
#' elements in the MAgPIE object
#' @param ndigit Number of digits the output should have
#' @param append Logical which decides whether data should be added to an
#' existing file or an existing file should be overwritten
#' @param skipempty Determines whether empty entries (all data NA) should be
#' written to file or not.
#' @param extracols names of dimensions which should appear in the output as additional columns
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{read.report}}
#' @examples
#'
#' write.report(maxample("pop"))
#' @family FileIO
#' @importFrom utils write.table
#' @export
write.report <- function(x, # nolint: object_name_linter.
                         file = NULL, model = NULL, scenario = NULL, unit = NULL, ndigit = 4,
                         append = FALSE, skipempty = TRUE, extracols = NULL) {

  scenarioCall <- scenario
  modelCall <- model
  if (is.list(x)) {
    if (!is.list(x[[1]])) stop("Wrong format. x must be either a list of lists or a MAgPIE object!")
    if (is.null(file)) stop("file = NULL not supported for lists!")
    for (.scenario in names(x)) {
      for (.model in names(x[[.scenario]])) {
        scenario <- ifelse(is.null(scenarioCall), .scenario, scenario)
        model <- ifelse(is.null(modelCall), .model, model)
        write.report(x[[.scenario]][[.model]], file = file, model = model, scenario = scenario, unit = unit,
                     ndigit = ndigit, append = append, skipempty = skipempty, extracols = extracols)
        append <- TRUE
      }
    }
  } else {
    if (!is.magpie(x)) stop("Input is not a MAgPIE object!")

    if (all(is.na(x)) && isTRUE(skipempty)) {
      msg <- "magclass object contains only NAs, returning empty data table."
      if (!is.null(file)) {
        msg <- paste0(msg, " No file was written.")
      }
      message(msg)
      return(data.table::data.table())
    }

    x <- prepareData(x, model = model, scenario = scenario, unit = unit, skipempty = skipempty,
                     ndigit = ndigit, extracols = extracols)
    if (is.null(file)) return(x)
    if (!file.exists(file)) append <- FALSE
    if (append) {
      # check header for consistency
      header <- read.table(file, nrows = 1, sep = ";", stringsAsFactors = FALSE)
      years1 <- as.numeric(header[vapply(header, is.numeric, logical(1))])
      years2 <- as.numeric(colnames(x)[!is.na(suppressWarnings(as.numeric(colnames(x))))])
      union <- sort(union(years1, years2))
      addycols <- function(data, years) {
        ycols <- !is.na(suppressWarnings(as.numeric(colnames(data))))
        tmp <- data[ycols]
        data <- data[!ycols]
        data[as.character(sort(years))] <- "N/A"
        data[names(tmp)] <- tmp
        return(data)
      }
      if (length(union) > length(years1)) {
        data <- read.table(file, sep = ";", stringsAsFactors = FALSE, header = TRUE, check.names = FALSE)
        data <- data[-length(data)]
        write.table(addycols(data, union), file, quote = FALSE, sep = ";", row.names = FALSE,
                    col.names = TRUE, append = FALSE, eol = ";\n")
      }
      if (length(union) > length(years2)) {
        x <- addycols(as.data.frame(x), union)
      }
    }
    write.table(x, file, quote = FALSE, sep = ";", row.names = FALSE, col.names = !append,
                append = append, eol = ";\n")
  }
}

prepareData <- function(x, model = NULL, scenario = NULL, unit = NULL, skipempty = FALSE,
                        ndigit = 4, extracols = NULL) {

  sep <- "."
  # clean data
  x <- round(clean_magpie(x, what = "sets"), digits = ndigit)
  names(dimnames(x))[1] <- "Region"
  dimnames(x)[[1]] <- sub("^GLO(\\.{0,1}[0-9]*)$", "World\\1", dimnames(x)[[1]])
  dimnames(x)[[2]] <- substring(dimnames(x)[[2]], 2)

  # check for duplicates and possibly remove duplicates
  d <- duplicated(as.data.table(getNames(x)))
  if (any(d)) {
    warning("Data contains duplicate entries (", paste(getNames(x)[d], collapse = ", "),
            "), only first found entries will be written to file!")
    x <- x[, , which(!d)]
  }

  # convert to data.table, reshape and convert to data.frame
  x <- data.table::setDF(data.table::dcast(data.table::as.data.table(x, na.rm = skipempty),
                                           eval(parse(text = paste0("...~", names(dimnames(x))[2])))))

  # split data and dimension information
  data <- x[3:length(x)]
  x <- x[1:2]

  # split subdimensions
  colsplit <- function(x, col, sep = ".") {
    if (all(grepl(sep, x[[col]], fixed = TRUE, useBytes = TRUE))) {
      tmp <- data.frame(t(matrix(unlist(strsplit(as.character(x[[col]]), split = sep, fixed = TRUE, useBytes = TRUE)),
                                 ncol = length(x[[col]]))), stringsAsFactors = FALSE)
      names(tmp) <- strsplit(names(x)[col], split = sep, fixed = TRUE, useBytes = TRUE)[[1]]
      x <- cbind(tmp, x[setdiff(seq_len(ncol(x)), col)])
    }
    return(x)
  }
  for (i in grep(sep, names(x), fixed = TRUE, useBytes = TRUE)) x <- colsplit(x, i, sep = sep)

  for (i in seq_along(x)) {
    if (tolower(names(x)[i]) %in% c("scenario", "model", "region")) next
    if (!inherits(x[[i]], "character")) next
    if (!any(grepl(" \\(.*\\)$", x[[i]]))) next
    x <- unitsplit(x, i)
  }

  correctNames <- function(x, name = "Scenario", replacement = NULL) {
    if (is.null(replacement)) replacement <- "N/A"
    w <- which(tolower(names(x)) == tolower(name))
    if (length(w) == 0) {
      x <- cbind(replacement, x, stringsAsFactors = FALSE)
    } else if (length(w) == 1) {
      x <- cbind(x[w], x[-w], stringsAsFactors = FALSE)
    } else {
      warning("Found ", name, " more than once! First occurrence will be used")
      w <- w[1]
      x <- cbind(x[w], x[-w], stringsAsFactors = FALSE)
    }
    if (is.factor(x[[1]])) x[[1]] <- as.character(x[[1]])
    x[1][x[1] == "NA"] <- "N/A"
    names(x)[1] <- name
    return(x)
  }

  x <- correctNames(x, name = "Unit", replacement = unit)
  if (!is.null(extracols)) {
    for (i in extracols) {
      x <- correctNames(x, name = i)
    }
  }
  x <- correctNames(x, name = "Region", replacement = NULL)
  x <- correctNames(x, name = "Scenario", replacement = scenario)
  x <- correctNames(x, name = "Model", replacement = model)

  nxcol <- length(extracols)

  if (length(x) == (4 + nxcol)) {
    tmp <- "N/A"
  } else {
    tmp <- do.call(paste, c(x[(5 + nxcol):length(x)], sep = "."))
  }
  x <- cbind(x[1:(3 + nxcol)], Variable = tmp, x[4 + nxcol], stringsAsFactors = FALSE)

  data[is.na(data)] <- "N/A"
  x <- cbind(x, data)
  x <- x[do.call("order", x[c("Scenario", "Model", "Variable", "Region")]), ]
  return(x)
}

p <- maxample("pop")
attr(p, "Metadata") <- NULL

test_that("read/write report works", {
  ref <- structure(list(Model = "N/A", Scenario = "N/A", Region = "World", Variable = "1",
                        Unit = "N/A", `1` = 1), row.names = 1L, class = "data.frame")
  expect_error(read.report("bla"), "could not be found")
  f <- tempfile()
  getSets(p) <- c("region", "year", "scenario")
  expect_silent(write.report(p, f))
  expect_silent(p2 <- read.report(f, as.list = FALSE))
  p2 <- collapseDim(p2)
  expect_identical(round(p2, 4), round(p, 4))
  expect_silent(write.report(p, f, append = TRUE))
  expect_warning(p3 <- read.report(f, as.list = TRUE), "Duplicate entries")
  expect_identical(names(p3), c("A2", "B1"))
  expect_identical(names(p3[[1]]), "NA")

  d <- tempdir()
  f1 <- file.path(d, "bla1.mif")
  f2 <- file.path(d, "bla2.mif")
  getSets(p)[3] <- "variable"
  expect_silent(write.report(p, f1, scenario = "A"))
  expect_warning(write.report2(p, f2, scenario = "B"), "Deprecated")
  # capitalize header of f1
  tmp <- readLines(f1)
  tmp[1] <- toupper(tmp[1])
  writeLines(tmp, f1)
  expect_silent(p4 <- read.report(file.path(d, "bla*.mif")))
  expect_identical(names(p4), c("A", "B"))
  expect_identical(p4$A, p4$B)

  write.report(p[, 1:3, ], f)
  # remove header of f
  tmp <- readLines(f)
  writeLines(tmp[2:length(tmp)], f)
  expect_error(read.report(f), "No header given")

  write.report(p[, 1:11, ], f)
  # remove header of f
  tmp <- readLines(f)
  writeLines(tmp[2:length(tmp)], f)
  expect_warning(pguess <- read.report(f, as.list = FALSE), "Years are being guessed")
  years <- c(2005, seq(2010, 2100, 10))
  expect_equal(getYears(pguess, as.integer = TRUE), years)


  getNames(p)[1] <- c("A.2")
  write.report(p, f)
  expect_warning(p4 <- read.report(f, as.list = FALSE), "Replaced")
  expect_identical(getItems(p4, dim = 3), c("NA.NA.Ap2 (NA)", "NA.NA.B1 (NA)"))

  # test list format
  p <- maxample("pop")
  write.report(p, f1)
  pl <- read.report(f1, as.list = TRUE)
  write.report(pl, f2)
  pl2 <- read.report(f2, as.list = TRUE)
  expect_identical(pl, pl2)

  expect_error(write.report(as.list(1)), "Wrong format")
  expect_error(write.report(list(list(as.magpie(1)))), "not supported for lists")
  expect_error(write.report(1), "not a MAgPIE object")
  expect_warning(write.report(p[, , rep(1, 2)], f), "duplicate entries")

  unlink(c(f, f1, f2))
})

test_that("append works", {
  f <- tempfile()
  expect_silent(write.report(p[, 1:3, ], f, model = "A", append = TRUE))
  expect_silent(write.report(p[, 1:2, ], f, model = "B", append = TRUE))
  expect_silent(write.report(p[, , ], f, model = "C", append = TRUE))
  expect_silent(r <- read.report(f, as.list = FALSE))
  expect_equal(dim(r), c(10, 16, 6))
  expect_true(all(is.na(r[, -1:-3, "A"])))
  expect_true(!anyNA(r[, 1:3, "A"]))
  expect_true(all(is.na(r[, -1:-2, "B"])))
  expect_true(!anyNA(r[, 1:2, "B"]))
  expect_true(!anyNA(r[, , "C"]))
  unlink(f)
})

test_that("multidim handling works", {
  p <- addDim(p, 3.2, "Scenario", "blub")
  expect_warning(r <- write.report(p[, 1:2, ]), "Found Scenario more than once")
  expect_identical(names(r), c("Model", "Scenario", "Region", "Variable", "Unit", "1995",
                               "2005"))
  getSets(p)[4] <- "Xtra"
  ref <- structure(list(Model = c("N/A", "N/A", "N/A", "N/A"),
                        Scenario = c("A2", "A2", "B1", "B1"),
                        Region = c("AFR", "CPA", "AFR", "CPA"),
                        Xtra = c("blub", "blub", "blub", "blub"),
                        Variable = c("N/A", "N/A", "N/A", "N/A"),
                        Unit = c("N/A", "N/A", "N/A", "N/A"),
                        `1995` = c(552.6664, 1280.635, 552.6664, 1280.635),
                        `2005` = c(696.44, 1429.53, 721.85, 1429.26)),
                   row.names = c(1L, 3L, 2L, 4L), class = "data.frame")
  expect_identical(write.report(p[1:2, 1:2, ], extracols = "Xtra"), ref)
  ref2 <- ref[-4]
  ref2$Variable <- "blub"
  expect_identical(write.report(p[1:2, 1:2, ]), ref2)
})

test_that("read/write report works with braces", {
  f <- tempfile()
  foo <- new.magpie("DEU", c(2015, 2020),
                    "Emissions|CO2|Energy|Demand|Transportation (w/ bunkers) (Mt CO2/yr)",
                    fill = 0)
  foo["DEU", 2020, "Emissions|CO2|Energy|Demand|Transportation (w/ bunkers) (Mt CO2/yr)"] <- 10
  expect_silent(write.report(foo, f))
  df <- utils::read.csv(f, sep = ";", stringsAsFactors = FALSE)
  expect_identical(df$Unit, "Mt CO2/yr")
})

test_that("write report does not crash with only NAs", {
  f <- tempfile()
  foo <- new.magpie("DEU", c(2015, 2020),
                    "Emissions|CO2|Energy|Demand|Transportation (w/ bunkers) (Mt CO2/yr)",
                    fill = NA)
  expect_identical(write.report(foo, f), data.table::data.table())
  expect_message(write.report(foo, f),
                 "magclass object contains only NAs, returning empty data table. No file was written.")
})

#' mcalc
#'
#' Select values from a MAgPIE-object
#'
#' This functions only work for MAgPIE objects with named dimensions as the
#' dimension name (set_name) has to be used to indicate in which dimension the
#' entries should be searched for!
#'
#' @aliases mcalc mcalc<-
#' @param x MAgPIE object
#' @param f A formula describing the calculation that should be performed
#' @param dim The dimension in which the manipulation should take place. If set
#' to NULL function tries to detect the dimension automatically.
#' @param append If set to TRUE the result will be appended to x, otherwise the
#' result will be returned.
#' @return The calculated MAgPIE object in the case that append is set to
#' FALSE. Otherwise nothing is returned (as x is appended in place)
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{mselect}}
#' @examples
#'
#' pop <- maxample("pop")
#' pop
#' mcalc(pop, X12 ~ A2 * B1, append = TRUE)
#' pop
#' mcalc(pop, `Nearly B1` ~ 0.5 * A2 + 99.5 * B1)
#' @family SelectionCalculation
#' @export
#' @importFrom stats as.formula
mcalc <- function(x, f, dim = NULL, append = FALSE) {
  x <- clean_magpie(x)
  f <- as.formula(f)
  vars <- all.vars(f[[3]])

  if (is.null(dim)) dim <- getDim(vars, x, fullmatch = FALSE, dimCode = FALSE)
  if (length(dim) == 0) stop("Dimension not provided and automatic detection failed (no match)!")
  if (length(dim) > 1) stop("Dimension not provided and automatic detection failed (multiple matches)!")

  for (v in vars) {
    l <- list()
    l[dim] <- v
    tmp <- mselect(x, l, collapseNames = FALSE)
    getNames(tmp, dim = dim) <- all.vars(f[[2]])
    assign(v, tmp)
  }
  if (append) {
    assign(as.character(as.list(match.call())$x), mbind(x, eval(f[[3]])), envir =  parent.frame())
  } else {
    out <- eval(f[[3]])
    return(out)
  }
}

#' dimExists
#'
#' This function checks whether a dimension given by name or number exists.
#'
#' @param dim A vector of dimension numbers or dimension names which should be
#' checked for.
#' @param x MAgPIE object in which the dimensions should be searched for.
#' @param sep A character separating joined dimension names.
#' @return Boolean indicating whether the dimension exists or not.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{dimCode}}
#' @examples
#'
#' pop <- maxample("pop")
#' dimExists(c("t", "scenario", "blablub"), pop)
#'
#' @family ObjectInfo
#' @export
dimExists <- function(dim, x, sep = ".") {
  if (is.null(dim)) {
    return(FALSE)
  }
  if (length(dim) > 1) {
    return(vapply(dim, dimExists, x, sep = sep, FUN.VALUE = logical(1)))
  }

  d <- dimCode(dim, x, sep = sep)
  if (d == 0) {
    return(FALSE)
  }
  if (d %in% c(1:3, (1:3) + 0.1)) {
    return(TRUE)
  }

  maindim <- as.integer(d)
  subdim <- as.integer(substring(d, 3))

  maxsubdim <- nchar(gsub(paste0("[^\\", sep, "]*"), "", dimnames(x)[[maindim]][1])) + 1
  return(subdim <= maxsubdim)
}

context("Dimension existance Test")

pop <- maxample("pop")

test_that("existance of dimensions is properly detected", {
  expect_true(dimExists(1, pop))
  expect_true(dimExists(1.1, pop))
  expect_true(dimExists("t", pop))
  expect_true(dimExists("scenario", pop))
  expect_false(dimExists(3.2, pop))
  expect_false(dimExists("blablub", pop))
  expect_false(dimExists(4, pop))
  expect_false(dimExists(NULL, pop))
  expect_equivalent(dimExists(c("t", "scenario", "blablub"), pop), c(TRUE, TRUE, FALSE))
  expect_equivalent(dimExists(c(4, 1.2, 3.1), pop), c(FALSE, FALSE, TRUE))

  x <- new.magpie("GLO", NULL, NULL)
  expect_true(dimExists(3.1, x))
})

#' Round-method for MAgPIE objects
#'
#' Round-method for MAgPIE-objects respectively. Works exactly as for arrays.
#'
#'
#' @name round-methods
#' @aliases round-methods round,magpie-method
#' @param x a magpie object
#' @param digits integer indicating the number of decimal places (round) or significant
#' digits (signif) to be used. Negative values are allowed.
#' @docType methods
#' @section Methods: \describe{
#'
#' \item{x = "magpie"}{ works as round(x) for arrays. }
#'
#' }
#' @exportMethod round


setMethod("round",
  signature(x = "magpie"),
  function(x, digits = 0) {
    x@.Data <- round(x@.Data, digits = digits)
    return(x)
  }
)

#' Data Class and Tools for Handling Spatial-Temporal Data
#'
#' Data class for increased interoperability working with spatial-temporal
#' data together with corresponding functions and methods (conversions,
#' basic calculations and basic data manipulation). The class distinguishes
#' between spatial, temporal and other dimensions to facilitate the development
#' and interoperability of tools build for it. Additional features are name-based
#' addressing of data and internal consistency checks (e.g. checking for the
#' right data order in calculations).
#'
#' @name magclass-package
#' @aliases magclass-package magclass
#' @author
#' Maintainer: Jan Philipp Dietrich <dietrich@pik-potsdam.de>
"_PACKAGE"

.mselectSupport <- function(search, where, ndim, names) {
  search <- escapeRegex(search)
  matches <- NULL
  while (length(search) > 0) {
    end <- min(length(search), 1000)
    subsearch <- search[1:end]
    search <- search[-(1:end)]
    subsearch <- paste0("(", paste(subsearch, collapse = "|"), ")")
    subsearch <- paste0("^", paste(rep("[^\\.]*\\.", where - 1), collapse = ""), subsearch,
                        paste(rep("\\.[^\\.]*", ndim - where), collapse = ""), "$")
    matches <- union(matches, grep(subsearch, names))
  }
  return(names[sort(matches)])
}

.mselectCoords <- function(x, ...) {
  if (is.null(names(dimnames(x)))) stop("Dimnames must have names in order to use mselect!")
  args <- list(...)
  if (length(args) == 1) if (is.list(args[[1]])) args <- args[[1]]
  sep <- "\\."
  sets <- strsplit(names(dimnames(x)), sep)

  .getItemsX <- function(x, dim) {
    out <- getItems(x, dim = dim)
    if (is.null(out)) out <- 1
    return(out)
  }

  i <- .getItemsX(x, dim = 1)
  j <- .getItemsX(x, dim = 2)
  k <- .getItemsX(x, dim = 3)

  for (n in names(args)) {
    where <- grep(paste0("^", n, "$"), unlist(sets))
    if (length(where) > 1) stop(paste0("set name \"", n, "\" found more than once!"))
    if (length(where) == 0) stop(paste0("set name \"", n, "\" not found!"))

    if (where <= length(sets[[1]])) {
      # spatial
      ndim <- nchar(gsub("[^\\.]", "", getCells(x)[1])) + 1
      i <- .mselectSupport(args[[n]], where, ndim, i)
    } else if (where <= length(unlist(sets[1:2]))) {
      # temporal
      j <- .mselectSupport(args[[n]], where - length(sets[[1]]), length(sets[[2]]), j)
    } else {
      # data
      k <- .mselectSupport(args[[n]], where - length(unlist(sets[1:2])), length(sets[[3]]), k)
    }
  }
  m <- list(i = i, j = j, k = k)
  m[lapply(m, length) == 0] <- NULL
  if (is.null(m$j) && dim(x)[2] > 0) m$j <- -seq_len(dim(x)[2])
  return(m)
}





#' MSelect
#'
#' Select values from a MAgPIE-object
#'
#' This functions only work for MAgPIE objects with named dimensions as the
#' dimension name (set_name) has to be used to indicate in which dimension the
#' entries should be searched for!
#'
#' @aliases mselect mselect<-
#' @param x MAgPIE object
#' @param ... entry selections of the form
#' \code{set_name=c(set_elem1,set_elem2)}. Alternatively a single list element
#' containing these selections can be provided.
#' @param collapseNames Boolean which decides whether names should be collapsed
#' or not.
#' @param value values on which the selected magpie entries should be set.
#' @return The reduced MAgPIE object containing only the selected entries or
#' the full MAgPIE object in which a selection of entries was manipulated.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{collapseNames}}, \code{"\linkS4class{magpie}"}
#' @examples
#'
#' pop <- maxample("pop")
#' mselect(pop, i = c("AFR", "EUR"), scenario = "A2", t = "y2035")
#' @family SelectionCalculation
#' @export
mselect <- function(x, ..., collapseNames = FALSE) {
  m <- .mselectCoords(x, ...)
  if (collapseNames) return(collapseNames(x[m$i, m$j, m$k]))
  return(x[m$i, m$j, m$k])
}

#' @describeIn mselect replace values in magpie object
#' @export
`mselect<-` <- function(x, ..., value) {
  m <- .mselectCoords(x, ...)
  x[m$i, m$j, m$k] <- value
  return(x)
}

#' ~~ Methods for Function rowSums and rowMeans ~~
#'
#' ~~ Methods for function \code{rowSums} and \code{rowMeans}~~
#'
#'
#' @name rowSums-methods
#' @aliases rowSums-methods rowSums,ANY-method rowSums,magpie-method
#' rowMeans-methods rowMeans,ANY-method rowMeans,magpie-method
#' @param x object on which calculation should be performed
#' @param na.rm logical. Should missing values (including NaN) be omitted from the calculations?
#' @param dims integer: Which dimensions are regarded as "rows" or "columns" to sum over. For row*,
#' the sum or mean is over dimensions dims+1, ...; for col* it is over dimensions 1:dims.
#' @param ... further arguments passed to other colSums/colMeans methods
#' @docType methods
#' @section Methods: \describe{
#'
#' \item{list("signature(x = \"ANY\")")}{ normal rowSums and rowMeans method }
#'
#' \item{list("signature(x = \"magpie\")")}{ classical method prepared to
#' handle MAgPIE objects } }
#' @keywords methods ~~ other possible keyword(s) ~~
#' @family Aggregation
#' @exportMethod rowSums

setMethod("rowSums",
  signature(x = "magpie"),
  function(x, na.rm = FALSE, dims = 1, ...) { # nolint: object_name_linter.
    out <- rowSums(as.array(x), na.rm = na.rm, dims = dims, ...)
    out <- as.magpie(as.array(out), spatial = 1)
    return(out)
  }
)

#' complete_magpie
#'
#' MAgPIE objects can be incomplete to reduce memory. This function blows up a
#' magpie object to its real dimensions, so you can apply unwrap.
#'
#'
#' @param x MAgPIE object which should be completed.
#' @param fill Value that shall be written into the missing entries
#' @param dim dimensions in which the completion should take place (1, 2 and/or 3). For
#' full completion use \code{1:3}
#' @return The completed MAgPIE object
#' @author Jan Philipp Dietrich, Benjamin Bodirsky
#' @seealso \code{\link{addDim}},\code{\link{clean_magpie}}
#' @examples
#'
#' pop <- maxample("pop")
#' complete_magpie(pop)
#'
#' ani <- maxample("animal")
#' complete_magpie(ani)
#' @family ObjectCreation
#' @export
complete_magpie <- function(x, fill = NA, dim = 3) { # nolint: object_name_linter.
  .expand <- function(x) {
    grid <- expand.grid(x, stringsAsFactors = FALSE)
    return(sort(apply(grid, 1, paste, collapse = ".")))
  }
  .completeSize <- function(x, dimIn, dim) {
    .prod <- function(x) return(prod(vapply(x, length, integer(1))))
    out <- dimIn
    out[dim] <- vapply(x[dim], .prod, double(1))
    out[out == 0] <- 1
    return(out)
  }
  dim <- sort(unique(dim))
  if (any(!is.element(dim, seq_len(3)))) stop("Invalid dim selection (can only be set to 1, 2, or 3)!")
  items <- getItems(x, split = TRUE)
  dimOut <- .completeSize(items, dim(x), dim)
  if (all(dimOut == dim(x))) return(x)
  dimnamesOut <-  dimnames(x)
  for (i in dim) {
    dimnamesOut[[i]] <- .expand(items[[i]])
  }
  out <- new("magpie", array(data = fill, dim = dimOut, dimnames = dimnamesOut))
  dimnamesIn <- dimnames(x)
  for (i in 1:3) {
    if (is.null(dimnamesIn[[i]])) dimnamesIn[[i]] <- seq_len(dim(x)[i])
  }
  out[dimnamesIn[[1]], dimnamesIn[[2]], dimnamesIn[[3]]] <- x
  getSets(out) <- getSets(x)
  return(out)
}

test_that("complete_magpie works", {
  pop <- maxample("pop")
  expect_identical(complete_magpie(pop), pop)

  a <- maxample("animal")
  ref <- new("magpie",
             .Data = structure(c(0, 4, NA, NA, 6, 10, NA, NA, NA, NA, 0, 0, NA, NA, NA,
                                 NA, 0, 0, NA, NA, NA, NA, 6, 14),
                               .Dim = c(1L, 2L, 12L),
                               .Dimnames = list(x.y.country.cell = "5p75.53p25.NLD.14084",
                                                year.month.day = c("y2000.april.20", "y2001.may.20"),
                                                type.species.color = c("animal.bird.black", "animal.bird.brown",
                                                                       "animal.bird.red", "animal.bird.white",
                                                                       "animal.dog.black", "animal.dog.brown",
                                                                       "animal.dog.red", "animal.dog.white",
                                                                       "animal.rabbit.black", "animal.rabbit.brown",
                                                                       "animal.rabbit.red", "animal.rabbit.white"))))
  a2 <- a[1, c(1, 5), ]
  expect_identical(complete_magpie(a2), ref)

  expect_error(complete_magpie(pop, dim = 4), "Invalid dim selection")
  expect_identical(complete_magpie(a, dim = 1:3), complete_magpie(complete_magpie(a, dim = 1:3), dim = 1:3))
  expect_identical(complete_magpie(a, dim = 1:3), complete_magpie(complete_magpie(a), dim = 1:3))
  expect_identical(dim(complete_magpie(a, dim = 1:3)), c(6720L, 12L, 12L))

  getItems(a2, dim = 1) <- NULL
  getItems(ref, dim = 1) <- NULL
  expect_identical(complete_magpie(a2), ref)

})

#' Class "magpie" ~~~
#'
#' The MAgPIE class is a data format for cellular MAgPIE data with a close
#' relationship to the array data format. \code{is.magpie} tests if \code{x} is
#' an MAgPIE-object, \code{as.magpie} transforms \code{x} to an MAgPIE-object
#' (if possible).
#'
#'
#' @name magpie-class
#' @aliases magpie-class as.magpie as.magpie-methods as.magpie,magpie-method
#' as.magpie,array-method as.magpie,lpj-method as.magpie,data.frame-method
#' as.magpie,numeric-method as.magpie,NULL-method as.magpie,quitte-method
#' as.magpie,tbl_df-method as.magpie,RasterBrick-method as.magpie,logical-method
#' as.magpie,RasterLayer-method as.magpie,RasterStack-method as.magpie,SpatRaster-method
#' as.magpie,LPJmLData-method as.magpie,SpatVector-method
#' is.magpie [,magpie-method [,magpie,ANY,ANY-method [<-,magpie,ANY,ANY-method
#' [<-,magpie-method Ops,magpie,magpie-method Ops,magpie,numeric-method Ops,numeric,magpie-method
#' ifelse,magpie-method is.na,magpie-method is.nan,magpie-method
#' is.infinite,magpie-method is.finite,magpie-method
#' @docType class
#' @param x An object that should be either tested or transformed as/to an
#' MAgPIE-object.
#' @param ... additional arguments supplied for the conversion to a MAgPIE
#' object. Allowed arguments for arrays and dataframes are \code{spatial} and
#' \code{temporal} both expecting a vector of dimension or column numbers which
#' contain the spatial or temporal information. By default both arguments are
#' set to NULL which means that the \code{as.magpie} will try to detect
#' automatically the temporal and spatial dimensions. The arguments will just
#' overwrite the automatic detection. If you want to specify that the data does
#' not contain a spatial or temporal dimension you can set the corresponding
#' argument to 0. In addition \code{as.magpie} for data.frames is also
#' expecting an argument called \code{datacol} which expects a number stating
#' which is the first column containing data. This argument should be used if
#' the dimensions are not detected corretly, e.g. if the last dimension column
#' contains years which are then detected as values and therefore interpreted
#' as first data column. In addition an argument \code{tidy=TRUE} can be used
#' to indicate that the data.frame structure is following the rules of tidy
#' data (last column is the data column all other columns contain dimension
#' information). This information will help the conversion. \code{sep} defines
#' the dimension separator (default is ".") and \code{replacement} defines how
#' the separator as a reserved character should be converted in order to not
#' mess up with the object (default "_").
#' Another available argument for conversions of data.frames and quitte
#' objects to magpie is \code{filter} if set to TRUE (default)
#' "." (separator) will be replaced withe the \code{replacement} character and
#' empty entries will be replaced with a single space. If set to FALSE no filter
#' will be applied to the data.
#' @section Objects from the Class: Objects can be created by calls of the form
#' \code{new("magpie", data, dim, dimnames, ...)}. MAgPIE objects have three
#' dimensions (cells,years,datatype) and the dimensionnames of the first
#' dimension have the structure "REGION.cellnumber". MAgPIE-objects behave the
#' same like array-objects with 2 exceptions: \cr 1.Dimensions of the object
#' will not collapse (e.g. \code{x[1,1,1]} will remain 3D instead of becoming
#' 1D)\cr 2.It is possible to extract full regions just by typing
#' \code{x["REGIONNAME",,]}. \cr\cr
#'
#' Please mind following standards: \cr Header must not contain any purely
#' numeric entries, but combinations of characters and numbers are allowed
#' (e.g. "bla","12" is forbidden, wheras "bla","b12" is allowed)\cr Years
#' always have the structure "y" + 4-digit number, e.g. "y1995"\cr Regions
#' always have the structure 3 capital letters, e.g. "AFR" or "GLO"\cr\cr This
#' standards are necessary to allow the scripts to detect headers, years and
#' regions properly and to have a distinction to other data.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{read.magpie}}, \code{\link{write.magpie}},
#' \code{\link{getRegions}}, \code{\link{getYears}}, \code{\link{getNames}},
#' \code{\link{getCPR}}, \code{\link{ncells}}, \code{\link{nyears}},
#' \code{\link{ndata}}
#' @keywords classes
#' @examples
#'
#' showClass("magpie")
#'
#' pop <- maxample("pop")
#'
#' # returning PAO and PAS for 2025
#' pop["PA", 2025, , pmatch = "left"]
#'
#' # returning CPA for 2025
#' pop["PA", 2025, , pmatch = "right"]
#'
#' # returning CPA PAO and PAS for 2025
#' pop["PA", 2025, , pmatch = TRUE]
#'
#' # returning PAS and 2025
#' pop["PAS", 2025, ]
#'
#' # return all entries for year 2025
#' pop[2025, dim = 2]
#'
#' # returning everything but values for PAS or values for 2025
#' pop["PAS", 2025, , invert = TRUE]
#'
#' # accessing subdimension via set name
#'
#' a <- maxample("animal")
#' a[list(country = "NLD", y = "53p25"), , list(species = c("rabbit", "dog"))]
#'
#' # please note that the list elements act as filter. For instance, the
#' # following example will not contain any dogs as the data set does
#' # not contain any dogs which are black.
#' a[list(country = "NLD", y = "53p25"), , list(species = c("rabbit", "dog"), color = "black")]
#'
#' # it is also possible to extract given combinations of subdimensions
#' # via a data-frame
#' df <- data.frame(getItems(a, 3, split = TRUE, full = TRUE))[c(1, 3, 4), ][3:2]
#' getItems(a[df], 3)
#'
#' # Unknown dimensions to be added in output!
#' df$blub <- paste0("bl", 1:dim(df)[1])
#' getItems(a[df], 3)
#' @exportClass magpie
#' @importFrom data.table as.data.table
#' @importFrom methods setClass


setClass("magpie", contains = "array", prototype = array(0, c(0, 0, 0)))

.mselectDF <- function(x, df) { # nolint: cyclocomp_linter.
  if (is.null(names(dimnames(x)))) {
    stop("Dimnames must have names in order to use mselect!")
  }

  dims <- dimCode(names(df), x)
  dims[dims %in% 1:3] <- dims[dims %in% 1:3] + 0.1

  if (all(dims == 0)) {
    stop("None of the dimensions in the mapping could be found in the magpie object!")
  }
  dfmissing <- NULL
  if (any(dims == 0)) {
    dfmissing <- df[dims == 0]
    df <- df[dims != 0]
    dims <- dims[dims > 0]
  }

  if (anyDuplicated(dims)) {
    stop('Dimension(s) "', paste(names(dims)[duplicated(dims)], collapse = '", "'),
         '" appear(s) more than once in the given mapping!')
  }

  maindim <- unname(round(dims[1]))
  if (any(round(dims) != maindim)) {
    stop("Data.frame must only contain subdimensions with a shared main dimension.",
         " Mixtures across main dimensions are not allowed!")
  }

  sdims <- as.integer(substring(dims, 3))
  maxdim <- nchar(gsub("[^\\.]", "", names(dimnames(x))[maindim])) + 1
  if (nrow(df) == 1) {
    df[1, ] <- escapeRegex(df[1, ])
  }
  if (nrow(df) > 1) {
    df <- data.frame(lapply(df, escapeRegex))
  }
  dmissing <- which(!(1:maxdim %in% sdims))
  sdims <- c(sdims, dmissing)
  for (d in dmissing) df <- cbind(df, "[^\\.]*")
  df <- df[order(sdims)]
  search <-  paste0("^", apply(df, 1, paste, collapse = "\\."), "$")
  .subset <- function(x, f, dim) {
    if (all(f == 1) && dim(x)[dim] == 0) {
      d <- getItems(x, full = TRUE)
      d[[dim]] <- "dummy"
      x <- new.magpie(d[[1]], d[[2]], d[[3]], sets = getSets(x, fulldim = FALSE))
    }
    if (dim == 1) {
      return(x[f, , ])
    }
    if (dim == 2) {
      return(x[, f, ])
    }
    return(x[, , f])
  }
  found <- lapply(search, grep, getItems(x, dim = maindim, full = TRUE), perl = TRUE)
  x <- .subset(x, unlist(found), maindim)
  length <- unlist(lapply(found, length))

  if (!is.null(dfmissing)) {
    if (length(dfmissing) > 1) {
      nameExtRaw <- do.call("paste", c(dfmissing, sep = "."))
    } else {
      nameExtRaw <- dfmissing[[1]]
    }
    nameExtensions <- nameExtRaw[rep(seq_along(nameExtRaw), length)]
    getItems(x, dim = maindim, raw = TRUE) <- paste(getItems(x, dim = maindim, full = TRUE), nameExtensions,
                                                    sep = ".")
    names(dimnames(x))[maindim] <- paste(getSets(x, fulldim = FALSE)[maindim], paste(names(dfmissing), collapse = "."),
                                         sep = ".")
  }
  if (any(length == 0) && nrow(df) > 0) {
    rowExtensions <- gsub("\\.", ".", sub("[^\\.]*", "NA", sub("^\\^", "", sub("\\$$", "", search[length == 0])),
                                          fixed = TRUE), fixed = TRUE)
    if (!is.null(dfmissing)) {
      rowExtensions <- paste(rowExtensions, nameExtRaw[length == 0], sep = ".")
    }
    elems <- rep(1, length(rowExtensions))
    ext <- .subset(x, elems, maindim)
    getItems(ext, dim = maindim, raw = TRUE) <- rowExtensions
    ext[, , ] <- NA
    x <- suppressWarnings(mbind(x, ext))
    if (!is.null(getOption("magclass.verbosity")) && getOption("magclass.verbosity") > 1) {
      message("NOTE (.mselectDF): The following elements were added to x as they appeared",
              " in the mapping but not in x: ", paste0(rowExtensions, collapse = ", "),
              " (values set to NA)\n")
    }
  }
  return(x)
}

.dimextract <- function(x, i, dim, pmatch = FALSE, invert = FALSE) { # nolint: cyclocomp_linter.

  if (is.magpie(i) && is.logical(i)) {
    # check whether input is a 1D magpie object
    if (dim(i)[dim] == length(i)) {
      elemsX <- getItems(x, dim = dim)
      elemsI <- getItems(i, dim = dim)
      i <- as.vector(i)
      if (dim(x)[dim] > 1 && setequal(elemsX, elemsI)) {
        # reorder, if necessary
        i <- i[match(elemsX, elemsI)]
      }
    } else if (dim == 1) {
      i <- magpie_expand(i, x)
    }
  }
  if (is.factor(i)) {
    i <- as.character(i)
  }
  if (invert && is.numeric(i)) {
    i <- -i
  }
  if (!is.character(i) && !is.list(i)) {
    return(i)
  }

  if (length(i) == 0) {
    return(NULL)
  }
  dimnames <- dimnames(x)[[dim]]
  if (is.null(dimnames)) {
    stop("Missing element names in dimensions ", dim, "!")
  }
  .countdots <- function(i) {
    return(nchar(gsub("[^\\.]", "", i)))
  }
  if (!is.list(i) && .countdots(i[1]) == .countdots(dimnames[1]) && pmatch == FALSE) {
    # i vector seems to specify the full dimname
    if (!anyDuplicated(dimnames)) {
      if (invert) {
        return(which(!(dimnames %in% i)))
      } else if (identical(i, dimnames)) {
        return(seq_along(i))
      } else {
        match <- match(i, dimnames)
        if (any(is.na(match))) {
          stop("subscript out of bounds (\"", paste0(i[is.na(match)], collapse = "\", \""), "\")")
        }
        return(match)
      }
    } else {
      warning("Your dimnames in dim=", dim, " contain duplicates! This might lead to erronous results",
              " and bad code performance. Please try to avoid duplicates in dimnames under all circumstances!")
    }
  }

  pmatch1 <- ifelse(pmatch == TRUE | pmatch == "right", "[^.]*", "")
  pmatch2 <- ifelse(pmatch == TRUE | pmatch == "left", "[^.]*", "")

  if (!is.list(i)) {
    i <- list(i)
  }
  elems <- seq_len(dim(x)[dim])
  if (!is.null(names(i))) {
    if (is.null(names(dimnames(x))[dim])) {
      stop("subdimension does not exist (missing set names)!")
    }
    nameOrder <- strsplit(names(dimnames(x))[dim], ".", fixed = TRUE)[[1]]
    if (!all(names(i) %in% nameOrder)) {
      stop("subdimension does not exist (\"", paste(names(i)[!(names(i) %in% nameOrder)], collapse = "\", \""), "\")")
    }
  }
  k <- 1
  for (j in i) {
    if (is.factor(j)) {
      j <- as.character(j)
    }
    if (!is.null(names(i))) {
      subdim <- which(nameOrder == names(i)[k])
      startpattern <- paste0("^", strrep("[^.]*\\.", subdim - 1))
    } else {
      startpattern <- "(^|\\.)"
    }
    tmp <- lapply(paste0(startpattern, pmatch1, escapeRegex(j), pmatch2, "(\\.|$)"),
                  grep, dimnames[elems], perl = TRUE)
    if (any(vapply(tmp, length, length(tmp)) == 0)) {
      stop("subscript out of bounds (\"", paste(j[vapply(tmp, length, length(tmp)) == 0], collapse = "\", \""), "\")")
    }
    tmp <- unlist(tmp)
    if (invert) {
      tmp <- setdiff(seq_along(elems), tmp)
    }
    elems <- elems[tmp]
    k <- k + 1
  }
  return(elems)
}

#' @exportMethod [
setMethod("[", # nolint: cyclocomp_linter.
  signature(x = "magpie"),
  function(x, i, j, k, ..., pmatch = FALSE, invert = FALSE, dim = NULL, drop = FALSE) {
    if (...length() > 0) {
      stop("unknown argument(s) supplied!")
    }
    if (!is.null(dim)) {
      if (!is.element(dim, 1:3)) {
        stop("Invalid dim selection (allowed: 1, 2 or 3)")
      }
      if (!missing(j) || (!missing(k) && !is.null(k) && k != FALSE)) {
        stop("Only single dimension selection allowed when dim is set!")
      }
      if (dim == 1) {
        return(x[i, , , drop = drop, pmatch = pmatch, invert = invert])
      } else if (dim == 2) {
        return(x[, i, , drop = drop, pmatch = pmatch, invert = invert])
      } else {
        return(x[, , i, drop = drop, pmatch = pmatch, invert = invert])
      }
    }
    if (is.null(dim(x))) {
      return(x@.Data[i])
    }
    if (!missing(i)) {
      if (is.data.frame(i)) {
        return(.mselectDF(x, i))
      }
      i <- .dimextract(x, i, 1, pmatch = pmatch, invert = invert)
    }
    if (!missing(j)) {
      .addY <- function(j, n) {
        if (is.list(j)) {
          return(lapply(j, .addY, n))
        }
        if (is.numeric(j) && any(j > n)) {
          return(paste("y", j, sep = ""))
        }
        return(j)
      }
      j <- .addY(j, dim(x)[2])
      if (is.null(j)) {
        j <- seq_len(dim(x)[2])
      }
      j <- .dimextract(x, j, 2, pmatch = pmatch, invert = invert)
    }
    if (!missing(k)) {
      k <- .dimextract(x, k, 3, pmatch = pmatch, invert = invert)
    }
    # if [] contains only 1 argument, the number of elements in sys.call(-1)
    # is 3. This indicates that it is a vectorised call of the form x[i]
    if (length(sys.call(-1)) == 3) {
      return(x@.Data[i])
    }
    x@.Data <- x@.Data[i, j, k, drop = FALSE]
    if (drop) {
      x <- collapseNames(x)
    }
    return(x)
  }
)

#' @exportMethod [<-
setMethod("[<-", # nolint: cyclocomp_linter.
  signature(x = "magpie"),
  function(x, i, j, k, ..., pmatch = FALSE, dim = NULL, value) {
    if (...length() > 0) {
      stop("unknown argument(s) supplied!")
    }
    if (!is.null(dim)) {
      if (!is.element(dim, 1:3)) {
        stop("Invalid dim selection (allowed: 1, 2 or 3)")
      }
      if (!missing(j) || (!missing(k) && !missing(value))) {
        stop("Only single dimension selection allowed when dim is set!")
      }
      if (dim == 1) {
        x[i, , , pmatch = pmatch] <- value
      } else if (dim == 2) {
        x[, i, , pmatch = pmatch] <- value
      } else {
        x[, , i, pmatch = pmatch] <- value
      }
      return(x)
    }
    if (is.null(dim(x))) {
      tmp <- x@.Data
      tmp[i] <- k
      return(tmp)
    }
    if (!missing(i)) {
      if (is.factor(i)) {
        i <- as.character(i)
      }
      if (is.character(i) || is.list(i)) {
        i <- .dimextract(x, i, 1, pmatch = pmatch)
      }
    }
    if (!missing(j)) {
      if (is.factor(j)) {
        j <- as.character(j)
      }
      if (is.numeric(j) & any(j > dim(x)[2])) {
        j <- paste("y", j, sep = "")
      } else if (is.null(j)) {
        j <- seq_len(dim(x)[2])
      } else if ((is.character(j) || is.list(j)) && grepl(".", dimnames(x)[[2]][1], fixed = TRUE)) {
        j <- .dimextract(x, j, 2, pmatch = pmatch)
      }
    }
    if (!missing(k)) {
      if (is.factor(k)) {
        k <- as.character(k)
      }
      if (is.character(k) || is.list(k)) {
        k <- .dimextract(x, k, 3, pmatch = pmatch)
      }
    }
    # if [] contains only 1 argument, the number of elements in sys.call(-1)
    # is 4. This indicates that it is a vectorised call of the form x[i] <- value
    if (length(sys.call(which = -1)) == 4) {
      x@.Data[i] <- value
      return(x)
    }
    if (is.magpie(value)) {
      ii <- if (missing(i)) seq_len(dim(x)[1]) else i
      jj <- if (missing(j)) seq_len(dim(x)[2]) else j
      kk <- if (missing(k)) seq_len(dim(x)[3]) else k
      value <- magpie_expand(value, x[ii, jj, kk])
    } else if (length(value) != length(x@.Data[i, j, k]) && length(value) != 1) {
      # dangerous writing of value as order might be wrong!
      stop("Replacement does not work! Different replacement length!")
    } else if (length(value) != 1 && getOption("magclass.verbosity", default = 0) > 1) {
      message("NOTE ([<-): Dangerous replacement! As replacement value is not",
              " an MAgPIE object name checking is deactivated!\n")
    }
    x@.Data[i, j, k] <- value
    return(x)
  }
)

#' print
#'
#' print method for MAgPIE objects for conventient display of magpie data.
#'
#' @param x MAgPIE object
#' @param drop argument which controls whether empty dimensions should be
#' skipped or not.
#' @param reshape argument that controls tabular representation of nested data
#' dimension cross tables, FALSE will reproduce standard print behavior
#' any pair of two dimension numbers will create a table for these two dims,
#' and loop over the other dimensions
#'
#' @param ... arguments to be passed to or from other methods.
#' @return Invisibly, the MAgPIE object x.
#' @author Jan Philipp Dietrich, Kristine Karstens, Felicitas Beier
#' @seealso \code{\link[base]{print}}
#' @examples
#'
#' pop <- maxample("pop")
#' print(pop)
#' print(pop[, 1, ], drop = FALSE)
#' print(pop[, 1, ])
#' @family Display
#' @export
print.magpie <- function(x, drop = TRUE, reshape = FALSE, ...) {
  if (length(reshape) != 2 || !is.vector(reshape, mode = "numeric")) {
    reshape <- FALSE
  }
  if (any(!reshape)) {
    print(as.array(x)[, , , drop = drop], ...)
  } else if (drop) {
    dims <- as.numeric(substring(names(getSets(x)), 2))
    names(dims) <- getSets(x)
    reshape[reshape %% 1 == 0] <- reshape[reshape %% 1 == 0] + 0.1

    # check if specified subdims exists in x
    if (!all(reshape %in% dims)) {
      warning("Non-existing subdims specified in 'reshape'. Call print again with reshape=FALSE.")
      return(print.magpie(x, reshape = FALSE, drop = drop))
    }

    names(reshape) <- c(names(dims[dims == reshape[1]]), names(dims[dims == reshape[2]]))
    loop           <- setdiff(dims, reshape)
    names(loop)    <- setdiff(names(dims), names(reshape))

    header         <- array(dim = 3)
    names(header)  <- vapply(1:3,
                             function(dim, i) paste(names(dim[round(dim) == i]), collapse = "."),
                             character(1),
                             dim = loop)


    nestedLoop <- function(toPrint, missing, header) {
      i <- missing[1]
      missing <- tail(missing, length(missing) - 1)

      for (j in getItems(toPrint, dim = i)) {
        toPrintReduced <- toPrint[j, dim = floor(i)]
        if (is.na(header[trunc(i)])) {
          header[trunc(i)] <- j
        } else {
          header[trunc(i)] <- paste(c(header[trunc(i)], j), collapse = ".")
        }

        if (length(missing) != 0) {
          nestedLoop(toPrintReduced, missing, header)
        } else {
          header[is.na(header)] <- " "
          writeLines(paste(paste(names(header), collapse = ", "), " = ",
                           paste(header,        collapse = ", ")), sep = "\n\n")
          writeLines(paste("\t", names(reshape[1])))
          print(reshape(subset(as.data.frame(toPrintReduced, rev = 2), select = c(names(reshape), ".value")),
                        timevar = names(reshape[1]), idvar = names(reshape[2]),
                        direction = "wide",
                        new.row.names = getItems(toPrintReduced, dim = reshape[2]),
                        varying = list(getItems(toPrintReduced, dim = reshape[1]))), row.names = FALSE)
          writeLines("")
        }

        if (grepl("\\.", header[trunc(i)])) {
          header[trunc(i)] <- gsub(".([^\\.]+$)", "", header[trunc(i)])
        } else {
          header[trunc(i)] <- NA
        }
      }
    }
    nestedLoop(x, loop, header)
  } else {
    warning("reshape option can just be used with drop=TRUE. Execute with reshape=FALSE:")
    print.magpie(x = x, drop = drop, reshape = FALSE, ...)
  }
  return(invisible(x))
}

#' convergence
#'
#' Cross-Fades the values of one MAGPIE object into the values of another over
#' a certain time
#'
#'
#' @param origin an object with one name-column
#' @param aim Can be twofold: An magpie object or a numeric value.
#' @param start_year year in which the convergence from origin to aim starts.
#' If set to NULL the the first year of aim is used as start_year
#' @param end_year year in which the convergence from origin to aim shall be
#' (nearly) reached. If set to NULL the the last year of aim is used as
#' end_year.
#' @param direction NULL, "up" or "down". NULL means normal convergence in both
#' directions, "up" is only a convergence if origin<aim, "down" means only a
#' convergence if origin>aim
#' @param type "smooth", "s", "linear" or "decay". Describes the function used
#' for convergence (x starts at 0 and ends at 1):
#' "linear" means linear interpolation.
#' "s" is an s-curve which starts from origin in start_year and reaches aim
#' precisely in end_year. After 50\% of the convergence time, it reaches about
#' the middle of the two values. It's using the function x^4/(0.07+x^4)*1.07).
#' "smooth" is a conversion based on the function x^3/(0.1+x^3). With "smooth"
#' only 90\% of convergence will be reached in the end year, full
#' convergence is approached as x approaches infinity.
#' "decay" is a conversion based on the function x / (par + x) * (par + 1).
#' @param par parameter value for convergence function; currently only used for
#' type="decay"
#' @return returns a time-series with the same timesteps as origin, which
#' lineary fades into the values of the aim object
#' @author Benjamin Bodirsky, Jan Philipp Dietrich
#' @examples
#'
#' pop <- maxample("pop")
#' population <- add_columns(pop, "MIX")
#' population[, , "MIX"] <- convergence(population[, , "A2"], population[, , "B1"])
#' @family TemporalOperations
#' @export
convergence <- function(origin, aim,
                        start_year = NULL, end_year = NULL, # nolint: object_name_linter.
                        direction = NULL, type = "smooth", par = 1.5) {
  ### Basic checks ###
  if (!is.magpie(origin)) stop("origin is no magpie object")

  if (is.null(dim(aim))) aim <- as.magpie(array(aim, dim(origin), dimnames(origin)),
                                          spatial = 1, temporal = 2)
  if (!is.magpie(aim)) stop("aim is no magpie object")

  if (!identical(dimnames(aim)[[1]], dimnames(origin)[[1]])) stop("regions have to be the same")

  if (ndata(origin) != 1 && !identical(getNames(origin), getNames(aim))) {
    stop("If there ist more than one name-column, dimnames have to be the same")
  }

  .yearFill <- function(x, ref) {
    if (nyears(x) == 1) {
      tmp <- setYears(x, NULL)
      x <- ref
      x[, , ] <- tmp
    }
    return(x)
  }
  aim    <- .yearFill(aim, origin)
  origin <- .yearFill(origin, aim)

  if (!identical(getYears(aim), getYears(origin))) {
    stop("Objects need the same timesteps, or aim has to have only one timestep")
  }

  .prepYear <- function(year, refyear) {
    if (is.null(year)) year <- refyear
    if (isYear(year, with_y = TRUE)) year <- substr(year, 2, 5)
    year <- as.numeric(year)
    if (!isYear(year, with_y = FALSE)) stop("wrong year format for convergence aim")
    return(year)
  }

  startYear <- .prepYear(start_year, getYears(aim)[1])
  endYear   <- .prepYear(end_year,   getYears(aim)[nyears(aim)])

  # In the case of direction up or down data should only be manipulated in one
  # direction. Therefor, the aim object is manipulated accordingly
  if (!is.null(direction)) {
    if (direction == "up") aim[aim < origin] <- origin[aim < origin]
    else if (direction == "down") aim[aim > origin] <- origin[aim > origin]
    else stop("Illegal direction setting, only up and down are allowed arguments!")
  }

  years <- new.magpie("GLO", getYears(origin), NULL, getYears(origin, as.integer = TRUE))
  pos <- (years - startYear) / (endYear - startYear)
  pos[pos < 0] <- 0
  pos[pos > 1] <- 1
  if (type == "linear") {
    mix <- pos
  } else if (type == "s") {
    mix <- pos^4 / (0.07 + pos^4) * 1.07
  } else if (type == "smooth") {
    mix <- pos^3 / (0.1 + pos^3)
  } else if (type == "decay") {
    mix <- pos / (par + pos) * (par + 1) # nolint: undesirable_function_linter.
  } else {
    stop("type does not exist")
  }
  converged <- aim * mix + origin * (1 - mix)

  return(converged)
}

p <- maxample("pop")
a <- maxample("animal")

test_that("convergence calculation works", {
  expect_error(convergence(1, 1), "no magpie object")
  expect_error(convergence(as.magpie(1), array(1)), "aim is no magpie object")
  expect_error(convergence(p, a), "regions have to be the same")
  expect_error(convergence(a, a[, , 1:3]), "dimnames have to be the same")
  expect_error(convergence(a, a[, 1:3, ]), "need the same timesteps")
  expect_error(convergence(p, p, 123345), "wrong year format")
  expect_error(convergence(p, p, direction = "unknown"), "Illegal direction")
  expect_error(convergence(p, p, type = "doesnotexist"), "does not exist")
  ref <- new("magpie", .Data = structure(c(553, 1281, 554, 349, 716, 293, 1, 2, 3),
                                         .Dim = c(3L, 3L, 1L),
                                         .Dimnames = list(i = c("AFR", "CPA", "EUR"),
                                                          t = c("y1995", "y2005", "y2015"), scenario = "A2")))
  p0 <- p[1:3, 1:3, 1]
  expect_identical(round(convergence(p0, 1:3, type = "linear")), ref)
  ref[, , ] <- c(553, 1281, 554, 345, 709, 290, 1, 2, 3)
  expect_identical(round(convergence(p0, 1:3, type = "s")), ref)
  ref[, , ] <- c(553, 1281, 554, 310, 636, 260, 82, 140, 57)
  expect_identical(round(convergence(p0, 1:3, type = "smooth")), ref)
  ref[, , ] <- c(553, 1281, 554, 262, 537, 220, 1, 2, 3)
  expect_identical(round(convergence(p0, 1:3, type = "decay")), ref)
  ref[, , ] <- c(553, 1281, 554, 848, 1430, 791, 1000, 1518, 1000)
  expect_identical(round(convergence(p0, 1000, direction = "up", type = "linear")), ref)
  ref[, , ] <- c(553, 1281, 554, 696, 1215, 582, 889, 1000, 594)
  p1000 <- p0[, 1, ]
  p1000[, , ] <- 1000
  expect_identical(round(convergence(p0, p1000, direction = "down", type = "linear")), ref)

  # test atypical region names in combination with numeric aim
  p2 <- p[, , 1]
  getItems(p2, dim = 1) <- paste0(getItems(p2, dim = 1), "XYZ")
  p20 <- p2
  p20[, , ] <- 0
  expect_identical(convergence(p2, 0), convergence(p2, p20))
})

tidy2magpie <- function(x, spatial = NULL, temporal = NULL) { # nolint: cyclocomp_linter.
  # assumption: dataframe format in which only the very last
  #             column contains values!
  if ("data.frame" %in% class(x)) {
    class(x) <- "data.frame"
  } else {
    stop("Data does not seem to be a data.frame!")
  }
  sep <- "."

  for (i in seq_len(ncol(x))) {
    if (is.factor(x[[i]])) x[[i]] <- as.character(x[[i]])
  }

  if (is.null(colnames(x))) colnames(x) <- paste0("col", seq_len(dim(x)[2]))
  if (anyNA(colnames(x))) colnames(x)[is.na(colnames(x))] <- "NA"
  colnames(x) <- make.unique(colnames(x), sep = "")

  if (dim(x)[1] == 0) return(copy.attributes(x, new.magpie(NULL)))

  if (is.null(spatial)) spatial  <- colnames(x[-length(x)])[apply(x[-length(x)], 2, is.spatial)]
  if (is.null(temporal)) temporal <- colnames(x[-length(x)])[apply(x[-length(x)], 2, is.temporal)]
  if (is.numeric(spatial)) spatial  <- colnames(x)[spatial]
  if (is.numeric(temporal)) temporal <- colnames(x)[temporal]

  .collapsecol <- function(x, which, sep = ".") {
    xname <- paste(colnames(x)[which], collapse = sep)
    args <- list()
    for (i in seq_along(which)) {
      args[[i]] <- x[, which[i]]
    }
    args["sep"] <- sep
    out <- as.data.frame(do.call(paste, args))
    colnames(out) <- xname
    return(out)
  }

  if (sum(colnames(x) %in% temporal) > 1) {
    t <- .collapsecol(x, which(colnames(x) %in% temporal), sep)
  } else if (sum(colnames(x) %in% temporal) == 1) {
    t <- x[, which(colnames(x) %in% temporal), drop = FALSE]
  } else {
    t <- data.frame(year = rep("NOTIME", dim(x)[1]))
  }
  t[[1]] <- as.character(t[[1]])

  if (sum(colnames(x) %in% spatial) > 1) {
    s <- .collapsecol(x, which(colnames(x) %in% spatial), sep)
  } else if (sum(colnames(x) %in% spatial) == 1) {
    s <- x[, which(colnames(x) %in% spatial), drop = FALSE]
  } else {
    s <- data.frame(region = rep("GLO", dim(x)[1]))
  }
  s[[1]] <- as.character(s[[1]])

  if (sum(!(colnames(x)[-dim(x)[2]] %in% c(temporal, spatial))) > 1) {
    d <- .collapsecol(x, which(!(colnames(x)[-dim(x)[2]] %in% c(temporal, spatial))), sep)
  } else if (sum(!(colnames(x)[-dim(x)[2]] %in% c(temporal, spatial))) == 1) {
    d <- x[, which(!(colnames(x)[-dim(x)[2]] %in% c(temporal, spatial))), drop = FALSE]
  } else {
    d <- data.frame(data = rep(tail(colnames(x), 1), dim(x)[1]))
  }
  d[[1]] <- as.character(d[[1]])

  uSpat <- as.character(unique(s[, 1]))
  uTemp <- as.character(unique(t[, 1]))
  uData <- as.character(unique(d[, 1]))
  dimnames <- list(uSpat, uTemp, uData)
  m <- array(dim = c(length(uSpat), length(uTemp), length(uData)), dimnames = dimnames)
  coord <- as.matrix(cbind(s, t, d))
  .duplicates_check(coord)
  m[coord] <- x[, dim(x)[2]]
  if (dim(m)[2] == 1) if (dimnames(m)[[2]] == "NOTIME") dimnames(m) <- list(dimnames(m)[[1]], NULL, dimnames(m)[[3]])
  if (dim(m)[3] == 1) if (dimnames(m)[[3]] == "NODATA") dimnames(m) <- list(dimnames(m)[[1]], dimnames(m)[[2]], NULL)

  names(dimnames(m)) <- c(names(s), names(t), names(d))
  m <- as.magpie(m, spatial = 1, temporal = 2)
  return(copy.attributes(x, m))
}

#' MAgPIE-Clean
#'
#' Function cleans MAgPIE objects so that they follow some extended magpie
#' object rules (currently it makes sure that the dimnames have names and
#' removes cell numbers if it is purely regional data)
#'
#'
#' @param x MAgPIE object which should be cleaned.
#' @param what term defining what type of cleaning should be performed. Current
#' modes are "cells" (removes cell numbers if the data seems to be regional -
#' this should be used carefully as it might remove cell numbers in some cases
#' in which they should not be removed), "sets" (making sure that all
#' dimensions have names), "items" (replace empty elements with single spaces " ")
#' and "all" (performing all available cleaning methods)
#' @param maindim main dimension(s)  the cleaning should get applied to.
#' @return The eventually corrected MAgPIE object
#' @author Jan Philipp Dietrich
#' @seealso \code{"\linkS4class{magpie}"}
#' @examples
#'
#' pop <- maxample("pop")
#' a <- clean_magpie(pop)
#' @family ObjectCreation
#' @export
clean_magpie <- function(x, what = "all", maindim = 1:3) { # nolint: object_name_linter, cyclocomp_linter.
  availableTypes <- c("cells", "items", "sets")
  if ("all" %in% what) what <- availableTypes
  if (any(!is.element(what, availableTypes))) stop('Unknown setting for argument what ("', what, '")!')

  # remove cell numbers if data is actually regional
  if (is.element("cells", what) && ncells(x) == nregions(x)) {
    items <- getItems(x, dim = 1.1, full = TRUE)
    names(items) <- NULL
    if (!is.null(items)) getItems(x, dim = 1) <- items
  }
  # make sure that all dimensions have names
  if ("sets" %in% what) {

    if (is.null(names(dimnames(x)))) names(dimnames(x)) <- rep(NA, 3)

    .countSubdim <- function(x, sep = "\\.") {
      o <- nchar(gsub(paste0("[^", sep, "]*"), "", x)) + 1
      if (length(o) == 0) o <- 0
      return(o)
    }

    .fixNames <- function(names, ndim, key = "data") {
      if (is.na(names) || names == "" || names == "NA") {
        tmp <- rep(key, max(1, ndim))
        names <- paste(make.unique(tmp, sep = ""), collapse = ".")
      } else {
        cdim <- .countSubdim(names)
        if (ndim != cdim) {
          if (ndim > cdim) {
            names <- paste(c(names, rep(key, ndim - cdim)), collapse = ".")
          } else {
            search <- paste0(c(rep("\\.[^\\.]*", cdim - ndim), "$"), collapse = "")
            names <- sub(search, "", names)
          }
          names <- paste0(make.unique(strsplit(names, "\\.")[[1]], sep = ""), collapse = ".")
        }
      }
      return(names)
    }

    names <- names(dimnames(x))
    keys <- c("region", "year", "data")

    for (i in maindim) {
      names[i] <- .fixNames(names[i], ndim = .countSubdim(dimnames(x)[[i]][1]), key = keys[i])
    }
    names(dimnames(x)) <- names
  }

  if ("items" %in% what) {
    .fixEmptySubims <- function(x, dim) {
      if (is.null(dimnames(x)[[dim]])) return(x)
      pattern <- "(^|\\.)(\\.|$)"
      while (any(grepl(pattern, dimnames(x)[[dim]]))) {
        dimnames(x)[[dim]] <- gsub(pattern, "\\1 \\2", dimnames(x)[[dim]], perl = TRUE)
      }
      return(x)
    }
    for (i in maindim) {
      x <- .fixEmptySubims(x, dim = i)
    }
  }

  return(x)
}

p <- maxample("pop")

test_that("clean_magpie works", {
  expect_identical(clean_magpie(p), p)
  expect_error(clean_magpie(p, what = "themess"), "Unknown setting")
  names(dimnames(p)) <- NULL
  expect_identical(getSets(clean_magpie(p)), c(d1.1 = "region", d2.1 = "year", d3.1 = "data"))
  p0 <- dimSums(maxample("pop"), dim = 1)
  expect_silent(clean_magpie(p0))
  p <- maxample("pop")
  names(dimnames(p)[[1]]) <- paste0("n", 1:10)
  expect_silent(clean_magpie(p))
})

#' Match dimensions of a magpie object to those of a reference object
#'
#' A helper that restricts and expands a magpie object `x` to the size of a
#' magpie object `ref`.  Dimension names not present in `x` are added and set to
#' the value provided by `fill`.  Dimension names not present in `ref` are
#' cropped.
#'
#' @md
#' @param x A `magpie` object to be modified.
#' @param ref A `magpie` object used as a reference for the modification.
#'   Returns `x` if `ref` is `NULL`.
#' @param fill Value to be set in new dimensions.
#' @param dim Subset of dimensions for which the matching should be done.  Can
#'   be either a number between 1 and 3 or a vector of these.  Defaults to all
#'   dimensions (i.e. `1:3`).
#'
#' @return The modified `magpie` object.
#'
#' @author Falk Benke
#'
#' @family DataBinding
#' @export
matchDim <- function(x, ref, dim = 1:3, fill = NA) {

  if (is.null(ref)) {
    return(x)
  }

  if (!is.magpie(x)) {
    stop("`x` must be a magpie object, not `", paste(class(x), collapse = ", "),
         "`")
  }

  if (!is.magpie(ref)) {
    stop("`ref` must be a magpie object or NULL, not `",
         paste(class(ref), collapse = ", "), "`")
  }

  if (!is.numeric(dim) || any(!dim %in% 1:3)) {
    stop("`dim` must be a numeric vector containing only numbers between 1 ",
         "and 3, not `", utils::capture.output(dput(dim)), "`")
  }

  for (i in dim) {
    if (ndim(x, dim = i) != ndim(ref, dim = i)) {
      stop(
        paste0(
          "Unsupported case: magclass objects x and ref have different number ",
          "of subdimensions in dimension ", i
        )
      )
    }
  }

  # extend new object to the union of both objects
  d1 <- if (1 %in% dim) union(getItems(x, dim = 1), getItems(ref, dim = 1)) else getItems(x, dim = 1)
  d2 <- if (2 %in% dim) union(getItems(x, dim = 2), getItems(ref, dim = 2)) else getItems(x, dim = 2)
  d3 <- if (3 %in% dim) union(getItems(x, dim = 3), getItems(ref, dim = 3)) else getItems(x, dim = 3)

  r <- new.magpie(
    cells_and_regions = d1,
    years = d2,
    names = d3,
    fill = fill,
    sets = names(dimnames(x))
  )

  # copy over values from x
  r[getItems(x, dim = 1), getItems(x, dim = 2), getItems(x, dim = 3)] <- x

  # restrict object to dimensions of ref
  if (1 %in% dim) r <- r[getItems(ref, dim = 1), , ]
  if (2 %in% dim) r <- r[, getItems(ref, dim = 2), ]
  if (3 %in% dim) r <- r[, , getItems(ref, dim = 3)]

  return(r)
}

test_that("matchDim works", {

  mp <- new.magpie(
    cells_and_regions = c("AAA", "BBB", "CCC"),
    years = paste0("y", c(2000, 2005, 2010)),
    names = c("foo", "bar", "baz"),
    sets = c("region", "t", "name")
  )

  mp2 <- new.magpie(
    cells_and_regions = c("BBB", "CCC", "DDD"),
    years = paste0("y", c(2000, 2010, 2020)),
    names = c("foo", "bar", "qux"),
    sets = c("region", "t", "name")
  )

  # reject non-magpie objects ----
  expect_error(matchDim(data.frame(), mp2),
               regexp = "must be a magpie object, not")
  expect_error(matchDim(mp, data.frame()),
               regexp = "must be a magpie object or NULL, not")

  # ignore NULL reference ----
  expect_identical(matchDim(mp, NULL), mp)

  # match all dimensions ----
  x <- matchDim(mp, mp2)
  expect_equal(getItems(x, dim = 1), c("BBB", "CCC", "DDD"))
  expect_equal(getItems(x, dim = 2), c("y2000", "y2010", "y2020"))
  expect_equal(getItems(x, dim = 3), c("foo", "bar", "qux"))

  # match only spatial ----
  x <- matchDim(mp, mp2, dim = 1)
  expect_equal(getItems(x, dim = 1), c("BBB", "CCC", "DDD"))
  expect_equal(getItems(x, dim = 2), c("y2000", "y2005", "y2010"))
  expect_equal(getItems(x, dim = 3), c("foo", "bar", "baz"))

  # match only temporal ----
  x <- matchDim(mp, mp2, dim = 2)
  expect_equal(getItems(x, dim = 1), c("AAA", "BBB", "CCC"))
  expect_equal(getItems(x, dim = 2), c("y2000", "y2010", "y2020"))
  expect_equal(getItems(x, dim = 3), c("foo", "bar", "baz"))

  # match only third ----
  x <- matchDim(mp, mp2, dim = 3)
  expect_equal(getItems(x, dim = 1), c("AAA", "BBB", "CCC"))
  expect_equal(getItems(x, dim = 2), c("y2000", "y2005", "y2010"))
  expect_equal(getItems(x, dim = 3), c("foo", "bar", "qux"))

  # reject objects with varying subdimensions ----
  mp3 <- addDim(mp2, dim = 1.2, dimName = "subregion", "yyy")
  expect_error(matchDim(mp, mp3, dim = 1),
               regexp = "magclass objects x and ref have different number of subdimensions in dimension 1")
  expect_no_error(matchDim(mp, mp3, dim = 2))

})

#' maxample
#'
#' A collection of magclass example data sets
#'
#'
#' @param data name of the example data set. Currently available are "pop" (regional population data,
#' previously named "population_magpie"), "animal" (fictional, high-dimensional animal sighting data set)
#' and "bilateral" (fictional, bilateral trade cost data set).
#' @return the chosen example data set
#' @author Jan Philipp Dietrich
#' @examples
#'
#' p <- maxample("pop")
#' str(p)
#'
#' a <- maxample("animal")
#' str(a)
#' getItems(a, split = TRUE)
#' @family Display
#' @export
maxample <- function(data) {
  if (!(data %in% c("pop", "animal", "bilateral"))) stop("Unknown data set \"", data, "\"!")
  .read <- function(data) return(readRDS(system.file(paste0("extdata/examples/", data, ".rds"), package = "magclass")))
  return(.read(data))
}

context("Dimension detection Test")

pop <- maxample("pop")
animal <- maxample("animal")

test_that("getDim matches whole strings", {
  expect_identical(getDim(c("AFR", "CPA"), pop,  dimCode = FALSE), "i")
  expect_identical(getDim(c("AFR", "CPA"), pop,  dimCode = TRUE), 1.1)
  expect_identical(getDim(c("AF", "CP"),   pop), numeric(0))
  expect_identical(getDim("bird", animal,  dimCode = FALSE), "species")
  expect_identical(getDim("bird", animal,  dimCode = TRUE), 3.2)
  expect_identical(getDim("april", animal,  dimCode = FALSE), "month")
  expect_identical(getDim("april", animal,  dimCode = TRUE), 2.2)
  expect_identical(getDim(c("AFR", "CPA"), pop,  fullmatch = TRUE), numeric(0))
  expect_identical(getDim(getItems(pop, 1.1), pop,  fullmatch = TRUE), 1.1)
})

context("dimCode Test")

x   <- maxample("animal")
pop <- maxample("pop")

test_that("dimension codes are correctly extracted", {
  expect_equivalent(dimCode(1, x), 1)
  expect_equivalent(dimCode(1.5, x), 1.5)
  expect_equivalent(dimCode(4.2, x), 0)
  expect_equivalent(dimCode(3.5, x, strict = FALSE), 3.5)
  expect_equivalent(dimCode(3.5, x, strict = TRUE), 0)
  expect_equivalent(dimCode("month", x), 2.2)
  expect_equivalent(dimCode("x", x), 1.1)
  expect_equivalent(dimCode("country", x), 1.3)
  expect_equivalent(dimCode(c("y", "year", "color"), x), c(1.2, 2.1, 3.3))
  expect_equivalent(dimCode(NULL, x), NULL)
  expect_equivalent(dimCode("t", pop), 2)
  expect_equivalent(dimCode("notavail", pop), 0)
})

test_that("illegal dim values are properly detected", {
  expect_error(dimCode("a.b", x), "separator must not be used in dimension name")
  getSets(pop) <- rep("same", 3)
  expect_error(dimCode("same", pop), "found more than once")
  getSets(x, fulldim = FALSE)[3] <- "species.species.color"
  expect_error(dimCode("species", x), "more than once")
  expect_error(dimCode(5, x, missing = "stop"), "illegal dimension")
})

context("Set Manipulation Test")

test_that("sets names can be overwritten", {
  x <- new.magpie("GLO.REG", "y1995", "BLA.BLUB", sets = c("glo.reg", "year", "glo.data"))
  expect_identical(getSets(x), c("d1.1" = "glo", "d1.2" = "reg", "d2.1" = "year", "d3.1" = "glo", "d3.2" = "data"))
  getSets(x)[4] <- "bla"
  expect_identical(getSets(x), c("d1.1" = "glo", "d1.2" = "reg", "d2.1" = "year", "d3.1" = "bla", "d3.2" = "data"))
  getSets(x)["d1.2"] <- "region"
  expect_identical(getSets(x), c("d1.1" = "glo", "d1.2" = "region", "d2.1" = "year", "d3.1" = "bla", "d3.2" = "data"))
  x2 <- x
  expect_silent(getSets(x2) <- NULL)
  expect_identical(x, x2)
  expect_error(getSets(x)[6:7] <- 1:2, "length does not agree")
  names(dimnames(x)) <- "bla"
  expect_identical(getSets(x), c(d1.1 = "bla", d2.1 = NA, d3.1 = NA))
  getSets(x) <- c("", "bla", "blub")
  expect_identical(getSets(x), c(d1.1 = NA, d2.1 = "bla", d3.1 = "blub"))

  a <- maxample("animal")
  getSets(a, fulldim = FALSE)[3] <- "type.species"
  expect_silent(getSets(a)[3] <- "bla")
})

test_that("handling of faulty set names work", {
  a <- maxample("animal")
  b <- a
  getSets(b, fulldim = FALSE)[3] <- "x.y.z"
  getSets(b) <- getSets(a)
  expect_identical(b, a)

  getSets(b, fulldim = FALSE)[3] <- "x.y"
  uniqueSets <- make.unique(getSets(b), sep = "")
  getSets(b) <- uniqueSets
  expect_identical(unname(getSets(b)), c(uniqueSets, "data"))

  a <- maxample("animal")
  b <- a
  sets <- getSets(a)
  names(dimnames(b)) <- NULL
  expect_silent(getSets(b) <- sets)
  expect_identical(b, a)
})

a <- maxample("animal")
attr(a, "Metadata") <- NULL

test_that("add_columns works", {
  expect_error(add_columns(a, dim = "bla"), "does not exist")
  expect_error(add_columns(a, dim = 1:2), "must be a single \\(sub\\)dimension")
  expect_error(add_columns(a, dim = 3.1, fill = 1:10), "fill value must be of length 1")
  a2 <- add_columns(a, addnm = c("horse", "magpie"), dim = "species", fill = 42)
  items <- c("animal.rabbit.black", "animal.rabbit.white", "animal.bird.black",
             "animal.bird.red", "animal.dog.brown", "animal.horse.black",
             "animal.horse.white", "animal.horse.red", "animal.horse.brown",
             "animal.magpie.black", "animal.magpie.white", "animal.magpie.red",
             "animal.magpie.brown")
  expect_identical(getItems(a2, dim = 3), items)
  expect_true(all(a2[, , "horse"] == 42))
  expect_identical(add_columns(a, dim = 3.2, addnm = character(0)), a)
  expect_identical(add_columns(a, dim = 3.2, addnm = NULL), a)
})

context("Print test")

pop <- maxample("pop")

test_that("order in reshape does not effect header", {
  headerReshape12Print <- capture.output(print(pop, reshape = c(1, 2)))[1]
  headerReshape21Print <- capture.output(print(pop, reshape = c(2, 1)))[1]
  expect_identical(headerReshape21Print, headerReshape12Print)
})

test_that("reshape FALSE and reshape with c(2,1) are of same length for pure 3 dimensional data", {
  printReshape21 <- capture.output(print(pop[1:4, 1:4, ], reshape = c(2, 1)))
  printPure      <- capture.output(print(pop[1:4, 1:4, ]))
  expect_identical(length(printPure), length(printReshape21))
})

test_that("reshape can table subdims, other dims will be printed in header", {
  animal <- maxample("animal")
  d <- c(3.1, 3.2, 3.3)
  headerReshapesubdim3print <- capture.output(print(animal, reshape = d))[1]
  expect_true(all(vapply(getSets(animal)[paste0("d", d)], grepl, logical(1), headerReshapesubdim3print)))
})

test_that("everything else works", {
  expect_warning(capture.output(print(pop, drop = FALSE, reshape = 1:2)), "Execute with reshape=FALSE")
  expect_warning(capture.output(print(pop, reshape = c(2.3, 2.1))), "Call print again with reshape=FALSE")
})

context("Coordinate handling test")


test_that("getCoords can read and write coordinates", {
  x <- maxample("pop")
  coords <- data.frame(x = 1:ncells(x) + 0.1, y = ncells(x):1 + 0.1)
  expect_error(getCoords(x), "No coordinates found")
  expect_silent(getCoords(x) <- coords)
  expect_identical(getCoords(x), coords)
  expect_silent(getCoords(x) <- as.matrix(coords))
  expect_identical(getCoords(x), coords)
  expect_error(getCoords(x) <- 1:10, "unsupported format")
  expect_error(getCoords(x) <- head(coords), "Number of rows does not match")
  expect_identical(getItems(x, "x"), sub(".", "p", coords$x, fixed = TRUE))
  coords2 <- data.frame(x = coords$y, y = coords$x)
  expect_silent(getCoords(x) <- coords2)
  expect_identical(getItems(x, "x"), sub(".", "p", coords2$x, fixed = TRUE))
  expect_identical(getCoords(x), coords2)
  expect_silent(getCoords(x, xlab = "x2", ylab = "y2") <- coords)
  expect_identical(getItems(x, "y2"), sub(".", "p", coords$y, fixed = TRUE))
  expect_identical(getCoords(x, xlab = "x2", ylab = "y2"), coords)
})

noC <- function(x) {
  getComment(x) <- NULL
  return(x)
}

test_that("ifelse method works correctly", {
  p <- maxample("pop")
  p[1, 1, 1] <- Inf
  p[2, 1, 1] <- -Inf
  p2 <- p[10:1, , ]
  expect_identical(noC(ifelse(p2 > 1000, p, 0)), noC(ifelse(p2 > 1000, p2, 0)))
  p[c(1, 4, 6)] <- NA
  p2 <- p[10:1, , ]
  expect_identical(noC(ifelse(!is.na(p2) & !is.infinite(p2), p, 0)), noC(ifelse(!is.na(p2)  & !is.infinite(p2), p2, 0)))
  expect_error(ifelse(p, 1, 9), "must only contain booleans")
})

test_that("years is not converted if conversion to int fails", {
  df <- as.data.frame(new.magpie("AFR.1", "2000-02-02", "test", 123))
  expect_identical(as.character(df$Year), "2000-02-02")

  df <- as.data.frame(new.magpie("AFR.1", "some text", "test", 123))
  expect_identical(as.character(df$Year), "some text")

  df <- as.data.frame(new.magpie("AFR.1", "2000", "test", 123))
  expect_identical(as.character(df$Year), "2000")

  df <- as.data.frame(new.magpie("AFR.1", "y2000", "test", 123))
  expect_identical(as.character(df$Year), "2000")

  df <- as.data.frame(new.magpie("AFR.1", "yy2000", "test", 123))
  expect_identical(as.character(df$Year), "yy2000")

})

test_that("data.frame conversion works", {
  a <- maxample("animal")
  getItems(a, "x")[2] <- "-6p25"
  expect_silent(a2 <- as.data.frame(a, rev = 2))
  expect_silent(a3 <- as.data.frame(a, rev = 3))
  expect_silent(a3r <- as.data.frame(a, rev = 3, raw = TRUE))

  ea2 <- structure(list(x = c(5.75, -6.25), y = c(53.25, 53.25),
                        country = structure(c(1L, 1L), .Label = c("NLD", "BEL", "LUX"), class = "factor"),
                        cell = c(14084L, 14113L), year = c(2000L, 2000L),
                        month = structure(c(1L, 1L), .Label = c("april", "may", "june", "august"), class = "factor"),
                        day = c(20L, 20L), type = structure(c(1L, 1L), .Label = "animal", class = "factor"),
                        species = structure(c(1L, 1L), .Label = c("rabbit", "bird", "dog"), class = "factor"),
                        color = structure(c(1L, 1L), .Label = c("black", "white", "red", "brown"), class = "factor"),
                        .value = c(0, 0)), dimtype = c(".spat1", ".spat2", ".spat3", ".spat4", ".temp1", ".temp2",
                                                       ".temp3", ".data1", ".data2", ".data3", ".value"),
                   row.names = 1:2, class = "data.frame")

  ea3 <- structure(list(x = c(5.75, -6.25), y = c(53.25, 53.25), country = c("NLD", "NLD"),
                        cell = c(14084L, 14113L), year = c(2000L, 2000L),
                        month = c("april", "april"), day = c(20L, 20L),
                        type = c("animal", "animal"), species = c("rabbit", "rabbit"),
                        color = c("black", "black"), .value = c(0, 0)),
                   dimtype = c(".spat1", ".spat2", ".spat3", ".spat4", ".temp1", ".temp2",
                               ".temp3", ".data1", ".data2", ".data3", ".value"),
                   row.names = 1:2, class = "data.frame")

  ea3r <- structure(list(x = c("5p75", "-6p25"), y = c("53p25", "53p25"), country = c("NLD", "NLD"),
                         cell = c(14084L, 14113L), year = c("y2000", "y2000"), month = c("april", "april"),
                         day = c(20L, 20L), type = c("animal", "animal"), species = c("rabbit", "rabbit"),
                         color = c("black", "black"), .value = c(0, 0)),
                    dimtype = c(".spat1", ".spat2", ".spat3", ".spat4", ".temp1", ".temp2",
                                ".temp3", ".data1", ".data2", ".data3", ".value"),
                    row.names = 1:2, class = "data.frame")

  expect_identical(head(a2, n = 2), ea2)
  expect_identical(head(a3, n = 2), ea3)
  expect_identical(head(a3r, n = 2), ea3r)

  # test removal of empty dimensions
  b <- a[, 1, 1]
  getItems(b, dim = 2) <- NULL

  eb3 <- structure(list(x = c(5.75, -6.25, 6.75, 4.75, 5.75, -6.25), y = c(53.25, 53.25, 53.25, 52.75, 52.75, 52.75),
                        country = c("NLD", "NLD", "NLD", "NLD", "NLD", "NLD"),
                        cell = c(14084L, 14113L, 14141L, 14040L, 14083L, 14112L),
                        type = c("animal", "animal", "animal", "animal", "animal", "animal"),
                        species = c("rabbit", "rabbit", "rabbit", "rabbit", "rabbit", "rabbit"),
                        color = c("black", "black", "black", "black", "black", "black"),
                        .value = c(0, 0, 8, 3, 4, 3)),
                   dimtype = c(".spat1", ".spat2", ".spat3", ".spat4", ".data1", ".data2", ".data3", ".value"),
                   row.names = c(NA, 6L), class = "data.frame")

  expect_silent(b3 <- as.data.frame(b, rev = 3))
  expect_identical(head(b3), eb3)

})

test_that("extend works", {
  animal <- maxample("animal")
  # shuffle first dim to ensure order does not matter
  animal <- animal[sample(seq_along(getItems(animal, 1))), , ]
  animalCoords <- getCoords(animal)
  expect_identical(guessResolution(animal), 0.5)

  x <- extend(animal)
  expectedCoords <- expand.grid(x = seq(min(animalCoords$x), max(animalCoords$x), 0.5),
                                y = seq(max(animalCoords$y), min(animalCoords$y), -0.5),
                                KEEP.OUT.ATTRS = FALSE)
  expect_identical(getCoords(x), expectedCoords)
  expect_identical(getSets(x), getSets(animal))
  expect_identical(getItems(x[getItems(animal, 1), , ], 1), getItems(animal, 1))

  expect_error(extend(animal, gridDefinition = c(3.25, 6.75, 53.25, 49.75, 1)),
               "The coordinates of the input object are not a subset of the extended coordinates")

  x <- extend(animal, gridDefinition = c(3.25, 6.75, 53.25, 49.75, 1), crop = TRUE)
  expectedCoords <- expand.grid(x = seq(3.25, 6.75, 1),
                                y = seq(53.25, 49.75, -1),
                                KEEP.OUT.ATTRS = FALSE)
  expect_identical(getCoords(x), expectedCoords)

  x <- extend(animal, gridDefinition = c(-3.25, 16.75, 63.25, -59.75, 0.25))

  expectedCoords <- expand.grid(x = seq(-3.25, 16.75, 0.25),
                                y = seq(63.25, -59.75, -0.25),
                                KEEP.OUT.ATTRS = FALSE)
  expect_identical(getCoords(x), expectedCoords)

  # ensure there are no assumptions about position of subdims x and y in first dim
  animal2 <- dimOrder(animal, c(3, 4, 2, 1), dim = 1)
  expect_identical(names(dimnames(animal2))[1], "country.cell.y.x")
  x <- extend(animal2)
  expectedCoords <- expand.grid(x = seq(min(animalCoords$x), max(animalCoords$x), 0.5),
                                y = seq(max(animalCoords$y), min(animalCoords$y), -0.5),
                                KEEP.OUT.ATTRS = FALSE)
  expect_identical(getCoords(x), expectedCoords)
})

a <- maxample("animal")

test_that("getNames works", {
  expect_identical(getNames(a, dim = 2), c("rabbit", "bird", "dog"))
  expect_silent(getNames(a, dim = 2) <- c("hase", "vogel", "hund"))
  expect_identical(getNames(a), c("animal.hase.black", "animal.hase.white", "animal.vogel.black",
                                  "animal.vogel.red", "animal.hund.brown"))
  expect_identical(getNames(setNames(a, paste0("a.", 1:5))),  paste0("a.", 1:5))
  expect_error(getNames(a) <- NULL, "Cannot unset")
  expect_error(getNames(a) <- "a", "Wrong number of items")
  expect_silent(getNames(a[, , NULL]) <- character())
  expect_error(getNames(a, dim = 2:3) <- 1, "Unsupported dim selection")
  expect_error(getNames(a, dim = 5) <- 1, "Wrong number of items")
  expect_error(getNames(a, dim = "blub") <- 1, "does not exist")
  expect_silent(getNames(a, dim = "type") <- "tier")
  expect_identical(getNames(a, dim = "type"), "tier")

})

context("Collapse Test")

test_that("arguments (collapsedim and preservedim) work in preserving and collapsing as specified dims", {
  x <- maxample("animal")
  expect_identical(collapseNames(x, "type"), collapseDim(x, "type"))
  expect_identical(x, collapseNames(x, preservedim = 1))
  expect_identical(x, collapseNames(x, preservedim = "type"))
  expect_identical(collapseNames(x), collapseNames(x, collapsedim = 1))
  expect_identical(collapseNames(x, 3), collapseNames(x, 3.3))
  x <- x[, , 1]
  getNames(x) <- NULL
  expect_identical(x, collapseNames(x))
  expect_null(collapseNames(NULL))
})

p <- maxample("pop")
attr(p, "Metadata") <- NULL
getSets(p)[1:2] <- c("d1", "d2")

test_that("dimReduce works", {
  p2 <- addDim(p, item = paste0("scen", 1:5))
  expect_identical(dimReduce(p2), p)

  p2[, , ] <- setYears(p2[, 2, ], NULL)

  expect_identical(dimReduce(p2), setYears(p[, 2, ], NULL))

  p2[, , ] <- setCells(p2[1, , ], "GLO")
  expect_identical(dimReduce(p2), setYears(setCells(p[1, 2, ], "GLO"), NULL))


})

test_that("commonYears works", {
  pop <- maxample("pop")
  expect_identical(commonYears(pop), getYears(pop))
  expect_identical(commonYears(pop, pop), getYears(pop))
  expect_identical(commonYears(pop, pop, pop), getYears(pop))
  expect_identical(commonYears(pop[, 1:3, ], pop[, 2:4, ], pop[, 3:5, ]), getYears(pop)[3])
  expect_identical(commonYears(pop[, 1:3, ], pop[, 2:4, ], pop[, 3, ]), getYears(pop)[2:3])
  expect_identical(commonYears(pop[, 1:3, ], pop[, 2:4, ], pop[, 3, ], asInteger = TRUE),
                   getYears(pop, as.integer = TRUE)[2:3])
  expect_identical(commonYears(list(pop[, 1:3, ], pop[, 2:4, ], pop[, 3, ])), getYears(pop)[2:3])
  expect_identical(commonYears(pop[, 1:3, ], pop[, 4, ], pop[, 3, ]), getYears(pop)[1:3])
  expect_identical(commonYears(pop[, 1, ], pop[, 4, ], pop[, 3, ]), NULL)
})

context("Conversion Test")

mag <- maxample("pop")

test_that("conversions do not affect content", {
  expect_identical(as.magpie(mag), mag)
  expect_equivalent(as.magpie(as.array(mag)), mag)
  expect_equivalent(as.magpie(as.data.frame(mag)), mag)
  expect_null(as.magpie(NULL))

  empty <- new("magpie", .Data = structure(logical(0), .Dim = c(0L, 1L, 1L),
                                           .Dimnames = list(region = NULL, year = NULL,
                                                            data = NULL)))
  expect_identical(as.magpie(data.frame()), empty)
  df <- data.frame(x = 1:3, y = 4:6, row.names = c("a", "b", "c"))
  expect_identical(as.magpie(df), as.magpie(as.matrix(df)))

  ref <- new("magpie", .Data = structure(TRUE, .Dim = c(1L, 1L, 1L),
                                         .Dimnames = structure(list("GLO", NULL, "1.a"),
                                                               .Names = c("fake", NA, NA))))
  expect_identical(as.magpie(data.frame(a = TRUE)), ref)
  skip_if_not_installed("quitte")
  expect_equivalent(as.magpie(quitte::as.quitte(mag)), mag)
})

test_that("special cases work", {
  a <- as.array(mag)
  m2 <- as.magpie(a, spatial = 0, temporal = 0)
  expect_equal(dim(m2), c(1, 1, 320))

  names(dimnames(a)) <- NULL
  attr(a, "sets") <- c("spatial", "temporal", "data")
  m2 <- as.magpie(a)
  expect_equal(unname(getSets(m2)), attr(a, "sets"))

  bla <- data.frame(from = c("bla", "blub"), to = c("ble", "blo"), stringsAsFactors = FALSE)
  bla2 <- data.frame(from = c("bla", "blub"), to = c("ble", "blo"), stringsAsFactors = TRUE)
  blaExpect <- new("magpie",
                   .Data = structure(c("ble", "blo"),
                                     .Dim = c(1L, 1L, 2L),
                                     .Dimnames = list(region = "GLO", year = NULL,
                                                      from = c("bla", "blub"))))
  expect_identical(as.magpie(bla), blaExpect)
  expect_identical(as.magpie(bla2), blaExpect)
})

test_that("underscores are preserved", {
  mag2 <- maxample("pop")
  getCells(mag2) <- paste(getCells(mag2), 1:ncells(mag2), sep = "_")
  arr <- as.array(mag2)
  mag3 <- as.magpie(arr, spatial = 1)
  expect_identical(getCells(mag3), getCells(mag2))
})

test_that("error detection works", {
  expect_warning(m <- as.magpie(array(1, c(2, 2), list(c("ABC", "DEF"), c("GHI", "JKL")))), "No clear mapping")

  ref <- new("magpie", .Data = structure(c(1, 1, 1, 1), .Dim = c(2L, 1L, 2L),
                                         .Dimnames = structure(list(c("ABC", "DEF"), NULL, c("GHI", "JKL")),
                                                               .Names = c("fake", NA, NA))))
  expect_identical(m, ref)

  skip_if_not_installed("quitte")
  qi <- quitte::as.quitte(mag)
  qi$unit <- NULL
  expect_warning(as.magpie(qi), "does not follow the full quitte class definition")

})

context("CollapseDim Test")

test_that("arguments (dim and keepdim) work", {
  x <- maxample("animal")
  expect_identical(x, collapseDim(x, keepdim = c(2.3, 3.1)))
  expect_identical(collapseDim(x), collapseDim(x, dim = c(3.1, 2.3)))
  expect_identical(getItems(collapseDim(x[, , "rabbit"]), dim = 3), c("black", "white"))
  expect_identical(collapseDim(x, keepdim = 1:2), collapseDim(x, dim = 3.1))
  expect_identical(collapseDim(x, keepdim = c(1, 3)), collapseDim(x, dim = 2.3))
  expect_identical(getItems(collapseDim(x, dim = 3.2), dim = 3, split = TRUE),
                   getItems(x, dim = c(3.1, 3.3), split = TRUE))
  expect_identical(getItems(collapseDim(x, dim = c("type", "species")), dim = 3), getItems(x, dim = 3.3, full = TRUE))

  expect_warning(collapseDim(x, dim = 3.2, keepdim = 1:2), "keepdim argument is ignored")
  expect_warning(collapseDim(x, dim = "notthere"), "could not be found")

  x2 <- x
  names(dimnames(x2))[3] <- "species.species.color"
  expect_identical(collapseDim(x2), collapseDim(x))

  x <- x[, , 1]
  getItems(x, dim = 3) <- NULL
  expect_identical(x, collapseDim(x, keepdim = 1:2))
  expect_null(collapseDim(NULL))

  p <- maxample("pop")
  names(dimnames(p)) <- NULL
  expect_identical(getSets(collapseDim(p)), c(d1.1 = "region", d2.1 = "year", d3.1 = "data"))
})

test_that("collapseDim works for misconfigured objects", {
  a <- maxample("animal")
  b <- a
  getSets(b, fulldim = FALSE)[3] <- "x.y"
  expect_silent(a <- collapseDim(b))

  p <- maxample("pop")
  p2 <- p
  dimnames(p2)[[3]] <- paste0(dimnames(p2)[[3]], ".blub..")
  expect_identical(collapseDim(p2), p)
})

test_that("Turn magpie into tibble", {
  tib <- tibble::as_tibble(new.magpie("AFR.1", "2000", "test", 123))
  expect_s3_class(tib, "tbl_df")

  tib <- tibble::as_tibble(maxample("pop"))
  expect_s3_class(tib, "tbl_df")
})

a  <- maxample("animal")
p  <- maxample("pop")
attr(p, "Metadata") <- NULL # nolint: object_name_linter

test_that("mbind works", {
  expect_identical(mbind(p, p), p[, , c(1:2, 1:2)])
  expect_identical(mbind(p[, 1:3, ], p[, 4:16, ]), p)
  expect_identical(mbind(p[1:4, , ], p[5:10, , ]), p)
  expect_identical(mbind(p, NULL, p), mbind(list(p, p)))
  expect_identical(mbind(p, NULL, p), mbind(p, p))
  expect_null(mbind(NULL, NULL))

  expect_error(mbind(p, 12), "must all be MAgPIE-objects")
  expect_error(mbind(p[2:3, , ], p[c(1, 1), , ]), "occur more than once")
  expect_error(mbind(p[, 2:3, ], p[, c(1, 1), ]), "occur more than once")

  expect_error(mbind(a, p), "Cannot handle")
  expect_error(mbind(p[1:3, 2:4, ], p), "Cannot handle")
  expect_error(mbind(p[, 1:3, 1], p), "Cannot handle")
  expect_error(mbind(p[1:3, , 1], p), "Cannot handle")
  expect_error(suppressWarnings(mbind(p[, , c(1, 1)])), "Invalid object")

  for (i in 1:3) {
    p0 <- p[1, dim = i]
    getItems(p0, dim = i) <- NULL
    expect_null(getItems(mbind(p0, p0), dim = i))
  }
})

test_that("cbind issues a warning but works otherwise", {
  expect_warning(cbind(new.magpie(), new.magpie()))
  expect_warning(cbind(new.magpie(), c()))
  expect_warning(cbind(c(), new.magpie()))

  cbindResult <- suppressWarnings(cbind(c(1, 2), p[1, , ]))
  expect_is(cbindResult, "matrix")
  expect_equal(cbindResult[2:3, 1],
               c(2, 1))
  expect_equal(cbindResult[, 2],
               as.vector(p[1, , ]))
})

context("Subsetting test")

p <- maxample("pop")
a <- as.array(p)

mv <- getOption("magclass.verbosity")
on.exit(options(magclass.verbosity = mv))
options(magclass.verbosity = 2)

test_that("single element subsetting works", {
  expect_identical(p[11], a[11])
  expect_identical(p[3], a[3])
})

test_that("multi element subsetting works", {
  expect_equivalent(as.array(p[3, , ]),                a[3, , , drop = FALSE])
  expect_equivalent(as.array(p["FSU", , ]),            a[4, , , drop = FALSE])
  expect_equivalent(as.array(p[, 2005, ]),             a[, 2, , drop = FALSE])
  expect_equivalent(as.array(p[as.factor("PAS"), , "B1"]),     a[9, , 2, drop = FALSE])
  expect_equivalent(as.array(p["PAS", "y2005", "B1"]),         a[9, 2, 2, drop = FALSE])
  expect_equivalent(as.array(p[c("CPA", "CPA"), , ]),           a[c(2, 2), , , drop = FALSE])
  expect_equivalent(as.array(p[list(c("CPA", "CPA")), , ]),     a[c(2, 2), , , drop = FALSE])
  expect_equivalent(as.array(p[list(i = c("CPA", "CPA")), , ]),   a[c(2, 2), , , drop = FALSE])
  expect_equivalent(as.array(p[c("EUR", "CPA"), , ]),           a[c(3, 2), , , drop = FALSE])
  expect_equivalent(as.array(p[list(c("EUR", "CPA")), , ]),     a[c(3, 2), , , drop = FALSE])
  expect_equivalent(as.array(p[list(i = c("EUR", "CPA")), , ]),   a[c(3, 2), , , drop = FALSE])
  expect_equivalent(as.array(p[character(0), character(0), character(0)]), a[NULL, NULL, NULL, drop = FALSE])
  expect_identical(p[, NULL, ], p)
})

test_that("subsetting via dim argument works", {
  expect_identical(p[, 1:3, ], p[1:3, dim = 2])
  expect_identical(p[, , "B1"], p["B1", dim = 3])
  expect_identical(p[c("FSU", "EUR"), , ], p[c("FSU", "EUR"), dim = 1])
  expect_error(p[1, 2, dim = 2], "Only single dimension selection allowed")
  expect_error(p[1, 2, 3, dim = 2], "Only single dimension selection allowed")
  expect_error(p[1, dim = 1.2], "Invalid dim selection")
  p3 <- p2 <- p
  expect_silent(p2[, , "A2"] <- 99)
  expect_silent(p3["A2", dim = 3] <- 99)
  expect_identical(p2, p3)
  expect_silent(p2[c("FSU", "EUR"), , ] <- 42)
  expect_silent(p3[c("FSU", "EUR"), dim = 1] <- 42)
  expect_identical(p2, p3)
  expect_silent(p2[, 2015, ] <- -99)
  expect_silent(p3[2015, dim = 2] <- -99)
  expect_identical(p2, p3)
  expect_error(p3[1, 2, dim = 2] <- 1, "Only single dimension selection allowed")
  expect_error(p3[1, 2, 3, dim = 2] <- 1, "Only single dimension selection allowed")
  expect_error(p[1, dim = 1.2] <- 1, "Invalid dim selection")

})

test_that("boolean subsetting works", {
  expect_identical(p[p > 1000], p[p[10:1, , ] > 1000])
  p2 <- p[, 1, 1]
  expect_identical(p[p2 > 1000, , ], p[as.vector(p2 > 1000), , ])
  expect_identical(p[p2[10:1, , ] > 1000, , ], p[as.vector(p2 > 1000), , ])
  p2 <- p[1, , 1]
  expect_identical(p[, p2 > 1000, ], p[, as.vector(p2 > 1000), ])
  expect_identical(p[, p2[, 16:1, ] > 1000, ], p[, as.vector(p2 > 1000), ])
  expect_identical(p[, p2[, c(16:10, 1:9), ] > 1000, ], p[, as.vector(p2 > 1000), ])
})

test_that("error detection works", {
  expect_error(p[, , , ], "argument")
  expect_error(p[, , , blub = 42], "unknown argument\\(s\\) supplied")
  expect_error(p[, , , blub = 42] <- 42, "unknown argument\\(s\\) supplied")
  expect_error(p[, , "A3"], "out of bounds")
  expect_error(p[, , list("A3")], "out of bounds")
  expect_error(p[, , list(scenario = "A3")], "out of bounds")
  expect_error(p[, , list(blub = "A2")], "subdimension does not exist")

  names(dimnames(p)) <- NULL
  expect_error(p[, , list(scenario = "A2")], "subdimension does not exist \\(missing set names\\)")

  dimnames(p)[[3]] <- NULL
  expect_error(p[, , "A2"], "Missing element names")
})


test_that("invert argument works", {
  expect_identical(p[-1, , ],    p["AFR", invert = TRUE])
  expect_identical(p[-1, , ],    p["AFR", , invert = TRUE])
  expect_identical(p[-1, , ],    p["AFR", , , invert = TRUE])
  expect_identical(p[-9, -4, ],  p["PAS", 2025, invert = TRUE])
  expect_identical(p[-9, -4, ],  p["PAS", 2025, , invert = TRUE])
  expect_identical(p[, -4, ],    p[, 2025, , invert = TRUE])
  expect_identical(p[-1:-3, , ], p[1:3, , , invert = TRUE])
})

test_that("drop works", {
  a <- maxample("animal")
  expect_identical(getItems(a[, , , drop = TRUE], dim = 3)[1], "rabbit.black")
})

test_that("pmatch argument works", {
  expect_identical(getItems(p[, list("y1"), , pmatch = TRUE], 2), "y1995")
  expect_identical(getItems(p[, list(as.factor("y1")), , pmatch = TRUE], 2), "y1995")
  expect_identical(getItems(p[, "y1", , pmatch = TRUE], 2), "y1995")
  expect_error(getItems(p[, "y1", , pmatch = "right"], 2), "out of bounds")
  expect_identical(getItems(p[, "y1", , pmatch = "left"], 2), "y1995")
  expect_error(getItems(p[, "05", , pmatch = "left"], 2), "out of bounds")
  expect_identical(getItems(p[, "05", , pmatch = "right"], 2), c("y2005", "y2105"))
})

test_that("multiple subdimensions work", {
  getItems(p, "j", maindim = 1) <- 1:10
  expect_identical(p["AFR", , ], p[1, , ])
  expect_identical(p[list(i = "CPA"), , ], p[2, , ])
  expect_silent(p[list(i = "AFR"), , ] <- 99)
  expect_equal(as.vector(p["AFR", 1, 1]), 99)
  expect_silent(p[list(i = "AFR"), , list(scenario = "A2")] <- 100)
  expect_equal(as.vector(p["AFR", 1, "A2"]), 100)
  expect_silent(p[list("AFR"), 2145, list("A2")] <- 101)
  expect_equal(as.vector(p["AFR", 16, "A2"]), 101)

  # testing that years are properly handled
  t <- c(1995, 2005)
  yt <- paste0("y", t)
  expect_identical(p[, list(yt), ], p[, t, ])
  expect_identical(p[, list(t), ], p[, yt, ])
  expect_identical(getYears(p[, list(t), , invert = TRUE]), setdiff(getYears(p), yt))
})

test_that("value assignment works", {
  a <- maxample("animal")
  expect_silent(a[, NULL, as.factor("rabbit")] <- as.magpie(99))
  expect_true(all(a[, , "rabbit"] == 99))
  expect_silent(a[as.factor("NLD"), as.factor(c("april", "june")), as.factor("rabbit")] <- 12)
  expect_true(all(a["NLD", "june", "rabbit"] == 12))
  b <- a
  expect_silent(b[, , ] <- 0)
  expect_true(all(b[, , ] == 0))
  expect_silent(b[, , ] <- as.magpie(99))
  expect_true(all(b[, , ] == 99))
  expect_message(b[1:2, 1, 1] <- 1:2, "Dangerous replacement")
  expect_error(b[1:2, 1:2, 1:2] <- 1:7, "Different replacement length!")

  expect_silent(b["NLD", c("april", "june"), list("rabbit", "black")] <- a["NLD", c("april", "june"),
                                                                           list("rabbit", "black")])
  expect_identical(b["NLD", c("april", "june"), list("rabbit", "black")], a["NLD", c("april", "june"),
                                                                            list("rabbit", "black")])
})


test_that("data.frame subsetting works", {
  a <- maxample("animal")
  df <- data.frame(getItems(a, 3, split = TRUE, full = TRUE), stringsAsFactors = FALSE)
  w <- c(1, 3, 4)
  expect_identical(getItems(a[df[w, ]], 3), getItems(a, 3)[w])
  expect_identical(getItems(a[df[3:1][w, ]], 3), getItems(a, 3)[w])
  expect_identical(getItems(a[df[3:2][w, ]], 3), getItems(a, 3)[w])

  # Unknown dimensions to be added in output!
  df$blub <- paste0("bl", seq_len(dim(df)[1]))
  expect_identical(getItems(a[df[w, ]], 3), paste(getItems(a, 3), df$blub, sep = ".")[w])

  df2 <- df
  df2$ble <- paste0("ble", seq_len(dim(df2)[1]))
  expect_identical(getItems(a[df2[w, ]], 3), paste(getItems(a, 3), df2$blub, df2$ble, sep = ".")[w])

  # subselections work
  df$species <- NULL
  expect_identical(getItems(a[df[1, ]], 3), c("animal.rabbit.black.bl1", "animal.bird.black.bl1"))
  expect_identical(getItems(a[df[w, ]], 3), c("animal.rabbit.black.bl1", "animal.bird.black.bl1",
                                              "animal.rabbit.black.bl3", "animal.bird.black.bl3",
                                              "animal.bird.red.bl4"))

  df2 <- df
  df2$type <- NULL
  expect_identical(getItems(a[df2[1, ]], 3), c("animal.rabbit.black.bl1", "animal.bird.black.bl1"))


  # rows in df but not in a will get added with value NA
  df[3, 1] <- "car"
  expect_message(b <- a[df[w, ]], "elements were added")
  expect_identical(getItems(b, 3), c("animal.rabbit.black.bl1", "animal.bird.black.bl1",
                                     "animal.bird.red.bl4", "car.NA.black.bl3"))
  expect_true(all(is.na(b[, , "car.NA.black.bl3"])))

  df[4, 1] <- "house"
  expect_message(b <- a[df[w, ]], "elements were added")
  expect_identical(getItems(b, 3), c("animal.rabbit.black.bl1", "animal.bird.black.bl1",
                                     "car.NA.black.bl3", "house.NA.red.bl4"))


  df1 <- data.frame(getItems(a, 1, split = TRUE, full = TRUE))
  expect_identical(getItems(a[df1[w, ]], 1), getItems(a, 1)[w])
  df2 <- data.frame(getItems(a, 2, split = TRUE, full = TRUE))
  expect_identical(getItems(a[df2[w, ][c(3, 1, 2)]], 2), getItems(a, 2)[w])

  names(df2)[2] <- names(df2)[1]
  expect_error(a[df2], "more than once")

  names(df2)[2] <- "country"
  expect_error(a[df2], "must only contain subdimensions with a shared main dimension")


  names(df2) <- paste0("bla", seq_along(df2))
  expect_error(a[df2], "None of the dimensions in the mapping could be found")
  names(dimnames(a)) <- NULL
  expect_error(a[df], "must have names")

  # check that it works for single subdimension
  p <- maxample("pop")
  df <- data.frame(getItems(p, 3, split = TRUE, full = TRUE), stringsAsFactors = FALSE)
  df$blub <- paste0("bla", seq_len(nrow(df)))
  expect_identical(getItems(p[df], 3), paste0(getItems(p, 3), ".", df$blub))

  # check that it works if not element can be found in object
  df$scenario <- c("C1", "D2")
  expect_identical(getItems(p[df], 3), c("C1.bla1", "D2.bla2"))
  expect_true(all(is.na(p[df])))

  # check that empty magclass objects can get expanded
  p0 <- p[, , NULL]
  expect_true(all(is.na(p0[df])))
  expect_identical(getItems(p0[df], 3), c("C1.bla1", "D2.bla2"))
})

test_that("duplicates detection works", {
  a <- maxample("animal")
  expect_warning(a[, c(1, 1, 2, 3), ][, "y2000.april.20", ], "contain duplicates")
})

a  <- maxample("animal")

test_that("subdimension permutation works", {
  expect_identical(dimOrder(a, perm = 3:1, dim = 1), dimOrder(a, perm = c(3:1, 4), dim = 1))
  expect_identical(dimOrder(a, perm = 1:4, dim = 1), a)
  expect_identical(dimOrder(a, perm = 1, dim = 1), a)
  expect_identical(dimOrder(a, perm = 1, dim = 3), a)
  expect_identical(getItems(dimOrder(a, perm = c(2, 1, 3), dim = 3), dim = 3)[1], "rabbit.animal.black")
})

test_that("edge cases in dimOrder work", {
  expect_error(dimOrder(a, perm = c(3, 100)), "Values of perm must be integers in the range from 1 to 3")
  expect_identical(getSets(dimOrder(a[NULL, , ], perm = 3:1, dim = 1), fulldim = FALSE)[1], "country.y.x.cell")
})

a  <- maxample("animal")
p  <- maxample("pop")

test_that("subdimension counting works", {
  expect_identical(ndim(a, 1), 4)
  expect_identical(ndim(a, 2), 3)
  expect_identical(ndim(a, 3), 3)
  expect_identical(ndim(a), 10)
  expect_identical(ndim(p, 1), 1)
  expect_identical(ndim(p), 3)

  expect_identical(ndim(a[NULL, , ], 1), 4)
  names(dimnames(a)) <- NULL
  expect_identical(ndim(a[NULL, , ], 1), 0)

  expect_error(ndim(a, "bla"), "Invalid dim selection")
  expect_error(ndim(a, 1.5), "Invalid dim selection")
  expect_error(ndim(a, 0), "Invalid dim selection")
  expect_error(ndim(a, 10), "Invalid dim selection")
})

a <- maxample("animal")
attr(a, "Metadata") <- NULL

test_that("addDim works", {
  aMini <- a[1, 1:3, 1:2]
  attr(aMini, ".internal.selfref") <- NULL
  ref1 <- new("magpie",
              .Data = structure(c(0, 0, 0, 6, 7, 9),
                                .Dim = c(1L, 3L, 2L),
                                .Dimnames = list(x.y.country.cell = "5p75.53p25.NLD.14084",
                                                 year.month.day = c("y2000.april.20", "y2000.may.20", "y2000.june.20"),
                                                 new.type.species.color = c("dummy.animal.rabbit.black",
                                                                            "dummy.animal.rabbit.white"))))
  expect_identical(addDim(aMini), ref1)
  ref2 <- new("magpie",
              .Data = structure(c(0, 0, 0, 6, 7, 9),
                                .Dim = c(1L, 3L, 2L),
                                .Dimnames = list(x.y.country.cell = "5p75.53p25.NLD.14084",
                                                 year.new.month.day = c("y2000.dummy.april.20", "y2000.dummy.may.20",
                                                                        "y2000.dummy.june.20"),
                                                 type.species.color = c("animal.rabbit.black",
                                                                        "animal.rabbit.white"))))
  expect_identical(addDim(aMini, dim = 2.2), ref2)
  ref3 <- new("magpie",
              .Data = structure(c(0, 0, 0, 0, 0, 0, 6, 7, 9, 6, 7, 9),
                                .Dim = c(1L, 6L, 2L),
                                .Dimnames = list(x.y.country.cell = "5p75.53p25.NLD.14084",
                                                 year.new.month.day = c("y2000.d1.april.20", "y2000.d1.may.20",
                                                                        "y2000.d1.june.20", "y2000.d2.april.20",
                                                                        "y2000.d2.may.20",  "y2000.d2.june.20"),
                                                 type.species.color = c("animal.rabbit.black",
                                                                        "animal.rabbit.white"))))
  expect_identical(addDim(aMini, dim = 2.2, item = paste0("d", 1:2)), ref3)

  a0 <- dimSums(a[, 1, 1], dim = 1)
  ref4 <- new("magpie", .Data = structure(190, .Dim = c(1L, 1L, 1L),
                                          .Dimnames = list(new = "dummy", year.month.day = "y2000.april.20",
                                                           type.species.color = "animal.rabbit.black")))
  expect_identical(addDim(a0, dim = 1), ref4)
  expect_identical(addDim(a0, dim = 1.2), ref4)

  p <- maxample("pop")
  expect_error(addDim(p, dim = 3.2, dimName = "scenario"), "Dimension .* does already exist")
  expect_silent(p <- addDim(p, 3.1))
  expect_silent(p <- addDim(p, 3.2))
  expect_identical(getSets(p, fulldim = FALSE)[3], "new.new1.scenario")
  expect_silent(p <- addDim(p, 3.4))
  expect_identical(getSets(p, fulldim = FALSE)[3], "new.new1.scenario.new2")

  p <- addDim(dimSums(p, 3), item = c("a.b", "c.d"))
  expect_identical(getItems(p, 3), c("a.b", "c.d"))
})

test_that("addDim's expand argument works", {
  p <- maxample("pop")
  p <- addDim(dimSums(p, 3), item = c("a.b", "c.d"))
  p2 <- addDim(p, item = c("e", "f"), expand = FALSE)
  expect_identical(getItems(p2, 3), c("e.a.b", "f.c.d"))
  p3 <- addDim(p, item = c("e", "f"), expand = TRUE)
  expect_identical(getItems(p3, 3), c("e.a.b", "e.c.d", "f.a.b", "f.c.d"))
})

test_that("addDim works for objects with inconsistent set information", {
  a <- maxample("animal")
  expect_silent(b <- addDim(a))
  expect_identical(getItems(b, 3.1), "dummy")
})

test_that("getCPR works", {
  ref <- structure(c(AFR = 8224L, CPA = 4795L, EUR = 3448L, FSU = 14048L, LAM = 7226L, MEA = 3957L,
                     NAM = 10985L, PAO = 3124L, PAS = 1326L, SAS = 2066L), .Dim = 10L,
                   .Dimnames = list(c("AFR", "CPA", "EUR", "FSU", "LAM", "MEA", "NAM", "PAO", "PAS", "SAS")),
                   class = "table")
  expect_identical(getCPR(0.5), ref)
  a <- maxample("animal")
  ref2 <- structure(c(black = 2L, brown = 1L, red = 1L, white = 1L), .Dim = 4L,
                    .Dimnames = list(c("black", "brown", "red", "white")), class = "table")
  expect_identical(getCPR(a, dim = "color"), ref2)
  expect_identical(getCPR(a, dim = 3.3), ref2)

  ref3 <- structure(c(1L, 0L, 1L, 0L, 0L, 1L, 0L, 0L, 1L, 0L, 0L, 1L), .Dim = 4:3,
                    .Dimnames = list(c("black", "brown", "red", "white"),
                                     c("bird", "dog", "rabbit")), class = "table")
  expect_identical(getCPR(a, dim = c("color", "species")), ref3)
  expect_error(getCPR(a, dim = c("country", "color")), "Selected subdimensions must belong to the same main dimension")
  expect_error(getCPR(42), "Cannot extract cpr information")
})

test_that("maxample runs without error for all combinations of global and total", {
  # Make sure no output is rendered
  # Open a null graphics device (e.g., PDF without a file)
  withr::local_pdf(NULL)

  for (example in c("pop", "animal", "bilateral")) {
    for (global in c(TRUE, FALSE)) {
      for (total in c(TRUE, FALSE)) {
        capture_output(expect_no_error(mplot(maxample(!!example), global = !!global, total = !!total)))
      }
    }
  }
})

#' spatRasterToDataset
#'
#' Convert a SpatRaster to a SpatRasterDataset.
#'
#' @param x SpatRaster with names of the form "y[0-9]+..[varname]"
#' @return SpatRasterDataset
#' @author Pascal Sauer
#' @family SpatialOperations
#' @export
spatRasterToDataset <- function(x) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("The package \"terra\" is required!")
  }

  if (all(grepl("^y[0-9]+\\.\\.", names(x)))) {
    varnames <- unique(sub("^y[0-9]+\\.\\.", "", names(x)))
    datasets <- lapply(varnames, function(varname) {
      spatRaster <- x[paste0("\\.\\.", varname, "$")]
      terra::varnames(spatRaster) <- varname
      return(spatRaster)
    })
  } else {
    varnames <- names(x)
    datasets <- lapply(varnames, function(varname) {
      spatRaster <- x[varname]
      terra::varnames(spatRaster) <- varname
      return(spatRaster)
    })
  }
  x <- terra::sds(datasets)
  names(x) <- varnames
  return(x)
}

#' mplot
#'
#' Get a quick visualization of the content of most magpie objects.
#'
#' @param px The magpie object to be visualized
#' @param global Whether data should be aggregated over regions to global values.
#' @param total Whether the total of all data values should also be visualized.
#' @author Pascal Sauer, Patrick Rein
#' @family Display
#' @importFrom rlang .data
#' @export
mplot <- function(px, global = TRUE, total = FALSE) {

  rlang::check_installed("ggplot2")

  originalDimNames <- names(dimnames(px))
  firstDimensionPart <- function(aName) strsplit(aName, ".", fixed = TRUE)[[1]][[1]]
  temporalCategorial <- FALSE

  # Ensure all dimensions have items
  # or are somewhat normalized
  if (is.null(getItems(px, 1))) {
    getItems(px, 1) <- "global"
  }
  if (is.null(getItems(px, 2))) {
    getItems(px, 2) <- "timeless"
    temporalCategorial <- TRUE
  }
  if (is.null(getItems(px, 3))) {
    getItems(px, 3) <- "total"
  }

  # Rewrite names of data dimension to make them symbols
  # Simplify and keep the data dimension name
  getNames(px) <- gsub(".", "_", getNames(px), fixed = TRUE)
  dataDimName <- firstDimensionPart(originalDimNames[[3]])
  names(dimnames(px))[[3]] <- dataDimName

  # Retrieve the temporal dimension name and potentially
  # make temporal items symbols
  temporalDimName <- originalDimNames[[2]]
  if (grepl(".", temporalDimName, fixed = TRUE)) {
    # We have subdimensions, convert to symbols
    dimnames(px)[[2]] <- gsub(".", "_", dimnames(px)[[temporalDimName]], fixed = TRUE)
    temporalDimName <- "temporal"
    names(dimnames(px))[[2]] <- temporalDimName
    temporalCategorial <- TRUE
  }

  spatialDimName <- originalDimNames[[1]]
  if (global) {
    # Transform to global values
    px <- dimSums(px, 1)
    spatialDimName <- names(dimnames(px))[[1]]
  } else {
    # Rewrite names of spatial dimension to symbols
    if (grepl(".", spatialDimName, fixed = TRUE)) {
      # We have subdimensions, convert to symbols
      dimnames(px)[[1]] <- gsub(".", "_", dimnames(px)[[spatialDimName]], fixed = TRUE)
      spatialDimName <- "spatial"
      names(dimnames(px))[[1]] <- spatialDimName
    }
  }

  # Calculate and add total
  # If temporal is categorial, we will render stacked bar charts and
  # do not need the total
  if (total && !temporalCategorial) {
    if (ndata(px) > 1) {
      px <- mbind(px, magclass::setNames(dimSums(px, 3), "total"))
    } else {
      getNames(px) <- "total"
    }
  }

  # Convert to data.frame
  px <- as.data.frame(px, rev = 3)

  # Plot
  plot <- ggplot2::ggplot(data = px, ggplot2::aes(x = .data[[temporalDimName]],
                                                  color = .data[[dataDimName]],
                                                  group = .data[[dataDimName]]))

  if (!temporalCategorial) {
    plot <- plot + ggplot2::geom_line(linewidth = 1.5, ggplot2::aes(y = .data$.value))
  } else {
    plot <- plot + ggplot2::geom_bar(linewidth = 1.5, ggplot2::aes(weight = .data$.value))
  }

  if (!global) {
    plot <- plot + ggplot2::facet_wrap(~.data[[spatialDimName]])
  }

  print(plot)
}

#' as.RasterBrick
#'
#' Convert magclass object to a RasterBrick object
#'
#'
#' @param x MAgPIE object
#' @param res spatial data resolution. If not provided it will be guessed.
#' @return A RasterBrick object
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getCoords}}
#' @examples
#' \dontrun{
#' if (requireNamespace("raster", quietly = TRUE)) {
#'    r <- raster::brick(ncols = 36, nrows = 18, nl = 4)
#'    r[14:18, 25:28] <- (1:20 %*% t(1:4))
#'    names(r) <- c("y2000..bla", "y2001..bla", "y2000..blub", "y2001..blub")
#'    m <- as.magpie(r)
#'    r2 <- as.RasterBrick(m)
#' }
#' }
#' @family MAgPIE-Conversions
#' @export
as.RasterBrick <- function(x, res = NULL) { # nolint: object_name_linter.
  if (!is.magpie(x)) stop("Input is not a magpie object")
  if (!requireNamespace("raster", quietly = TRUE)) stop("The package \"raster\" is required!")

  if (!hasCoords(x) && dimExists(1.2, x)) {
    items <- getItems(x, dim = 1.2)
    if (length(items) == 59199 && all(items == seq_len(59199))) {
      # special treatment for data with 59199 cells as this
      # is the originally used 0.5deg resolution in magclass
      # before it had been generalized
      getCoords(x) <- magclassdata$half_deg[c("lon", "lat")]
    }
  }
  xy <- getCoords(x)
  if (is.null(res)) res <- guessResolution(xy)
  out <- raster::brick(ncols = 360 / res, nrows = 180 / res, nl = nyears(x) * ndata(x))
  m   <- wrap(as.array(x), list(1, 2:3), sep = "..")
  layerNames <- colnames(m)
  colnames(m) <- NULL
  raster::values(out)[raster::cellFromXY(out, xy), ] <- m
  names(out) <- layerNames
  return(out)
}

skip_if_not_installed("terra")

test_that("terra conversion does not alter data", {
  expect_error(as.SpatRaster(1), "not a magpie object")
  for (i in c(0.5, 2)) {
    for (j in c(1, 4)) {
      for (t in c(1, 3)) {
        r <- terra::rast(ncols = 360 / i, nrows = 180 / i, nlyrs = j * t)
        if (t > 1) {
          years <- paste0("y", 1900 + 1:t)
          data  <- paste0("bla", 1:j)
          names(r) <- paste0(rep(years, each = j), "..", rep(data, t))
        } else {
          names(r) <- paste0("bla", 1:j)
        }
        r[85:89, 176:179] <- (1:20 %*% t(1:(j * t)))
        m <- as.magpie(r)
        expect_equal(ndata(m), j)
        expect_equal(nyears(m), t)
        r2 <- as.SpatRaster(m)

        srd <- as.SpatRasterDataset(m)
        expect_identical(names(srd), getItems(m, 3))

        noTime <- function(a) {
          terra::time(a) <- NULL
          return(a)
        }

        # comparing times does not work here, so compare separately
        expect_equal(noTime(r), noTime(terra::extend(r2[[names(r)]], terra::ext(r))))
        if (t > 1) {
          expect_equal(terra::time(r2), rep_len(1900 + 1:t, terra::nlyr(r)))
        }
        m2 <- as.magpie(r2)
        expect_identical(m, m2)
        v <- terra::as.polygons(r)
        m3 <- as.magpie(v)
        v2 <- as.SpatVector(m3)
        m4 <- as.magpie(v2)
        expect_equal(m3, m4)
        v3 <- as.SpatVector(m4)
        expect_equal(v2, v3)
      }
    }
  }
})

skip_if_not_installed("raster")

test_that("raster conversion does not alter data", {
  expect_error(as.RasterBrick(1), "not a magpie object")
  for (i in c(0.5, 2)) {
    for (j in c(1, 4)) {
      for (t in c(1, 3)) {
        r <- raster::brick(ncols = 360 / i, nrows = 180 / i, nl = j * t)
        if (t > 1) {
          years <- paste0("y", 1900 + 1:t)
          data  <- paste0("bla", 1:j)
          names(r) <- paste0(rep(years, each = j), "..", rep(data, t))
        } else {
          names(r) <- paste0("bla", 1:j)
        }
        r[85:89, 176:179] <- (1:20 %*% t(1:(j * t)))
        m <- as.magpie(r)
        expect_equal(ndata(m), j)
        expect_equal(nyears(m), t)
        r2 <- as.RasterBrick(m)
        expect_identical(r[[names(r)]], r2[[names(r)]])
        m2 <- as.magpie(r2)
        expect_identical(m, m2)
      }
    }
  }
})

#' Unwrap
#'
#' Creates a higher dimensional array by separating all subdimensions in
#' the third dimension of a MAgPIE object and returning them as
#' separate dimension.
#'
#'
#' @param x A MAgPIE object
#' @param sep deprecated, please do not use anymore
#' @return An array with the full dimensionality of the original data
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{wrap}},\code{\link{fulldim}}
#' @examples
#'
#' a <- as.magpie(array(1:6, c(3, 2), list(c("bla", "blub", "ble"), c("up", "down"))))
#' unwrap(a)
#' @family MAgPIE-Conversions
#' @export
unwrap <- function(x, sep = NULL) {
  if (!is.null(sep)) warning("\"sep\" argument is not supported anymore.")
  if (!is.magpie(x)) stop("Input is not a MAgPIE object. unwrap works only for MAgPIE objects")
  dim3 <- unname(getItems(x, dim = 3, split = TRUE))
  dim <- c(unname(dimnames(x)[1:2]), dim3)
  ndim3 <- vapply(dim3, length, integer(1))
  if (length(ndim3) == 1 && ndim3 == 0) ndim3 <- dim(x)[3]
  ndim <- c(dim(x)[1:2], ndim3)
  if (anyDuplicated(as.data.table(getItems(x, dim = 3)))) stop("Malformed MAgPIE object. Duplicated names detected!")
  if (prod(ndim3) != dim(x)[3]) {
    stop("MAgPIE object is sparse in 3rd dimension but needs to be complete in order to be unwrapped!")
  }
  reorder <- dimnames(wrap(array(NA, ndim, dim), list(1, 2, NA), sep = "."))[[3]]
  if (!is.null(reorder)) x <- x[, , reorder]
  return(array(as.vector(x), ndim, dim))
}

#' Get Items
#'
#' Extract items of a given (sub-)dimension of a MAgPIE-object
#'
#' @aliases getItems<-
#' @param x MAgPIE object
#' @param dim Dimension for which the items should be returned. Either number or
#' name of dimension or a vector of these. See \code{\link{dimCode}} for more details.
#' @param split Boolean which determines whether a main dimension should be split in subdimensions.
#' Only applicable to main dimensions (1,2,3) and ignored for all other.
#' @param full if TRUE dimension names are returned as they are (including repetitions), if FALSE only
#' the dimension elements (unique list of entries) are returned.
#' @param maindim main dimension the data should be added to (does not need to be set if \code{dim} exists
#' in the data. Should be set if \code{dim} might not exist, or if \code{dim} might potentially exist
#' in a different main dimension than the one anticipated).
#' @param raw if set to FALSE inputs will be corrected (e.g. dots replaced by the letter "p") if necessary. If
#' TRUE data will be written as is (risking the creation of inconsistent objects).
#' @param value a vector with the length of the main dimension the dimnames should be replaced in / added to.
#' If set to NULL the corresponding dimension will be removed.
#' @return items of the requested dimension in the MAgPIE-object. If split=TRUE and applied to a
#' main dimension (1,2,3) a list of items for each sub-dimension.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{dimCode}}
#' @importFrom data.table tstrsplit
#' @examples
#' x <- maxample("pop")
#' getItems(x, "scenario")
#' getItems(x, 3.1)
#' getItems(x, "i") <- paste0("REG", seq_len(ncells(x)))
#' getItems(x, "i")
#' y <- x[, 1, ]
#' getItems(y, "t") <- NULL
#' @family GetterSetter
#' @export
getItems <- function(x, dim = NULL, split = FALSE, full = FALSE) { # nolint: cyclocomp_linter.
  if (is.null(dim)) dim <- 1:3
  dim <- dimCode(dim, x, missing = "stop")
  if (setequal(dim, 1:3) && !split) return(dimnames(x))
  if (length(dim) > 1) {
    out <- list()
    for (i in dim) out[[as.character(i)]] <- getItems(x, dim = i, split = split, full = full)
    if (all(dim == floor(dim))) {
      names(out) <- NULL
    } else if (hasSets(x)) {
      sets <- c(d1 = "", d2 = "", d3 = "", getSets(x))
      names(out) <- sets[paste0("d", dim)]
    }
    return(out)
  }

  if (dim == floor(dim) && !split) return(dimnames(x)[[dim]])
  if (dim == floor(dim) && split) {
    if (is.null(dimnames(x)[[dim]])) {
      out <- list(NULL)
      if (hasSets(x)) {
        names(out) <- getSets(x, fulldim = FALSE)[dim]
      }
      return(out)
    }
    tmp <- tstrsplit(dimnames(x)[[dim]], ".", fixed = TRUE, names = TRUE)
    if (!isTRUE(full)) tmp <- lapply(tmp, unique)
    if (hasSets(x)) {
      tmp2 <- strsplit(getSets(x, fulldim = FALSE)[dim], ".", fixed = TRUE)[[1]]
      if (length(tmp2) == length(tmp)) names(tmp) <- tmp2
    }
    return(tmp)
  }
  if (!dimExists(dim, x)) stop("Subdimension ", dimCode(dim, x), " does not exist!")
  tmp <- dimnames(x)[[as.integer(dim)]]
  if (is.null(tmp)) return(NULL)
  subdim <- as.integer(substring(dim, 3))
  reg <- paste0(rep("([^\\.]*)", subdim), collapse = "\\.")
  out <- sub(paste0("^", reg, ".*$"), paste0("\\", subdim), tmp)
  if (isTRUE(full)) return(out)
  return(unique(out))
}

#' @describeIn getItems set dimension names
#' @export
`getItems<-` <- function(x, dim, full = NULL, maindim = NULL, raw = FALSE, value) { # nolint: cyclocomp_linter.
  if (!is.null(full) && !isTRUE(full)) stop("Currently only full = NULL or TRUE supported!")
  if (length(dim) > 1) stop("dim with length > 1 is currently not supported when setting items.")
  dc <- dimCode(dim, x)
  if (dc == 0 && is.null(maindim)) {
    stop("Dimension does not exist in object and cannot be added as main dimension is not specified!")
  }
  if (!is.null(maindim)) {
    if (!(maindim %in% 1:3)) stop("Unsupported maindim (can only be 1, 2 or 3!)")
    if (dc == 0) {
      # make sure that number of set names aggres with number of sub-dimensions
      x <- clean_magpie(x, what = "sets", maindim = maindim)
      dc <- maindim + 0.99999
    } else {
      if (floor(dc) != maindim) stop("Specified dimension (dim) found in main dimension different to maindim!")
    }
  } else {
    maindim <- floor(dc)
  }

  if (!is.null(value)) {
    if (length(value) != dim(x)[maindim]) {
      elems <- try(getItems(x, dim = dim), silent = TRUE)
      if ("try-error" %in% class(elems) || length(elems) != length(value)) stop("Wrong number of items supplied!")
      # try to expand value input to full length of main dimension
      if (is.null(names(value))) {
        names(value) <- elems
      } else if (!setequal(names(value), elems)) {
        stop("Names of input vector do not match existing dimension names!")
      }
      elemsFull <- getItems(x, dim = dim, full = TRUE)
      value <- unname(value[elemsFull])
    }
    if (!isTRUE(raw)) {
      nv           <- names(value)
      value        <- gsub(".", "p", value, fixed = TRUE)
      names(value) <- nv
    }

    .sortvalues <- function(value, x, dim) {
      if (length(value) == 1) return(list(value))
      if (!is.null(names(value))) {
        order <- getItems(x, dim)
        if (!all(order %in% names(value))) {
          stop("Input vector is named but not all names match items of the dimension to be replaced!")
        }
        value <- value[order]
        names(value) <- NULL
      }
      return(value)
    }
    if (dc == maindim) {
      dimnames(x)[[maindim]] <- .sortvalues(value, x, maindim)
      if (!isTRUE(raw)) names(dimnames(x))[maindim] <- gsub(".", "p", names(dimnames(x))[maindim], fixed = TRUE)
    } else if (!dimExists(dc, x)) {
      if (!is.null(names(value))) {
        warning("Names of input vector are being ignored as dimension is not yet existing!")
        names(value) <- NULL
      }
      dimnames(x)[[maindim]] <- paste0(dimnames(x)[[maindim]], ".", value)
      if (!is.character(dim)) dim <- "newdim"
      names(dimnames(x))[maindim] <- paste0(names(dimnames(x))[maindim], ".", dim)
    } else {
      tmp <- getItems(x, maindim, split = TRUE, full = TRUE)
      subdim <- as.integer(substring(dc, 3))
      tmp[[subdim]] <- .sortvalues(value, x, dc)
      .paste <- function(...) return(paste(..., sep = "."))
      dimnames(x)[[maindim]] <- do.call(.paste, tmp)
    }
  } else {
    if (dc == maindim) {
      if (isFALSE(raw) && dim(x)[maindim] > 1) {
        stop("Cannot unset dimension names for dimensions with more than 1 element!")
      }
      if (maindim == 1) dimnames(x) <- c(d1 = list(NULL), dimnames(x)[2:3])
      if (maindim == 2) dimnames(x) <- c(dimnames(x)[1], d2 = list(NULL), dimnames(x)[3])
      if (maindim == 3) dimnames(x) <- c(dimnames(x)[1:2], d3 = list(NULL))
    } else if (!dimExists(dc, x)) {
      warning("Nothing to do here as dimension does not exist!")
    } else {
      tmp <- getItems(x, maindim, split = TRUE, full = TRUE)
      subdim <- as.integer(substring(dc, 3))
      tmp[[subdim]] <- NULL
      .paste <- function(...) return(paste(..., sep = "."))
      dimnames(x)[[maindim]] <- do.call(.paste, tmp)
      subdim <- as.integer(substring(dc, 3))
      names(dimnames(x))[maindim] <- paste(strsplit(names(dimnames(x))[maindim], ".", fixed = TRUE)[[1]][-subdim],
                                           collapse = ".")
    }
  }
  return(x)
}

context("MAgPIE creation test")

test_that("magpie objects are created correctly", {
  reg <- c("AFR", "EUR")
  expect_silent(x <- new.magpie(reg, 1, 1))
  expect_identical(getItems(x, 1), reg)
  reg2 <- reg
  names(reg2) <- reg
  expect_silent(x <- new.magpie(reg2, 1, 1))
  expect_identical(getItems(x, 1), reg)

  expect_silent(o <- new.magpie(10:3, sort = TRUE))
  expect_identical(getItems(o, dim = 1), paste0("GLO.", 3:10))

})

#' MAgPIE-Sort
#'
#' Brings the spatial and temporal structure of MAgPIE objects in the right
#' order. This function is especially useful when you create new MAgPIE objects
#' as the order typically should be correct for MAgPIE objects.
#'
#' @param x MAgPIE object which might not be in the right order.
#' @return The eventually corrected MAgPIE object (right order in spatial in
#' temporal dimension)
#'
#' @author Jan Philipp Dietrich
#' @seealso \code{"\linkS4class{magpie}"}
#' @examples
#' pop <- maxample("pop")
#' a <- magpiesort(pop)
#' @family DataBinding
#' @export
magpiesort <- function(x) {
  if (!is.magpie(x)) {
    stop("Input is not a MAgPIE object!")
  }

  if (dim(x)[1] == 1) {
    spatialOrder <- 1
  } else if (length(grep("\\.[0-9]*$", dimnames(x)[[1]])) == dim(x)[1]) {
    spatialOrder <- order(as.numeric(gsub("^.*\\.", "", dimnames(x)[[1]])))
  } else {
    spatialOrder <- order(dimnames(x)[[1]])
  }

  if (!is.null(dimnames(x)[[2]])) {
    temporalOrder <- order(dimnames(x)[[2]])
  } else {
    temporalOrder <- seq_len(dim(x)[2])
  }

  return(x[spatialOrder, temporalOrder, ])
}

.readComment <- function(fileName, commentChar = "*", metaChar = "#") {
  comment <- NULL
  if (!is.null(commentChar)) {
    if (commentChar != "") {
      zz <- file(fileName)
      open(zz)
      readRepeat <- TRUE
      while (readRepeat) {
        tmp <- readLines(zz, 1)
        if (length(grep(paste("^", escapeRegex(commentChar), sep = ""), tmp)) &&
              !grepl(metaChar, substr(tmp, 3, 3), fixed = TRUE)) {
          comment <- c(comment, tmp)
        } else {
          readRepeat <- FALSE
        }
      }
      close(zz)
    }
  }
  return(substring(comment, 2))
}

#' Get Coordinates
#'
#' Extracts spatial coordinates of a MAgPIE-object
#'
#'
#' @aliases getCoords getCoords<-
#' @param x MAgPIE object
#' @param xlab label of x-dimension
#' @param ylab label of y-dimension
#' @param value coordinates as two column data.frame the data should be set to
#' (first column = x, second column = y).
#' @return coordinates of the MAgPIE-object
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{as.RasterBrick}}, \code{\link{getItems}}, \code{"\linkS4class{magpie}"}
#' @examples
#' a <- maxample("animal")
#' getCoords(a)
#' @family GetterSetter
#' @export
getCoords <- function(x, xlab = "x", ylab = "y") {
  if (!hasCoords(x, xlab, ylab)) stop("No coordinates found in object!")
  .tmp <- function(x, n) {
    return(as.numeric(sub("p", ".", getItems(x, n, full = TRUE), fixed = TRUE)))
  }
  return(data.frame(x = .tmp(x, xlab), y = .tmp(x, ylab)))
}

#' @describeIn getCoords set coordinates
#' @export
`getCoords<-` <- function(x, xlab = "x", ylab = "y", value) {
  if (is.null(ncol(value)) || ncol(value) != 2) {
    stop("Coordinates supplied in unsupported format (must be a two column data.frame)")
  }
  if (!is.data.frame(value)) value <- as.data.frame(value)
  if (all(c("x", "y") %in% names(value))) value <- value[c("x", "y")]
  if (ncells(x) != nrow(value)) stop("Number of rows does not match number of cells of the MAgPIE object!")
  getItems(x, xlab, maindim = 1) <- value[[1]]
  getItems(x, ylab, maindim = 1)  <- value[[2]]
  return(x)
}

#' commonYears
#'
#' Given (a list of) magclass objects, return the years (getYears) they have in common,
#' ignoring magclass objects with 0 or 1 years.
#'
#' @param ... One or more magclass objects, or alternatively a list of magclass objects.
#' @param asInteger Passed on to getYears. If TRUE the output format is 1995 otherwise "y1995".
#' @return The years (as returned by getYears) the input objects have in common.
#'
#' @author Pascal Sauer
#' @family TemporalOperations
#' @export
commonYears <- function(..., asInteger = FALSE) {
  x <- list(...)
  if (length(x) == 1 && is.list(x[[1]])) {
    x <- x[[1]]
  }
  stopifnot(vapply(x, is.magpie, logical(1)))
  x <- Filter(function(i) nyears(i) > 1, x)
  return(Reduce(intersect, lapply(x, getYears, as.integer = asInteger)))
}

#' log-method for MAgPIE objects
#'
#' log-method for MAgPIE-objects respectively. Works exactly as for arrays.
#'
#'
#' @name log-methods
#' @aliases log-methods log,magpie-method logb-methods logb,magpie-method
#' @param x a magpie object
#' @param base ia positive or complex number: the base with respect to which
#' logarithms are computed. Defaults to e=exp(1).
#' @docType methods
#' @section Methods: \describe{
#'
#' \item{x = "magpie"}{ works as log(x) for arrays. }
#'
#' }
#' @exportMethod log


setMethod("log",
  signature(x = "magpie"),
  function(x, base = exp(1)) {
    x@.Data <- log(x@.Data, base = base)
    return(x)
  }
)

# from R documentation: logb is a wrapper for log for compatibility with S
#' @exportMethod logb
setMethod("logb",
  signature(x = "magpie"),
  function(x, base = exp(1)) {
    x@.Data <- log(x@.Data, base = base)
    return(x)
  }
)

#' splits a data.frame or vector of strings with form 'variable (unit)' into
#' a data.frame with variable and unit separated
#'
#' @param x data.frame or vector of strings
#' @param col column name. Default: variable
#' @author Jan Philipp Dietrich, Mika Pflüger, Oliver Richters
#'
#' @family Utility
#' @export
unitsplit <- function(x, col = "variable") {
  # structure of this regex (flavor: PCRE2):
  # '^' start of the string
  # '(?P<varname>…)' a named capture group for the varname. It contains:
  #   '.*' greedy everything
  # ' ' separator: a space
  # '(?P<recurse>…)' a named capture group for recursion to match balanced parentheses in the unit. It contains:
  #   '\\(' a literal opening parenthesis
  #   '(?P<unit>…)' a named capture group for the unit. It contains:
  #      '(?>…)*' an atomic non-capturing group which is repeated zero or more times. It contains:
  #          `[^()|]` any character which is neither a parenthesis or the pipe symbol
  #          `|` or
  #          `(?P>recurse)` a match for the named capture group "recurse".
  #          As the named capture group "recurse" contains exactly one literal opening and closing parenthesis, this
  #          recursion makes sure that parentheses in the unit are always balanced - if a parenthesis needs to be
  #          matched, we have to repeat the whole "recurse" capture group, and therefore, we need exactly one literal
  #          opening and closing parenthesis (or another recursion).
  #   '\\)' a literal closing parenthesis
  # '$' end of the string
  # If you have to understand the regex, I can warmly recommend putting it into https://regex101.com/ or a similar
  # service. Just replace all '\\' by '\', double backslashes are an R thing.
  if (is.null(ncol(x))) x <- data.frame(variable = x)
  pattern <- "^(?P<varname>.*) (?P<recurse>\\((?P<unit>(?>[^()|]|(?P>recurse))*)\\))$"
  # Even though we use named capture groups in the pattern, we can't use names in the replacement, because sub() only
  # supports numbered replacements. \\1 is the named capture group "varname", \\3 is the named capture group "unit".
  # Note that if the pattern does not match (e.g. because the value in question does not contain a unit), sub will not
  # replace anything, effectively passing everything through unchanged to varName and unit. That's correct for varName,
  # but for the unit, we have to overwrite non-matches with "N/A".
  varName <- sub(pattern, "\\1", x[[col]], perl = TRUE)
  unit <- sub(pattern, "\\3", x[[col]], perl = TRUE)
  unit[grep(pattern, x[[col]], invert = TRUE, perl = TRUE)] <- "N/A"
  tmp <- data.frame(varName, unit, stringsAsFactors = is.factor(x[[col]]))
  names(tmp) <- c(names(x[col]), "unit")
  x <- cbind(tmp, x[setdiff(names(x), names(x[col]))])
  return(x)
}

test_that("unitsplit and unitjoin on vectors and strings works", {
  teststring <- "Emi|CO2|+|Energy (Mt CO2/yr)"
  expected <- data.frame(
    variable = "Emi|CO2|+|Energy",
    unit = "Mt CO2/yr"
  )
  expect_identical(unitsplit(teststring), expected)
  expect_identical(unitsplit(teststring, 1), expected)
  expect_identical(unitsplit(teststring, "variable"), expected)
  expect_equal(unitjoin(expected)$variable, teststring)
  testvector <- c("Emi|CO2|+|Energy (Mt CO2/yr)", "Emi|CO|Land Use (Mt CO/yr)")
  expected <- data.frame(
    variable = c("Emi|CO2|+|Energy", "Emi|CO|Land Use"),
    unit = c("Mt CO2/yr", "Mt CO/yr")
  )
  expect_identical(unitsplit(testvector), expected)
  expect_identical(unitsplit(testvector, 1), expected)
  expect_identical(unitsplit(testvector, "variable"), expected)
  expect_equal(unitjoin(expected)$variable, testvector)
  expect_identical(unitjoin(as.factor("V"), as.factor("U")), as.factor("V (U)"))
  expect_identical(unitjoin("V", "U"), "V (U)")
})

test_that("simple unitsplit works", {
  df <- data.frame(
    Model = c("REMIND", "REMIND", "REMIND"),
    Scenario = c("everything nice", "everything awful", "middle of the road"),
    Region = c("GLO", "GLO", "GLO"),
    Data = c("floor covering|textile|carpet|red|length (m)",
             "floor covering|textile|carpet|red|length - for our american friends (inch)",
             "floor covering|textile|carpet|red|length (cm)"),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )
  expected <- data.frame(
    Data = c("floor covering|textile|carpet|red|length",
             "floor covering|textile|carpet|red|length - for our american friends",
             "floor covering|textile|carpet|red|length"),
    unit = c("m", "inch", "cm"),
    Model = c("REMIND", "REMIND", "REMIND"),
    Scenario = c("everything nice", "everything awful", "middle of the road"),
    Region = c("GLO", "GLO", "GLO"),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )
  expect_identical(unitsplit(df, 4), expected)
  expect_identical(unitsplit(df, "Data"), expected)
  expect_equal(unitjoin(expected, col = "Data")[colnames(df)], df)
  expect_error(unitjoin(expected, col = "doesnotexist"))
})

test_that("unitsplit works with braces", {
  df <- data.frame(
    Model = c("REMIND", "REMIND", "REMIND"),
    Scenario = c("everything nice", "everything awful", "middle of the road"),
    Region = c("GLO", "GLO", "GLO"),
    Data = c("floor covering|textile|carpet|red|length (m)",
             "floor covering|textile|carpet|red|length (for our american friends) (inch)",
             "floor covering|textile|carpet|red|length (cm)"),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )
  expected <- data.frame(
    Data = c("floor covering|textile|carpet|red|length",
             "floor covering|textile|carpet|red|length (for our american friends)",
             "floor covering|textile|carpet|red|length"),
    unit = c("m", "inch", "cm"),
    Model = c("REMIND", "REMIND", "REMIND"),
    Scenario = c("everything nice", "everything awful", "middle of the road"),
    Region = c("GLO", "GLO", "GLO"),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )
  expect_identical(unitsplit(df, 4), expected)
  expect_identical(unitsplit(df, "Data"), expected)
})

test_that("unitsplit handles all cases", {
  wrapper <- function(inputstr) {
    df <- data.frame(c(inputstr), stringsAsFactors = FALSE)
    splitted <- unitsplit(df, 1)
    return(c(splitted[[1]], splitted[[2]]))
  }
  expect_identical(wrapper("length"), c("length", "N/A"))
  expect_identical(wrapper("length (m)"), c("length", "m"))
  expect_identical(wrapper("carpet (the good one)|length"),
                   c("carpet (the good one)|length", "N/A"))
  expect_identical(wrapper("carpet (the good one)|length (US) (inch)"),
                   c("carpet (the good one)|length (US)", "inch"))
  expect_identical(wrapper("carpet (the good one)|make (as given) ()"),
                   c("carpet (the good one)|make (as given)", ""))
  expect_identical(wrapper("Price|Agriculture|Corn|Index (Index (2020 = 1))"),
                   c("Price|Agriculture|Corn|Index", "Index (2020 = 1)"))
  expect_identical(wrapper("Price|Agriculture (with cows)|Corn|Index (Index (2020 = 1))"),
                   c("Price|Agriculture (with cows)|Corn|Index", "Index (2020 = 1)"))
  expect_identical(wrapper("Price|Agriculture|Corn|Index (based on 2020) ((1 + 2) / (6 - 3))"),
                   c("Price|Agriculture|Corn|Index (based on 2020)", "(1 + 2) / (6 - 3)"))
  expect_identical(wrapper("Price|Agriculture|Corn|Index (based on 2020 (-;) ((1 + 2) / (6 - 3))"),
                   c("Price|Agriculture|Corn|Index (based on 2020 (-;)", "(1 + 2) / (6 - 3)"))
  expect_identical(wrapper("Price|Agriculture|Corn|Index (based on 2020 :-)) ((1 + 2) / (6 - 3))"),
                   c("Price|Agriculture|Corn|Index (based on 2020 :-))", "(1 + 2) / (6 - 3)"))
})

#' suppressSpecificWarnings
#'
#' Like \code{\link{suppressWarnings}}, but instead of suppressing all warnings this only suppresses
#' warnings if they match the given pattern.
#'
#' @param expr The expression/code to evaluate, can be a block of code inside curly braces.
#' @param regularExpr Only warnings matching this regular expression are suppressed.
#' @param fixed Match the literal string given by regularExpr instead of interpreting it
#' as a regular expression. Passed to \code{\link{grepl}}.
#' @return The result of evaluating expr.
#'
#' @author Pascal Sauer
#' @seealso \code{\link{suppressWarnings}}
#' @export
suppressSpecificWarnings <- function(expr, regularExpr, fixed = FALSE) {
  withCallingHandlers(expr, warning = function(m) {
    if (grepl(regularExpr, m[["message"]], fixed = fixed)) {
      invokeRestart("muffleWarning")
    }
  })
}

#' Collapse dataset dimensions
#'
#' This function will remove names in the data dimension which are the same for
#' each element (meaning that this data dimension contains exactly one element)
#' or, if forced, remove any other subdimension. It is a generalized version
#' of the function \code{\link{collapseNames}}
#'
#'
#' @param x MAgPIE object
#' @param dim Either NULL, dimension code or
#' name of dimension or a vector of these. If set to NULL all single entry subdimensions will
#' be removed as they are irrelevant to uniquely identfy a data element. If specified, only the specified
#' subdimensions will be removed (See \code{\link{dimCode}} for more details how to specify a subdimension).
#' CAUTION: The function also allows to specify subdimensions which are otherwise needed to clearly identify
#' an entry. By removing these subdimensions duplicates in the data will be created potentially causing
#' problems in the further use of the data set. Be careful in removing subdimensions.
#' @param keepdim (only considered if \code{dim} is not specified) Can be used to converse
#' single element subdimension which otherwise would get deleted. If \code{dim} is specified
#' this setting will not have any effect.
#' @note This function has some similarities to \code{\link{dimReduce}}, but
#' serves a different purpose. While \code{\link{collapseDim}} only removes
#' dimensions which contain only a single element or which it is
#' specifically told to remove, \code{\link{dimReduce}} looks whether the
#' entries of a multi-entry dimension are all the same and removes dimensions
#' for which this is the case. In some cases both will lead to the same result
#' but in many other cases the results will differ.
#' @return The provided MAgPIE object with collapsed dimensions
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getItems}}
#' \code{"\linkS4class{magpie}"}
#' @examples
#'
#' x <- new.magpie(c("GLO.1", "GLO.2"), 2000, c("bla.a", "bla.b"))
#' collapseDim(x)
#' collapseDim(x, keepdim = 1:2)
#' collapseDim(x, dim = 1.1)
#' collapseDim(x, dim = 3.2)
#' @family DimensionManipulation
#' @export

collapseDim <- function(x, dim = NULL, keepdim = NULL) {

  if (is.null(x)) return(NULL)

  x <- clean_magpie(x, what = c("sets", "items"))

  if (is.null(dim)) {
    sets <- getSets(x)
    if (anyDuplicated(sets)) getSets(x) <- make.unique(sets, sep = "")
    tmp <- vapply(unlist(getItems(x, split = TRUE), recursive = FALSE), length, integer(1))
    dim <- dimCode(names(tmp)[tmp == 1], x)
    if (anyDuplicated(sets)) getSets(x) <- sets
    if (!is.null(keepdim)) {
      dim <- setdiff(dim, dimCode(keepdim, x))
      dim <- dim[!(floor(dim) %in% keepdim)]
    }
    dim <- dim[dim > 0]
  } else {
    if (!is.null(keepdim)) warning("keepdim argument is ignored as dim has been specified!")
    dim <- dimCode(dim, x)
    if (any(dim == 0)) {
      warning("Some dimensions could not be found in the object and will be ignored!")
      dim <- dim[dim > 0]
    }
  }

  for (d in sort(dim, decreasing = TRUE)) getItems(x, dim = d) <- NULL
  return(x)
}

testMagpie <- function(regions = "AFR.1", years = "y2000", names = "test", values = c(123), ...) {
  # The special handling of null is necessary, as new.magpie creates 0-length
  # magpie objects when arguments are NULL
  if (is.null(regions)) regions <- "dummy"
  if (is.null(years)) years <- 0
  if (is.null(names)) names <- "dummy"
  newObject <- new.magpie(regions, years, names, values, ...)
  if (all(regions == "dummy")) newObject <- collapseDim(newObject, dim = 1)
  if (all(years == 0)) newObject <- collapseDim(newObject, dim = 2)
  if (all(names == "dummy")) newObject <- collapseDim(newObject, dim = 3)
  names(dimnames(newObject)) <- c("region", "year", "data")
  return(newObject)
}

expect_pmin_result <- function(a, b, result) { # nolint: object_name_linter. Mimics testthat style.
  pminResult <- pmin(a, b)
  expect_equal(!!result, !!pminResult)
  return(pminResult)
}

test_that("pmin on simple magpie objects", {
  # Equal
  expect_pmin_result(testMagpie(values = 123),
                     testMagpie(values = 123),
                     testMagpie(values = 123))

  # One is smaller
  expect_pmin_result(testMagpie(values = 121),
                     testMagpie(values = 123),
                     testMagpie(values = 121))
})

test_that("variadic pmin", {
  # Single argument is fine, but simply returns the same object
  expect_equal(pmin(testMagpie()), testMagpie())

  # More than two arguments are also fine
  expect_equal(pmin(testMagpie(values = 123),
                    testMagpie(values = 1),
                    testMagpie(values = 123)),
               testMagpie(values = 1))
})

test_that("pmin on magpie objects with different orderings in one dimension", {
  expect_pmin_result(testMagpie(regions = c("AFR.1", "AFR.2"), values = c(120, 345)),
                     testMagpie(regions = c("AFR.2", "AFR.1"), values = c(340, 123)),
                     testMagpie(regions = c("AFR.1", "AFR.2"),
                                values = c(120, 340)))

  expect_pmin_result(testMagpie(years = c("2000", "2001"), values = c(120, 345)),
                     testMagpie(years = c("2001", "2000"), values = c(340, 123)),
                     testMagpie(years = c("2000", "2001"), values = c(120, 340)))

  expect_pmin_result(testMagpie(names = c("test1", "test2"), values = c(120, 345)),
                     testMagpie(names = c("test2", "test1"), values = c(340, 123)),
                     testMagpie(names = c("test1", "test2"), values = c(120, 340)))
})

test_that("pmin on magpie objects with different orderings in all dimensions", {
  expect_pmin_result(testMagpie(regions = c("AFR.1", "AFR.2"),
                                years = c("2000", "2001"),
                                names = c("test1", "test2"),
                                values = c(1, 2, 3, 4, 5, 6, 7, 8)),
                     testMagpie(regions = c("AFR.2", "AFR.1"),
                                years = c("2001", "2000"),
                                names = c("test2", "test1"),
                                values = c(1, 2, 3, 4, 5, 6, 7, 8)),
                     testMagpie(regions = c("AFR.1", "AFR.2"),
                                years = c("2000", "2001"),
                                names = c("test1", "test2"),
                                values = c(1, 2, 3, 4, 4, 3, 2, 1)))

  # Sanity check: base::pmin fails in the same scenario as above
  expect_failure(expect_equal(base::pmin(testMagpie(regions = c("AFR.1", "AFR.2"),
                                                    years = c("2000", "2001"),
                                                    names = c("test1", "test2"),
                                                    values = c(1, 2, 3, 4, 5, 6, 7, 8)),
                                         testMagpie(regions = c("AFR.2", "AFR.1"),
                                                    years = c("2001", "2000"),
                                                    names = c("test2", "test1"),
                                                    values = c(1, 2, 3, 4, 5, 6, 7, 8))),
                              testMagpie(regions = c("AFR.1", "AFR.2"),
                                         years = c("2000", "2001"),
                                         names = c("test1", "test2"),
                                         values = c(1, 2, 3, 4, 4, 3, 2, 1))))
})

test_that("pmin allows different items, if there is only one item in that dimension", {
  # A single item in one dimension
  expect_pmin_result(testMagpie(regions = c("A"), years = 2001:2002, names = c("test1", "test2"),
                                values = c(3, 3, 3, 3)),
                     testMagpie(regions = c("B"), years = 2001:2002, names = c("test1", "test2"),
                                values = c(1, 2, 3, 4)),
                     testMagpie(regions = NULL, years = 2001:2002, names = c("test1", "test2"),
                                values = c(1, 2, 3, 3)))

  expect_pmin_result(testMagpie(regions = c("A", "B"), years = 2001, names = c("test1", "test2"),
                                values = c(3, 3, 3, 3)),
                     testMagpie(regions = c("A", "B"), years = 2002, names = c("test1", "test2"),
                                values = c(1, 2, 3, 4)),
                     testMagpie(regions = c("A", "B"), years = NULL, names = c("test1", "test2"),
                                values = c(1, 2, 3, 3)))

  expect_pmin_result(testMagpie(regions = c("A", "B"), years = 2001:2002, names = c("test1"),
                                values = c(3, 3, 3, 3)),
                     testMagpie(regions = c("A", "B"), years = 2001:2002, names = c("test2"),
                                values = c(1, 2, 3, 4)),
                     testMagpie(regions = c("A", "B"), years = 2001:2002, names = NULL,
                                values = c(1, 2, 3, 3)))

  # A single item in two dimensions
  expect_pmin_result(testMagpie(regions = c("A"), years = 2001:2002, names = c("test1"),
                                values = c(3, 3)),
                     testMagpie(regions = c("B"), years = 2001:2002, names = c("test2"),
                                values = c(1, 4)),
                     testMagpie(regions = NULL, years = 2001:2002, names = NULL,
                                values = c(1, 3)))

  expect_pmin_result(testMagpie(regions = c("A"), years = 2001, names = c("test1", "test2"),
                                values = c(3, 3)),
                     testMagpie(regions = c("B"), years = 2002, names = c("test1", "test2"),
                                values = c(1, 4)),
                     testMagpie(regions = NULL, years = NULL, names = c("test1", "test2"),
                                values = c(1, 3)))

  expect_pmin_result(testMagpie(regions = c("A", "B"), years = 2001, names = c("test1"),
                                values = c(3, 3)),
                     testMagpie(regions = c("A", "B"), years = 2002, names = c("test2"),
                                values = c(1, 4)),
                     testMagpie(regions = c("A", "B"), years = NULL, names = NULL,
                                values = c(1, 3)))


  # A single item in all three dimensions
  expect_pmin_result(testMagpie(regions = c("A"), years = 2001, names = c("test1"),
                                values = c(3)),
                     testMagpie(regions = c("B"), years = 2002, names = c("test2"),
                                values = c(2)),
                     testMagpie(regions = NULL, years = NULL, names = NULL,
                                values = c(2)))

  # This also holds if one object has only one item and the other one has none
  expect_pmin_result(testMagpie(regions = NULL, years = 2001:2002, names = c("test1", "test2"),
                                values = c(3, 3, 3, 3)),
                     testMagpie(regions = c("B"), years = 2001:2002, names = c("test1", "test2"),
                                values = c(1, 2, 3, 4)),
                     testMagpie(regions = NULL, years = 2001:2002, names = c("test1", "test2"),
                                values = c(1, 2, 3, 3)))

  expect_pmin_result(testMagpie(regions = NULL, years = NULL, names = c("test1", "test2"),
                                values = c(3, 3)),
                     testMagpie(regions = c("B"), years = 2001, names = c("test1", "test2"),
                                values = c(1, 4)),
                     testMagpie(regions = NULL, years = NULL, names = c("test1", "test2"),
                                values = c(1, 3)))

  # This also holds if a dimension is not present in one object and the other contains only one value
  # The following tests use collapseDim to create objects with "removed" dimensions. The single items
  # in these testMagpie objects are only placeholders so that collapseDim "removes" the dimension.
  expect_pmin_result(collapseDim(testMagpie(regions = "GLO", years = 2001:2002, names = c("test1", "test2"),
                                            values = c(3, 3, 3, 3))),
                     testMagpie(regions = c("B"), years = 2001:2002, names = c("test1", "test2"),
                                values = c(1, 2, 3, 4)),
                     collapseDim(testMagpie(regions = "GLO", years = 2001:2002, names = c("test1", "test2"),
                                            values = c(1, 2, 3, 3))))

  expect_pmin_result(collapseDim(testMagpie(regions = "GLO", years = 1, names = c("test1", "test2"),
                                            values = c(3, 3))),
                     testMagpie(regions = c("B"), years = 2001, names = c("test1", "test2"),
                                values = c(1, 4)),
                     collapseDim(testMagpie(regions = "GLO", years = 1, names = c("test1", "test2"),
                                            values = c(1, 3))))

})

test_that("pmin special cases", {
  # pmin with a single number
  expect_pmin_result(testMagpie(years = 2001:2002, values = c(1, 3)),
                     2,
                     testMagpie(years = 2001:2002, values = c(1, 2)))
  expect_pmin_result(2,
                     testMagpie(years = 2001:2002, values = c(1, 3)),
                     c(1, 2))

  # pmin with a vector
  expect_pmin_result(c(2, 5),
                     testMagpie(years = 2001:2002, values = c(1, 3)),
                     c(1, 3))

  expect_error(pmin(testMagpie(regions = "A"),
                    testMagpie(regions = c("A", "B"))),
               "pmin expects magpie objects with equal dimensions")
  expect_error(pmin(testMagpie(years = "2000"),
                    testMagpie(years = c("2000", "2001"))),
               "pmin expects magpie objects with equal dimensions")

  expect_error(pmin(testMagpie(names = "A"),
                    testMagpie(names = c("A", "B"))),
               "pmin expects magpie objects with equal dimensions")

  expect_error(pmin(testMagpie(names = c("A", "B")),
                    testMagpie(names = c("C", "D"))),
               "pmin expects magpie objects with the same items in all dimensions")
})

test_that("pmax simple test", {
  # As pmin and pmax use the same underlying function to rearrange data,
  # a basic test whether pmax works is sufficient.
  expect_equal(pmax(testMagpie(values = 5), testMagpie(values = 10), testMagpie(values = 5)),
               testMagpie(values = 10))
})

#' Count sub-dimensions
#'
#' Functions to count the subdimensions of an
#' MAgPIE-object
#'
#' @param x A MAgPIE-object
#' @param dim main dimension in which the sub-dimensions should be counted. If NULL
#' the sum of all subdimensions is returned
#' @return Number of subdimensions
#' @author Jan Philipp Dietrich
#' @examples
#'
#'   a <- maxample("animal")
#'   ndim(a)
#'   ndim(a,1)
#'   ndim(a,2)
#'   ndim(a,3)
#'
#' @family ObjectInfo
#' @export
ndim <- function(x, dim = NULL) {
  if (is.null(dim)) return(ndim(x, 1) + ndim(x, 2) + ndim(x, 3))
  if (!(dim %in% 1:3)) stop("Invalid dim selection (Allowed: 1, 2 or 3).")
  tmp <- dimnames(x)[[dim]][1]
  if (is.null(tmp)) {
    # fallback to set information
    tmp <- names(dimnames(x))[dim]
  }
  if (is.null(tmp)) return(0)
  ndim <- nchar(gsub("[^.]", "", tmp)) + 1
  return(ndim)
}

#' Has Sets
#'
#' Checks, whether set names have been set
#'
#'
#' @param x MAgPIE object
#' @return Boolean indicating whether coordinates were found or not
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getCoords}}
#' @examples
#'
#' hasSets(maxample("pop"))
#' hasSets(maxample("animal"))
#'
#' @family ObjectInfo
#' @export
hasSets <- function(x) {
  return(!is.null(names(dimnames(x))))
}

#' isYear
#'
#' Function to find out whether a vector consists of strings in the format
#' "yXXXX" or "XXXX" with X being a number
#'
#'
#' @param x A vector
#' @param with_y indicates which dataformat years have to have (4-digit without
#' y (e.g.1984) or 5digit including y (y1984))
#' @return Returns a vector of the length of x with TRUE and FALSE
#' @author Benjamin Bodirsky
#' @examples
#'
#' x <- c("1955", "y1853", "12a4")
#' isYear(x, with_y = TRUE)
#' isYear(x, with_y = FALSE)
#' @family ObjectInfo
#' @export
isYear <- function(x, with_y = TRUE) { # nolint: object_name_linter.
  if (is.null(x)) return(FALSE)
  if (!is.vector(x)) {
    stop("Year Object is no Vector")
  }
  returnVector <- rep(TRUE, length(x))
  for (i in seq_along(x)) {
    if (with_y == FALSE) {
      if (nchar(x[i]) != 4) {
        returnVector[i] <- FALSE
      }
      if (grepl("[0-9]{4}", x[i]) == FALSE) {
        returnVector[i] <- FALSE
      }
    } else        {
      if (nchar(x[i]) != 5) {
        returnVector[i] <- FALSE
      }
      if (grepl("y[0-9]{4}", x[i]) == FALSE) {
        returnVector[i] <- FALSE
      }
    }
  }
  return(returnVector)
}

#' @title dimOrder
#' @description Changes the order of the sub-dimension in a magpie object
#' similar to unwrapping and applying the aperm command, but more efficient.
#'
#' @param x magpie object
#' @param perm vector with the new order of the sub-dimension. Missing
#' sub-dimensions will added automatically at the end
#' @param dim main dimension in which the order of sub-dimensions should
#' be changed (1, 2 or 3)
#'
#' @return magpie object
#' @author Jan Philipp Dietrich, Benjamin Leon Bodirsky
#' @examples
#' a <- maxample("animal")
#' head(a)
#' head(dimOrder(a, perm = 3:1, dim = 1))
#' head(dimOrder(a, perm = c(2,1,3), dim = 3))
#' @family DimensionManipulation
#' @export

dimOrder <- function(x, perm, dim = 3) {
  ndim <- ndim(x, dim)
  if (any(!(perm %in% 1:ndim))) {
    stop("Values of perm must be integers in the range from 1 to ", ndim)
  }
  if (ndim != length(perm)) {
    perm <- c(perm, setdiff(1:ndim, perm))
  }
  if (ndim > 1) {
    pattern <- paste0("^([^\\.]*)\\.", paste0(rep("([^\\.]*)\\.", ndim - 2), collapse = ""), "([^\\.]*)$")
    order <- paste0(paste0("\\", perm), collapse = ".")
    if (!is.null(dimnames(x)[[dim]])) {
      dimnames(x)[[dim]] <- gsub(pattern = pattern, replacement = order, dimnames(x)[[dim]], perl = TRUE)
    }
    if (!is.null(names(dimnames(x))[dim])) {
      names(dimnames(x))[dim] <- gsub(pattern = pattern, replacement = order, names(dimnames(x))[dim], perl = TRUE)
    }
  }
  return(x)
}

#' joins a data.frame or vector of strings with variable and unit separated into
#' a data.frame with variable and unit joined as 'variable (unit)'.
#' Use magclass::unitsplit to split them again
#'
#' @param x data.frame or vector of strings
#' @param col column name. Default: variable
#' @param unit vector of strings. If NULL, col 'unit' in x is used
#' @return data.frame or vector of strings, dependent on x
#'
#' @family Utility
#' @export
unitjoin <- function(x, unit = NULL, col = "variable") {
  # check whether is a data.frame
  if (is.null(ncol(x))) {
    if (length(x) == 0 || is.null(unit)) {
      # return empty variable vector or if no unit specified
      return(x)
    } else {
      #  paste unit to it, keeping factor type
      if (is.factor(x)) {
        return(as.factor(paste0(x, " (", unit, ")")))
      } else {
        return(paste0(x, " (", unit, ")"))
      }
    }
  } else { # is data.frame
    # no unit is specified -> take from unit column
    stopifnot(col %in% colnames(x))
    if (is.null(unit)) {
      x[col] <- unitjoin(x[[col]], x[["unit"]])
      x["unit"] <- NULL
    } else {
      # use what is specified
      x[col] <- unitjoin(x[[col]], unit)
    }
    return(x)
  }
}

#' @importFrom methods new setGeneric
#' @importFrom data.table as.data.table tstrsplit melt

#' @exportMethod as.magpie
setGeneric("as.magpie", function(x, ...) standardGeneric("as.magpie")) # nolint: object_name_linter

setMethod("as.magpie", signature(x = "magpie"), function(x) return(x))


tmpfilter <- function(x, sep = ".", replacement = "_") {
  .tmpfilter <- function(x, sep = ".", replacement = "_") {
    .tmp <- function(x, sep, replacement) {
      if (sep != replacement) x <- gsub(sep, replacement, x, fixed = TRUE, useBytes = TRUE)
      x[x == ""] <- " "
      return(x)
    }
    if (is.factor(x)) {
      levels(x) <- .tmp(levels(x), sep, replacement)
    } else if (is.character(x)) {
      x <- .tmp(x, sep, replacement)
    }
    return(x)
  }
  cl <- class(x)
  cn <- colnames(x)
  rn <- rownames(x)
  x <- as.data.frame(lapply(x, .tmpfilter, sep = sep, replacement = replacement))
  colnames(x) <- cn
  rownames(x) <- rn
  class(x) <- cl
  return(x)
}

setMethod("as.magpie",
  signature(x = "lpj"),
  function(x, unit = "unknown", ...) {
    xdimnames <- dimnames(x)
    xdim <- dim(x)
    x <- array(x[magclassdata$half_deg$lpj_index, , , ], dim = c(dim(x)[1:2], dim(x)[3] * dim(x)[4]))
    # convert to previous default standard 59199 cells half degree resolution
    dimnames(x) <- list(paste(magclassdata$half_deg$region, 1:59199, sep = "."),
                        xdimnames[[2]],
                        paste(rep(xdimnames[[3]], xdim[4]), rep(xdimnames[[4]], each = xdim[3]), sep = "."))
    out <- new("magpie", x)
    return(out)
  }
)

setMethod("as.magpie",
          signature(x = "LPJmLData"),
          function(x) {
            if (!requireNamespace("lpjmlkit", quietly = TRUE)) {
              stop("The package \"lpjmlkit\" is required for LPJmLData conversions!")
            }
            y <- lpjmlkit::transform(x, to = c("cell", "time"))
            y <- lpjmlkit::as_array(y)
            if (!is.null(x$grid)) {
              grid <- paste0(sub(".", "p", x$grid$data[, 1], fixed = TRUE), ".",
                             sub(".", "p", x$grid$data[, 2], fixed = TRUE))
              names(grid) <- rownames(x$grid$data)
              dimnames(y)[[1]] <- paste0(grid[dimnames(y)[[1]]], ".", dimnames(y)[[1]])
              names(dimnames(y))[1] <- c("x.y.cell")
            }
            out <- new("magpie", y)
            return(out)
          })

setMethod("as.magpie", # nolint: cyclocomp_linter.
  signature(x = "array"),
  function(x, spatial = NULL, temporal = NULL, unit = "unknown", ...) {
    storeAttributes <- copy.attributes(x, 0)

    # Add the sets as name to the dimnames, if existent
    if (is.null(names(dimnames(x))) & !is.null(attr(x, "sets"))) {
      tmp <- dimnames(x)
      names(tmp) <- attr(x, "sets")
      dimnames(x) <- tmp
    }
    # This part of the function analyses what structure the input has
    d <- list()  # list of dimension types found in the array
    if (!is.null(temporal)) d$temporal <- temporal
    if (!is.null(spatial)) d$regiospatial <- spatial
    for (i in seq_along(dim(x))) {
      if (!is.null(dimnames(x)[[i]])) {
        if (is.null(spatial)) {
          if (length(grep("^(([A-Z]{3})|(glob))$", dimnames(x)[[i]])) == dim(x)[i]) {
            d$regional <- c(d$regional, i)  # regional information
          }
          if (length(grep("^[A-Z]+[\\._][0-9]+$", dimnames(x)[[i]])) == dim(x)[i]) {
            d$regiospatial <- c(d$regiospatial, i)  # regio-spatial information
          }
        }
        if (is.null(temporal)) {
          if (is.temporal(dimnames(x)[[i]])) d$temporal <- c(d$temporal, i) # temporal information
        }
      } else if (dim(x)[i] == 1) {
        d$nothing <- c(d$nothing, i)   # dimension with no content
      }
    }

    if (!is.null(spatial)) {
      if (spatial == 0) {
        d$regiospatial <- NULL
        d$regional <- NULL
      }
    }

    if (!is.null(temporal)) {
      if (temporal == 0) {
        d$temporal <- NULL
      }
    }

    # Write warning when any type (except type "nothing") is found more than once
    nOccurrence <- lapply(d, length) > 1
    nOccurrence <- nOccurrence[names(nOccurrence) != "nothing"]
    if (any(nOccurrence) == TRUE) {
      warning("No clear mapping of dimensions to dimension types. First detected ",
              "possibility is used! Please use arguments temporal and spatial to specify",
              " which dimensions are what!")
    }
    for (i in which(nOccurrence)) {
      d[[i]] <- d[[i]][1]
    }

    # If a regional dimension exists, test whether "glob" appears in the dimnames and rename it with "GLO"
    if (!is.null(d$regional)) {
      for (i in d$regional) {
        dimnames(x)[[i]] <- sub("^glob$", "GLO", dimnames(x)[[i]])
      }
    }

    # make sure that temporal dimension uses dimnames of the form y0000
    if (!is.null(d$temporal)) {
      for (i in d$temporal) {
        dimnames(x)[[i]] <- sub("^[a-z]?([0-9]{4})$", "y\\1", dimnames(x)[[i]])
      }
    }

    # make sure that spatial dimension uses dimnames of the form XXX.123
    if (!is.null(d$regiospatial)) {
      for (i in d$regiospatial) {
        ntmp <- names(dimnames(x))[1]
        if (!is.null(ntmp) && !is.na(ntmp) && names(dimnames(x))[1] == "j") names(dimnames(x))[1] <- "i.j"
      }
    }


    # If no temporal dimension is defined, but a dimension of type nothing exists,
    # use this dimension as temporal dimension
    if (is.null(d$temporal)) {
      if (length(d$nothing) > 0) {
        d$temporal <- d$nothing[1]
        d$nothing <- d$nothing[-1]
        if (length(d$nothing) == 0) d$nothing <- NULL
      } else {
        d$temporal <- 0
      }
    }

    # try to create regiospatial dimension if possible
    if (is.null(d[["regiospatial"]])) {
      # regional dimension exists
      if (!is.null(d$regional)) {
        d$regiospatial <- d$regional
      } else {
        d$regiospatial <- 0
      }
    }
    d$regional <- NULL

    # Starting from here d$temporal and d$regiospatial should be defined both
    # If any of these two could neither be found nor created the value should be 0

    if (d$regiospatial == 0) {
      if (is.null(dimnames(x))) {
        x <- array(x, c(dim(x), 1))
        dimnames(x)[[length(dim(x))]] <- list("GLO")
      } else {
        x <- array(x, c(dim(x), 1), c(dimnames(x), "GLO"))
      }
      d$regiospatial <- length(dim(x))
    }

    if (d$temporal == 0) {
      x <- array(x, c(dim(x), 1), c(dimnames(x), NULL))
      d$temporal <- length(dim(x))
    }

    # Check if third dimension exists. If not, create it
    if (length(dim(x)) == 2) {
      x <- array(x, c(dim(x), 1), c(dimnames(x), NULL))
    }

    # Now temporal and regiospatial dimension should both exist
    # Return MAgPIE object
    out <- copy.attributes(storeAttributes, new("magpie", wrap(x, list(d$regiospatial, d$temporal, NA))))
    return(out)
  }
)

setMethod("as.magpie",
          signature(x = "numeric"),
          function(x, unit = "unknown", ...) {
            return(copy.attributes(x, as.magpie(as.array(x), ...)))
          })

setMethod("as.magpie",
          signature(x = "logical"),
          function(x, unit = "unknown", ...) {
            return(copy.attributes(x, as.magpie(as.array(x), ...)))
          })

setMethod("as.magpie",
  signature(x = "NULL"),
  function(x) {
    return(NULL)
  }
)

setMethod("as.magpie",
  signature(x = "data.frame"),
  function(x, datacol = NULL, tidy = FALSE, sep = ".", replacement = "_", unit = "unknown", filter = TRUE, ...) {
    # filter illegal characters
    if (isTRUE(filter)) {
      x <- tmpfilter(x, sep = sep, replacement = replacement)
    }

    if (tidy) return(tidy2magpie(x, ...))
    if (dim(x)[1] == 0) return(copy.attributes(x, new.magpie(NULL)))
    if (is.null(datacol)) {
      isNumericlike <- function(x) {
        .tmp <- function(x) return(all(!is.na(suppressWarnings(as.numeric(x[!is.na(x)])))))
        if (isFALSE(.tmp(x[1]))) return(FALSE)
        return(.tmp(x))
      }
      for (i in rev(seq_len(dim(x)[2]))) {
        if (!is.factor(x[[i]]) && isNumericlike(x[[i]]) && !is.temporal(x[[i]])) {
          datacol <- i
        } else {
          break
        }
      }
    }
    if (!is.null(datacol)) {
      if (is.character(datacol)) {
        datacol <- which(colnames(x) == datacol)
        if (length(datacol) != 1) {
          stop("datacol does not clearly match one column name.")
        }
      }
      if (datacol == 1) return(copy.attributes(x, as.magpie(as.matrix(x), ...)))
      if (datacol == dim(x)[2]) return(tidy2magpie(x, ...))
      x[[datacol - 1]] <- as.factor(x[[datacol - 1]])
    }

    helper <- suppressWarnings(data.table::setDF(data.table::melt(data.table::as.data.table(x))))
    # The warning that is suppressed above is:
    #  In melt.data.table(data.table::as.data.table(x)) :
    #   id.vars and measure.vars are internally guessed when both are 'NULL'. All non-numeric/integer/logical type
    #   columns are considered id.vars, which in this case are columns {...}. Consider providing at least one of
    #   'id' or 'measure' vars in future.

    out <- copy.attributes(x, tidy2magpie(suppressMessages((helper)), ...))
    return(out)
  }
)

setMethod("as.magpie",
  signature(x = "quitte"),
  function(x, sep = ".", replacement = "_", filter = TRUE, ...) {
    isQuitte <- function(x) {
      # object is formally defined as quitte but it has to
      # be checked whether it follows all structural
      # rules of a quitte object

      mandatoryColumns <- c("model", "scenario", "region", "variable", "unit", "period", "value")
      factorColumns    <- c("model", "scenario", "region", "variable", "unit")
      isQuitte <- (all(mandatoryColumns %in% names(x)) &&
                     all(vapply(x[factorColumns], is.factor, logical(1))) &&
                     is.numeric(x$value) &&
                     (methods::is(x$period, "POSIXct") || is.integer(x$period)))
      return(isQuitte)
    }

    if (!isQuitte(x)) {
      warning("Input does not follow the full quitte class definition! Fallback to data.frame conversion.")
      class(x) <- "data.frame"
      return(as.magpie(x, ...))
    }
    x$period <- format(x$period, format = "y%Y")
    # filter illegal characters

    if (isTRUE(filter)) {
      x <- tmpfilter(x, sep = sep, replacement = replacement)
    }

    if (length(grep("^cell$", names(x), ignore.case = TRUE)) > 0) {
      i <- grep("^cell$", names(x), ignore.case = TRUE, value = TRUE)
      x$region <- paste(x$region, x[[i]], sep = ".")
      x <- x[names(x) != i]
    }
    # remove NA columns and <NA> columns that have been replaced by forcats::fct_explicit_na()
    naString <- "(Missing)"
    x <- x[colSums(!(naString == x | is.na(x))) != 0]

    # put value column as last column
    x <- x[c(which(names(x) != "value"), which(names(x) == "value"))]
    out <- tidy2magpie(x, spatial = "region", temporal = "period")
    return(out)
  }
)

setMethod("as.magpie",
  signature(x = "tbl_df"),
  function(x, unit = "unknown", ...) {
    if ("quitte" %in% class(x)) {
      class(x) <- c("quitte", "data.frame")
      out <- as.magpie(x, ...)
      return(out)
    } else {
      class(x) <- "data.frame"
      out <- as.magpie(x, ...)
      return(out)
    }
  }
)


.raster2magpie <- function(x, unit = "unknown", temporal = NULL, spatial = NULL) {
  geometry <- NULL
  idVars <- NULL
  crs <- NULL
  if (inherits(x, "SpatRaster")) {
    if (!requireNamespace("terra", quietly = TRUE)) stop("The package \"terra\" is required for raster conversions!")
    df <- as.data.frame(x, xy = TRUE)
    df <- df[rowSums(is.na(df)) != ncol(df), ] # remove rows with all NAs
    df$x <- sub(".", "p", df$x, fixed = TRUE)
    df$y <- sub(".", "p", df$y, fixed = TRUE)
    idVars <- c("x", "y")
    crs <- terra::crs(x)
  } else if (inherits(x, "SpatVector")) {
    if (!requireNamespace("terra", quietly = TRUE)) stop("The package \"terra\" is required for raster conversions!")
    df <- as.data.frame(x, geom = "WKT")
    geometry <- df$geometry
    df$geometry <- NULL
    crs <- terra::crs(x)
  } else {
    if (!requireNamespace("raster", quietly = TRUE)) stop("The package \"raster\" is required for raster conversions!")
    # na.rm = TRUE seems to remove all cells in which at least one layer has an NA. Hence, use na.rm = FALSE
    # and remove all cells which have NAs in ALL layers afterwards!
    df <- as.data.frame(x, na.rm = FALSE)
    df <- df[rowSums(!is.na(df)) != 0, , drop = FALSE]

    co <- raster::coordinates(x)[as.integer(rownames(df)), ]
    co <- matrix(sub(".", "p", co, fixed = TRUE), ncol = 2)
    colnames(co) <- c("x", "y")
    df <- cbind(co, df)
    idVars <- c("x", "y")
  }

  # check for additional spatial identifiers
  if (is.null(spatial)) {
    # select all layers with a leading dot in its name as spatial dimension
    spatial <- grep("^\\.", names(x), value = TRUE)
  } else {
    if (all(is.integer(spatial))) {
      spatial <- names(x)[spatial]
    }
  }
  names(df) <- sub("^\\.", "", names(df))
  spatial  <- sub("^\\.", "", spatial)

  idVars <- c(spatial, idVars)

  if (length(idVars) == 0) {
    df <- cbind(id = df[[1]], df)
    idVars <- "id"
  }

  setDT(df) # No copy is made, df is changed in-place to data.tabe
  df <- melt(df, id.vars = idVars)

  # We first only work on a sample to determine the proper
  # variable names. Then we do the variable transformation in-place.
  variable <- data.frame(tstrsplit(df$variable[1], "..", fixed = TRUE))
  if (!is.null(temporal)) temporal <- temporal + length(idVars)
  if (ncol(variable) == 1) {
    if (all(grepl("^[yX][0-9]*$", variable[[1]], perl = TRUE))) {
      variable[[1]] <- sub("^X", "y", variable[[1]], perl = TRUE)
      names(variable) <- "year"
      if (is.null(temporal)) temporal <- length(idVars) + 1
    } else {
      names(variable) <- "data"
    }
  } else if (ncol(variable) == 2) {
    names(variable) <- c("year", "data")
    if (is.null(temporal)) temporal <- length(idVars) + 1
  } else {
    stop("Reserved dimension separator \"..\" occurred more than once in layer names! Cannot convert raster object")
  }

  # Now, we actually transform the variable columns
  df[, names(variable) := tstrsplit(df$variable, "..", fixed = TRUE)]
  df[, variable := NULL] # variable column is not needed anymore

  baseDims <- colnames(df)[!(colnames(df) %in% c("value", names(variable)))]
  setcolorder(df, neworder = c(baseDims, names(variable), "value"))
  setDF(df)
  out <- tidy2magpie(df, spatial = idVars, temporal = temporal)
  if (!is.null(geometry)) {
    names(geometry) <- getItems(out, dim = 1)
    attr(out, "geometry") <- geometry
  }
  if (!is.null(crs)) {
    attr(out, "crs") <- crs
  }
  return(out)
}

setMethod("as.magpie",
  signature(x = "RasterBrick"),
  function(x, unit = "unknown", temporal = NULL, ...) {
    return(.raster2magpie(x, unit = unit, temporal = temporal))
  }
)

setMethod("as.magpie",
  signature(x = "RasterStack"),
  function(x, unit = "unknown", temporal = NULL, ...) {
    return(.raster2magpie(x, unit = unit, temporal = temporal))
  }
)

setMethod("as.magpie",
  signature(x = "RasterLayer"),
  function(x, unit = "unknown", temporal = NULL, ...) {
    return(.raster2magpie(x, unit = unit, temporal = temporal))
  }
)

setMethod("as.magpie",
          signature(x = "SpatRaster"),
          function(x, unit = "unknown", temporal = NULL, ...) {
            return(.raster2magpie(x, unit = unit, temporal = temporal))
          })

setMethod("as.magpie",
          signature(x = "SpatVector"),
          function(x, unit = "unknown", temporal = NULL, spatial = NULL, ...) {
            return(.raster2magpie(x, unit = unit, temporal = temporal, spatial = spatial))
          })

#' ~~ Methods for Function as.array ~~
#'
#' ~~ Methods for function \code{as.array} ~~
#'
#'
#' @name as.array-methods
#' @aliases as.array-methods as.array,ANY-method as.array,magpie-method
#' @param x object which should be converted to an array
#' @docType methods
#' @section Methods: \describe{
#'
#' \item{list("signature(x = \"ANY\")")}{ standard as.array-method }
#'
#' \item{list("signature(x = \"magpie\")")}{ Conversion takes place just by
#' removing MAgPIE-object specific elements } }
#' @keywords methods
#' @family MAgPIE-Conversions
#' @importFrom methods as setMethod signature
#' @exportMethod as.array

setMethod("as.array",
  signature(x = "magpie"),
  function(x) {
    return(as(x, "array"))
  }
)

#' mbind
#'
#' Merges MAgPIE-objects with identical structure in two dimensions. If data
#' differs in the temporal or spatial dimension each year or region/cell must
#' appear only once!
#'
#' @param ... MAgPIE objects or a list of MAgPIE objects that should be merged.
#' @return The merged MAgPIE object
#' @author Jan Philipp Dietrich, Misko Stevanovic
#' @seealso \code{"\linkS4class{magpie}"}
#' @examples
#'
#' m <- new.magpie(c("AFR", "CPA", "EUR"), c(1995, 2005), "Data1", fill = c(1, 2, 3, 4, 5, 6))
#' ms <- dimSums(m, dim = 3.1)
#' mbind(m, ms)
#' my <- new.magpie(getRegions(m), 2010, getNames(m), fill = c(6, 6, 4))
#' mbind(m, my)
#' md <- new.magpie(getRegions(m), getYears(m), "Data2", fill = c(7, 6, 5, 7, 8, 9))
#' mbind(m, md)
#'
#' pop <- maxample("pop")
#' a <- mbind(pop, pop)
#' dim(pop)
#' dim(a)
#' @family DataBinding
#' @importFrom methods new
#' @importFrom abind abind
#' @export
mbind <- function(...) { # nolint: cyclocomp_linter.
  inputs <- list(...)
  if (length(inputs) == 1 && is.list(inputs[[1]])) {
    inputs <- inputs[[1]]
  }
  # Remove NULL elements from list
  for (i in rev(seq_along(inputs))) {
    if (is.null(inputs[[i]])) {
      inputs[[i]] <- NULL
    } else if (prod(dim(inputs[[i]])) == 0) {
      inputs[[i]] <- NULL
      warning("You are trying to mbind an empty magclass object. Is that really intended?")
    }
  }

  # if all inputs are NULL, return NULL
  if (0 == length(inputs)) {
    return(NULL)
  }

  # store total number of elements to ensure that they remain unchanged
  nElems <- sum(vapply(inputs, length, integer(1)))

  cells <- NULL
  elems <- NULL
  years <- NULL
  diffspat <- FALSE
  difftemp <- FALSE
  diffdata <- FALSE
  for (i in seq_along(inputs)) {
    if (!is.magpie(inputs[[i]])) {
      stop("Inputs must all be MAgPIE-objects")
    }
    for (j in 1:3) {
      if (is.null(dimnames(inputs[[i]])[[j]])) {
        dimnames(inputs[[i]])[[j]] <- paste("dummy", c("", seq_len(dim(inputs[[i]])[j] - 1)), sep = "")
      }
    }
    # Check which dimensions differ
    diffspat <- !setequal(dimnames(inputs[[1]])[[1]], dimnames(inputs[[i]])[[1]])
    difftemp <- !setequal(dimnames(inputs[[1]])[[2]], dimnames(inputs[[i]])[[2]])
    diffdata <- !setequal(dimnames(inputs[[1]])[[3]], dimnames(inputs[[i]])[[3]])

    years <- c(years, getYears(inputs[[i]]))
    elems <- c(elems, getNames(inputs[[i]]))
    cells <- c(cells, getCells(inputs[[i]]))
    if (!diffspat && ncells(inputs[[1]]) > 1) {
      inputs[[i]] <- inputs[[i]][getCells(inputs[[1]]), , ]
    }
    if (!difftemp && nyears(inputs[[1]]) > 1) {
      inputs[[i]] <- inputs[[i]][, getYears(inputs[[1]]), ]
    }
    if (!diffdata &&  ndata(inputs[[1]]) > 1) {
      inputs[[i]] <- inputs[[i]][, , getNames(inputs[[1]])]
    }
  }

  if (!(length(grep(".", cells, fixed = TRUE)) %in% c(0, length(cells)))) {
    stop("Mixture of regional (no cell numbers) and cellular (with cell numbers)",
         " data objects! Cannot handle this case!")
  }

  diffall <- c("spatial" = diffspat, "temporal" = difftemp, "data" = diffdata)
  if (1 < sum(diffall)) {
    msg <- c("Cannot handle objects!",
             paste(paste(head(names(diffall)[diffall], n = -1),
                         collapse = ", "),
                   tail(names(diffall)[diffall], n = 1),
                   sep = " as well as "),
             "dimensions differ!")
    substr(msg[2], 1, 1) <- toupper(substr(msg[2], 1, 1))

    msg <- paste(c(paste(msg, collapse = " "),
                   "  Differences from first mbind() input:",
                   vapply(which(diffall), function(i) {

                     lhs <- getItems(inputs[[1]], dim = i)
                     rhs <- Reduce(union, lapply(tail(inputs, n = -1),
                                                 getItems, dim = i))

                     # do a kind of three-way diff
                     diff <- list("missing" = setdiff(lhs, rhs),
                                  " having" = intersect(lhs, rhs),
                                  " adding" = setdiff(rhs, lhs))

                     diff <- lapply(diff, function(x) {
                       if (4 < length(x)) {
                         # shorten to four elements, like str() does
                         x <- c(paste0("`", x[1:4], "`"), "...")
                       } else if (0 < length(x)) {
                         x <- paste0("`", x, "`")
                       }
                       return(paste(x, collapse = ", "))
                     })

                     # paste diff
                     d <- nzchar(diff)
                     paste(c(paste0("  ", names(diffall)[i], ":"),
                             paste(paste0("      ", names(diff)[d], ":"),
                                   diff[d])),
                           collapse = "\n")
                   }, character(1))), collapse = "\n")
    stop(msg)
  }

  if (difftemp) {
    if (length(years) != length(unique(years))) {
      stop("Some years occur more than once! Cannot handle this case!")
    }
    output <- new("magpie", abind::abind(inputs, along = 2))
  } else if (diffspat) {
    if (length(cells) != length(unique(cells))) {
      stop("Some regions/cells occur more than once! Cannot handle this case!")
    }
    output <- new("magpie", abind::abind(inputs, along = 1))
  } else {
    tmp <- function(x) return(length(getNames(x, fulldim = TRUE)))
    tmp <- vapply(inputs, tmp, integer(1))
    if (length(unique(tmp)) > 1) {
      warning("mbind most likely returned an erronous magpie object due to",
              " different numbers of data subdimensions in inputs!")
    }
    output <- new("magpie", abind::abind(inputs, along = 3))
  }
  for (j in 1:3) {
    if (length(grep("^dummy[0-9]*$", getItems(output, dim = j), perl = TRUE)) == dim(output)[j]) {
      getItems(output, dim = j, raw = TRUE) <- NULL
    }
  }
  names(dimnames(output)) <- names(dimnames(inputs[[1]]))

  if (length(output) != nElems) {
    stop("Invalid object (number of values changed during mbind). Does the data contain duplicates?")
  }
  return(output)
}

#' dimCode
#'
#' Function converts a dimension name or number to a dimension Code used for
#' MAgPIE objects
#'
#'
#' @param dim A vector of dimension numbers or dimension names which should be
#' translated
#' @param x MAgPIE object in which the dimensions should be searched for.
#' @param missing Either a value to which a dimension should be set in case
#' that it is not found (default is 0), or "stop" indicating that the function
#' should throw an error in these cases.
#' @param strict if set to TRUE also properly set dimension names which refer
#' to non-existing subdimensions will be treated as missing, otherwise these
#' dimension codes will be returned, even if the subdimension does not exist
#' @param sep  A character separating joined dimension names
#' @return A dimension code identifying the dimension. Either a integer which
#' represents the main dimensions (1=spatial, 2=temporal, 3=data) or a numeric,
#' representing the subdimensions of a dimension (e.g. 3.2 for the second data
#' dimension).
#' @author Jan Philipp Dietrich, Kristine Karstens
#' @seealso \code{\link{mselect}}, \code{\link{getDim}}
#' @examples
#'
#' pop <- maxample("pop")
#' dimCode(c("t", "scenario", "blablub"), pop)
#' @family DimensionManipulation
#' @export
dimCode <- function(dim, x, missing = 0, strict = FALSE, sep = ".") {

  if (all(is.character(dim)) && any(grepl(sep, dim, fixed = TRUE))) {
    stop("Dimension separator must not be used in dimension name!")
  }

  # function to translate dim to dim code
  if (is.character(dim)) {

    # get super dims and initialize dim number array
    set     <- getSets(x, fulldim = FALSE, sep = sep)
    dnames  <- dim
    dim     <- numeric()

    # loop over all entries for dim
    for (i in seq_along(dnames)) {

      # get superdim and check for appearance and uniqueness
      superdim    <- grep(paste0("(\\.|^)", dnames[i], "(\\.|$)"), set)

      if (length(superdim) > 1) {
        stop("One or more elements were found more than once in x!")
      } else if (length(superdim) == 0) {
        dim[i] <- 0
      } else {

        # split in subdims and calculate dimension Code
        tmp      <- unlist(strsplit(set[superdim], split = sep, fixed = TRUE))

        if (length(tmp) > 1) {
          subdim   <- grep(paste0("^", dnames[i], "($)"), tmp)
          if (length(subdim) > 1) stop("One or more elements were found more than once or not at all in x!")
        } else {
          subdim <- 0
        }

        dim[i]   <- as.numeric(superdim + subdim / 10)
      }
    }

    names(dim) <- dnames
  }

  if (strict) dim[!dimExists(dim, x)] <- 0

  # check for errors and set "missing"
  if (any(dim >= 4) || any(dim < 1)) {
    if (missing == "stop") {
      stop("illegal dimension. Use either dimension 1, 2, or 3, or address subdimensions via 3.1, 3.2, ...")
    }
    dim[dim >= 4 | dim < 1] <- missing
  }

  return(dim)
}

#' @importFrom methods Ops callGeneric new
#' @exportMethod Ops
setMethod(Ops, signature(e1 = "magpie", e2 = "magpie"),
  function(e1, e2) {
    e2 <- magpie_expand(e2, e1)
    e1 <- magpie_expand(e1, e2)
    # if one object is empty, return it without additional calculations
    if (length(e1) == 0) return(e1)
    if (length(e2) == 0) return(e2)
    if (any(unlist(dimnames(e1)) != unlist(dimnames(e2)))) {
      stop("MAgPIE objects after MAgPIE object expansion do not agree in dimnames! magpie_expand seems to be bugged!")
    }
    return(new("magpie", callGeneric(e1@.Data, e2@.Data)))
  }
)

setMethod(Ops, signature(e1 = "magpie", e2 = "numeric"),
  function(e1, e2) {
    return(new("magpie", callGeneric(e1@.Data, e2)))
  }
)

setMethod(Ops, signature(e1 = "numeric", e2 = "magpie"),
  function(e1, e2) {
    return(new("magpie", callGeneric(e1, e2@.Data)))
  }
)

#' Get cells per region
#'
#' Counts how often each element of the provided subdimension exists in the given
#' data set. Originally created to count the number of cells in a region (this
#' is also where its name originates from) it can now be used to count elements
#' of any subdimension via the dim argument.
#'
#' @param x MAgPIE object or a resolution written as numeric (currently only
#' data for 0.5 degree resolution is available).
#' @param dim Dimension for which the items should be returned. Either number or
#' name of dimension or a vector of these (in case of a vector all subimensions
#' must belong to the same main dimension!). See \code{\link{dimCode}} for more details.
#' @return cells per region
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getRegions}}, \code{\link{read.magpie}},
#' \code{\link{write.magpie}}
#' @examples
#'
#' getCPR(0.5)
#' a <- maxample("animal")
#' getCPR(a, dim = "color")
#' getCPR(a, dim = 3.2)
#' getCPR(a, dim = "country")
#' getCPR(a, dim = c("color", "species"))
#' @family GetterSetter
#' @export

getCPR <- function(x, dim = 1.1) {
  if (is.magpie(x)) {
    dim <- dimCode(dim, x)
    if (length(dim) != 1 && length(unique(floor(dim))) > 1) {
      stop("Selected subdimensions must belong to the same main dimension!")
    }
    items <- getItems(x, dim = dim, full = TRUE)
  } else if (x == 0.5) {
    items <- magclassdata$half_deg$region
  } else {
    stop("Cannot extract cpr information for x!")
  }
  table <- table(items)
  names(dimnames(table)) <- NULL
  return(table)
}

#'
#' @title pmin/pmax
#' @name magpie-pmin-pmax
#' @aliases pmin pmax pmin,magpie-method pmin,ANY-method pmax,magpie-method pmax,ANY-method
#' @description magclass-aware pmin/pmax, which calculate the parallel minima/maxima of the input values
#' @details
#' \code{\link[base]{pmin}}/\code{\link[base]{pmax}} do not care about the order of items in a magclass
#' object's dim. magclass pmin/pmax reorder items in each dim so they are in the same order.
#' @param ... Multiple magpie objects with dimensions that each are either the same or only
#' contain one or no item among all objects.
#' @param na.rm Passed to \code{\link[base]{pmin}}
#' @return A new magpie object that contains the minimum/maximum of values between all the passed magpie objects.
#' @usage NULL
#' @docType methods
#' @keywords methods
#' @family Comparison
#' @author Pascal Sauer, Patrick Rein
NULL

#
# Start Preamble to set up generic pmin
#
existingPminGeneric <- getGeneric("pmin")
if (!is.null(existingPminGeneric)) {
  if (any(existingPminGeneric@signature != "...")) {
    warning("Redefining pmin generic function with new signature \"...\", ",
            "while there already is a pmin generic with signature: ",
            paste0(existingPminGeneric@signature, collapse = ", "))
  }
}
setGeneric("pmin", signature = "...")
#
# End Preamble
#

#' @rdname magpie-pmin-pmax
#' @exportMethod pmin
setMethod("pmin", "magpie", function(..., na.rm = FALSE) { # nolint: object_name_linter.
  return(withAlignedDims(function(...) base::pmin(..., na.rm = na.rm),
                         "pmin",
                         ...))
})

#' new.magpie
#'
#' Creates a new MAgPIE object
#'
#'
#' @param cells_and_regions Either the region names (e.g. "AFR"), or the cells
#' (e.g. 1:10), or both in combination (e.g. "AFR.1"). NULL means no spatial
#' element.
#' @param years dimnames for years in the format "yXXXX" or as integers. NULL
#' means one year which is not further specified
#' @param names dimnames for names. NULL means one data element which is not
#' further specified
#' @param fill Default value for the MAgPIE object
#' @param sort Bolean. Decides, wheher output should be sorted or not.
#' @param sets A vector of dimension names. See \code{\link{getSets}} for more
#' information.
#' @param unit deprecated
#' @return an empty magpie object filled with fill, with the given dimnames
#' @author Benjamin Bodirsky, Jan Philipp Dietrich
#' @seealso \code{\link{as.magpie}}
#' @examples
#'
#' a <- new.magpie(1:10, 1995:2000)
#' b <- new.magpie(c("AFR", "CPA"), "y1995", c("bla", "blub"), sets = c("i", "t", "value"))
#' c <- new.magpie()
#' @family ObjectCreation
#' @export
#' @importFrom methods new
new.magpie <- function(cells_and_regions = "GLO", # nolint: object_name_linter.
                       years = NULL, names = NULL, fill = NA, sort = FALSE, sets = NULL, unit = NULL) {
  cellsAndRegions <- cells_and_regions
  if (!is.null(unit)) warning("Argument \"unit\" is deprecated and will be ignored!")
  ncells <- length(cellsAndRegions)
  nyears <- ifelse(is.null(years), 1, length(years))
  ndata  <- ifelse(is.null(names), 1, length(names))
  if (all(!grepl("\\.", cellsAndRegions)) && !is.null(cellsAndRegions)) {
    if (all(is.numeric(cellsAndRegions))) {
      cellsAndRegions <- paste("GLO", cellsAndRegions, sep = ".")
    }
  }
  if (all(is.numeric(years)) && all(nchar(years) <= 4)) {
    years <- gsub(" ", "0", format(years, width = 4))
    years <- paste("y", years, sep = "")
  }

  object <- new("magpie", array(fill, dim = c(ncells, nyears, ndata)))
  getItems(object, dim = 1, raw = TRUE) <- as.vector(cellsAndRegions)
  getItems(object, dim = 2, raw = TRUE) <- as.vector(years)
  getItems(object, dim = 3, raw = TRUE) <- as.vector(names)
  names(dimnames(object)) <- NULL
  if (sort) object <- magpiesort(object)
  object <- clean_magpie(object, "sets")
  if (!is.null(sets)) getSets(object) <- sets
  return(object)
}

context("Test replace_non_finite()")

test_that(
  desc = "all instances of non-finite data are replaced",
  code = {
    a <- new.magpie(letters[1:5], years = "y1995", names = "foo")
    b <- new.magpie(letters[1:5], years = "y1995", names = "foo", fill = 0)
    a[, , ] <- c(0, NA, NaN, -Inf, Inf)

    expect_equal(object = replace_non_finite(a), b)
  }
)

#' Summation over dimensions
#'
#' This function sums over any (sub-)dimension of a magpie object
#'
#' @param x A MAgPIE-object
#' @param dim The dimensions(s) to sum over. A vector of dimension codes or dimension names.
#' See \code{\link{dimCode}} for more information
#' @param na.rm logical. Should missing values (including NaN) be omitted from
#' the calculations?
#' @return A MAgPIE object with values summed over the specified dimensions
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{rowSums}}, \code{\link{getItems}}, \code{\link{dimCode}}
#' @examples
#' a <- maxample("animal")
#' dimSums(a, dim = c(1, 2, 3.2))
#' dimSums(a, dim = c("x", "y", "cell", "month"))
#' @family Aggregation
#' @export
dimSums <- function(x, dim = 3, na.rm = FALSE) { # nolint: object_name_linter.
  return(magpply(X = x, FUN = sum, DIM = dim, na.rm = na.rm))
}

#' Get dataset names
#'
#' Extracts dataset names of a MAgPIE-object
#'
#' setNames is a shortcut to use a MAgPIE object with manipulated data names.
#' The setNames method uses the variable names "object" and "nm" in order to be
#' consistent to the already existing function setNames.
#'
#' @aliases getNames getNames<-
#' @param x MAgPIE object
#' @param fulldim specifies, how the object is treated. In case of FALSE, it is
#' assumed that x is 3 dimensional and dimnames(x)[[3]] is returned. In case of
#' TRUE, the dimnames of the real third dimension namesare returned
#' @param dim Argument to choose a specific data dimension either by name of
#' the dimension or by number of the data dimension.
#' @param value a vector of names current names should be replaced with. If
#' only one data element exists you can also set the name to NULL.
#' @return getNames returns data names of the MAgPIE-object, whereas setNames
#' returns the MAgPIE object with the manipulated data names.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{setNames-methods}}, \code{\link{getRegions}}, \code{\link{getYears}},
#' \code{\link{getCPR}}, \code{\link{read.magpie}},
#' \code{\link{write.magpie}},\code{\link{ndata}},
#' \code{"\linkS4class{magpie}"}
#' @examples
#'
#' a <- as.magpie(1)
#' getNames(a)
#' setNames(a, "bla")
#'
#' x <- new.magpie("GLO", 2000, c("a.o1", "b.o1", "a.o2"))
#' getNames(x, dim = 2)
#'
#' getSets(x, fulldim = FALSE)[3] <- "bla.blub"
#' getNames(x, dim = "bla")
#'
#' getSets(x)[4] <- "ble"
#' getNames(x, dim = "ble") <- c("Hi", "Bye")
#' x
#' @family GetterSetter
#' @export
getNames <- function(x, fulldim = FALSE, dim = NULL) {
  return(getItems(x, dim = .convertDim(dim), split = fulldim)) # nolint: undesirable_function_linter.
}

#' @describeIn getNames set names
#' @export
`getNames<-` <- function(x, dim = NULL, value) {
  if (is.null(dim) && is.null(value) && dim(x)[3] > 1) stop("Cannot unset names!")
  return(setItems(x, dim = .convertDim(dim), value = value, raw = TRUE))
}

.convertDim <- function(dim) {
  if (is.null(dim)) {
    dim <- 3
  } else if (is.numeric(dim)) {
    if (length(dim) != 1 || !is.element(dim, 1:9)) stop("Unsupported dim selection!")
    dim <- 3 + dim / 10
  }
  return(dim)
}

#' getComment
#'
#' Extracts the comment from a MAgPIE-object
#'
#'
#' @aliases getComment getComment<- setComment
#' @param x,object MAgPIE object
#' @param value,nm A vector containing the comment.
#' @return getComment returns the comment attached to a MAgPIE-object, NULL if
#' no comment is present. setComment returns the magpie object with the
#' modified comment.
#' @author Markus Bonsch
#' @seealso \code{\link{getRegions}}, \code{\link{getNames}},
#' \code{\link{getYears}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
#' \code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}
#' @examples
#'
#' a <- as.magpie(1)
#'  # returns NULL
#'  getComment(a)
#'  # set the comment
#'  getComment(a) <- c("bla", "blubb")
#'  getComment(a)
#'
#' @family GetterSetter
#' @export
getComment <- function(x) {
  return(attr(x, "comment"))
}

#' @describeIn getComment set comment
#' @export
`getComment<-` <- function(x, value) {
  attr(x, "comment") <- value
  return(x)
}

#' addDim
#'
#' This function adds a named dimension as dimension number "dim" with
#' the name "dimName" with an empty data column with the name "item".
#'
#' @param x MAgPIE object which should be extended.
#' @param dim The dimension number of the new dimension (e.g. 3.1)
#' @param dimName The name of the new dimension
#' @param item One or more names of items in the new dimension. If more than one
#' is given, behavior depends on the expand argument.
#' @param expand If TRUE, each item from the item argument is added to each item
#' already present, resulting in e.g. `c("A.d1", "B.d1", "A.d2", "B.d2")`.
#' Otherwise, length of item must equal the number of items already present and
#' they are simply added, resulting in e.g. `c("A.d1", "B.d2")`.
#'
#' @return The extended MAgPIE object
#' @author Jan Philipp Dietrich, Benjamin Bodirsky, Pascal Sauer
#' @seealso \code{\link{add_columns}},\code{\link{mbind}}
#' @examples
#'
#' a <- maxample("animal")
#' str(addDim(a, dim = 3.2))
#' str(addDim(a, dim = 2.3, item = paste0("d", 1:3)))
#' @family DimensionManipulation
#' @export
addDim <- function(x, dim = 3.1, dimName = NULL, item = "dummy", expand = TRUE) {
  x <- clean_magpie(x, what = "sets")
  if (is.null(dimName)) {
    # create non-existing variant of dimension name starting with "new"
    sets <- getSets(x, fulldim = TRUE)
    dimName <- tail(make.unique(c(sets, "new"), sep = ""), 1)
  } else if (dimName %in% getSets(x, fulldim = TRUE)) {
    stop("Dimension \"", dimName, "\" does already exist. Please use a different name!")
  }
  maindim <- floor(dim)
  subdim  <- as.integer(sub("^.\\.", "", as.character(dim)))
  if (length(item) > 1) {
    if (expand) {
      expansion <- rep(seq_len(dim(x)[maindim]), length(item))
      x <- x[expansion, dim = maindim]
      item <- rep(item, each = dim(x)[maindim] / length(item))
    } else if (length(item) != dim(x)[maindim]) {
      stop("length(item) != number of items already present in x in dim ", maindim,
           "; actual numbers: ", length(item), "!=", dim(x)[maindim])
    }
  }
  if (is.null(getItems(x, dim = maindim))) {
    getItems(x, dim = maindim, raw = TRUE) <- item
    getSets(x, fulldim = FALSE)[maindim] <- dimName
  } else if (subdim == 1) {
    getItems(x, dim = maindim, raw = TRUE) <- paste0(item, ".", getItems(x, dim = maindim, full = TRUE))
    getSets(x, fulldim = FALSE)[maindim] <- paste0(dimName, ".", getSets(x, fulldim = FALSE)[maindim])
  } else if (subdim > ndim(x, maindim)) {
    getItems(x, dim = maindim, raw = TRUE) <- paste0(getItems(x, dim = maindim, full = TRUE), ".", item)
    getSets(x, fulldim = FALSE)[maindim] <- paste0(getSets(x, fulldim = FALSE)[maindim], ".", dimName)
  } else {
    # this else branch can solve any case, the previous 3 are just faster implementations for common special cases
    items <- getItems(x, dim = maindim, split = TRUE, full = TRUE)
    before <- seq_len(subdim - 1)
    after <- setdiff(seq_along(items), before)
    items <- c(items[before], list(item), items[after])
    names(items)[subdim] <- dimName
    items <- Filter(Negate(is.null), items)
    getItems(x, dim = maindim, raw = TRUE) <- do.call(function(...) paste(..., sep = "."), items)
    getSets(x, fulldim = FALSE)[maindim] <- paste(names(items), collapse = ".")
  }
  return(x)
}

#' add_dimension
#'
#' This function adds a named dimension as dimension number "dim" with
#' the name "add" with an empty data column with the name "nm".
#' Please start using this function's successor \code{\link{addDim}}. add_dimension
#' might eventually become deprecated.
#'
#' @inherit addDim
#' @param add The name of the new dimension
#' @param nm One or more names of items in the new dimension.
#' @export
add_dimension <- function(x, dim = 3.1, add = NULL, nm = "dummy", expand = TRUE) { # nolint: object_name_linter.
  return(addDim(x = x, dim = dim, dimName = add, item = nm, expand = expand))
}

#' Get sets
#'
#' Extracts sets of a MAgPIE-object if available
#'
#'
#' @aliases getSets getSets<-
#' @param x MAgPIE object
#' @param sep A character separating joined dimension names
#' @param fulldim bool: Consider dimension 3 as a possible aggregate of more
#' dimensions (TRUE) or stick to it as one dimension (FALSE)
#' @param value A vector with set names you want to replace the current set
#' names of the object with.
#' @return Sets of the MAgPIE-object. If no information about contained sets is
#' available NULL
#' @author Markus Bonsch, Jan Philipp Dietrich
#' @seealso \code{\link{getRegions}},
#' \code{\link{getNames}},\code{\link{getYears}}, \code{\link{getCPR}},
#' \code{\link{read.magpie}}, \code{\link{write.magpie}},
#' \code{"\linkS4class{magpie}"}
#' @examples
#'
#' a <- new.magpie("GLO.1", 2000, c("a.o1", "b.o1", "a.o2"))
#'  getSets(a) <- c("reg", "cell", "t", "bla", "blub")
#'  getSets(a)
#'
#'  getSets(a)["d3.1"] <- "BLA"
#'  getSets(a, fulldim = FALSE)
#'  getSets(a)
#' @family GetterSetter
#' @export
getSets <- function(x, fulldim = TRUE, sep = ".") {
  out <- names(dimnames(x))[drop = FALSE]
  if (is.null(out)) return(NULL)

  if (fulldim) { # nolint: undesirable_function_linter.
    tmp <- strsplit(out, split = sep, fixed = TRUE)
    tmp <- lapply(tmp, FUN = function(x) {
      if (length(x) == 0) {
        x <- NA
      }
      return(x)
    })
    addDimCode <- function(x) {
      names(x) <- seq_along(x)
      return(x)
    }
    tmp <- lapply(tmp, addDimCode)
    names(tmp) <- paste0("d", seq_along(tmp))

    out <- unlist(tmp)
  }
  return(out)
}



#' @describeIn getSets replace set names
#' @export
`getSets<-` <- function(x, fulldim = TRUE, sep = ".", value) {
  fullDim <- fulldim # nolint: undesirable_function_linter.
  uncleanSets <- getSets(x)
  x <- clean_magpie(x, what = "sets")
  if (is.null(value)) {
    return(x)
  }
  if (is.null(names(dimnames(x))) || length(value) %in% c(0, 3)) {
    fullDim <- FALSE
  }
  if (!fullDim) {
    names(dimnames(x)) <- value
    return(x)
  } else {
    sNow <- getSets(x, fulldim = TRUE)
    if (length(value) != length(sNow)) {
      if (length(value) != length(uncleanSets)) {
        stop("Input length does not agree with the number of sets in x!")
      }
      # clean value
      oldValue <- value
      names(oldValue) <- names(uncleanSets)
      value <- sNow
      value[names(oldValue)] <- oldValue
    }
    mainDim <- as.integer(substring(names(sNow), 2, 2))
    sNew <- NULL
    for (i in 1:3) {
      sNew <- c(sNew, paste(value[mainDim == i], collapse = sep))
    }
    getSets(x, fulldim = FALSE, sep = sep) <- sNew
    return(x)
  }
}

#' dimReduce
#'
#' Remove dimensions which contain identical data for all elements in it
#'
#'
#' @param x MAgPIE object which should be reduced
#' @param dim_exclude Vector with names of dimensions which must not be reduced
#' @return The reduced MAgPIE object
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{addDim}}
#' @note This function has some similarities to \code{\link{collapseDim}}, but
#' serves a different purpose. While \code{\link{collapseDim}} only removes
#' dimensions which contain only a single element or which it is
#' specifically told to remove, \code{\link{dimReduce}} looks whether the
#' entries of a multi-entry dimension are all the same and removes dimensions
#' for which this is the case. In some cases both will lead to the same result
#' but in many other cases the results will differ.
#' @examples
#'
#' # create data with 5 identical scenarios
#' p <- addDim(maxample("pop")[1:3, 1:3, ], item = paste0("scen", 1:2))
#' str(p)
#' str(dimReduce(p))
#'
#' # set years to same value
#' p[, , ] <- setYears(p[, 1, ], NULL)
#' str(p)
#' str(dimReduce(p))
#'
#' # set regions to same value
#' p[, , ] <- setCells(p[1, , ], "GLO")
#' str(p)
#' str(dimReduce(p))
#' @family DimensionManipulation
#' @export
dimReduce <- function(x, dim_exclude = NULL) { # nolint: object_name_linter.
  x <- collapseDim(x)
  x <- clean_magpie(x)
  fd <- unlist(getItems(x, split = TRUE), recursive = FALSE)
  for (s in setdiff(getSets(x), dim_exclude)) {
    tmp <- list()
    tmp[[s]] <- fd[[s]][1]
    xSingle <- mselect(x, tmp)
    # same information in all dimension entries?
    if (all(x - xSingle  == 0, na.rm = TRUE)) {
      x <- xSingle
    }
  }
  x <- collapseDim(x)
  if (dim(x)[1] == 1) getCells(x) <- "GLO"
  return(x)
}

# full import of data.table package is required as there seems to be no other option
# to include the data.table method for the duplicated function (which is considerably faster
# than the data.frame counterpart!)
#' @import data.table

.duplicates_check <- function(coord) { # nolint: object_name_linter.
  duplicates <- duplicated(as.data.table(coord))
  if (any(duplicates)) {
    coord <- as.data.frame(coord)
    warning("Duplicate entries found, only the last entry will be used (duplicate entries: ",
            paste(apply(rbind(NULL, unique(coord[duplicates, ])),
                        1, paste, collapse = "|"), collapse = ", "), ")!")
  }
}

#' cbind method for MAgPIE objects
#'
#' cbind method for MAgPIE-objects. Issues a warning that mbind should be used
#' if all objects are magpie objects and continues afterwards.
#'
#' @aliases cbind.magpie
#' @param ... see \link[base]{cbind}
#' @param deparse.level see \link[base]{cbind}
#' @return A matrix as if cbind was applied to the internal vector of the MAgPIE objects
#' @author Patrick Rein
#' @family DataBinding
#' @export

cbind.magpie <- function(..., deparse.level = 1) { # nolint: object_name_linter.
  warning("Using cbind on magpie objects is not adviced as it is not merging based on element names. Use mbind instead.") # nolint: line_length_linter.
  objectsToCbind <- lapply(list(...), function(o) {
    if (inherits(o, "magpie")) {
      # Unwrap MAgPie objects so we only pass the internal vectors
      # to the cbind below
      return(o@.Data)
    } else {
      return(o)
    }
  })
  return(do.call(cbind, objectsToCbind, deparse.level))
}

#' sameDims
#'
#' Checks whether magpie objects have the same items in every dimension (ignoring order).
#'
#' @param ... Various magpie objects
#'
#' @returns A logical indicating whether the magpie objects have the same items in every dimension.
#' @author Pascal Sauer, Patrick Rein
#' @family ObjectInfo
#' @export
sameDims <- function(...) {
  magpies <- list(...)
  if (!all(vapply(magpies, is.magpie, logical(1)))) {
    stop("sameDims cannot handle non-magpie objects")
  }

  if (length(magpies) < 2) {
    return(TRUE)
  }

  .allSetequal <- function(vs) all(vapply(vs[-1], function(x) setequal(vs[[1]], x), logical(1)))
  return(.allSetequal(lapply(magpies, function(m) getItems(m, 1))) &&
           .allSetequal(lapply(magpies, function(m) getItems(m, 2))) &&
           .allSetequal(lapply(magpies, function(m) getItems(m, 3))))
}

#' magpie_expand_dim
#'
#' Expands a single MAgPIE object dimension
#'
#' Expansion means here that the dimensions of x are expanded acordingly to
#' ref. Please note that this is really only about expansion. In the case that
#' one dimension of ref is smaller than of x nothing happens with this
#' dimension. At the moment magpie_expand is only internally available in the
#' magclass library
#'
#' In contrast to \code{\link{magpie_expand}} this function is expanding only a single
#' dimension. It is meant as a support function for \code{\link{magpie_expand}} itself.
#'
#' @param x MAgPIE object that should be expanded
#' @param ref MAgPIE object that serves as a reference
#' @param dim dimension that should be expanded
#' @return An expanded version of x.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{as.magpie}}, \code{\link[base]{options}}
#' @examples
#' d <- new.magpie(c("AFR.BLUB.1", "AFR.BLUB.2", "EUR.BLUB.1",
#'                     "AFR.BLA.1", "AFR.BLA.2", "EUR.BLA.1"), fill = 1)
#'  getSets(d)[1:3] <- c("reg", "b", "i")
#'  e <- new.magpie(c("BLA.AFR.A", "BLA.EUR.A", "BLUB.AFR.A", "BLUB.EUR.A",
#'                     "BLA.AFR.B", "BLA.EUR.B", "BLUB.AFR.B", "BLUB.EUR.B"), fill = 2)
#'  getSets(e)[1:3] <- c("b", "reg", "a")
#'  magclass:::magpie_expand_dim(d, e, dim = 1)
magpie_expand_dim <- function(x, ref, dim = 1) { # nolint: object_name_linter.

  if (!(dim %in% seq_len(3))) {
    stop("Unsupported dim setting (dim = ", dim, ")")
  }

  dimnames2df <- function(x, dim = 1) {
    splits <- strsplit(dimnames(x)[[dim]], ".", fixed = TRUE)
    xd <- data.frame(do.call(rbind, splits), stringsAsFactors = TRUE)

    rownames(xd) <- NULL
    if (!is.null(names(dimnames(x)))) {
      tmp <- strsplit(names(dimnames(x))[dim], ".", fixed = TRUE)[[1]]
      if (length(tmp) == ncol(xd)) {
        names(xd) <- tmp
      }
    }
    xd$".line" <- seq_len(nrow(xd))
    return(xd)
  }

  df2dimnames <- function(x) {
    cols <- grep("^\\.", names(x), invert = TRUE, value = TRUE)
    tmp <- lapply(x[cols], as.character)
    out <- do.call("paste", c(tmp, sep = "."))
    return(list(dimnames = out, name = paste(sub("\\.", "", cols), collapse = ".")))
  }

  dx <- dimnames2df(x, dim = dim)
  dref <- dimnames2df(ref, dim = dim)

  # detect matching columns
  if (!isTRUE(getOption("magclass_setMatching"))) {
    names(dref)[is.na(names(dref))] <- "NA"
    names(dx)[is.na(names(dx))] <- "NA"
    lx <- lapply(dx[names(dx) != ".line"], levels)
    lref <- lapply(dref[names(dref) != ".line"], levels)
    if (anyDuplicated(lx) == 0 && anyDuplicated(lref) == 0) {
      # matching will be based on dimension content rather
      # than set names (in case of duplicated columns,
      # set matching will be used instead)

      # temporarily split .line col from rest
      drefLine <- dref[".line"]
      dxLine   <- dx[".line"]
      dref[".line"] <- NULL
      dx[".line"] <- NULL
      # ensure unique set names
      tmp <- make.unique(c(names(dref), names(dx)), sep = "")
      names(dref) <- tmp[seq_len(ncol(dref))]
      names(dx) <- tmp[(ncol(dref) + 1):length(tmp)]
      m <- match(lx, lref)
      tmp <- names(dref)[m]
      tmp[is.na(tmp)] <- names(dx)[is.na(tmp)]
      names(dx) <- tmp
      dref[".line"] <- drefLine
      dx[".line"] <- dxLine
    }
  }

  m <- merge(dref, dx, sort = FALSE, suffixes = c("_ref", "_x"),
             by = setdiff(intersect(names(dref), names(dx)), ".line"),
             all.x = TRUE, all.y = TRUE)

  if (anyNA(c(m$.line_ref, m$.line_x))) {
    stop(" Identical set names but different content. Correct set names!")
  }

  # reorder m so that dref columns appear first
  m <- m[union(names(dref)[seq_len((ncol(dref) - 1))], names(m))]
  m <- m[order(m$.line_ref), ]

  tmpdim <- dim(x)
  tmpdim[dim] <- nrow(m)
  sizeCheck(tmpdim)

  out <- x[m$".line_x", dim = dim]
  tmp <- df2dimnames(m)
  dimnames(out)[[dim]] <- tmp$dimnames
  names(dimnames(out))[dim] <- tmp$name
  return(out)
}

#' Get years
#'
#' Extracts years of a MAgPIE-object
#'
#' setYears is a shortcut to use a MAgPIE object with manipulated year names.
#' setYears uses the variable names "object" and "nm" in order to be consistent
#' to the already existing function setNames.
#'
#' @aliases getYears getYears<- setYears
#' @param x,object MAgPIE object
#' @param as.integer Switch to decide, if output should be the used year-name
#' (e.g. "y1995") or the year as integer value (e.g. 1995)
#' @param value,nm Years the data should be set to. Either supplied as a vector
#' of integers or a vector of characters in the predefined year format
#' ("y0000"). If only 1 year exist you can also set the name of the year to
#' NULL.
#' @return getYears returns years of the MAgPIE-object, whereas setYears
#' returns the MAgPIE object with the manipulated years.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getRegions}}, \code{\link{getNames}},
#' \code{\link{setNames}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
#' \code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}
#' @examples
#'
#' a <- as.magpie(1)
#' getYears(a)
#'  setYears(a, 1995)
#'
#' @family GetterSetter
#' @export
getYears <- function(x,
                     as.integer = FALSE) { # nolint: object_name_linter.
  if (as.integer) {
    return(as.integer(sub("^y", "", dimnames(x)[[2]])))
  } else {
    return(dimnames(x)[[2]])
  }
}

#' @describeIn getYears rename years
#' @export
`getYears<-` <- function(x, value) {
  if (!is.null(value)) if (length(value) != nyears(x)) stop("Wrong number of years supplied!")
  if (nyears(x) == 0) return(x)
  if (is.null(value) && nyears(x) != 1) {
    stop("Setting years to NULL is not possible as the number of years is not 1!")
  }
  if (is.null(value)) {
    tmp <- list(NULL, NULL, NULL)
    if (!is.null(dimnames(x)[[1]])) tmp[[1]] <- dimnames(x)[[1]]
    if (!is.null(dimnames(x)[[3]])) tmp[[3]] <- dimnames(x)[[3]]
    names(tmp) <- names(dimnames(x))
    dimnames(x) <- tmp
  } else {
    if (all(is.numeric(value))) value <- gsub(" ", "0", format(value, width = 4))
    if (all(nchar(value) == 4)) value <- paste("y", value, sep = "")
    if (any(nchar(value) != 5) || any(substr(value, 1, 1) != "y")) {
      stop("Wrong year format. Please supply either integer values or years in the format y0000!")
    }
    dimnames(x)[[2]] <- value
  }
  return(x)
}

#' Write file in report format
#'
#' This function is deprecated, please use \code{\link{write.report}} instead.
#'
#'
#' @param ... arguments are forwarded to \code{\link{write.report}}
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{write.report}}
#' @family FileIO
#' @export
write.report2 <- function(...) { # nolint: object_name_linter.
  .Deprecated("write.report")
  return(write.report(...))
}

#' Copy Attributes
#'
#' This function copies attributes from one object and assigns them to another.
#'
#'
#' @aliases copy.attributes copy.attributes<-
#' @param from object from which the attributes should be taken
#' @param value Same as "from" (object from which the attributes should be
#' taken)
#' @param to object to which the attributes should be written
#' @param delete attributes which should not be copied. By default this are
#' class specific attributes which might cause problems if copied to another
#' object. But you can add or remove attributes from the vector.
#' @param delete2 Identical to delete and just added for convenience for the
#' case that you want to delete additional attributes but do not want to repeat
#' the vector given in delete. In the function both vectors, delete and
#' delete2, are just merged to one deletion vector.
#' @author Jan Philipp Dietrich
#' @examples
#'
#' from <- array(12)
#' attr(from, "blablub") <- "I am an attribute!"
#' attr(from, "blablub2") <- "I am another attribute!"
#'
#' print(attributes(from))
#'
#' to <- as.magpie(0)
#' print(attributes(to))
#'
#' copy.attributes(to) <- from
#' print(attributes(to))
#'
#' @family ObjectCreation
#' @export
copy.attributes <- function(from, # nolint: object_name_linter.
                            to, delete = c("names", "row.names", "class", "dim", "dimnames"), delete2 = NULL) {
  a <- attributes(from)
  a[c(delete, delete2)] <- NULL
  attributes(to) <- c(attributes(to), a)
  return(to)
}

#' @describeIn copy.attributes assign attributes from object "value"
#' @export
`copy.attributes<-` <- function(to, # nolint: object_name_linter.
                                delete = c("names", "row.names", "class", "dim", "dimnames"), delete2 = NULL, value) {
  return(copy.attributes(from = value, to = to, delete = delete, delete2 = delete2))
}

#' getDim
#'
#' Function which tries to detect the dimension to which the given elems belong
#'
#'
#' @param elems A vector of characters containing the elements that should be
#' found in the MAgPIE object
#' @param x MAgPIE object in which elems should be searched for.
#' @param fullmatch If enabled, only dimensions which match exactly the elements
#' provided will be returned. Otherwise, it is sufficient if elems contains a subset
#' of the dimension.
#' @param dimCode If enabled, the dimCode will be returned, otherwise the name
#' of the dimension.
#' @return The name or dimCode of the dimensions in which elems were found.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{mcalc}},\code{\link{dimCode}}
#' @examples
#'
#' pop <- maxample("pop")
#' getDim(c("AFR", "CPA"), pop)
#'  getDim(c("AFR", "CPA"), pop, fullmatch = TRUE)
#'  getDim(c("AFR", "CPA"), pop, dimCode = FALSE)
#'
#' @family GetterSetter
#' @export
getDim <- function(elems, x, fullmatch = FALSE, dimCode = TRUE) {

  tmpfun <- function(x, elems) {
    return(all(elems %in% x))
  }
  tmpfun2 <- function(x, elems, fullmatch, dimCode) {
    if (fullmatch) {
      tmp <- vapply(x, setequal, logical(1), elems)
    } else {
      tmp <- vapply(x, tmpfun, logical(1), elems)
    }
    if (dimCode) names(tmp) <- seq_along(tmp)
    return(tmp)
  }
  tmp <- getItems(x, split = TRUE)
  tmp2 <- lapply(tmp, tmpfun2, elems, fullmatch, dimCode)
  if (dimCode) {
    names(tmp2) <- seq_along(tmp2)
    tmp2 <- unlist(tmp2)
    out <- as.numeric(names(tmp2)[tmp2])
  } else {
    tmp2 <- unlist(tmp2)
    out <- names(tmp2)[tmp2]
  }
  if (length(out) == 0) {
    tmp <- getItems(x, split = FALSE)
    tmp2 <- tmpfun2(tmp, elems, fullmatch, dimCode)
    out <- names(tmp2)[tmp2]
    if (dimCode) out <- as.numeric(out)
  }
  return(out)
}

#' Replace Non-Finite Data
#'
#' Replaces all instances of non-finite data (`NA`, `NaN`, `Inf`, and `-Inf`).
#'
#' @param x A vector or [`magpie`][magclass::magclass] object.
#' @param replace A value to replace non-finite data with.
#'
#' @return A vector or [`magpie`][magclass::magclass] object, same as `x`.
#'
#' @author Michaja Pehl
#' @md
#'
#' @examples
#' part  <- new.magpie(letters[1:3], years = "y1995", names = "foo")
#' total <- new.magpie(letters[1:3], years = "y1995", names = "foo")
#'
#' part[, , ]  <- c(0, 1, 2)
#' total[, , ] <- c(0, 10, 10)
#'
#' part / total
#'
#' replace_non_finite(part / total)
#'
#' @family Utility
#' @export
replace_non_finite <- function(x, replace = 0) { # nolint: object_name_linter.
  x[!is.finite(x)] <- replace

  return(x)
}

readMagpieMZ <- function(fileName, compressed) {
  fileFunction <- if (compressed) gzfile else file
  zz <- withr::local_connection(fileFunction(fileName, "rb"))

  newestFformatVersion <- 6
  fformatVersion <- readBin(zz, integer(), 1, size = 2)
  if (fformatVersion > newestFformatVersion) {
    stop("File format is newer (v", fformatVersion, ") than the formats supported (v1-v",
         newestFformatVersion, "). ", "\nPlease update your magclass package to the newest version!")
  }
  ncharComment <- readBin(zz, integer(), 1, size = 4)
  empty <- 94
  if (fformatVersion > 2) {
    nbyteMetadata <- readBin(zz, integer(), 1, size = 4)
  }
  if (fformatVersion > 1) {
    ncharSets <- readBin(zz, integer(), 1, size = 2)
    empty <- empty - 2
  }
  if (fformatVersion < 6) {
    readBin(zz, integer(), empty, size = 1)
  }

  nyears    <- readBin(zz, integer(), 1, size = 2)
  yearList <- readBin(zz, integer(), nyears, size = 2)
  useBytes  <- (fformatVersion > 4)
  if (fformatVersion > 5) {
    ncells      <- readBin(zz, integer(), 1, size = 4)
    ncharCells  <- readBin(zz, integer(), 1, size = 4)
    cellnames   <- strsplit(readChar(zz, ncharCells, useBytes = TRUE), "\n", useBytes = TRUE)[[1]]
  } else {
    nregions      <- readBin(zz, integer(), 1, size = 2)
    ncharRegions  <- readBin(zz, integer(), 1, size = ifelse(fformatVersion > 3, 4, 2))
    regions       <- strsplit(readChar(zz, ncharRegions, useBytes = useBytes), "\n", useBytes = useBytes)[[1]]
    cpr           <- readBin(zz, integer(), nregions, size = 4)
    ncells <- sum(cpr)
    if (any(cpr != 1)) {
      cellnames   <- paste(rep(regions, cpr), seq_len(ncells), sep = ".")
    } else {
      cellnames   <- regions
    }
  }
  nelem <- readBin(zz, integer(), 1, size = 4)
  ncharData <- readBin(zz, integer(), 1, size = 4)

  datanames <- strsplit(readChar(zz, ncharData, useBytes = useBytes), "\n", useBytes = useBytes)[[1]]

  output <- array(readBin(zz, numeric(), nelem, size = 4), c(ncells, nyears, nelem / ncells / nyears))
  output[is.nan(output)] <- NA
  if (length(cellnames) == 1) {
    cellnames <- list(cellnames)
  }
  dimnames(output)[[1]] <- cellnames
  if (yearList[1] > 0) {
    dimnames(output)[[2]] <- paste("y", yearList, sep = "")
  }
  if (length(datanames) > 0) {
    dimnames(output)[[3]] <- datanames
  }

  if (fformatVersion > 0 && ncharComment > 0) {
    attr(output, "comment") <- strsplit(readChar(zz, ncharComment, useBytes = useBytes), "\n",
                                        useBytes = useBytes)[[1]]
  }
  if (fformatVersion > 1 && ncharSets > 0) {
    names(dimnames(output)) <- strsplit(readChar(zz, ncharSets, useBytes = useBytes), "\n",
                                        useBytes = useBytes)[[1]]
  }
  if (fformatVersion > 2) {
    readBin(zz, raw(), nbyteMetadata)
  }
  attr(output, "FileFormatVersion") <- fformatVersion
  return(new("magpie", output))
}

#' Count elements
#'
#' Functions to count the number of cells/years/datasets/regions of an
#' MAgPIE-object
#'
#' @aliases ncells nyears ndata nregions
#' @param x A MAgPIE-object
#' @return \item{value}{The number of cells/years/datasets/regions of \code{x}}
#' @author Jan Philipp Dietrich
#' @examples
#'
#' a <- is.magpie(NULL)
#' ncells(a)
#' nyears(a)
#' ndata(a)
#' nregions(a)
#' @family ObjectInfo
#' @export
ncells <- function(x) {
  return(dim(x)[1])
}

#' ~~ Methods for Function as.data.frame ~~
#'
#' ~~ Methods for function \code{as.data.frame} ~~
#'
#'
#' @name as.data.frame-methods
#' @aliases as.data.frame as.data.frame-methods as.data.frame,ANY-method
#' as.data.frame,magpie-method
#' @docType methods
#' @param x A MAgPIE-object
#' @param rev The revision of the algorithm that should be used for conversion.
#' rev=1 creates columns with the predefined names Cell, Region, Year, Data1,
#' Data2,... and Value, rev=2 uses the set names of the MAgPIE object for
#' naming and adds an attribute "dimtype" to the data.frame which contains
#' information about the types of the different columns (spatial, temporal,
#' data or value), rev=3 is identical to rev=2 except that characters are
#' not being converted to factors (stringsAsFactors = FALSE).
#' @param raw Logical to control whether years beginning with "y" should be
#' converted to integers (without "y") and coordinates should be converted to
#' numerics. If set to raw columns are returned as they are in the initial
#' object.
#' @section Methods: \describe{
#'
#' \item{list("signature(x = \"magpie\")")}{ Conversion creates columns for
#' Cell, Region, Year, Data1, Data2,... and Value } }
#' @keywords methods
#' @examples
#'
#' pop <- maxample("pop")
#' head(as.data.frame(pop))
#' head(as.data.frame(pop, rev = 2))
#'
#' a <- maxample("animal")
#' head(as.data.frame(a, rev = 3))
#' head(as.data.frame(a, rev = 3, raw = TRUE))
#' attr(as.data.frame(a, rev = 3), "dimtype")
#'
#' @family MAgPIE-Conversions
#' @importFrom utils type.convert
#' @exportMethod as.data.frame

setMethod("as.data.frame",
  signature(x = "magpie"),
  function(x, rev = 1, raw = FALSE) {
    if (rev == 1) {
      yearsAsIntegers <- suppressWarnings(getYears(x, as.integer = TRUE))
      if (raw || any(is.na(getYears(x)) != is.na(yearsAsIntegers))) {
        dimnames(x)[[2]] <- getYears(x)
      } else {
        dimnames(x)[[2]] <- yearsAsIntegers
      }
      if (is.null(dimnames(x)[[2]])) dimnames(x)[[2]] <- 0
      if (is.null(dimnames(x)[[3]])) dimnames(x)[[3]] <- "NA"
      if (any(dim(x) == 0)) {
        return(data.frame())
      } else {
        x <- as.data.frame(as.table(x))
      }
      if (all(grepl(".", x[[3]], fixed = TRUE))) {
        levels(x[[3]]) <- gsub("\\.$", "\\.STRINGTORESETAG", levels(x[[3]]))
        tmp <- data.frame(t(matrix(unlist(strsplit(as.character(x[[3]]), split = ".", fixed = TRUE)), ncol = nrow(x))),
                          stringsAsFactors = FALSE)
        for (i in seq_len(ncol(tmp))) {
          tmp[[i]] <- factor(tmp[[i]], unique(tmp[[i]]))
          levels(tmp[[i]]) <- gsub("STRINGTORESETAG", "", levels(tmp[[i]]))
        }
        x <- cbind(x[, 1:2], tmp, x[4])
      }
      colnames(x) <- c("Region", "Year", paste("Data", 1:(dim(x)[2] - 3), sep = ""), "Value")
      x <- cbind(Cell = suppressWarnings(as.integer(gsub("^[^\\.]*\\.", "", x$Region))), x)
      x$Region <- gsub("\\..*$", "", x$Region)
      return(x)
    } else if (rev == 2) {
      return(asDataFrameX(x, raw = raw, stringsAsFactors = TRUE))
    } else if (rev == 3) {
      return(asDataFrameX(x, raw = raw, stringsAsFactors = FALSE))
    } else {
      stop('Unknown revision "', rev, '"!')
    }
  }
)

asDataFrameX <- function(x, raw, stringsAsFactors = FALSE) { # nolint: cyclocomp_linter.
  x <- clean_magpie(x, what = "sets")
  if (any(dim(x) == 0)) {
    return(data.frame())
  }
  dimnames <- dimnames(x)
  x <- as.data.frame(as.table(x), stringsAsFactors = FALSE)
  names(x)[4] <- ".value"
  what <- ".value"
  types <- c(".spat", ".temp", ".data")
  for (i in 3:1) {
    if (grepl(".", names(x)[i], fixed = TRUE)) {
      tmp <- data.frame(t(matrix(unlist(strsplit(as.character(x[[i]]), split = ".", fixed = TRUE)), ncol = nrow(x))),
                        stringsAsFactors = FALSE)
      names(tmp) <-  strsplit(names(x)[i], split = "\\.")[[1]]
      if (i == 1) {
        x <- cbind(tmp, x[2:length(x)])
      } else {
        x <- cbind(x[1:(i - 1)], tmp, x[(i + 1):length(x)])
      }
      what <- c(paste0(types[i], seq_len(dim(tmp)[2])), what)
    } else if (is.null(dimnames[[i]])) {
      x <- x[-i]
    } else {
      what <- c(paste0(types[i], 1), what)
    }
  }
  # use other types than character if possible
  for (i in seq_len(ncol(x))) {
    if (is.character(x[[i]])) {
      if (!raw) {
        # convert years
        if (grepl("^y[0-9]{1,4}$", x[[i]][1]) && all(grepl("^y[0-9]{1,4}$", x[[i]]))) {
          x[[i]] <- substring(x[[i]], 2)
        }
        # convert coordinates
        if (grepl("^-?[0-9]+p-?[0-9]+", x[[i]][1]) && all(grepl("^-?[0-9]+p-?[0-9]+", x[[i]]))) {
          x[[i]] <- sub("p", ".", x[[i]], fixed = TRUE)
        }
      }
      x[[i]] <- type.convert(x[[i]], as.is = TRUE)
      if (stringsAsFactors && is.character(x[[i]])) x[[i]] <- factor(x[[i]], unique(x[[i]]))
    }
  }
  attr(x, "dimtype") <- what
  return(x)
}

#' Lowpass Filter
#'
#' Filters high frequencies out of a time series. The filter
#' has the structure `x'(n) = (x(n-1)+2*x(n)+x(n+1))/4`
#'
#' @param x Vector of data points, that should be filtered or MAgPIE object
#' @param i number of iterations the filter should be applied to the data
#' @param fix Fixes the starting and/or ending data point. Default value is
#' \code{NULL} which doesn't fix any point. Available options are:
#' \code{"start"} for fixing the starting point, \code{"end"} for fixing the
#' ending point and \code{"both"} for fixing both ends of the data.
#' @param altFilter set special filter rule to indexes defined in this
#' parameter. Special filter structure: `x'(n) = (2*x(n)+x(n+1))/3`
#' @param warn boolean deciding whether lowpass issues a warning for
#' critical parameter choices or not
#' @return The filtered data vector or MAgPIE object
#' @author Jan Philipp Dietrich, Misko Stevanovic
#' @examples
#'
#' lowpass(c(1, 2, 11, 3, 4))
#' # to fix the starting point
#' lowpass(c(0, 9, 1, 5, 14, 20, 6, 11, 0), i = 2, fix = "start")
#' @family TemporalOperations
#' @export
lowpass <- function(x, # nolint: cyclocomp_linter.
                    i = 1, fix = NULL, altFilter = NULL, warn = TRUE) {

  if (!is.null(fix) && !(fix %in% c("start", "end", "both"))) {
    stop("Option \"", fix, "\" is not available for the \"fix\" argunemt!")
  }
  if (warn && !is.null(fix)) {
    warning("Fixing start or end does modify the total sum of values! Use fix=NULL to let the total sum unchanged!")
  }

  if (i == 0) return(x)

  if (is.magpie(x)) {
    years <- getYears(x, as.integer = TRUE)
    if (is.unsorted(years)) {
      x <- x[, sort(years), ]
    }
    l <- nyears(x)
    x2 <- x
    x <- as.array(x)
    for (j in 1:i) {
      y <- x
      x[, 2:(l - 1), ] <- (y[, 1:(l - 2), ] + 2 * y[, 2:(l - 1), ] + y[, 3:l, ]) / 4
      if (!is.null(altFilter)) {
        x[, altFilter, ] <- (2 * y[, altFilter, ] + y[, altFilter + 1, ]) / 3
      }
      if (is.null(fix) || !(fix %in% c("start", "both"))) {
        x[, 1, ] <- (3 * y[, 1, ] + y[, 2, ]) / 4
      }
      if (is.null(fix) || !(fix %in% c("end", "both"))) {
        x[, l, ] <- (3 * y[, l, ] + y[, l - 1, ]) / 4
      }
    }
    x <- as.magpie(x)
    copy.attributes(x) <- x2
  } else {
    l <- length(x)
    for (j in 1:i) {
      y <- x
      x[2:(l - 1)] <- (y[1:(l - 2)] + 2 * y[2:(l - 1)] + y[3:l]) / 4
      if (!is.null(altFilter)) {
        x[altFilter] <- (2 * y[altFilter] + y[altFilter + 1]) / 3
      }
      if (is.null(fix) || !(fix %in% c("start", "both"))) {
        x[1] <- (3 * y[1] + y[2]) / 4
      }
      if (is.null(fix) || !(fix %in% c("end", "both"))) {
        x[l] <- (3 * y[l] + y[l - 1]) / 4
      }
    }
  }
  return(x)
}

#' Get Cells
#'
#' Extracts cell names of a MAgPIE-object
#'
#' setCells is a shortcut to use a MAgPIE object with manipulated cell names.
#' setCells uses the variable names "object" and "nm" in order to be consistent
#' to the already existing function setNames.
#'
#' @aliases getCells getCells<- setCells
#' @param x,object MAgPIE object
#' @param value,nm cell names the data should be set to.
#' @return getCells returns cell names of the MAgPIE-object, whereas setCells
#' returns the MAgPIE object with the manipulated cell names.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getRegions}}, \code{\link{getNames}},
#' \code{\link{setNames}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
#' \code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}
#' @examples
#'
#' a <- as.magpie(1)
#'  getCells(a)
#'  setCells(a, "AFR")
#' @family GetterSetter
#' @export
getCells <- function(x) {
  return(dimnames(x)[[1]])
}

#' @describeIn getCells set cell names
#' @export
`getCells<-` <- function(x, value) {
  if (length(value) != ncells(x)) stop("Wrong number of cell names supplied!")
  if (ncells(x) == 0) return(x)
  if (length(value) == 1) value <- list(value)
  dimnames(x)[[1]] <- value
  return(x)
}

#' Reconstructs full dimensionality of MAgPIE objects
#'
#' Deprecated: Use \code{\link{getItems}} instead.
#'
#' If a MAgPIE object is created from a source with more than one data
#' dimension, these data dimensions are combined to a single dimension. fulldim
#' reconstructs the original dimensionality and reports it.
#'
#' @param x A MAgPIE-object
#' @param sep A character separating joined dimension names
#' @return A list containing in the first element the dim output and in the
#' second element the dimnames output of the reconstructed array.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{as.magpie}},\code{\link{unwrap}},\code{\link{wrap}}
#' @family ObjectInfo
#' @export
fulldim <- function(x, sep = ".") { # nolint: undesirable_function_linter.
  .Deprecated("getItems")
  if (!is.null(dimnames(x)[[3]])) {
    elemsplit <- strsplit(dimnames(x)[[3]], sep, fixed = TRUE)
    tmp <- vapply(elemsplit, length, integer(1))
  } else {
    tmp <- 0
  }
  if (length(tmp) == 1 && tmp == 0) tmp <- 1
  if (tmp[1] == 1) {
    # no need for splitting
    return(list(dim(x), dimnames(x)))
  } else if (any(tmp != rep(tmp[1], length(tmp)))) {
    # data dimension cannot be splitted return dim(x)
    warning("Data dimension cannot be splitted due to inconsistent number of subdimensions!")
    return(list(dim(x), dimnames(x)))
  } else {
    nElemDims <- tmp[1]
    tmp <- t(matrix(unlist(elemsplit), nElemDims))
    dimnames <- list()
    dimnames[[1]] <- dimnames(x)[[1]]
    dimnames[[2]] <- dimnames(x)[[2]]
    dim <- dim(x)[1:2]
    for (i in seq_len(nElemDims)) {
      dimnames[[i + 2]] <- unique(tmp[, i])
      dim <- c(dim, length(dimnames[[i + 2]]))
    }
    tmp <- getSets(x, sep = sep)
    if (length(tmp) == length(dimnames) || is.null(tmp)) {
      names(dimnames) <- tmp
    } else if (length(tmp) == length(dimnames) + 1 && length(grep(sep, names(dimnames(x))[1], fixed = TRUE)) > 0) {
      names(dimnames) <- c(names(dimnames(x))[1], tmp[3:length(tmp)])
    }
    return(list(dim, dimnames))
  }
}

#' sizeCheck
#'
#' Calculates expected magclass object length and checks that it stays below
#' the limit defined with magclass_sizeLimit (default = 10^9). This is useful
#' to prevent out of memory errors in case of unwanted object expansions
#' Ignored if \code{getOption("magclass_sizeLimit")} is negative.
#'
#' @param dim dimensions of the current object as returned by function \code{dim}
#' @author Jan Philipp Dietrich
#' @examples
#' pop <- maxample("pop")
#' magclass:::sizeCheck(dim(pop))
#'
#' \dontrun{
#' magclass:::sizeCheck(c(6765L, 10946L, 17711L))
#' }
sizeCheck <- function(dim) {
  if (!is.vector(dim, "integer")) {
    stop("Numeric vector of dimension sizes expected.")
  }

  if (is.null(getOption("magclass_sizeLimit"))) {
    options(magclass_sizeLimit = 10^9) # nolint: undesirable_function_linter.
  }
  # estimate new object size and check against size limit
  if (getOption("magclass_sizeLimit") > 0) {
    size <- prod(dim)
    if (size > getOption("magclass_sizeLimit")) {
      stop("magclass object size ",
           "(", paste(dim, collapse = " x "), " = ", size, ") ",
           "exceeds limit (getOption(\"magclass_sizeLimit\") = ",
           getOption("magclass_sizeLimit"), ").")
    }
  }
}

#' as.SpatVector
#'
#' Convert magclass object to a SpatVector object. Requires the terra package and
#' requires the magclass object to provide the geometry of the spatial entities
#' as "geometry" attribute in "WKT" format. (see object "m" in example).
#'
#' @param x MAgPIE object
#' @return A SpatVector object
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{as.SpatRaster}}
#' @examples
#'
#' if (requireNamespace("terra", quietly = TRUE)) {
#'    r <- terra::rast(ncols = 360, nrows = 180, nlyrs = 4)
#'    r[85:89, 176:179] <- (1:20 %*% t(1:4))
#'    r[15:19, 76:79] <-   (10 + 1:20 %*% t(1:4))
#'    names(r) <- c("y2000..bla", "y2001..bla", "y2000..blub", "y2001..blub")
#'    v <- terra::as.polygons(r)
#'    m <- as.magpie(v)
#'    attr(m, "geometry")
#'    attr(m, "crs")
#'    v2 <- as.SpatVector(m)
#' }
#' @family MAgPIE-Conversions
#' @export

as.SpatVector <- function(x) { # nolint: object_name_linter.
  if (!is.magpie(x)) stop("Input is not a magpie object")
  if (!requireNamespace("terra", quietly = TRUE)) stop("The package \"terra\" is required!")
  if (is.null(attr(x, "geometry"))) {
    stop("x must have an attribute geometry containing the polygon structure in WKT format!")
  }
  out <- terra::vect(attr(x, "geometry"), crs = attr(x, "crs"))
  spatDims <- getItems(x, dim = 1, split = TRUE, full = TRUE)
  names(spatDims) <- paste0(".", names(spatDims))
  data <- as.data.frame(magclass::wrap(as.array(x), list(1, 2:3), sep = ".."))
  terra::values(out) <- cbind(spatDims, data)
  return(out)
}

#' Collapse dataset names
#'
#' This function has been superseded by \code{\link{collapseDim}} which is a
#' more generalized version of this function. Please use this one instead!
#'
#' This function will remove names in the data dimension which are the same for
#' each element (meaning that this data dimension contains exactly one element)
#'
#'
#' @param x MAgPIE object
#' @param collapsedim If you want to remove the names of particular dimensions
#' provide the dimensions here. Since the function only works in the third dimension,
#' you have to count from there on (e.g. dim = 3.2 refers to collapsedim = 2).
#' Alternatively, you can also specify the name of the dimension. Default: NULL.
#' CAUTION with parameter collapsedim! You could also force him to remove dimnames,
#' which are NOT the same for each element and so create duplicates in dimnames.
#' @param preservedim If you want to remove the name of particular dimensions except some,
#' you can specify the dimension(s) to preserve here. See collapsedim for naming convention.
#' Note that preservedim will be ignored in the case, of a specified collapsedim
#' @return The provided MAgPIE object with collapsed names
#' @author Jan Philipp Dietrich, David Klein, Xiaoxi Wang
#' @seealso \code{\link{collapseDim}}, \code{\link{getItems}},
#' \code{"\linkS4class{magpie}"}
#' @family DimensionManipulation
#' @export

collapseNames <- function(x, collapsedim = NULL, preservedim = NULL) {
  .dimCodeExtra <- function(dim, x) {
    if (is.null(dim)) return(NULL)
    if (is.character(dim)) return(dimCode(dim, x))
    if (all(dim > 3 & round(dim) == 3)) return(dim)
    # convert old dim convention in collapseNames (just specificantion of
    # sub-dimension) to new convention (maindim.subdim)
    return(dim / 10 + 3)
  }
  if (is.null(collapsedim)) return(collapseDim(x, keepdim = c(1:2, .dimCodeExtra(preservedim, x))))
  return(collapseDim(x, dim = .dimCodeExtra(collapsedim, x)))
}

#' Get a list of celluare region-belongings
#'
#' Extracts a vector containing the region of each cell of a MAgPIE-object
#'
#'
#' @aliases getRegionList getRegionList<-
#' @param x MAgPIE object
#' @param value A vector with ncell elements containing the regions of each
#' cell.
#' @return A vector with ncell elements containing the region of each cell.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getRegions}},\code{\link{getYears}},
#' \code{\link{getNames}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
#' \code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}
#' @examples
#'
#' # a <- read.magpie("example.mz")
#' # getRegionList(a)
#' @family GetterSetter
#' @export
getRegionList <- function(x) { # nolint: undesirable_function_linter.
  .Deprecated("getItems")
  return(factor(sub("\\..*$", "", dimnames(x)[[1]])))
}

#' @describeIn getRegionList set region names
#' @export
`getRegionList<-` <- function(x, value) {
  reg <- getRegionList(x) # nolint: undesirable_function_linter.
  if (length(reg) != length(value)) stop("Lengths of RegionLists do not agree!")
  tmp <- sub("^.*\\.", "", dimnames(x)[[1]])
  dimnames(x)[[1]] <- paste(as.vector(value), tmp, sep = ".")
  return(x)
}

#' head/tail
#'
#' head and tail methods for MAgPIE objects to extract the head or tail of an
#' object
#'
#'
#' @aliases head.magpie tail.magpie
#' @param x MAgPIE object
#' @param n1,n2,n3 number of lines in first, second and third dimension that
#' should be returned. If the given number is higher than the length of the
#' dimension all entries in this dimension will be returned.
#' @param ... arguments to be passed to or from other methods.
#' @return head returns the first n1 x n2 x n3 entries, tail returns the last
#' n1 x n2 x n3 entries.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link[utils]{head}}, \code{\link[utils]{tail}}
#' @examples
#'
#' pop <- maxample("pop")
#' head(pop)
#'  tail(pop, 2, 4, 1)
#'
#' @family Display
#' @importFrom utils head
#' @export
head.magpie <- function(x, n1 = 3L, n2 = 6L, n3 = 2L, ...) {
  if (dim(x)[1] < n1) n1 <- dim(x)[1]
  if (dim(x)[2] < n2) n2 <- dim(x)[2]
  if (dim(x)[3] < n3) n3 <- dim(x)[3]
  return(x[1:n1, 1:n2, 1:n3])
}

#' withAlignedDims
#'
#' (for internal use) Executes the passed function after reordering items
#' in each dimension for all passed magpie objects
#' @param func The function that should be executed with prior alignment of dimensions
#' @param funcName The name of the resulting function, to improve error messages
#' @author Pascal Sauer, Patrick Rein
#' @keywords internal
withAlignedDims <- function(func, funcName = "anonymous func", ...) {
  ms <- list(...)
  .allIdentical <- function(vs) all(vapply(vs[-1], function(x) identical(vs[[1]], x), logical(1)))

  if (!all(vapply(ms, is.magpie, logical(1)))) {
    stop(paste0(funcName, " cannot handle non-magpie objects"))
  }

  if (.allIdentical(lapply(ms, dimnames))) {
    return(func(...))
  }

  # collapseDim if the dim has a single item in all and they differ
  singleDifferingItemDims <- Filter(function(dimCode) {
    items <- lapply(ms, function(m) dimnames(m)[[dimCode]])
    return(all(lapply(items, length) <= 1) && !.allIdentical(items))
  }, 1:3)
  if (length(singleDifferingItemDims) > 0) {
    originalDimensionNames <- names(dimnames(ms[[1]]))
    ms <- lapply(ms, function(m) {
      m <- collapseDim(m, dim = singleDifferingItemDims)
      names(dimnames(m)) <- originalDimensionNames
      return(m)
    })
  }

  if (!.allIdentical(lapply(ms, dim))) {
    stop(paste0(funcName, " expects magpie objects with equal dimensions"))
  }

  if (!do.call(sameDims, ms)) {
    stop(paste0(funcName, " expects magpie objects with the same items in all dimensions"))
  }

  a <- ms[[1]]
  if (dim(a)[1] > 1) {
    ms <- lapply(ms, function(x) x[getItems(a, 1), , ])
  }
  if (dim(a)[2] > 1) {
    ms <- lapply(ms, function(x) x[, getItems(a, 2), ])
  }
  if (dim(a)[3] > 1) {
    ms <- lapply(ms, function(x) x[, , getItems(a, 3)])
  }

  return(do.call(func, ms))
}

#' magclassdata
#'
#' General magclass-dataset
#'
#' Please do not directly access that data. It should be only used by library
#' functions.
#' @name magclassdata
#' @author Jan Philipp Dietrich
NULL

#' Read MAgPIE-object from file
#'
#' Reads a MAgPIE-file and converts it to a 3D array of the structure
#' (cells,years,datacolumn)
#'
#' See \code{\link{write.magpie}} for a list of supported file formats.
#'
#' @param file_name file name including file ending (wildcards are supported).
#' Optionally also the full path can be specified here (instead of splitting it
#' to file_name and file_folder)
#' @param file_folder folder the file is located in (alternatively you can also
#' specify the full path in file_name - wildcards are supported)
#' @param file_type format the data is stored in. If file_type=NULL the file ending
#' of the file_name is used as format. If format is different to the formats
#' mentioned standard MAgPIE format is assumed. See \code{\link{write.magpie}}
#' for a list of supported file formats.
#' @param as.array Should the input be transformed to an array? This can be
#' useful for regional or global inputs, but all advantages of the magpie-class
#' are lost.
#' @param comment.char character: a character vector of length one containing a
#' single character or an empty string. Use "" to turn off the interpretation
#' of comments altogether. If a comment is found it will be stored in
#' attr(,"comment"). In text files the comment has to be at the beginning of
#' the file in order to be recognized by read.magpie.
#' @param check.names logical. If TRUE then the names of the variables in the
#' data frame are checked to ensure that they are syntactically valid variable
#' names. Same functionality as in read.table.
#' @param ... additional arguments passed to specific read functions (e.g.
#' \code{varname} for specifying the variable to be read in from a multi-variable
#' NCDF file.)
#' @return \item{x}{MAgPIE-object}
#' @note
#'
#' See \code{\link{write.magpie}} for the detailed structure of binary MAgPIE formats .m and .mz.
#'
#' @author Jan Philipp Dietrich, Stephen Bi, Florian Humpenoeder, Pascal Sauer
#' @seealso \code{"\linkS4class{magpie}"}, \code{\link{write.magpie}}
#' @family FileIO
#' @export
read.magpie <- function(file_name, file_folder = "", file_type = NULL, # nolint: object_name_linter, cyclocomp_linter.
                        as.array = FALSE, comment.char = "*", check.names = FALSE, ...) { # nolint: object_name_linter.

  .buildFileName <- function(fileName, fileFolder) {
    fileName <- paste0(fileFolder, fileName)
    fileNameOut <- Sys.glob(fileName)
    if (length(fileNameOut) > 1) {
      fileNameOut <- fileNameOut[1]
      warning("File name ", fileName, " is ambiguous, only first alternative is used!")
    } else if (length(fileNameOut) == 0) {
      stop("File ", fileName, " does not exist!")
    }
    return(fileNameOut)
  }
  fileName <- .buildFileName(file_name, file_folder)

  .getFileType <- function(fileType, fileName) {
    # if file-type is not mentioned file-ending is used as file-type
    fileType <- ifelse(is.null(fileType), tail(strsplit(fileName, "\\.")[[1]], 1), fileType)
    allowedTypes <- c("rds", "m", "mz", "csv", "cs2", "cs2b", "cs3", "cs4", "cs5", "csvr", "cs2r", "cs3r",
                      "cs4r", "put", "asc", "nc")
    if (!(fileType %in% allowedTypes)) stop("Unknown file type: ", fileType)
    return(fileType)
  }
  fileType <- .getFileType(file_type, fileName)

  if (fileType %in% c("m", "mz")) {
    readMagpie <- readMagpieMZ(fileName, compressed = (fileType == "mz"))
  } else if (fileType == "rds") {
    readMagpie <- readRDS(fileName)
    if (!is.magpie(readMagpie)) stop("File does not contain a magpie object!")
  } else if (fileType %in% c("cs3", "cs3r")) {
    x <- utils::read.csv(fileName, comment.char = comment.char, check.names = check.names, stringsAsFactors = TRUE)
    datacols <- grep("^dummy\\.?[0-9]*$", colnames(x))
    xdimnames <- lapply(x[datacols], function(x) return(as.character(unique(x))))
    xdimnames[[length(xdimnames) + 1]] <- colnames(x)[-datacols]
    names(xdimnames) <- NULL
    tmparr <- array(NA, dim = vapply(xdimnames, length, integer(1)), dimnames = xdimnames)
    for (i in xdimnames[[length(xdimnames)]]) {
      j <- vapply(cbind(x[datacols], i), as.character, character(nrow(x)))
      .duplicates_check(j)
      tmparr[j] <- x[, i]
    }
    readMagpie <- as.magpie(tmparr)
    if (length(grep("^[A-Z]+_[0-9]+$", getCells(readMagpie))) == ncells(readMagpie)) {
      getCells(readMagpie) <- sub("_", ".", getCells(readMagpie))
    }
    attr(readMagpie, "comment") <- .readComment(fileName, commentChar = comment.char)
  } else if (fileType %in% c("cs4", "cs4r")) {
    x <- utils::read.csv(fileName, comment.char = comment.char, header = FALSE,
                         check.names = check.names, stringsAsFactors = TRUE)
    readMagpie <- as.magpie(x, tidy = TRUE)
    attr(readMagpie, "comment") <- .readComment(fileName, commentChar = comment.char)
  } else if (fileType == "cs5") {
    .metaExtract <- function(fileName, commentChar) {
      comment <- .readComment(fileName, commentChar = commentChar)
      m <- grep("^META ", comment)
      metadata <- comment[m]
      comment <- comment[-m]
      pattern <- "^META (.*?): (.*)$"
      mNames <- sub(pattern, "\\1", metadata)
      mData  <- strsplit(sub(pattern, "\\2", metadata), ", ")
      names(mData) <- mNames
      return(list(comment = comment, metadata = mData))
    }
    m <- .metaExtract(fileName, comment.char)
    x <- utils::read.csv(fileName, comment.char = comment.char, header = FALSE,
                         check.names = check.names, stringsAsFactors = FALSE)
    colnames(x) <- m$metadata$names
    readMagpie <- as.magpie(x, tidy = TRUE,
                            spatial = grep(".spat", m$metadata$dimtype, fixed = TRUE),
                            temporal = grep(".temp", m$metadata$dimtype, fixed = TRUE),
                            datacol = grep(".data", m$metadata$dimtype, fixed = TRUE))
    attr(readMagpie, "comment") <- m$comment
  } else if (fileType %in% c("asc", "nc", "grd", "tif")) {
    if (fileType == "nc") {
      if (!requireNamespace("terra", quietly = TRUE)) {
        stop("The package \"terra\" is required!")
      }

      x <- terra::rast(fileName)
      if (all(grepl("Time=[0-9]+", names(x)))) {
        names(x) <- sub("(.+)_Time=([0-9]+)", "y\\2..\\1", names(x))
      } else {
        parts <- strsplit(names(x), "_")
        lastParts <- vapply(parts, function(p) p[length(p)], character(1))
        timeIndices <- suppressSpecificWarnings(as.numeric(lastParts), "NAs introduced by coercion")

        terraTime <- terra::time(x)

        if (any(is.na(terraTime)) && requireNamespace("ncdf4", quietly = TRUE)) {
          nc <- ncdf4::nc_open(fileName)
          withr::defer(ncdf4::nc_close(nc))
          if ("time" %in% names(nc$dim)) {
            terraTime <- rep_len(nc$dim$time$vals, terra::nlyr(x))
            if (terra::nlyr(x) %% nc$dim$time$len != 0) {
              warning("Found ", terra::nlyr(x), " layers, but ",
                      nc$dim$time$len, " time steps. Now using ",
                      terraTime)
            }
          }
        }

        if (all(timeIndices %in% seq_along(terraTime))) {
          names(x) <- paste0("y", terraTime[timeIndices], "..",
                             vapply(parts, function(p) paste0(p[-length(p)], collapse = "_"), character(1)))
        } else if (all(!is.na(terraTime))) {
          names(x) <- paste0("y", terraTime, "..", names(x))
        }
      }

      readMagpie <- clean_magpie(as.magpie(x))
      attr(readMagpie, "crs") <- NULL
    } else {
      if (!requireNamespace("raster", quietly = TRUE)) {
        stop("The package \"raster\" is required!")
      }
      readMagpie <- as.magpie(raster::brick(fileName, ...))
    }
  } else {
    readMagpie <- readMagpieOther(fileName, fileType, comment.char = comment.char, check.names = check.names)
  }
  if (as.array) readMagpie <- as.array(readMagpie)[, , ]
  return(readMagpie)
}

#
# Start Preamble to set up generic pmax
#
existingPmaxGeneric <- getGeneric("pmax")
if (!is.null(existingPmaxGeneric)) {
  if (any(existingPmaxGeneric@signature != "...")) {
    warning("Redefining pmax generic function with new signature \"...\", ",
            "while there already is a pmax generic with signature: ",
            paste0(existingPmaxGeneric@signature, collapse = ", "))
  }
}
setGeneric("pmax", signature = "...")
#
# End Preamble
#

#' @rdname magpie-pmin-pmax
#' @exportMethod pmax
setMethod("pmax", "magpie", function(..., na.rm = FALSE) { # nolint: object_name_linter.
  return(withAlignedDims(function(...) base::pmax(..., na.rm = na.rm),
                         "pmax",
                         ...))
})

#' ~~ Method for function \code{show} ~~
#'
#' Show a magpie object by calling the default show method, print, or str,
#' depending on options("magclass_show_func").
#'
#' @name show-methods
#' @aliases show show-methods show,ANY-method show,magpie-method
#' @docType methods
#' @param object A MAgPIE-object
#' @keywords methods
#' @family Display
#' @author Pascal Sauer
#'
#' @exportMethod show
methods::setMethod("show", methods::signature(object = "magpie"), function(object) {
  # callNextMethod calls the show method that would be called if this method here would not exist
  switch(getOption("magclass_show_func", default = "show"),
         show = methods::callNextMethod(object),
         print = print(object),
         str = utils::str(object),
         stop("options('magclass_show_func') must be set to 'show', 'print', or 'str'"))
  return(invisible(NULL))
})

#' time_interpolate
#'
#' Function to extrapolate missing years in MAgPIE objects.
#'
#'
#' @param dataset An MAgPIE object
#' @param interpolated_year Vector of years, of which values are required. Can
#' be in the formats 1999 or y1999.
#' @param integrate_interpolated_years FALSE returns only the dataset of the
#' interpolated year, TRUE returns the whole dataset, including all years of
#' data and the itnerpolated year
#' @param extrapolation_type Determines what happens if extrapolation is
#' required, i.e. if a requested year lies outside the range of years in
#' \code{dataset}. Specify "linear" for a linear extrapolation. "constant" uses
#' the value from dataset closest in time to the requested year.
#' @return Uses linear extrapolation to estimate the values of the interpolated
#' year, using the values of the two surrounding years. If the value is before
#' or after the years in data, the two closest neighbours are used for
#' extrapolation.
#' @author Benjamin Bodirsky, Jan Philipp Dietrich
#' @seealso \code{\link{convergence}}
#' @examples
#'
#' p <- maxample("pop")
#' time_interpolate(p, "y2000", integrate = TRUE)
#' time_interpolate(p, c("y1980", "y2000"), integrate = TRUE, extrapolation_type = "constant")
#' @family TemporalOperations
#' @importFrom abind abind
#' @export
time_interpolate <- function(dataset, interpolated_year, # nolint: object_name_linter.
                             integrate_interpolated_years = FALSE, # nolint: object_name_linter.
                             extrapolation_type = "linear") { # nolint: object_name_linter.
  if (!is.magpie(dataset)) {
    stop("Invalid data format of measured data. Has to be a MAgPIE-object.")
  }
  interpolatedYear <- interpolated_year
  dataset <- clean_magpie(dataset, what = "sets")
  sets <- getSets(dataset)
  if (all(isYear(interpolatedYear, with_y = FALSE))) {
    interpolatedYear <- paste("y", interpolatedYear, sep = "")
  } else  {
    if (!any(isYear(interpolatedYear, with_y = TRUE))) {
      stop("year not in the right format")
    }
  }

  if (nyears(dataset) == 1) {
    tmp <- dataset
    dimnames(tmp)[[2]] <- "y0000"
    dataset <- mbind(tmp, dataset)
  }

  interpolatedYearFiltered <- interpolatedYear[!interpolatedYear %in% getYears(dataset)]
  datasetInterpolated       <- array(NA,
    dim = c(dim(dataset)[1], length(interpolatedYearFiltered), dim(dataset)[3]),
    dimnames = list(getCells(dataset), interpolatedYearFiltered, getNames(dataset))
  )
  dataset <- as.array(dataset)


  for (singleInterpolatedYear in interpolatedYearFiltered) {
    sortedYears                <-  sort(c(dimnames(dataset)[[2]], singleInterpolatedYear))
    if (sortedYears[1] == singleInterpolatedYear) {
      yearBefore <- sortedYears[2]
      yearAfter  <- sortedYears[3]
      yearExtrapolate <- ifelse(extrapolation_type == "constant", sortedYears[2], -1)
    } else if (sortedYears[length(sortedYears)] == singleInterpolatedYear) {
      yearBefore <- sortedYears[length(sortedYears) - 2]
      yearAfter  <- sortedYears[length(sortedYears) - 1]
      yearExtrapolate <- ifelse(extrapolation_type == "constant", sortedYears[length(sortedYears) - 1], -1)
    } else {
      yearBefore <- sortedYears[which(sortedYears == singleInterpolatedYear) - 1]
      yearAfter <- sortedYears[which(sortedYears == singleInterpolatedYear) + 1]
      yearExtrapolate <- -1
    }

    interpolatedYearInt       <- as.integer(substring(singleInterpolatedYear, 2))
    yearBeforeInt             <- as.integer(substring(yearBefore, 2))
    yearAfterInt              <- as.integer(substring(yearAfter, 2))

    datasetDifference        <-  dataset[, yearAfter, , drop = FALSE] - dataset[, yearBefore, , drop = FALSE]
    yearBeforeToAfter        <-  yearAfterInt        - yearBeforeInt
    yearBeforeToInterpolated <-  interpolatedYearInt - yearBeforeInt


    if (yearExtrapolate == -1) {
      datasetInterpolated[, singleInterpolatedYear, ] <- dataset[, yearBefore, , drop = FALSE] +
        yearBeforeToInterpolated * datasetDifference /
          yearBeforeToAfter
    } else {
      datasetInterpolated[, singleInterpolatedYear, ] <- dataset[, yearExtrapolate, , drop = FALSE]
    }
  }
  if (integrate_interpolated_years == FALSE) {
    addYears <- setdiff(interpolatedYear, interpolatedYearFiltered)
    if (length(addYears) > 0) {
      dataset <- abind::abind(datasetInterpolated, dataset[, addYears, , drop = FALSE], along = 2)
    } else {
      dataset <- datasetInterpolated
    }
  } else {
    if (any(getYears(dataset) == "y0000")) {
      dataset <- dataset[, -which(getYears(dataset) == "y0000"), , drop = FALSE]
    }
    dataset <- abind::abind(dataset, datasetInterpolated, along = 2)
  }
  dataset <- as.magpie(dataset, spatial = 1, temporal = 2)
  dataset <- dataset[, sort(getYears(dataset)), ]
  getSets(dataset) <- sets
  return(dataset)
}

#' escapeRegex
#'
#' Escapes all symbols in a string which have a special meaning in regular
#' expressions.
#'
#'
#' @param x String or vector of strings that should be escaped.
#' @return The escaped strings.
#' @author Jan Philipp Dietrich
#' @seealso \code{\link[base]{grep}}
escapeRegex <- function(x) {
  return(gsub("([.|()\\^{}+$*?]|\\[|\\])", "\\\\\\1", x))
}

#' guessResolution
#'
#' Guess the resolution of the given magpie object/coordinates by looking at the minimum
#' difference between unique sorted values. Fall back to 0.5 if guess is infinite.
#'
#' @param x A magpie object or the coordinates of a magpie object (the result
#' of \code{\link{getCoords}})
#' @return The guessed resolution of the data
#' @author Jan Philipp Dietrich, Pascal Sauer
#' @family SpatialOperations
#' @export
guessResolution <- function(x) {
  if (is.magpie(x)) {
    xy <- getCoords(x)
  } else {
    xy <- x
  }
  .tmp <- function(x) {
    return(min(diff(sort(unique(x)))))
  }
  guess <- min(.tmp(xy[[1]]), .tmp(xy[[2]]))
  # use 0.5deg as guess if it cannot be determined otherwise as this is
  # the default spatial resolution in the magpie universe.
  if (is.infinite(guess)) guess <- 0.5
  return(guess)
}

readMagpieOther <- function(fileName, fileType, # nolint: cyclocomp_linter.
                            comment.char = "*", check.names = FALSE) { # nolint: object_name_linter.
  sep <- ifelse(fileType == "put", "\t", ",")
  # check for header
  temp <- utils::read.csv(fileName, nrows = 1, header = FALSE, sep = sep, comment.char = comment.char,
                          check.names = check.names, stringsAsFactors = TRUE)

  # check for numeric elements in first row, which means a missing header
  header <- !any(vapply(temp, is.numeric, logical(1)))

  temp <- utils::read.csv(fileName, header = header, sep = sep, comment.char = comment.char,
                          check.names = check.names, stringsAsFactors = TRUE)

  # analyse column content
  coltypes <- rep("data", dim(temp)[2])
  for (column in seq_len(dim(temp)[2])) {
    if (sum(coltypes == "year") == 0 &&
          length(grep("^(y[0-9]{4}|[0-2][0-9]{3})$", temp[, column])) == dim(temp)[1]) {
      coltypes[column] <- "year"
    } else if (sum(coltypes == "region") == 0 && sum(coltypes == "regiospatial") == 0 &&
                 length(grep("^[A-Z]{3}$", temp[, column])) == dim(temp)[1]) {
      coltypes[column] <- "region"
    } else if (sum(coltypes == "regiospatial") == 0 && sum(coltypes == "region") == 0 &&
                 length(grep("^[A-Z]{3}_[0-9]+$", temp[, column])) == dim(temp)[1]) {
      coltypes[column] <- "regiospatial"
    } else if (!is.numeric(temp[1, column])) {
      coltypes[column] <- "other"
    } else if (sum(coltypes == "cell") == 0 && all(!is.na(temp[, column])) && all(temp[, column] != 0) &&
                 length(temp[, column]) %% max(temp[, column]) == 0 &&
                 suppressWarnings(try(all(unique(temp[, column]) == 1:max(temp[, column])), silent = TRUE) == TRUE)) {
      coltypes[column] <- "cell"
    }
  }

  if (any(coltypes == "year")) {
    temp <- temp[order(temp[, which(coltypes == "year")]), ]
    if (length(grep("y", temp[, which(coltypes == "year")])) == 0) {
      temp[, which(coltypes == "year")] <- as.factor(paste("y", temp[, which(coltypes == "year")], sep = ""))
    }
  }

  # backup check if cell column is really a cell column
  if (any(coltypes == "cell")) {
    if (dimnames(temp)[[2]][which(coltypes == "cell")] == "iteration") {
      temp[, which(coltypes == "cell")] <- paste("iter", format(temp[, which(coltypes == "cell")]), sep = "")
      coltypes[which(coltypes == "cell")] <- "other"
    } else if (header && !(dimnames(temp)[[2]][which(coltypes == "cell")] %in%
                             c("dummy", "dummy.1", "dummy.2", "dummy.3", "", " ",
                               "cell", "cells", "Cell", "Cells"))) {
      coltypes[which(coltypes == "cell")] <- "data"
    }
  }

  if (any(coltypes == "cell")) {
    ncells <- dim(temp)[1]
    if (any(coltypes == "year")) ncells <- ncells / length(unique(temp[, which(coltypes == "year")]))
    if (any(coltypes == "other")) ncells <- ncells / length(unique(temp[, which(coltypes == "other")]))
    if (!all(temp[1:ncells, which(coltypes == "cell")] == 1:ncells)) coltypes[which(coltypes == "cell")] <- "data"
  }

  if (all(coltypes != "data")) {
    if (coltypes[length(coltypes)] == "cell") coltypes[length(coltypes)] <- "data"
    else stop("No data column found! (coltypes: ", paste(coltypes, collapse = ", "))
  }

  if (sum(coltypes == "other") > 1) {
    stop("Invalid format. More than one \"other\" column is not allowed! (coltypes: ", paste(coltypes, collapse = ", "))
  }

  # set all coltypes after the first occurrence of "data" to "data"
  coltypes[min(which(coltypes == "data")):length(coltypes)] <- "data"

  # set first columntype from "cell" to "data" if it seems that the data set is just a vector of numbers
  if (all(coltypes == c("cell", rep("data", length(coltypes) - 1))) && dim(temp)[1] == 1) coltypes[1] <- "data"

  headertype <- "none"
  if (header) {
    if (length(grep("^y+[0-9]{4}$", dimnames(temp)[[2]][which(coltypes == "data")[1]])) == 1) {
      headertype <- "year"
    } else if (length(grep("^[A-Z]{3}$", dimnames(temp)[[2]][which(coltypes == "data")[1]])) == 1) {
      headertype <- "region"
    } else {
      headertype <- "other"
    }
  }

  if (any(coltypes == "other")) {
    othernames <- levels(as.factor(temp[, which(coltypes == "other")]))
    nother <- length(othernames)
    if (header) {
      if (headertype == "other") {
        elemnames <- dimnames(temp)[[2]][which(coltypes == "data")]
        elemnames <- paste(rep(othernames, each = length(elemnames)), elemnames, sep = ".")
      } else {
        elemnames <- othernames
      }
    } else {
      if (sum(coltypes == "data") == 1) {
        elemnames <- othernames
      } else {
        elemnames <- 1:sum(coltypes == "data")
        elemnames <- paste(rep(othernames, each = length(elemnames)), elemnames, sep = ".")
      }
    }
    ncols <- length(elemnames)
  } else {
    nother <- 1
    if (header) {
      if (headertype == "other") {
        elemnames <- dimnames(temp)[[2]][which(coltypes == "data")]
        ncols <- length(elemnames)
      } else {
        elemnames <- NULL
        ncols <- 1
      }
    } else {
      elemnames <- NULL
      ncols <- sum(coltypes == "data")
    }
  }


  if (any(coltypes == "year")) {
    yearnames <- levels(temp[, which(coltypes == "year")])
  } else if (headertype == "year") {
    yearnames <- dimnames(temp)[[2]][which(coltypes == "data")]
  } else {
    yearnames <- NULL
  }
  nyears <- max(1, length(yearnames))

  if (any(coltypes == "cell")) {
    ncells <- max(temp[, which(coltypes == "cell")])
  } else {
    if (headertype != "year") {
      ncells <- dim(temp)[1] / (nyears * nother)
    } else {
      ncells <- dim(temp)[1] / nother
    }
  }

  if (any(coltypes == "regiospatial")) {
    cellnames <- gsub("_", ".", temp[1:ncells, which(coltypes == "regiospatial")], fixed = TRUE)
  } else {
    if (any(coltypes == "region")) {
      tmpRegionnames <- levels(temp[, which(coltypes == "region")])
      regionnames <- tmpRegionnames[temp[, which(coltypes == "region")]]
      if (ncells == length(tmpRegionnames)) regionnames <- unique(regionnames)
    } else if (headertype == "region") {
      regionnames <- dimnames(temp)[[2]][which(coltypes == "data")]
      ncells <- ncells * length(regionnames)
    } else {
      regionnames <- "GLO"
    }
    if (length(unique(regionnames)) < length(regionnames)) {
      cellnames <- paste(regionnames, 1:ncells, sep = ".")
    } else {
      cellnames <- regionnames
    }
  }
  if (length(cellnames) == 1) cellnames <- list(cellnames)
  output <- array(NA, c(ncells, nyears, ncols))
  dimnames(output)[[1]] <- cellnames
  dimnames(output)[[2]] <- yearnames
  dimnames(output)[[3]] <- elemnames

  if (any(coltypes == "other") && (headertype == "other" || headertype == "none")) {
    counter <- 0
    for (other.elem in othernames) {
      output[, , (1:sum(coltypes == "data")) + counter] <- array(as.vector(
        as.matrix(temp[which(temp[, which(coltypes == "other")] == other.elem), which(coltypes == "data")])
      ), c(ncells, nyears, sum(coltypes == "data")))
      counter <- counter + sum(coltypes == "data")
    }
  } else if (!any(coltypes == "other") && headertype == "region") {
    for (i in seq_along(cellnames)) {
      output[i, , 1] <- temp[, which(coltypes == "data")[i]]
    }
  } else if (!any(coltypes == "other") && headertype == "year") {
    for (year in yearnames) {
      output[, year, 1] <- temp[, year]
    }
  } else if (any(coltypes == "other") && headertype == "region") {
    for (i in seq_along(cellnames)) {
      for (elem in elemnames) {
        output[i, , elem] <- temp[which(temp[, which(coltypes == "other")] == elem), which(coltypes == "data")[i]]
      }
    }
  } else if (any(coltypes == "other") && headertype == "year") {
    for (year in yearnames) {
      for (elem in elemnames) {
        output[, year, elem] <- temp[which(temp[, which(coltypes == "other")] == elem), year]
      }
    }
  } else {
    output[, , ] <- as.vector(as.matrix(temp[, which(coltypes == "data")]))
  }
  output <- as.magpie(output)
  attr(output, "comment") <- .readComment(fileName, commentChar = comment.char)
  return(output)
}

#' str
#'
#' str method for MAgPIE objects for conventient display of the structure of a magpie object.
#'
#' In contrast to the default str this will not show the attributes of object@.Data as these
#' contain only a duplicate of dimnames. Also, dim is not shown, because the information it
#' provides is implicitly included in dimnames.
#'
#' @param object MAgPIE object
#' @param ... arguments to be passed to or from other methods.
#' @author Pascal Sauer
#' @seealso \code{\link[utils]{str}}
#' @examples
#' str(maxample("pop"))
#' @family Display
#' @export
str.magpie <- function(object, ...) {
  cat("A magpie object (package: magclass)\n")
  cat(" @ .Data: ")
  utils::str(c(object@.Data)) # c() is important here to strip attributes
  attrib <- attributes(object)
  attrib <- attrib[setdiff(names(attrib), c("class", "dim"))]
  cat(paste(utils::capture.output({
    utils::str(attrib)
  })[-1], collapse = "\n"))
  cat("\n")
  return(invisible(NULL))
}

#' Get regions
#'
#' Extracts regions of a MAgPIE-object
#'
#'
#' @aliases getRegions getRegions<-
#' @param x MAgPIE object
#' @param value Vector containing the new region names of the MAgPIE objects.
#' If you also want to change the mapping of regions to cell please use
#' \code{\link{getRegionList}} instead.
#' @return Regions of the MAgPIE-object
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{getYears}}, \code{\link{getNames}},
#' \code{\link{getCPR}}, \code{\link{read.magpie}}, \code{\link{write.magpie}},
#' \code{"\linkS4class{magpie}"}
#' @examples
#'
#' # a <- read.magpie("example.mz")
#' # getRegions(a)
#' @family GetterSetter
#' @export
getRegions <- function(x) { # nolint: undesirable_function_linter.
  return(getItems(x, dim = 1.1))
}

#' @describeIn getRegions overwrite region names
#' @export
`getRegions<-` <- function(x, value) {
  .Deprecated("getItems")
  getCells(x) <- value
  return(x)
}

#' extend
#'
#' Extend a magpie object to a dense grid based on the given xRange, yRange and resolution.
#' This is e.g. required when writing netCDF files. Extending a sparse magpie object to a dense grid
#' requires much more memory, so use with caution.
#'
#' @param x A magpie object
#' @param gridDefinition A vector of 5 numeric values: c(xMin, xMax, yMin, yMax, resolution).
#' Use c(-179.75, 179.75, -89.75, 89.75, 0.5) to extend to a standard 0.5-degree-resolution
#' lon/lat grid. If NULL, use min/max of coordinates in x and guessResolution.
#' @param crop If TRUE, discard cells from x which are not in the gridDefinition grid. If FALSE,
#' throw an error if the coordinates of x are not a subset of the extended coordinates.
#' @return Magpie object x with dense grid according to gridDefinition, gaps filled with NA.
#' @author Pascal Sauer
#' @family DataBinding
#' @export
extend <- function(x, gridDefinition = NULL, crop = FALSE) {
  if (is.null(gridDefinition)) {
    coords <- getCoords(x)
    firstX <- min(coords$x)
    lastX <- max(coords$x)
    firstY <- max(coords$y)
    lastY <- min(coords$y)
    res <- guessResolution(coords)
  } else {
    stopifnot(length(gridDefinition) == 5)
    firstX <- gridDefinition[1]
    lastX <- gridDefinition[2]
    firstY <- gridDefinition[3]
    lastY <- gridDefinition[4]
    res <- gridDefinition[5]
  }

  coords <- expand.grid(x = seq(firstX, lastX, if (firstX < lastX) res else -res),
                        y = seq(firstY, lastY, if (firstY < lastY) res else -res))
  coords <- paste0(coords$x, "|", coords$y)
  coords <- gsub("\\.", "p", coords)
  coords <- sub("\\|", ".", coords)

  sparseCoords <- paste0(getItems(x, "x", full = TRUE),
                         ".",
                         getItems(x, "y", full = TRUE))
  if (crop) {
    x <- x[sparseCoords %in% coords, , ]
    if (length(x) > 0) {
      sparseCoords <- paste0(getItems(x, "x", full = TRUE),
                             ".",
                             getItems(x, "y", full = TRUE))
    } else {
      sparseCoords <- character(0)
    }
  } else {
    if (!all(sparseCoords %in% coords)) {
      stop("The coordinates of the input object are not a subset of the extended coordinates. ",
           "Try changing res, xRange or yRange.")
    }
  }

  subdims1 <- getSets(x)[grep("^d1\\.", names(getSets(x)))]
  notCoords <- subdims1[!subdims1 %in% c("x", "y")]

  dim1 <- paste0(coords, paste(rep(".NA", length(notCoords)), collapse = ""))

  extended <- new.magpie(dim1, getItems(x, 2), getItems(x, 3))

  # dimOrder subdims x and y to where they are in x
  perm <- match(getSets(x), c("x", "y"))[seq_along(subdims1)]
  perm[is.na(perm)] <- seq_along(notCoords) + 2
  extended <- dimOrder(extended, perm, dim = 1)

  getSets(extended) <- getSets(x)

  posInCoords <- match(sparseCoords, coords)

  # fill subdims for dim 1 other than x and y (e.g. country code) if they were set in x
  dimnames(extended)[[1]][posInCoords] <- getItems(x, 1)

  # fill extended with data from x where available
  extended[posInCoords, , ] <- x

  return(extended)
}

#' add_columns
#'
#' Function adds new columns to the existing magpie object.
#'
#' @param x MAgPIE object which should be extended.
#' @param dim The (sub)dimension to be filled either identified via
#' name or dimension code (see \code{\link{dimCode}} for more information)
#' @param addnm The new elements that should be added to the (sub)dimension
#' @param fill fill value of length 1 for the newly added columns (NA by default)
#' @return The extended MAgPIE object
#' @author Jan Philipp Dietrich, Benjamin Bodirsky
#' @seealso \code{\link{addDim}},\code{\link{dimCode}}
#' @examples
#' a <- maxample("animal")
#' a2 <- add_columns(a, addnm = c("horse", "magpie"), dim = "species", fill = 42)
#' getItems(a2, dim = 3)
#' getItems(a2, dim = 3, split = TRUE)
#' head(a2[, , "magpie"])
#' @family DimensionManipulation
#' @export
add_columns <- function(x, addnm = "new", dim = 3.1, fill = NA) { # nolint: object_name_linter.
  if (length(dim) != 1) stop("dim must be a single (sub)dimension!")
  if (length(fill) != 1) stop("fill value must be of length 1")
  if (!dimExists(dim, x)) stop("dim \"", dim, "\" does not exist")
  x <- clean_magpie(x, what = "sets")
  if (length(addnm) == 0) return(x)
  dim <- dimCode(dim, x)
  add <- addDim(dimSums(x, dim = dim), dim = dim, item = addnm)
  add[, , ] <- fill
  getSets(add) <- getSets(x)
  return(mbind(x, add))
}

#' as.SpatRasterDataset
#'
#' Convert magclass object to a SpatRasterDataset object. Requires the terra package.
#'
#' Calls \code{\link{as.SpatRaster}} and then \code{\link{spatRasterToDataset}}.
#'
#' @param ... arguments passed to as.SpatRaster
#' @return A SpatRasterDataset object
#' @author Pascal Sauer
#' @family MAgPIE-Conversions
#' @export
as.SpatRasterDataset <- function(...) { # nolint: object_name_linter.
  return(spatRasterToDataset(as.SpatRaster(...)))
}

#' Read file in report format
#'
#' This function reads the content of a reporting file (a file in the model
#' intercomparison file format *.mif) into a list of MAgPIE objects or a single
#' MAgPIE object.
#'
#'
#' @param file file name the object should be read from.
#' @param as.list if TRUE a list is returned (default), if FALSE it is tried to
#' merge all information in one MAgPIE object (still under development and
#' works currently only if the entries for the different models and scenarios
#' have exactly the same regions and years).
#' @param showSeparatorWarning Boolean (default value TRUE) that decides whether the warning
#' about the replacement of dots in variable names is displayed (default value) or not.
#'
#' @details
#'
#' The \strong{Model Intercomparison File Format (MIF)} is the default file
#' format for data produced by Integrated Assessment Models. It is based on the
#' common format used for Model Intercomparison Projects such as EMF and SSP
#' with some slight changes/clarifications in its definition. For interactions
#' between models this format should be used. For everything else it is at least
#' recommended to use this format, too.
#'
#' Aim of this standardization is to achieve a more flexible and smooth
#' communication between models and to facilitate the creation of aggregated
#' outputs from integrated assessment scenario runs which then can easily be
#' uploaded to external databases such as the EMF or SSP database. By using this
#' standard most of the required decisions for a working input output interface
#' between models have already been specified which significantly reduces the
#' required work to get a new interaction running.
#'
#' \strong{Definition}
#'
#' The format is characterized by the following features:
#'
#' \itemize{
#'   \item The file ending is ".mif"
#'   \item The file is written in ASCII format
#'   \item Entries are separated with ";", every line ends with a ";"
#'   \item The file always contains a header
#'   \item The format of the header is: \code{Model;Scenario;Region;Variable;Unit;<ADDITIONAL_COLUMNS>;<YEARS>;}
#' }
#'
#' The first 5 entries always have to exist, <ADDITIONAL_COLUMNS> is additional
#' information which can be added optionally (e.g. "Description") and <YEARS>
#' are the years for which data is delivered. <YEARS> are always written as 4
#' digit numbers. In the (very unlikely) case that a year before 1000 is used
#' the number has to start with a 0, e.g. 0950. <ADDITIONAL_COLUMNS> can be
#' anything, there are no further rules at the moment what it can contain.
#' However, there are strict rules for naming these columns. Allowed are single
#' names starting with a capital letter without special characters in it except
#' "_" which is allowed. Examples: "Description" allowed, "More Description" not
#' allowed, "More_Description" allowed, "123Description" not allowed,
#' "Description123" allowed. Scripts using this format must be able to ignore
#' additional columns. For years there are no specific limitations/requirements
#' which years should be reported. Scripts dealing with this data must be able
#' to work with different temporal resolutions. For variables basically
#' everything can be reported here. Missing values have to be marked with "N/A".
#'
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{write.report}}
#' @examples
#' \dontrun{
#' read.report("report.csv")
#' }
#'
#' @family FileIO
#' @export
#' @importFrom utils read.table
#'
read.report <- function(file, as.list = TRUE, # nolint: object_name_linter, cyclocomp_linter.
                        showSeparatorWarning = TRUE) {

  .trim <- function(a) return(gsub("(^ +)|( +$)", "", as.character(a)))

  .returnMagpie <- function(tmp, scenario, model) {
    # replace weird degree symbol in tables
    tmp$Unit <- enc2utf8(tmp$Unit)
    tmp$Unit <- sub(pattern = "\U{00B0}C", replacement = "K", x = tmp$Unit, useBytes = TRUE)
    regions <- unique(as.character(tmp$Region))
    names(regions) <- regions
    years <- sub("X", "y", grep("^X[0-9]{4}$", dimnames(tmp)[[2]], value = TRUE))
    names <- unique(paste(tmp$Variable, "#SPLITHERE# (", tmp$Unit, ")", sep = ""))
    names(names) <- sub("#SPLITHERE#", "", names)
    names <- sub("#SPLITHERE#", "", names)
    # delete dots if they are aparently not used as dimension separator
    ndots <- nchar(gsub("[^\\.]*", "", names))
    if (any(ndots != ndots[1])) {
      names <- gsub("\\.", "p", names)
      if (showSeparatorWarning) {
        warning("Replaced all \".\" with \"p\" to prevent misinterpretation as dim separator")
      }
    }
    # replace weird ° in tables after sub function evaluation
    names <- enc2utf8(names)
    names <- sub(pattern = "\U{00B0}C", replacement = "K", x = names, useBytes = TRUE)
    names(names) <- enc2utf8(names(names))
    names(names) <- sub(pattern = "\U{00B0}C", replacement = "K", x = names(names), useBytes = TRUE)
    mag <- new.magpie(sub("ZZZZZZGLO", "GLO", (sort(sub("GLO", "ZZZZZZGLO", regions)))), years, names)
    yearelems <- grep("^X[0-9]{4}$", dimnames(tmp)[[2]])
    regions[order(sub("GLO", "ZZZZZZGLO", regions))] <- dimnames(mag)[[1]]
    mag <- as.array(mag)
    coord <- cbind(regions[tmp$Region], rep(years, each = dim(tmp)[1]),
                   names[paste(tmp$Variable, " (", tmp$Unit, ")", sep = "")])
    if (dim(coord)[1] > length(mag)) {
      duplicates <- duplicated(as.data.table(coord))
      warning("Duplicate entries found for model \"", model, "\" and scenario \"", scenario,
              "\" and only the last entry found in the data will be used (duplicate entries: ",
              paste(apply(rbind(NULL, unique(coord[duplicates, c(1, 3)])), 1, paste, collapse = "|"),
                    collapse = ", "), ")!")
    }
    # suppress warnings from potential NA conversions
    mag[coord] <- suppressWarnings(as.numeric(as.vector(as.matrix(tmp[, yearelems]))))
    names(dimnames(mag)) <- c("region", "year", "variable")
    mag <- as.magpie(mag, spatial = 1, temporal = 2)
    return(mag)
  }

  .readmif <- function(file) {
    defaultHeader <- c("Model", "Scenario", "Region", "Variable", "Unit", "X2005",
                       "X2010", "X2020", "X2030", "X2040", "X2050", "X2060", "X2070",
                       "X2080", "X2090", "X2100")
    # determine seperator
    s <- read.table(file, sep = ";", header = FALSE, nrows = 1, stringsAsFactors = FALSE)
    if (all(names(s) == "V1")) sep <- "," else sep <- ";"
    # recognize header
    s <- read.table(file, sep = sep, header = FALSE, nrows = 1, stringsAsFactors = FALSE)
    header <- (.trim(s[, 1]) == "Model" | .trim(s[, 1]) == "MODEL")
    # read in raw data
    raw <- read.table(file, sep = sep, header = header, stringsAsFactors = FALSE, na.strings = "N/A")
    uglyFormat <-  all(is.na(raw[, dim(raw)[2]]))
    if (uglyFormat) raw <- raw[, -dim(raw)[2]]

    # rename from uppercase to lowercase
    if (header && .trim(s[, 1]) == "MODEL") {
      names(raw)[1:5] <- defaultHeader[1:5]
    }

    if (!header) {
      if (dim(raw)[2] == length(defaultHeader)) {
        warning("Header is missing. Years are being guessed based on structure!")
        dimnames(raw)[[2]] <- defaultHeader
      } else {
        stop("Cannot read report. No header given and report has not the standard size!")
      }
    }

    output <- list()
    raw$Scenario <- .trim(raw$Scenario)
    raw$Model    <- .trim(raw$Model)
    raw$Region   <- .trim(raw$Region)
    raw$Unit     <- .trim(raw$Unit)
    raw$Variable <- .trim(raw$Variable)

    raw$Model[is.na(raw$Model)] <- "NA"
    raw$Scenario[is.na(raw$Scenario)] <- "NA"

    raw$Region <- sub("R5\\.2", "", raw$Region)
    raw$Region <- sub("World|glob", "GLO", raw$Region)
    models <- unique(raw$Model)
    scenarios <- unique(raw$Scenario)
    for (scenario in scenarios) {
      output[[scenario]] <- list()
      for (model in models) {
        if (nrow(raw[raw$Model == model & raw$Scenario == scenario, ]) > 0) {
          output[[scenario]][[model]] <- .returnMagpie(raw[raw$Model == model & raw$Scenario == scenario, ],
                                                       scenario, model)
          if (!as.list) {
            getNames(output[[scenario]][[model]]) <- paste(scenario, model, getNames(output[[scenario]][[model]]),
                                                           sep = ".")
          }
        }
      }
    }
    return(output)
  }

  # expand wildcards
  fileNameUnexpanded <- file
  file <- Sys.glob(file)
  if (length(file) > 1) {
    output <- NULL
    for (f in file) {
      output <- c(output, .readmif(f))
    }
  } else if (length(file) == 0) {
    stop("File ", fileNameUnexpanded, " could not be found!")
  } else {
    output <- .readmif(file)
  }

  if (!as.list) {
    regions <- Reduce(union, lapply(unlist(output, recursive = FALSE), function(x) {
      getItems(x, dim = 1.1)
    })) # make sure that magpie objects to be merged share the same regions

    .tmpFunc <- function(x) {
      data <- new.magpie(regions, getYears(x), getNames(x), fill = NA)
      data[getItems(x, dim = 1.1), getYears(x), getNames(x)] <- x[getItems(x, dim = 1.1), getYears(x), getNames(x)]
      return(data)
    }

    output <- mbind(lapply(unlist(output, recursive = FALSE), .tmpFunc))
    names(dimnames(output))[3] <- "scenario.model.variable"
  }
  return(output)
}

#' magpie method for tibble::as_tibble
#'
#' @param x A magpie object
#' @inheritParams tibble::as_tibble
#' @return A tibble object
#' @family MAgPIE-Conversions
#' @exportS3Method tibble::as_tibble
as_tibble.magpie <- function(x,
                             ...,
                             .rows = NULL,
                             .name_repair = c("check_unique", # nolint: object_name_linter.
                                              "unique", "universal", "minimal"),
                             rownames = pkgconfig::get_config("tibble::rownames", NULL)) {
  # Turn the magpie object into a data.frame and rename ".value" to "value" - the default column name for values in
  # tibbles.
  y <- as.data.frame(x, rev = 2)
  names(y)[which(colnames(y) == ".value")] <- "value"
  # Turn the data.frame into a tibble
  tibble::as_tibble(x = y,
                    ...,
                    .rows = .rows,
                    .name_repair = .name_repair,
                    rownames = rownames)
}

#' Write a magpie object to a netCDF file
#'
#' @param x A magpie object
#' @param filename Name of the netCDF file to write
#' @param unit Unit of the data, to omit pass "" (empty string)
#' @param ... For future expansion
#' @param compression Level of compression to use (1-9), NA for no compression
#' @param missval The value that encodes NA in the resulting netCDF file
#' @param gridDefinition A vector of 5 numeric values: c(xMin, xMax, yMin, yMax, resolution).
#' Use c(-179.75, 179.75, -89.75, 89.75, 0.5) to write a standard 0.5-degree-resolution
#' lon/lat grid. If NULL, use min/max of coordinates in x and guessResolution
#' @param zname Name of the z dimension in the netCDF file
#' @param progress If TRUE, print progress messages
#' @author Pascal Sauer
#' @keywords internal
writeNC <- function(x, filename, unit, ..., compression = 2, missval = NA,
                    gridDefinition = NULL, zname = "time", progress = FALSE) {
  if (!requireNamespace("ncdf4", quietly = TRUE)) {
    stop("The ncdf4 package is required to write netCDF files, please install it.")
  }
  # fail immediately if arguments are not set
  stopifnot(is.character(filename), is.character(unit))
  if (!(...length() == 0 || all(...names() == "verbose"))) {
    stop("Unknown argument passed to writeNC: ", paste(...names(), collapse = ", "))
  }

  if (is.null(gridDefinition)) {
    coords <- getCoords(x)
    firstX <- min(coords$x)
    lastX <- max(coords$x)
    firstY <- max(coords$y)
    lastY <- min(coords$y)
    res <- guessResolution(coords)
  } else {
    stopifnot(length(gridDefinition) == 5,
              gridDefinition[1] < gridDefinition[2],
              gridDefinition[3] < gridDefinition[4])
    firstX <- gridDefinition[1]
    lastX <- gridDefinition[2]
    firstY <- gridDefinition[4]
    lastY <- gridDefinition[3]
    res <- gridDefinition[5]
    # swap y values before passing gridDefinition to magclass::extend
    gridDefinition[3] <- firstY
    gridDefinition[4] <- lastY
  }
  xCoords <- seq(firstX, lastX, res)
  yCoords <- seq(firstY, lastY, -res)

  if (zname != "time") {
    message("terra will not recognize zname != 'time' as time dimension")
  }

  if (is.null(getItems(x, 3))) {
    getItems(x, 3) <- sub("\\.nc$", "", basename(filename))
  }

  # create netCDF file
  dimVars <- list(ncdf4::ncdim_def("lon", "degrees_east", xCoords),
                  ncdf4::ncdim_def("lat", "degrees_north", yCoords))
  hasTime <- !is.null(getItems(x, 2))
  if (hasTime) {
    dimVars <- c(dimVars, list(ncdf4::ncdim_def(zname, "years since 0",
                                                getYears(x, as.integer = TRUE), unlim = TRUE)))
  }
  vars <- lapply(getItems(x, 3), function(vname) {
    return(ncdf4::ncvar_def(vname, units = unit, dim = dimVars,
                            missval = missval, compression = compression))
  })

  nc <- ncdf4::nc_create(filename, vars = vars)
  withr::defer(ncdf4::nc_close(nc))

  if (hasTime && zname == "time") {
    ncdf4::ncatt_put(nc, "time", "axis", "T")
  }

  messageIf(progress, "Writing ", filename)

  for (i in seq_along(getItems(x, 3))) {
    vname <- getItems(x, 3)[i]
    messageIf(progress, i, "/", length(getItems(x, 3)), " Writing ", vname)
    # memory problems due to the extend happening here? checkout commit
    # 9e780b5fe00f38b82d20edc245e34605c6eb9c46 to get a chunk-based solution
    ncdf4::ncvar_put(nc, vname, extend(x[, , vname], gridDefinition = gridDefinition))
  }
}

messageIf <- function(condition, ...) {
  if (condition) {
    message(...)
  }
}

#' Copy MAgPIE-files
#'
#' This function copies MAgPIE-files from one location to another. During the
#' copying it is also possible to change the file type (e.g. from 'mz' to
#' 'csv')
#'
#'
#' @aliases copy.magpie copy.magpie
#' @param input_file file, that should be copied
#' @param output_file copy destination
#' @param round number of digits the values should be rounded. NULL means no rounding
#' @author Jan Philipp Dietrich
#' @seealso \code{\link{read.magpie}},\code{\link{write.magpie}}
#' @examples
#'
#' # copy.magpie("bla.csv","blub.mz")
#' @family FileIO
#' @export
copy.magpie <- function(input_file, output_file, round = NULL) { # nolint: object_name_linter.
  inType  <- tail(strsplit(input_file, "\\.")[[1]], 1)
  outType <- tail(strsplit(output_file, "\\.")[[1]], 1)
  if (inType == outType && is.null(round)) {
    file.copy(input_file, output_file, overwrite = TRUE)
  } else if (is.null(round)) {
    write.magpie(read.magpie(input_file), output_file)
  } else {
    write.magpie(round(read.magpie(input_file), round), output_file)
  }
}

#' @importFrom methods new
#' @exportMethod ifelse
setMethod("ifelse",
  signature(test = "magpie"),
  function(test, yes, no) {
    if (!all(suppressWarnings(as.integer(test)) %in% c(0, 1))) {
      stop("'test' must only contain booleans!")
    }
    yes <- test * yes
    no <- (1 - test) * no
    yes[!test] <- 0
    no[test] <- 0
    return(yes + no)
  }
)

#' is.temporal, is.spatial
#'
#' Functions to find out whether a vector consists of strings consistent with
#' the definition for auto-detection of temporal or spatial data.
#'
#'
#' @aliases is.temporal is.spatial
#' @param x A vector
#' @return Returns TRUE or FALSE
#' @author Jan Philipp Dietrich
#' @examples
#'
#' is.temporal(1991:1993)
#' is.spatial(c("GLO", "AFR"))
#'
#' @family ObjectInfo
#' @export
is.temporal <- function(x) { # nolint: object_name_linter.
  return(length(grep("^[a-z]?[0-9]{4}$", x)) == length(x))
}

#' Get dataset names
#'
#' Extracts dataset names of a MAgPIE-object
#'
#' setNames is a shortcut to use a MAgPIE object with manipulated data names.
#' The setNames method uses the variable names "object" and "nm" in order to be
#' consistent to the already existing function setNames.
#'
#' @name setNames-methods
#' @aliases setNames setNames,magpie-method
#' setNames,NULL-method
#' @param object MAgPIE object
#' @param nm a vector of names current names should be replaced with. If
#' only one data element exists you can also set the name to NULL.
#' @docType methods
#' @section Methods: \describe{
#'
#' \item{list("signature(object = \"ANY\")")}{ normal setNames method }
#'
#' \item{list("signature(object = \"magpie\")")}{ setNames for MAgPIE objects} }
#' @seealso \code{\link{getNames}},
#' @keywords methods
#' @family GetterSetter
#' @exportMethod setNames

setMethod("setNames",
  signature(object = "magpie"),
  function(object, nm) {
    getNames(object) <- nm
    return(object)
  }
)

#' @importFrom methods new
#' @exportMethod colMeans
setMethod("colMeans",
  signature(x = "magpie"),
  function(x, na.rm = FALSE, dims = 1, ...) { # nolint: object_name_linter.
    xGlo <- colMeans(as.array(x), na.rm = na.rm, ...)
    out <- new("magpie", array(xGlo, dim = c(1, dim(xGlo)), dimnames = c("GLO", dimnames(xGlo))))
    return(out)
  }
)

---
title: "MAGPIE Class Tutorial"
author: "Jan Philipp Dietrich"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MAGPIE Class Tutorial}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

This tutorial provides a basic introduction to magpie objects in R. If you would like to get more details on the concept of and the idea behind magpie objects you can have a look at [magclass-concept](magclass-concept.html), where this is explained in more detail by comparing the magpie class to other data classes in R.

## Generate a magpie object

Generation of magpie objects can either be done from scratch, by reading in files or by converting objects from other classes to magpie objects. For creation of a new magpie object you can use `new.magpie`, conversion of an existing object happens with `as.magpie`:

```{r, echo = TRUE}
library(magclass)

# creating a magpie object with 2 regions, 2 years and 2 different values
m <- new.magpie(cells_and_regions = c("AFR", "CPA"),
                years = c(1995, 2000),
                names = c("bla", "blub"),
                sets = c("region", "year", "value"),
                fill = 0)
print(m)

# converting a simple vector with one value per region to a magpie object
v <- c(ENG = 10, USA = 20, BRA = 30, CHN = 40, IND = 50)
m2 <- as.magpie(v)
str(m2)
```


In the example above the names were automatically detected as regions, if for some reason the automatic detection fails you can also indicate to what type of dimension the data belongs. Lets assume in the following example, that the names are not representing regions but something else. The argument `spatial=0` indicates the non-existence of a spatial dimension, but it can also be used to point to the spatial dimension in the data.

```{r, echo = TRUE}
m3 <- as.magpie(v, spatial = 0)
str(m3)
```

## Accessing magpie objects

For the following a example data set containing population data is used.

```{r, echo = TRUE}
pm <- maxample("pop")
```

### General properties

Let's first have a look at the content of the magpie object:

Show the structure of the object:
```{r, echo = TRUE}
str(pm)
```

Show the first elements:
```{r, echo = TRUE}
head(pm)
```
Show the last elements:
```{r, echo = TRUE}
tail(pm)
```

Which elements are there?
```{r, echo = TRUE}
getItems(pm)
```

Which spatial elements are there?
```{r, echo = TRUE}
getItems(pm, dim = 1)
```

Which scenarios?
```{r, echo = TRUE}
getItems(pm, dim = 3)
```

`getItems` as well as most of the other functions allow to select a dimension either via its dimension code (as done in the previous example), or via its dimension name:

```{r, echo = TRUE}
getItems(pm, dim = "scenario")
```

In terms of readability it is recommended to use the dimension name where possible, but one has to keep in mind that sometimes the dimension name might not be well defined and vary from case to case. In these instances it is safer to use the dimension code.

What are the sets (dimensions names) of the data?
```{r, echo = TRUE}
getSets(pm)
```

are there any comments which come with the data?
```{r, echo = TRUE}
getComment(pm)
```

let's have a look at a higher dimensional object

```{r, echo = TRUE}
a <- maxample("animal")
```

what is the full dimensionality of this object?
```{r, echo = TRUE}
getItems(a)
```

...split in sub-dimensions
```{r, echo = TRUE}
getItems(a, split = TRUE)
```

These functions can also be used to manipulate the object:

set a comment
```{r, echo = TRUE}
getComment(pm) <- "This is a comment!"
getComment(pm)
```
...or alternatively
```{r, echo = TRUE}
pm2 <- setComment(pm, "This is comment for pm2!")
getComment(pm2)
```
rename 1st region in "RRR" 
```{r, echo = TRUE}
getItems(pm, dim = 1)[1] <- "RRR"
```
rename region set in "zones" 
```{r, echo = TRUE}
getSets(pm)[2] <- "year"
```

Finally, to get a quick impression of the content of the object, you can get a rough plot:
```{r, fig.show}
mplot(pm)
```

### Subsets

Now that we have had a look into the structure of the object let's extract some subsets out of it. There are different methods that can be used to extract data from a magpie object. Here are some examples:

Return all A2 related data for LAM and the years 2005 and 2015
```{r, echo = TRUE}
pm["LAM", c(2005, 2015), "A2"]
```

Return data for regions which have "AS" in its name (pmatch allows for partial matching of the given search string)
```{r, echo = TRUE}
pm["AS", , , pmatch = TRUE]
```

If you want to specifically select from one dimension from which you have the dimension name:
```{r, echo = TRUE}
mselect(pm, scenario = "B1", i = c("FSU", "LAM"))
```
Or you can use alternatively:
```{r, echo = TRUE}
pm[list(i = c("FSU", "LAM")), , list(scenario = "B1")]
```


### Data transformations / calculations

Now we can perform some calculations with it.

take a subset of the data as an example
```{r, echo = TRUE}
d <- head(pm)
```

create a new object with some fancy calculations
```{r, echo = TRUE}
d2 <- d^2 + 12 * d + 99 / exp(d)
getItems(d2, dim = 3) <- c("NEWSCEN1", "NEWSCEN2")
getSets(d2)[3] <- "newscen"
d2
```

multiply both data sets with each other
```{r, echo = TRUE}
d <- d * d2
d
```
Because we changed the names of the elements in the data dimension it is assumed that they reflect different dimensions. Therefor the object is blown up in size instead of just having the elements multiplied pairwise with each other as it happens when two objects with identical dimensionality are multiplied with each other:

```{r, echo = TRUE}
d2 * d2
```

sum over the data dimension
```{r, echo = TRUE}
dimSums(d, dim = 3)
```
..over the second data dimension only
```{r, echo = TRUE}
dimSums(d, dim = 3.2)
```
..or alternatively addressed by name
```{r, echo = TRUE}
dimSums(d, dim = "newscen")
```
sum over regions and first data dimension
```{r, echo = TRUE}
dimSums(d, dim = c(1, 3.1))
```

apply a lowpass filter on the data
```{r, echo = TRUE}
lowpass(d)
```

do a linear interpolation of the data over time (yearly)
```{r, echo = TRUE}
time_interpolate(d[, , 1], 2005:2030)
```

split the data, do some calculations and bind it back together
```{r, echo = TRUE}
d1 <- d[, 1:3, ] * 100
d2 <- d[, 4:6, ] * (-1)
dd <- mbind(d1, d2)
dd
```

set all values greater than 0.5 to 0.51
```{r, echo = TRUE}
d[d > 0.5] <- 0.51
d
```

round the data
```{r, echo = TRUE}
round(d, 0)
```

---
title: "magclass 6"
author: "Jan Philipp Dietrich"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{magclass 6}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

The step from magclass 5.29 to 6.00 marks a major rewrite of the whole package. While the definition of the magclass object has been refined and generalized over time (allowing multiple sub-dimensions in all main dimensions, introduction of set names, etc.) some functions of the package still relied on implicit assumptions which were not necessarily valid anymore for the new definitions. This led to situations in which functions might only work for specific cases (such as functions for sub dimension manipulation only working in the third dimension) or even return wrong results in some instances. 

This tutorial will go through the major changes coming with version 6 and what they mean for working with magclass objects.

## Testing

Magclass comes now with more than 600 unit tests which are covering more than 90% lines of code of the total package. These tests should help to detect code breaking changes already early in the process and reduce the risk of pushing incompatible modifications to our main repository. 
In addition, it is planned to expand the test collection with every new feature or newly observed bug so that if issues arise that they only arise once. 

## (soon to be) deprecated functions

There are some function in magclass building on the old standard of having exactly one spatial and one temporal dimension (no subdimensions), which has been generalized in the meantime (up to 9 subdimension in every of the three main dimensions allowed). So far only two of them received a deprecation warning, which are `fulldim` (does not split subdimensions in first two dimensions) and `getRegionList` (assumes to only have regions in the first dimension). 
Functionality can now be found in the new function `getItems` which allows to return elements of (sub-)dimensions in various ways:

```{r, echo = TRUE}
library(magclass)
a <- maxample("animal")
p <- maxample("pop")
```

instead of `fulldim(a)` to receive a list of all sub-dimensions use now:

```{r, echo = TRUE}
getItems(a, split = TRUE)
```


In contrast to `fulldim` this splits all dimensions in sub-dimensions and reports the corresponding set names.

It is also possible to split a specific main dimension into sub-dimensions:

```{r, echo = TRUE}
getItems(a, dim = 1, split = TRUE)
```

Instead of `getRegionList(a)` one should use `getItems` as well:

```{r, echo = TRUE}
getItems(a, dim = 1.1, full = TRUE)
```

This command gives the same results as `getRegionList` but makes it more obvious what the command is doing (returning the first spatial sub-dimension in full length)

Some other functions such as `getRegions` or `getCells` have not been formally deprecated due to that broad usage, but are encouraged to be replaced as well with `getItems`. 

Another function which recieved a deprecation warning is `write.report2`. So far the package contained two very similar functions `write.report` and `write.report2` from which the latter is the more efficient one. In the new version the old `write.report` function has been replaced with the code of `write.report2` so that in the future all computation are based on the same function. 

## New feature: extended sub-dimension support

So far `dimSums`, `add_columns`, `addDim` were only functioning for the sub-dimensions in the third dimension while `getCPR` were only functioning for the first dimension. All these functions have been completely rewritten and work now for all sub-dimensions and are able to address a dimension either by its `dimCode` or set name:

```{r, echo = TRUE}
dimSums(a, c("x", "y", "cell", "day", "month", "species", "color"))

getCPR(a, 3.2)

addDim(p[, 1, 1], 1.2)

add_columns(p[, 1, 1], dim = "i")
```
## New magclass subsetting variant

To ease the process to write generalized functions which work for all main dimensions a new sub-setting option has been introduced which allows to specify the main dimension in which the sub-setting should take place via a new `dim` argument:

```{r, echo = TRUE}
p[3, dim = 2] # equivalent to p[,3,]
```
## Generalized handling of spatial raster data

Previously spatial data could only be handled for a pre-defined set of 59199 0.5x0.5 degree cells for which cells were numbered from 1 to 59199 and coordinates had to be known by the user and properly assigned to these cells. 
With the update raster data instead gets assigned its coordinates in the spatial dimension allowing also to handle data for differing resolutions and/or coverage. Accordingly, `read.magpie` and `write.magpie` have been adapted to this new behavior and custom implemetations for `ncdf4` and `asc` file support have been replaced with the read/write routines provided by the `raster` package. 
For applications this means that the behavior of code reading/writing these files might change and might require some modifications, but for the future it should ease the handling and exchange of spatial data with magclass objects.

## One-stop-shop-function getItems

As indicated earlier `getItems` can be used to replace `fulldim` and `getRegionList`. But its functionality goes beyond it making it the function to be used when elements of (sub-)dimensions should be read, added, removed or modified. While so far many different functions had to be remembered for these applications (such as `getCells`, `getNames`, `addDim`) it is now recommended to first look at `getItems`. In the following some examples of its functionality:

Returning a sub-dimension:

```{r, echo = TRUE}
getItems(a, dim = 3.2)
getItems(a, dim = "color")
```

Returning the full vector of a sub-dimension:

```{r, echo = TRUE}
getItems(a, dim = 3.2, full = TRUE)
```

Returning a full main dimension...

```{r, echo = TRUE}
getItems(a, dim = 3)
```

...split into its sub-dimensions

```{r, echo = TRUE}
getItems(a, dim = 3, split = TRUE)
```

Replace elements element-wise:

```{r, echo = TRUE}
getItems(a, dim = "color") <- paste0("color", 1:4)
getItems(a, dim = 3)
```

Replace the whole vector:

```{r, echo = TRUE}
getItems(a, dim = "color", full = TRUE) <- paste0("color", 1:5)
getItems(a, dim = 3)
```

Delete a dimension:

```{r, echo = TRUE}
getItems(a, dim = "color") <- NULL
getItems(a, dim = 3)
```

Add a dimension:

```{r, echo = TRUE}
getItems(a, maindim = 3, dim = "newcolor") <- paste0("color", 1:5)
getItems(a, dim = 3)
```

All these manipulations work for all (sub-)dimensions of the object.

## Internal code modifications

As the complete package underwent a code review also many other functions which have not been mentioned here received modifications. While this should affect the behavior of these functions it improved in some cases the efficiency of the underlying computations.  

---
title: "Concept behind magclass"
author: "Jan Philipp Dietrich"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Concept behind magclass}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

# magclass vs other data-classes

In order to understand the magclass format it is helpful to know the basic data structures in R:

## dataclass: matrix

A matrix is a 2D table (columns and rows) with values in it which all belong to the same variable type (e.g. all numeric or all characters). It can have dimnames giving each row and each column a name.

```{r, echo = TRUE}
a <- matrix(1:9, 3, 3, dimnames = list(c("AFR", "CPA", "EUR"),
                                       paste0("y", 2000:2002)))
a
```

It is very handy for mathematical operations such as matrix multiplications and so on, but quite limited due to its restriction to 2 dimenions and only one allowed variable type for all entries.

## dataclass: array

An array is the extension of a matrix to more than two dimensions. A 2D array is identical to a matrix. As the matrix also the array is a very efficient structure for mathematical operations.

```{r, echo = TRUE}
b <- array(1:18, dim = c(3, 3, 2),
           dimnames = list(paste0("y", 2000:2002),
                           c("AFR", "CPA", "EUR"),
                           c("IndexA", "IndexB")))
b
```

## dataclass: dataframe

A dataframe is similar in its structure to a matrix but has quite different advantages and disadvantages. As the matrix also the data frame is only two dimensional. The major difference between matrices and dataframes is that dataframes allow for different variable types per column. That means that one column can contain characters describing the data whereas another column can contain the values itself. Having this extension it is much easier to store higher dimensional data in it, as you can always use a structure in which each column except the last represent one dimension and the last dimension contains the data itself. The higher flexibility means that you can store the same data in very different ways, making it often hard to predict how the data frame will look like.

```{r, echo = TRUE}
library(magclass)

as.data.frame(b)

as.data.frame(as.magpie(b))
```

In the first example the array is directly transformed to a data frame whereas in the second example the matrix is transformed first to a magclass object and then to a data frame. Both data frames still contain the same data but in a different representation.

## dataclass: magclass

The main idea behind the magclass object was to have a data class for which it is always quite clear how the data is organized. Having this knowledge independent of the data itself allows to write more generic functions which can treat various kinds of data sets without knowing anything about it in advance. 

The magclass data class is a child class of an array, meaning that it is based on most of its features. Contrary to an array the magpie object is always three dimensional, containing spatial information in the first dimension, temporal information in the second dimension and everything else in the third dimension.

```{r, echo = TRUE}
as.magpie(b)
```

However, that limitation to a 3D-structure does not mean that you are limited to data which only has 3 dimensions. Also data with more than one so-called "data dimension" (a dimension which is neither temporal nor spatial) can be handled:

```{r, echo = TRUE}
ar <- array(1:6, c(3, 2), list(c("bla", "blub", "ble"),
                               c("up", "down")))
a <- as.magpie(ar)
head(a)

getItems(a, dim = 3)

getItems(a[, , "up"], dim = 3)

getItems(a[, , "up", drop = TRUE], dim = 3)

getItems(a, dim = 3, split = TRUE)
```

If an array has three dimensions and you take in one dimension exactly one element the array will be reduced to an 2D object, whereas the magclass object will always remain as 3D object. In addition to that a lot of checks will be executed when you do calculations with magclass objects that should prevent you from doing unwanted things such as accidentally sum up the wrong numbers if your objects are ordered differently:

```{r, echo = TRUE}
c <- as.magpie(b)
# change order of spatial components
d <- c[3:1, , ]

# correct summation based on dimension names:
c + d

# wrong summation (dimension names ignored):
as.array(c) + as.array(d)
```

This additional comfort and safety is bought with performance meaning that if you are doing calculations which are very resource consuming you might want to convert your magclass object to an array, do your calculations (for which you then have to be very careful) and convert it back to a magclass object.

## Why use magclass for your work?

Magclass can be a good choice if you work with tempo-spatial data and want to establish a data standard for exchange between different R functions. If everybody is writing functions based on magclass objects all these functions will be directly compatible to each other. Having this compatibility your work will directly profit from the work of others, e.g. you are able to use plotting routines for your purposes even though they were developed for a completely different data source. Your code will be highly modular so that you can easily adapt it to modified requirements (e.g. exchanging a data read function by an API to a database).
Having everybody working with the same functions can also lower the number of bugs in your code as each part of the code will be tested by several people and bugs will be found easier and fixed faster.

But why use magclass objects as this standard class for all functions and not something else as for instance data frames?

The main argument is that magclass objects are more standardized than the default data classes in R. It is always clear that the first dimension contains the spatial information, the second dimension contains temporal information and the third dimension contains the data. This makes it easier to develop generic plots because it is clear how the data has to be accessed. 

---
title: "MAGPIE Class Object Expansion"
author: "Jan Philipp Dietrich"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MAGPIE Class Object Expansion}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
# set back to default
options(magclass_expand_version = 2.1)
options(magclass_setMatching = FALSE)
```

One of the most important features of magclass objects is their auto expansion when used in calculations with other magclass objects: Dimensions of the objects in the calculation are compared and tried to be matched accordingly. Afterwards all objects are expanded in dimensionality to the superset of dimensions involved in the calculation.

This tutorial will go through different variants of this auto expansion logic and explain the differences.

## Auto expansion v1 vs v2

This auto expansion can follow different rules. In magclass version 1 to 4 it was assumed that there is exactly one spatial and one temporal dimension, but various data dimensions (auto expand version 1). Beginning with version 5 of the package this logic has been harmonized and it is now assumed for all main dimensions (spatial, temporal, data) that they consist of more than one subdimension (auto expand version 2). 
With the updated auto expansion calculation, which previously have led to an error, will now work:

Creating two magpie objects with different world regions:
```{r, echo = TRUE}
library(magclass)
a <- b <- maxample("pop")
getItems(b, dim = 1)[1] <- "AAA"
```

The multiplication for auto expand version 1 (magclass < 5.0) failed, but the same calculation with auto expand version 2.1 (magclass >= 5.0) works:
```{r, echo = TRUE}

options(magclass_expand_version = 2.1) # default setting for magclass >= 5.0
head(a * b)
```

As this modification only changes cases which did not work previously it could be included without causing backwards compatibility issues. 

## Set matching

Another way of modifying auto expansion in magclass is the use of set matching. By default dimensions are compared based on their elements: If dimensions are found which contain the same elements it is assumed that it is the same dimension. While this assumption is correct in many cases it fails in some others, e.g. if an import-export matrix should be created. 
A better way of distinguishing whether dimensions are identical or not is the use of the set names (the names of the dimension). If they are identical it can be assumed that it is the same dimension otherwise it can be assumed that it is not.

Due to the absense of set names in initial versions of magclass the dimension matching was initially implement based on element comparison. With version 5 of the package set matching is now also supported but switched off by default due to backwards compatibility issues. If you want to switch it on (recommended) you can do so by setting the option `options(magclass_setMatching=TRUE)`. However, if you do so be prepared that the package behavior will be quite different in many cases as the following examples show:

### Idential set elements, different set names

Without set matching, identical set elements but different set names produce a simple entity matching without expansion:

```{r, echo = TRUE}
options(magclass_expand_version = 2.1) # default setting for magclass >= 5.0

a <- b <- maxample("pop")

getSets(a)[1] <- "import"
getSets(b)[1] <- "export"

options(magclass_setMatching = FALSE)
head(a * b)
```

With set matching the same calculation will expand the spatial dimension to a cross product of each region with each other:

```{r, echo = TRUE}
options(magclass_setMatching = TRUE)
head(a * b)
```

### Different set element, identical set names

If the set elements are different but the set names identical this will lead to an expansion of the spatial dimension in the case without set matching (as the set name is just ignored):

```{r, echo = TRUE}
options(magclass_expand_version = 2.1) # default setting for magclass >= 5.0

a <- b <- maxample("pop")

getItems(a, dim = 1)[1] <- "AAA"
getItems(b, dim = 1)[1] <- "BBB"

options(magclass_setMatching = FALSE)
head(a * b)
```

With set matching it will return an error as identical set names but differing elements indicates most likely an error.

```{r, echo = TRUE, error=TRUE}
options(magclass_setMatching = TRUE)
head(a * b)
```

To ensure that existing calculations continue to work as they did before use `options(magclass_setMatching=FALSE)`, to write new code more consistently with magclass objects use `options(magclass_setMatching=TRUE)`.

```{r, echo = FALSE}
# set back to default
options(magclass_expand_version = 2.1)
options(magclass_setMatching = FALSE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getYears.R, R/setYears.R
\name{getYears}
\alias{getYears}
\alias{getYears<-}
\alias{setYears}
\title{Get years}
\usage{
getYears(x, as.integer = FALSE)

getYears(x) <- value

setYears(object, nm = NULL)
}
\arguments{
\item{x, object}{MAgPIE object}

\item{as.integer}{Switch to decide, if output should be the used year-name
(e.g. "y1995") or the year as integer value (e.g. 1995)}

\item{value, nm}{Years the data should be set to. Either supplied as a vector
of integers or a vector of characters in the predefined year format
("y0000"). If only 1 year exist you can also set the name of the year to
NULL.}
}
\value{
getYears returns years of the MAgPIE-object, whereas setYears
returns the MAgPIE object with the manipulated years.
}
\description{
Extracts years of a MAgPIE-object
}
\details{
setYears is a shortcut to use a MAgPIE object with manipulated year names.
setYears uses the variable names "object" and "nm" in order to be consistent
to the already existing function setNames.
}
\section{Functions}{
\itemize{
\item \code{getYears(x) <- value}: rename years

\item \code{setYears()}: set years

}}
\examples{

a <- as.magpie(1)
getYears(a)
 setYears(a, 1995)

}
\seealso{
\code{\link{getRegions}}, \code{\link{getNames}},
\code{\link{setNames}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
\code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hasSets.R
\name{hasSets}
\alias{hasSets}
\title{Has Sets}
\usage{
hasSets(x)
}
\arguments{
\item{x}{MAgPIE object}
}
\value{
Boolean indicating whether coordinates were found or not
}
\description{
Checks, whether set names have been set
}
\examples{

hasSets(maxample("pop"))
hasSets(maxample("animal"))

}
\seealso{
\code{\link{getCoords}}

Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isYear.R
\name{isYear}
\alias{isYear}
\title{isYear}
\usage{
isYear(x, with_y = TRUE)
}
\arguments{
\item{x}{A vector}

\item{with_y}{indicates which dataformat years have to have (4-digit without
y (e.g.1984) or 5digit including y (y1984))}
}
\value{
Returns a vector of the length of x with TRUE and FALSE
}
\description{
Function to find out whether a vector consists of strings in the format
"yXXXX" or "XXXX" with X being a number
}
\examples{

x <- c("1955", "y1853", "12a4")
isYear(x, with_y = TRUE)
isYear(x, with_y = FALSE)
}
\seealso{
Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{ncells}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Benjamin Bodirsky
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/withAlignedDims.R
\name{withAlignedDims}
\alias{withAlignedDims}
\title{withAlignedDims}
\usage{
withAlignedDims(func, funcName = "anonymous func", ...)
}
\arguments{
\item{func}{The function that should be executed with prior alignment of dimensions}

\item{funcName}{The name of the resulting function, to improve error messages}
}
\description{
(for internal use) Executes the passed function after reordering items
in each dimension for all passed magpie objects
}
\author{
Pascal Sauer, Patrick Rein
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/population_magpie.R
\name{population_magpie}
\alias{population_magpie}
\title{population_magpie}
\value{
A2 and B1 population scenario from SRES
}
\description{
Example dataset for a regional MAgPIE object
}
\author{
Benjamin Bodirsky
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimExists.R
\name{dimExists}
\alias{dimExists}
\title{dimExists}
\usage{
dimExists(dim, x, sep = ".")
}
\arguments{
\item{dim}{A vector of dimension numbers or dimension names which should be
checked for.}

\item{x}{MAgPIE object in which the dimensions should be searched for.}

\item{sep}{A character separating joined dimension names.}
}
\value{
Boolean indicating whether the dimension exists or not.
}
\description{
This function checks whether a dimension given by name or number exists.
}
\examples{

pop <- maxample("pop")
dimExists(c("t", "scenario", "blablub"), pop)

}
\seealso{
\code{\link{dimCode}}

Other ObjectInfo: 
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setItems.R
\name{setItems}
\alias{setItems}
\title{Set Items}
\usage{
setItems(x, dim, value, maindim = NULL, raw = FALSE)
}
\arguments{
\item{x}{MAgPIE object}

\item{dim}{Dimension for which the items should be returned. Either number or
name of dimension or a vector of these. See \code{\link{dimCode}} for more details.}

\item{value}{a vector with the length of the main dimension the dimnames should be replaced in / added to.
If set to NULL the corresponding dimension will be removed.}

\item{maindim}{main dimension the data should be added to (does not need to be set if \code{dim} exists
in the data. Should be set if \code{dim} might not exist, or if \code{dim} might potentially exist
in a different main dimension than the one anticipated).}

\item{raw}{if set to FALSE inputs will be corrected (e.g. dots replaced by the letter "p") if necessary. If
TRUE data will be written as is (risking the creation of inconsistent objects).}
}
\value{
the manipulated MAgPIE object
}
\description{
Set items of a given (sub-)dimension of a MAgPIE-object
}
\examples{
x <- maxample("pop")
setItems(x, "i", paste0("REG", 1:ncells(x)))
}
\seealso{
\code{\link{getItems}}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/suppressSpecificWarnings.R
\name{suppressSpecificWarnings}
\alias{suppressSpecificWarnings}
\title{suppressSpecificWarnings}
\usage{
suppressSpecificWarnings(expr, regularExpr, fixed = FALSE)
}
\arguments{
\item{expr}{The expression/code to evaluate, can be a block of code inside curly braces.}

\item{regularExpr}{Only warnings matching this regular expression are suppressed.}

\item{fixed}{Match the literal string given by regularExpr instead of interpreting it
as a regular expression. Passed to \code{\link{grepl}}.}
}
\value{
The result of evaluating expr.
}
\description{
Like \code{\link{suppressWarnings}}, but instead of suppressing all warnings this only suppresses
warnings if they match the given pattern.
}
\seealso{
\code{\link{suppressWarnings}}
}
\author{
Pascal Sauer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSets.R
\name{getSets}
\alias{getSets}
\alias{getSets<-}
\title{Get sets}
\usage{
getSets(x, fulldim = TRUE, sep = ".")

getSets(x, fulldim = TRUE, sep = ".") <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{fulldim}{bool: Consider dimension 3 as a possible aggregate of more
dimensions (TRUE) or stick to it as one dimension (FALSE)}

\item{sep}{A character separating joined dimension names}

\item{value}{A vector with set names you want to replace the current set
names of the object with.}
}
\value{
Sets of the MAgPIE-object. If no information about contained sets is
available NULL
}
\description{
Extracts sets of a MAgPIE-object if available
}
\section{Functions}{
\itemize{
\item \code{getSets(x, fulldim = TRUE, sep = ".") <- value}: replace set names

}}
\examples{

a <- new.magpie("GLO.1", 2000, c("a.o1", "b.o1", "a.o2"))
 getSets(a) <- c("reg", "cell", "t", "bla", "blub")
 getSets(a)

 getSets(a)["d3.1"] <- "BLA"
 getSets(a, fulldim = FALSE)
 getSets(a)
}
\seealso{
\code{\link{getRegions}},
\code{\link{getNames}},\code{\link{getYears}}, \code{\link{getCPR}},
\code{\link{read.magpie}}, \code{\link{write.magpie}},
\code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Markus Bonsch, Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRegionList.R
\name{getRegionList}
\alias{getRegionList}
\alias{getRegionList<-}
\title{Get a list of celluare region-belongings}
\usage{
getRegionList(x)

getRegionList(x) <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{value}{A vector with ncell elements containing the regions of each
cell.}
}
\value{
A vector with ncell elements containing the region of each cell.
}
\description{
Extracts a vector containing the region of each cell of a MAgPIE-object
}
\section{Functions}{
\itemize{
\item \code{getRegionList(x) <- value}: set region names

}}
\examples{

# a <- read.magpie("example.mz")
# getRegionList(a)
}
\seealso{
\code{\link{getRegions}},\code{\link{getYears}},
\code{\link{getNames}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
\code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getComment.R, R/setComment.R
\name{getComment}
\alias{getComment}
\alias{getComment<-}
\alias{setComment}
\title{getComment}
\usage{
getComment(x)

getComment(x) <- value

setComment(object, nm = NULL)
}
\arguments{
\item{x, object}{MAgPIE object}

\item{value, nm}{A vector containing the comment.}
}
\value{
getComment returns the comment attached to a MAgPIE-object, NULL if
no comment is present. setComment returns the magpie object with the
modified comment.
}
\description{
Extracts the comment from a MAgPIE-object
}
\section{Functions}{
\itemize{
\item \code{getComment(x) <- value}: set comment

\item \code{setComment()}: set comment

}}
\examples{

a <- as.magpie(1)
 # returns NULL
 getComment(a)
 # set the comment
 getComment(a) <- c("bla", "blubb")
 getComment(a)

}
\seealso{
\code{\link{getRegions}}, \code{\link{getNames}},
\code{\link{getYears}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
\code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Markus Bonsch
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.SpatRasterDataset.R
\name{as.SpatRasterDataset}
\alias{as.SpatRasterDataset}
\title{as.SpatRasterDataset}
\usage{
as.SpatRasterDataset(...)
}
\arguments{
\item{...}{arguments passed to as.SpatRaster}
}
\value{
A SpatRasterDataset object
}
\description{
Convert magclass object to a SpatRasterDataset object. Requires the terra package.
}
\details{
Calls \code{\link{as.SpatRaster}} and then \code{\link{spatRasterToDataset}}.
}
\seealso{
Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as.data.frame-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{unwrap}()},
\code{\link{wrap}()}
}
\author{
Pascal Sauer
}
\concept{MAgPIE-Conversions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ndim.R
\name{ndim}
\alias{ndim}
\title{Count sub-dimensions}
\usage{
ndim(x, dim = NULL)
}
\arguments{
\item{x}{A MAgPIE-object}

\item{dim}{main dimension in which the sub-dimensions should be counted. If NULL
the sum of all subdimensions is returned}
}
\value{
Number of subdimensions
}
\description{
Functions to count the subdimensions of an
MAgPIE-object
}
\examples{

  a <- maxample("animal")
  ndim(a)
  ndim(a,1)
  ndim(a,2)
  ndim(a,3)

}
\seealso{
Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxample.R
\name{maxample}
\alias{maxample}
\title{maxample}
\usage{
maxample(data)
}
\arguments{
\item{data}{name of the example data set. Currently available are "pop" (regional population data,
previously named "population_magpie"), "animal" (fictional, high-dimensional animal sighting data set)
and "bilateral" (fictional, bilateral trade cost data set).}
}
\value{
the chosen example data set
}
\description{
A collection of magclass example data sets
}
\examples{

p <- maxample("pop")
str(p)

a <- maxample("animal")
str(a)
getItems(a, split = TRUE)
}
\seealso{
Other Display: 
\code{\link{head.magpie}()},
\code{\link{mplot}()},
\code{\link{print.magpie}()},
\code{\link{show-methods}},
\code{\link{str.magpie}()}
}
\author{
Jan Philipp Dietrich
}
\concept{Display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show.R
\docType{methods}
\name{show-methods}
\alias{show-methods}
\alias{show}
\alias{show,ANY-method}
\alias{show,magpie-method}
\title{~~ Method for function \code{show} ~~}
\usage{
\S4method{show}{magpie}(object)
}
\arguments{
\item{object}{A MAgPIE-object}
}
\description{
Show a magpie object by calling the default show method, print, or str,
depending on options("magclass_show_func").
}
\seealso{
Other Display: 
\code{\link{head.magpie}()},
\code{\link{maxample}()},
\code{\link{mplot}()},
\code{\link{print.magpie}()},
\code{\link{str.magpie}()}
}
\author{
Pascal Sauer
}
\concept{Display}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.array.R
\docType{methods}
\name{as.array-methods}
\alias{as.array-methods}
\alias{as.array,ANY-method}
\alias{as.array,magpie-method}
\title{~~ Methods for Function as.array ~~}
\usage{
\S4method{as.array}{magpie}(x)
}
\arguments{
\item{x}{object which should be converted to an array}
}
\description{
~~ Methods for function \code{as.array} ~~
}
\section{Methods}{
 \describe{

\item{list("signature(x = \"ANY\")")}{ standard as.array-method }

\item{list("signature(x = \"magpie\")")}{ Conversion takes place just by
removing MAgPIE-object specific elements } }
}

\seealso{
Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.data.frame-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{unwrap}()},
\code{\link{wrap}()}
}
\concept{MAgPIE-Conversions}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean_magpie.R
\name{clean_magpie}
\alias{clean_magpie}
\title{MAgPIE-Clean}
\usage{
clean_magpie(x, what = "all", maindim = 1:3)
}
\arguments{
\item{x}{MAgPIE object which should be cleaned.}

\item{what}{term defining what type of cleaning should be performed. Current
modes are "cells" (removes cell numbers if the data seems to be regional -
this should be used carefully as it might remove cell numbers in some cases
in which they should not be removed), "sets" (making sure that all
dimensions have names), "items" (replace empty elements with single spaces " ")
and "all" (performing all available cleaning methods)}

\item{maindim}{main dimension(s)  the cleaning should get applied to.}
}
\value{
The eventually corrected MAgPIE object
}
\description{
Function cleans MAgPIE objects so that they follow some extended magpie
object rules (currently it makes sure that the dimnames have names and
removes cell numbers if it is purely regional data)
}
\examples{

pop <- maxample("pop")
a <- clean_magpie(pop)
}
\seealso{
\code{"\linkS4class{magpie}"}

Other ObjectCreation: 
\code{\link{complete_magpie}()},
\code{\link{copy.attributes}()},
\code{\link{new.magpie}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectCreation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/magpie-class.R
\docType{class}
\name{magpie-class}
\alias{magpie-class}
\alias{as.magpie}
\alias{as.magpie-methods}
\alias{as.magpie,magpie-method}
\alias{as.magpie,array-method}
\alias{as.magpie,lpj-method}
\alias{as.magpie,data.frame-method}
\alias{as.magpie,numeric-method}
\alias{as.magpie,NULL-method}
\alias{as.magpie,quitte-method}
\alias{as.magpie,tbl_df-method}
\alias{as.magpie,RasterBrick-method}
\alias{as.magpie,logical-method}
\alias{as.magpie,RasterLayer-method}
\alias{as.magpie,RasterStack-method}
\alias{as.magpie,SpatRaster-method}
\alias{as.magpie,LPJmLData-method}
\alias{as.magpie,SpatVector-method}
\alias{is.magpie}
\alias{[,magpie-method}
\alias{[,magpie,ANY,ANY-method}
\alias{[<-,magpie,ANY,ANY-method}
\alias{[<-,magpie-method}
\alias{Ops,magpie,magpie-method}
\alias{Ops,magpie,numeric-method}
\alias{Ops,numeric,magpie-method}
\alias{ifelse,magpie-method}
\alias{is.na,magpie-method}
\alias{is.nan,magpie-method}
\alias{is.infinite,magpie-method}
\alias{is.finite,magpie-method}
\title{Class "magpie" ~~~}
\arguments{
\item{x}{An object that should be either tested or transformed as/to an
MAgPIE-object.}

\item{...}{additional arguments supplied for the conversion to a MAgPIE
object. Allowed arguments for arrays and dataframes are \code{spatial} and
\code{temporal} both expecting a vector of dimension or column numbers which
contain the spatial or temporal information. By default both arguments are
set to NULL which means that the \code{as.magpie} will try to detect
automatically the temporal and spatial dimensions. The arguments will just
overwrite the automatic detection. If you want to specify that the data does
not contain a spatial or temporal dimension you can set the corresponding
argument to 0. In addition \code{as.magpie} for data.frames is also
expecting an argument called \code{datacol} which expects a number stating
which is the first column containing data. This argument should be used if
the dimensions are not detected corretly, e.g. if the last dimension column
contains years which are then detected as values and therefore interpreted
as first data column. In addition an argument \code{tidy=TRUE} can be used
to indicate that the data.frame structure is following the rules of tidy
data (last column is the data column all other columns contain dimension
information). This information will help the conversion. \code{sep} defines
the dimension separator (default is ".") and \code{replacement} defines how
the separator as a reserved character should be converted in order to not
mess up with the object (default "_").
Another available argument for conversions of data.frames and quitte
objects to magpie is \code{filter} if set to TRUE (default)
"." (separator) will be replaced withe the \code{replacement} character and
empty entries will be replaced with a single space. If set to FALSE no filter
will be applied to the data.}
}
\description{
The MAgPIE class is a data format for cellular MAgPIE data with a close
relationship to the array data format. \code{is.magpie} tests if \code{x} is
an MAgPIE-object, \code{as.magpie} transforms \code{x} to an MAgPIE-object
(if possible).
}
\section{Objects from the Class}{
 Objects can be created by calls of the form
\code{new("magpie", data, dim, dimnames, ...)}. MAgPIE objects have three
dimensions (cells,years,datatype) and the dimensionnames of the first
dimension have the structure "REGION.cellnumber". MAgPIE-objects behave the
same like array-objects with 2 exceptions: \cr 1.Dimensions of the object
will not collapse (e.g. \code{x[1,1,1]} will remain 3D instead of becoming
1D)\cr 2.It is possible to extract full regions just by typing
\code{x["REGIONNAME",,]}. \cr\cr

Please mind following standards: \cr Header must not contain any purely
numeric entries, but combinations of characters and numbers are allowed
(e.g. "bla","12" is forbidden, wheras "bla","b12" is allowed)\cr Years
always have the structure "y" + 4-digit number, e.g. "y1995"\cr Regions
always have the structure 3 capital letters, e.g. "AFR" or "GLO"\cr\cr This
standards are necessary to allow the scripts to detect headers, years and
regions properly and to have a distinction to other data.
}

\examples{

showClass("magpie")

pop <- maxample("pop")

# returning PAO and PAS for 2025
pop["PA", 2025, , pmatch = "left"]

# returning CPA for 2025
pop["PA", 2025, , pmatch = "right"]

# returning CPA PAO and PAS for 2025
pop["PA", 2025, , pmatch = TRUE]

# returning PAS and 2025
pop["PAS", 2025, ]

# return all entries for year 2025
pop[2025, dim = 2]

# returning everything but values for PAS or values for 2025
pop["PAS", 2025, , invert = TRUE]

# accessing subdimension via set name

a <- maxample("animal")
a[list(country = "NLD", y = "53p25"), , list(species = c("rabbit", "dog"))]

# please note that the list elements act as filter. For instance, the
# following example will not contain any dogs as the data set does
# not contain any dogs which are black.
a[list(country = "NLD", y = "53p25"), , list(species = c("rabbit", "dog"), color = "black")]

# it is also possible to extract given combinations of subdimensions
# via a data-frame
df <- data.frame(getItems(a, 3, split = TRUE, full = TRUE))[c(1, 3, 4), ][3:2]
getItems(a[df], 3)

# Unknown dimensions to be added in output!
df$blub <- paste0("bl", 1:dim(df)[1])
getItems(a[df], 3)
}
\seealso{
\code{\link{read.magpie}}, \code{\link{write.magpie}},
\code{\link{getRegions}}, \code{\link{getYears}}, \code{\link{getNames}},
\code{\link{getCPR}}, \code{\link{ncells}}, \code{\link{nyears}},
\code{\link{ndata}}
}
\author{
Jan Philipp Dietrich
}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.report2.R
\name{write.report2}
\alias{write.report2}
\title{Write file in report format}
\usage{
write.report2(...)
}
\arguments{
\item{...}{arguments are forwarded to \code{\link{write.report}}}
}
\description{
This function is deprecated, please use \code{\link{write.report}} instead.
}
\seealso{
\code{\link{write.report}}

Other FileIO: 
\code{\link{copy.magpie}()},
\code{\link{read.magpie}()},
\code{\link{read.report}()},
\code{\link{write.magpie}()},
\code{\link{write.report}()}
}
\author{
Jan Philipp Dietrich
}
\concept{FileIO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/magpiesort.R
\name{magpiesort}
\alias{magpiesort}
\title{MAgPIE-Sort}
\usage{
magpiesort(x)
}
\arguments{
\item{x}{MAgPIE object which might not be in the right order.}
}
\value{
The eventually corrected MAgPIE object (right order in spatial in
temporal dimension)
}
\description{
Brings the spatial and temporal structure of MAgPIE objects in the right
order. This function is especially useful when you create new MAgPIE objects
as the order typically should be correct for MAgPIE objects.
}
\examples{
pop <- maxample("pop")
a <- magpiesort(pop)
}
\seealso{
\code{"\linkS4class{magpie}"}

Other DataBinding: 
\code{\link{cbind.magpie}()},
\code{\link{extend}()},
\code{\link{magpie_expand}()},
\code{\link{matchDim}()},
\code{\link{mbind}()}
}
\author{
Jan Philipp Dietrich
}
\concept{DataBinding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/escapeRegex.R
\name{escapeRegex}
\alias{escapeRegex}
\title{escapeRegex}
\usage{
escapeRegex(x)
}
\arguments{
\item{x}{String or vector of strings that should be escaped.}
}
\value{
The escaped strings.
}
\description{
Escapes all symbols in a string which have a special meaning in regular
expressions.
}
\seealso{
\code{\link[base]{grep}}
}
\author{
Jan Philipp Dietrich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapseNames.R
\name{collapseNames}
\alias{collapseNames}
\title{Collapse dataset names}
\usage{
collapseNames(x, collapsedim = NULL, preservedim = NULL)
}
\arguments{
\item{x}{MAgPIE object}

\item{collapsedim}{If you want to remove the names of particular dimensions
provide the dimensions here. Since the function only works in the third dimension,
you have to count from there on (e.g. dim = 3.2 refers to collapsedim = 2).
Alternatively, you can also specify the name of the dimension. Default: NULL.
CAUTION with parameter collapsedim! You could also force him to remove dimnames,
which are NOT the same for each element and so create duplicates in dimnames.}

\item{preservedim}{If you want to remove the name of particular dimensions except some,
you can specify the dimension(s) to preserve here. See collapsedim for naming convention.
Note that preservedim will be ignored in the case, of a specified collapsedim}
}
\value{
The provided MAgPIE object with collapsed names
}
\description{
This function has been superseded by \code{\link{collapseDim}} which is a
more generalized version of this function. Please use this one instead!
}
\details{
This function will remove names in the data dimension which are the same for
each element (meaning that this data dimension contains exactly one element)
}
\seealso{
\code{\link{collapseDim}}, \code{\link{getItems}},
\code{"\linkS4class{magpie}"}

Other DimensionManipulation: 
\code{\link{addDim}()},
\code{\link{add_columns}()},
\code{\link{collapseDim}()},
\code{\link{dimCode}()},
\code{\link{dimOrder}()},
\code{\link{dimReduce}()}
}
\author{
Jan Philipp Dietrich, David Klein, Xiaoxi Wang
}
\concept{DimensionManipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sameDims.R
\name{sameDims}
\alias{sameDims}
\title{sameDims}
\usage{
sameDims(...)
}
\arguments{
\item{...}{Various magpie objects}
}
\value{
A logical indicating whether the magpie objects have the same items in every dimension.
}
\description{
Checks whether magpie objects have the same items in every dimension (ignoring order).
}
\seealso{
Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{ndim}()}
}
\author{
Pascal Sauer, Patrick Rein
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hasCoords.R
\name{hasCoords}
\alias{hasCoords}
\title{Has Coordinates}
\usage{
hasCoords(x, xlab = "x", ylab = "y")
}
\arguments{
\item{x}{MAgPIE object}

\item{xlab}{label of x-dimension}

\item{ylab}{label of y-dimension}
}
\value{
Boolean indicating whether coordinates were found or not
}
\description{
Checks, whether object contains coordinates.
}
\examples{

hasCoords(maxample("pop"))
hasCoords(maxample("animal"))

}
\seealso{
\code{\link{getCoords}}

Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimReduce.R
\name{dimReduce}
\alias{dimReduce}
\title{dimReduce}
\usage{
dimReduce(x, dim_exclude = NULL)
}
\arguments{
\item{x}{MAgPIE object which should be reduced}

\item{dim_exclude}{Vector with names of dimensions which must not be reduced}
}
\value{
The reduced MAgPIE object
}
\description{
Remove dimensions which contain identical data for all elements in it
}
\note{
This function has some similarities to \code{\link{collapseDim}}, but
serves a different purpose. While \code{\link{collapseDim}} only removes
dimensions which contain only a single element or which it is
specifically told to remove, \code{\link{dimReduce}} looks whether the
entries of a multi-entry dimension are all the same and removes dimensions
for which this is the case. In some cases both will lead to the same result
but in many other cases the results will differ.
}
\examples{

# create data with 5 identical scenarios
p <- addDim(maxample("pop")[1:3, 1:3, ], item = paste0("scen", 1:2))
str(p)
str(dimReduce(p))

# set years to same value
p[, , ] <- setYears(p[, 1, ], NULL)
str(p)
str(dimReduce(p))

# set regions to same value
p[, , ] <- setCells(p[1, , ], "GLO")
str(p)
str(dimReduce(p))
}
\seealso{
\code{\link{addDim}}

Other DimensionManipulation: 
\code{\link{addDim}()},
\code{\link{add_columns}()},
\code{\link{collapseDim}()},
\code{\link{collapseNames}()},
\code{\link{dimCode}()},
\code{\link{dimOrder}()}
}
\author{
Jan Philipp Dietrich
}
\concept{DimensionManipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/where.R
\name{where}
\alias{where}
\title{where}
\usage{
where(x, plot = NULL)
}
\arguments{
\item{x}{A logical statement with a magpie object}

\item{plot}{deprecated. Use the function whereplot in package luplot.}
}
\value{
A list of analysis parameters
}
\description{
Analysis function for magpie objects
}
\examples{

p <- maxample("pop")
where(p > 500)
}
\seealso{
whereplot in package luplot

Other SelectionCalculation: 
\code{\link{magpply}()},
\code{\link{mcalc}()},
\code{\link{mselect}()}
}
\author{
Benjamin Leon Bodirsky, Jan Philipp Dietrich
}
\concept{SelectionCalculation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.SpatRaster.R
\name{as.SpatRaster}
\alias{as.SpatRaster}
\title{as.SpatRaster}
\usage{
as.SpatRaster(x, res = NULL)
}
\arguments{
\item{x}{MAgPIE object}

\item{res}{spatial data resolution. If not provided it will be guessed.}
}
\value{
A SpatRaster object
}
\description{
Convert magclass object to a SpatRaster object. Requires the terra package.
}
\examples{

if (requireNamespace("terra", quietly = TRUE)) {
   r <- terra::rast(ncols = 360, nrows = 180, nlyrs = 4)
   r[85:89, 176:179] <- (1:20 \%*\% t(1:4))
   r[15:19, 76:79] <-   (10 + 1:20 \%*\% t(1:4))
   names(r) <- c("y2000..bla", "y2001..bla", "y2000..blub", "y2001..blub")
   m <- as.magpie(r)
   r2 <- as.SpatRaster(m)
}
}
\seealso{
\code{\link{getCoords}}

Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as.data.frame-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{unwrap}()},
\code{\link{wrap}()}
}
\author{
Jan Philipp Dietrich
}
\concept{MAgPIE-Conversions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.RasterBrick.R
\name{as.RasterBrick}
\alias{as.RasterBrick}
\title{as.RasterBrick}
\usage{
as.RasterBrick(x, res = NULL)
}
\arguments{
\item{x}{MAgPIE object}

\item{res}{spatial data resolution. If not provided it will be guessed.}
}
\value{
A RasterBrick object
}
\description{
Convert magclass object to a RasterBrick object
}
\examples{
\dontrun{
if (requireNamespace("raster", quietly = TRUE)) {
   r <- raster::brick(ncols = 36, nrows = 18, nl = 4)
   r[14:18, 25:28] <- (1:20 \%*\% t(1:4))
   names(r) <- c("y2000..bla", "y2001..bla", "y2000..blub", "y2001..blub")
   m <- as.magpie(r)
   r2 <- as.RasterBrick(m)
}
}
}
\seealso{
\code{\link{getCoords}}

Other MAgPIE-Conversions: 
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as.data.frame-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{unwrap}()},
\code{\link{wrap}()}
}
\author{
Jan Philipp Dietrich
}
\concept{MAgPIE-Conversions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guessResolution.R
\name{guessResolution}
\alias{guessResolution}
\title{guessResolution}
\usage{
guessResolution(x)
}
\arguments{
\item{x}{A magpie object or the coordinates of a magpie object (the result
of \code{\link{getCoords}})}
}
\value{
The guessed resolution of the data
}
\description{
Guess the resolution of the given magpie object/coordinates by looking at the minimum
difference between unique sorted values. Fall back to 0.5 if guess is infinite.
}
\seealso{
Other SpatialOperations: 
\code{\link{spatRasterToDataset}()}
}
\author{
Jan Philipp Dietrich, Pascal Sauer
}
\concept{SpatialOperations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sysdata.R
\name{magclassdata}
\alias{magclassdata}
\title{magclassdata}
\description{
General magclass-dataset
}
\details{
Please do not directly access that data. It should be only used by library
functions.
}
\author{
Jan Philipp Dietrich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pmax-method.R, R/pmin-method.R
\docType{methods}
\name{pmax,magpie-method}
\alias{pmax,magpie-method}
\alias{magpie-pmin-pmax}
\alias{pmin}
\alias{pmax}
\alias{pmin,magpie-method}
\alias{pmin,ANY-method}
\alias{pmax,ANY-method}
\title{pmin/pmax}
\usage{
\S4method{pmax}{magpie}(..., na.rm = FALSE)

\S4method{pmin}{magpie}(..., na.rm = FALSE)
}
\arguments{
\item{...}{Multiple magpie objects with dimensions that each are either the same or only
contain one or no item among all objects.}

\item{na.rm}{Passed to \code{\link[base]{pmin}}}
}
\value{
A new magpie object that contains the minimum/maximum of values between all the passed magpie objects.
}
\description{
magclass-aware pmin/pmax, which calculate the parallel minima/maxima of the input values
}
\details{
\code{\link[base]{pmin}}/\code{\link[base]{pmax}} do not care about the order of items in a magclass
object's dim. magclass pmin/pmax reorder items in each dim so they are in the same order.
}
\author{
Pascal Sauer, Patrick Rein
}
\concept{Comparison}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/magclass-package.R
\docType{package}
\name{magclass-package}
\alias{magclass-package}
\alias{magclass}
\title{Data Class and Tools for Handling Spatial-Temporal Data}
\description{
Data class for increased interoperability working with spatial-temporal
data together with corresponding functions and methods (conversions,
basic calculations and basic data manipulation). The class distinguishes
between spatial, temporal and other dimensions to facilitate the development
and interoperability of tools build for it. Additional features are name-based
addressing of data and internal consistency checks (e.g. checking for the
right data order in calculations).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/pik-piam/magclass}
  \item \doi{10.5281/zenodo.1158580}
  \item Report bugs at \url{https://github.com/pik-piam/magclass/issues}
}

}
\author{
Maintainer: Jan Philipp Dietrich <dietrich@pik-potsdam.de>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unitsplit.R
\name{unitsplit}
\alias{unitsplit}
\title{splits a data.frame or vector of strings with form 'variable (unit)' into
a data.frame with variable and unit separated}
\usage{
unitsplit(x, col = "variable")
}
\arguments{
\item{x}{data.frame or vector of strings}

\item{col}{column name. Default: variable}
}
\description{
splits a data.frame or vector of strings with form 'variable (unit)' into
a data.frame with variable and unit separated
}
\seealso{
Other Utility: 
\code{\link{replace_non_finite}()},
\code{\link{unitjoin}()}
}
\author{
Jan Philipp Dietrich, Mika Pflüger, Oliver Richters
}
\concept{Utility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy.magpie.R
\name{copy.magpie}
\alias{copy.magpie}
\title{Copy MAgPIE-files}
\usage{
copy.magpie(input_file, output_file, round = NULL)
}
\arguments{
\item{input_file}{file, that should be copied}

\item{output_file}{copy destination}

\item{round}{number of digits the values should be rounded. NULL means no rounding}
}
\description{
This function copies MAgPIE-files from one location to another. During the
copying it is also possible to change the file type (e.g. from 'mz' to
'csv')
}
\examples{

# copy.magpie("bla.csv","blub.mz")
}
\seealso{
\code{\link{read.magpie}},\code{\link{write.magpie}}

Other FileIO: 
\code{\link{read.magpie}()},
\code{\link{read.report}()},
\code{\link{write.magpie}()},
\code{\link{write.report}()},
\code{\link{write.report2}()}
}
\author{
Jan Philipp Dietrich
}
\concept{FileIO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeNC.R
\name{writeNC}
\alias{writeNC}
\title{Write a magpie object to a netCDF file}
\usage{
writeNC(
  x,
  filename,
  unit,
  ...,
  compression = 2,
  missval = NA,
  gridDefinition = NULL,
  zname = "time",
  progress = FALSE
)
}
\arguments{
\item{x}{A magpie object}

\item{filename}{Name of the netCDF file to write}

\item{unit}{Unit of the data, to omit pass "" (empty string)}

\item{...}{For future expansion}

\item{compression}{Level of compression to use (1-9), NA for no compression}

\item{missval}{The value that encodes NA in the resulting netCDF file}

\item{gridDefinition}{A vector of 5 numeric values: c(xMin, xMax, yMin, yMax, resolution).
Use c(-179.75, 179.75, -89.75, 89.75, 0.5) to write a standard 0.5-degree-resolution
lon/lat grid. If NULL, use min/max of coordinates in x and guessResolution}

\item{zname}{Name of the z dimension in the netCDF file}

\item{progress}{If TRUE, print progress messages}
}
\description{
Write a magpie object to a netCDF file
}
\author{
Pascal Sauer
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rowSums-method.R
\docType{methods}
\name{rowSums-methods}
\alias{rowSums-methods}
\alias{rowSums,ANY-method}
\alias{rowSums,magpie-method}
\alias{rowMeans-methods}
\alias{rowMeans,ANY-method}
\alias{rowMeans,magpie-method}
\title{~~ Methods for Function rowSums and rowMeans ~~}
\usage{
\S4method{rowSums}{magpie}(x, na.rm = FALSE, dims = 1, ...)
}
\arguments{
\item{x}{object on which calculation should be performed}

\item{na.rm}{logical. Should missing values (including NaN) be omitted from the calculations?}

\item{dims}{integer: Which dimensions are regarded as "rows" or "columns" to sum over. For row*,
the sum or mean is over dimensions dims+1, ...; for col* it is over dimensions 1:dims.}

\item{...}{further arguments passed to other colSums/colMeans methods}
}
\description{
~~ Methods for function \code{rowSums} and \code{rowMeans}~~
}
\section{Methods}{
 \describe{

\item{list("signature(x = \"ANY\")")}{ normal rowSums and rowMeans method }

\item{list("signature(x = \"magpie\")")}{ classical method prepared to
handle MAgPIE objects } }
}

\seealso{
Other Aggregation: 
\code{\link{colSums-methods}},
\code{\link{dimSums}()}
}
\concept{Aggregation}
\keyword{keyword(s)}
\keyword{methods}
\keyword{other}
\keyword{possible}
\keyword{~~}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fulldim.R
\name{fulldim}
\alias{fulldim}
\title{Reconstructs full dimensionality of MAgPIE objects}
\usage{
fulldim(x, sep = ".")
}
\arguments{
\item{x}{A MAgPIE-object}

\item{sep}{A character separating joined dimension names}
}
\value{
A list containing in the first element the dim output and in the
second element the dimnames output of the reconstructed array.
}
\description{
Deprecated: Use \code{\link{getItems}} instead.
}
\details{
If a MAgPIE object is created from a source with more than one data
dimension, these data dimensions are combined to a single dimension. fulldim
reconstructs the original dimensionality and reports it.
}
\seealso{
\code{\link{as.magpie}},\code{\link{unwrap}},\code{\link{wrap}}

Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/magpply.R
\name{magpply}
\alias{magpply}
\title{magpply}
\usage{
magpply(X, FUN, MARGIN = NULL, DIM = NULL, ..., INTEGRATE = FALSE)
}
\arguments{
\item{X}{magpie object}

\item{FUN}{function that shall be applied X}

\item{MARGIN}{dimension over which FUN shall be applied (like a loop over that dimension).
This dimension will be preserved in the output object (see also \code{DIM}).}

\item{DIM}{dimension in which FUN shall be applied. This dimension will be missing in the output. DIM and MARGIN
are opposite ways of expressing the dimensions to be addressed and you must only use one of them with MARGIN
excluding dimensions from the calculation and DIM including them.}

\item{...}{further parameters passed on to FUN}

\item{INTEGRATE}{if TRUE, the output will be filled into an magpie object of the same dimensionality as X}
}
\value{
magpie object
}
\description{
apply command for magpieobjects. Very efficient for replacing loops.
}
\examples{
pop <- maxample("pop")
magpply(pop, FUN = sum, MARGIN = 2)
fourdim <- pop * setNames(pop, c("jkk", "lk"))
magpply(fourdim, FUN = sum, MARGIN = c(1, 3.1))
}
\seealso{
Other SelectionCalculation: 
\code{\link{mcalc}()},
\code{\link{mselect}()},
\code{\link{where}()}
}
\author{
Jan Philipp Dietrich, Benjamin Leon Bodirsky
}
\concept{SelectionCalculation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addDim.R
\name{add_dimension}
\alias{add_dimension}
\title{add_dimension}
\usage{
add_dimension(x, dim = 3.1, add = NULL, nm = "dummy", expand = TRUE)
}
\arguments{
\item{x}{MAgPIE object which should be extended.}

\item{dim}{The dimension number of the new dimension (e.g. 3.1)}

\item{add}{The name of the new dimension}

\item{nm}{One or more names of items in the new dimension.}

\item{expand}{If TRUE, each item from the item argument is added to each item
already present, resulting in e.g. `c("A.d1", "B.d1", "A.d2", "B.d2")`.
Otherwise, length of item must equal the number of items already present and
they are simply added, resulting in e.g. `c("A.d1", "B.d2")`.}
}
\value{
The extended MAgPIE object
}
\description{
This function adds a named dimension as dimension number "dim" with
the name "add" with an empty data column with the name "nm".
Please start using this function's successor \code{\link{addDim}}. add_dimension
might eventually become deprecated.
}
\examples{

a <- maxample("animal")
str(addDim(a, dim = 3.2))
str(addDim(a, dim = 2.3, item = paste0("d", 1:3)))
}
\seealso{
\code{\link{add_columns}},\code{\link{mbind}}

Other DimensionManipulation: 
\code{\link{add_columns}()},
\code{\link{collapseDim}()},
\code{\link{collapseNames}()},
\code{\link{dimCode}()},
\code{\link{dimOrder}()},
\code{\link{dimReduce}()}
}
\author{
Jan Philipp Dietrich, Benjamin Bodirsky, Pascal Sauer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.magpie.R
\name{print.magpie}
\alias{print.magpie}
\title{print}
\usage{
\method{print}{magpie}(x, drop = TRUE, reshape = FALSE, ...)
}
\arguments{
\item{x}{MAgPIE object}

\item{drop}{argument which controls whether empty dimensions should be
skipped or not.}

\item{reshape}{argument that controls tabular representation of nested data
dimension cross tables, FALSE will reproduce standard print behavior
any pair of two dimension numbers will create a table for these two dims,
and loop over the other dimensions}

\item{...}{arguments to be passed to or from other methods.}
}
\value{
Invisibly, the MAgPIE object x.
}
\description{
print method for MAgPIE objects for conventient display of magpie data.
}
\examples{

pop <- maxample("pop")
print(pop)
print(pop[, 1, ], drop = FALSE)
print(pop[, 1, ])
}
\seealso{
\code{\link[base]{print}}

Other Display: 
\code{\link{head.magpie}()},
\code{\link{maxample}()},
\code{\link{mplot}()},
\code{\link{show-methods}},
\code{\link{str.magpie}()}
}
\author{
Jan Philipp Dietrich, Kristine Karstens, Felicitas Beier
}
\concept{Display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extend.R
\name{extend}
\alias{extend}
\title{extend}
\usage{
extend(x, gridDefinition = NULL, crop = FALSE)
}
\arguments{
\item{x}{A magpie object}

\item{gridDefinition}{A vector of 5 numeric values: c(xMin, xMax, yMin, yMax, resolution).
Use c(-179.75, 179.75, -89.75, 89.75, 0.5) to extend to a standard 0.5-degree-resolution
lon/lat grid. If NULL, use min/max of coordinates in x and guessResolution.}

\item{crop}{If TRUE, discard cells from x which are not in the gridDefinition grid. If FALSE,
throw an error if the coordinates of x are not a subset of the extended coordinates.}
}
\value{
Magpie object x with dense grid according to gridDefinition, gaps filled with NA.
}
\description{
Extend a magpie object to a dense grid based on the given xRange, yRange and resolution.
This is e.g. required when writing netCDF files. Extending a sparse magpie object to a dense grid
requires much more memory, so use with caution.
}
\seealso{
Other DataBinding: 
\code{\link{cbind.magpie}()},
\code{\link{magpie_expand}()},
\code{\link{magpiesort}()},
\code{\link{matchDim}()},
\code{\link{mbind}()}
}
\author{
Pascal Sauer
}
\concept{DataBinding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unwrap.R
\name{unwrap}
\alias{unwrap}
\title{Unwrap}
\usage{
unwrap(x, sep = NULL)
}
\arguments{
\item{x}{A MAgPIE object}

\item{sep}{deprecated, please do not use anymore}
}
\value{
An array with the full dimensionality of the original data
}
\description{
Creates a higher dimensional array by separating all subdimensions in
the third dimension of a MAgPIE object and returning them as
separate dimension.
}
\examples{

a <- as.magpie(array(1:6, c(3, 2), list(c("bla", "blub", "ble"), c("up", "down"))))
unwrap(a)
}
\seealso{
\code{\link{wrap}},\code{\link{fulldim}}

Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as.data.frame-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{wrap}()}
}
\author{
Jan Philipp Dietrich
}
\concept{MAgPIE-Conversions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCPR.R
\name{getCPR}
\alias{getCPR}
\title{Get cells per region}
\usage{
getCPR(x, dim = 1.1)
}
\arguments{
\item{x}{MAgPIE object or a resolution written as numeric (currently only
data for 0.5 degree resolution is available).}

\item{dim}{Dimension for which the items should be returned. Either number or
name of dimension or a vector of these (in case of a vector all subimensions
must belong to the same main dimension!). See \code{\link{dimCode}} for more details.}
}
\value{
cells per region
}
\description{
Counts how often each element of the provided subdimension exists in the given
data set. Originally created to count the number of cells in a region (this
is also where its name originates from) it can now be used to count elements
of any subdimension via the dim argument.
}
\examples{

getCPR(0.5)
a <- maxample("animal")
getCPR(a, dim = "color")
getCPR(a, dim = 3.2)
getCPR(a, dim = "country")
getCPR(a, dim = c("color", "species"))
}
\seealso{
\code{\link{getRegions}}, \code{\link{read.magpie}},
\code{\link{write.magpie}}

Other GetterSetter: 
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.report.R
\name{write.report}
\alias{write.report}
\title{Write file in report format}
\usage{
write.report(
  x,
  file = NULL,
  model = NULL,
  scenario = NULL,
  unit = NULL,
  ndigit = 4,
  append = FALSE,
  skipempty = TRUE,
  extracols = NULL
)
}
\arguments{
\item{x}{MAgPIE object or a list of lists with MAgPIE objects as created by
read.report. In the latter case settings for model and scenario are
overwritten by the information given in the list.}

\item{file}{file name the object should be written to. If NULL the formatted
content is returned}

\item{model}{Name of the model which calculated the results}

\item{scenario}{The scenario which was used to get that results.}

\item{unit}{Unit of the data. Only relevant if unit is not already supplied
in Dimnames (format "name (unit)"). Can be either a single string or a
vector of strings with a length equal to the number of different data
elements in the MAgPIE object}

\item{ndigit}{Number of digits the output should have}

\item{append}{Logical which decides whether data should be added to an
existing file or an existing file should be overwritten}

\item{skipempty}{Determines whether empty entries (all data NA) should be
written to file or not.}

\item{extracols}{names of dimensions which should appear in the output as additional columns}
}
\description{
This function writes the content of a MAgPIE object into a file or returns
it directly using the reporting format as it is used for many model
inter-comparisons.
}
\examples{

write.report(maxample("pop"))
}
\seealso{
\code{\link{read.report}}

Other FileIO: 
\code{\link{copy.magpie}()},
\code{\link{read.magpie}()},
\code{\link{read.report}()},
\code{\link{write.magpie}()},
\code{\link{write.report2}()}
}
\author{
Jan Philipp Dietrich
}
\concept{FileIO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mbind.R
\name{mbind}
\alias{mbind}
\title{mbind}
\usage{
mbind(...)
}
\arguments{
\item{...}{MAgPIE objects or a list of MAgPIE objects that should be merged.}
}
\value{
The merged MAgPIE object
}
\description{
Merges MAgPIE-objects with identical structure in two dimensions. If data
differs in the temporal or spatial dimension each year or region/cell must
appear only once!
}
\examples{

m <- new.magpie(c("AFR", "CPA", "EUR"), c(1995, 2005), "Data1", fill = c(1, 2, 3, 4, 5, 6))
ms <- dimSums(m, dim = 3.1)
mbind(m, ms)
my <- new.magpie(getRegions(m), 2010, getNames(m), fill = c(6, 6, 4))
mbind(m, my)
md <- new.magpie(getRegions(m), getYears(m), "Data2", fill = c(7, 6, 5, 7, 8, 9))
mbind(m, md)

pop <- maxample("pop")
a <- mbind(pop, pop)
dim(pop)
dim(a)
}
\seealso{
\code{"\linkS4class{magpie}"}

Other DataBinding: 
\code{\link{cbind.magpie}()},
\code{\link{extend}()},
\code{\link{magpie_expand}()},
\code{\link{magpiesort}()},
\code{\link{matchDim}()}
}
\author{
Jan Philipp Dietrich, Misko Stevanovic
}
\concept{DataBinding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/new.magpie.R
\name{new.magpie}
\alias{new.magpie}
\title{new.magpie}
\usage{
new.magpie(
  cells_and_regions = "GLO",
  years = NULL,
  names = NULL,
  fill = NA,
  sort = FALSE,
  sets = NULL,
  unit = NULL
)
}
\arguments{
\item{cells_and_regions}{Either the region names (e.g. "AFR"), or the cells
(e.g. 1:10), or both in combination (e.g. "AFR.1"). NULL means no spatial
element.}

\item{years}{dimnames for years in the format "yXXXX" or as integers. NULL
means one year which is not further specified}

\item{names}{dimnames for names. NULL means one data element which is not
further specified}

\item{fill}{Default value for the MAgPIE object}

\item{sort}{Bolean. Decides, wheher output should be sorted or not.}

\item{sets}{A vector of dimension names. See \code{\link{getSets}} for more
information.}

\item{unit}{deprecated}
}
\value{
an empty magpie object filled with fill, with the given dimnames
}
\description{
Creates a new MAgPIE object
}
\examples{

a <- new.magpie(1:10, 1995:2000)
b <- new.magpie(c("AFR", "CPA"), "y1995", c("bla", "blub"), sets = c("i", "t", "value"))
c <- new.magpie()
}
\seealso{
\code{\link{as.magpie}}

Other ObjectCreation: 
\code{\link{clean_magpie}()},
\code{\link{complete_magpie}()},
\code{\link{copy.attributes}()}
}
\author{
Benjamin Bodirsky, Jan Philipp Dietrich
}
\concept{ObjectCreation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy.attributes.R
\name{copy.attributes}
\alias{copy.attributes}
\alias{copy.attributes<-}
\title{Copy Attributes}
\usage{
copy.attributes(
  from,
  to,
  delete = c("names", "row.names", "class", "dim", "dimnames"),
  delete2 = NULL
)

copy.attributes(
  to,
  delete = c("names", "row.names", "class", "dim", "dimnames"),
  delete2 = NULL
) <- value
}
\arguments{
\item{from}{object from which the attributes should be taken}

\item{to}{object to which the attributes should be written}

\item{delete}{attributes which should not be copied. By default this are
class specific attributes which might cause problems if copied to another
object. But you can add or remove attributes from the vector.}

\item{delete2}{Identical to delete and just added for convenience for the
case that you want to delete additional attributes but do not want to repeat
the vector given in delete. In the function both vectors, delete and
delete2, are just merged to one deletion vector.}

\item{value}{Same as "from" (object from which the attributes should be
taken)}
}
\description{
This function copies attributes from one object and assigns them to another.
}
\section{Functions}{
\itemize{
\item \code{copy.attributes(
  to,
  delete = c("names", "row.names", "class", "dim", "dimnames"),
  delete2 = NULL
) <- value}: assign attributes from object "value"

}}
\examples{

from <- array(12)
attr(from, "blablub") <- "I am an attribute!"
attr(from, "blablub2") <- "I am another attribute!"

print(attributes(from))

to <- as.magpie(0)
print(attributes(to))

copy.attributes(to) <- from
print(attributes(to))

}
\seealso{
Other ObjectCreation: 
\code{\link{clean_magpie}()},
\code{\link{complete_magpie}()},
\code{\link{new.magpie}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectCreation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getNames.R
\name{getNames}
\alias{getNames}
\alias{getNames<-}
\title{Get dataset names}
\usage{
getNames(x, fulldim = FALSE, dim = NULL)

getNames(x, dim = NULL) <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{fulldim}{specifies, how the object is treated. In case of FALSE, it is
assumed that x is 3 dimensional and dimnames(x)[[3]] is returned. In case of
TRUE, the dimnames of the real third dimension namesare returned}

\item{dim}{Argument to choose a specific data dimension either by name of
the dimension or by number of the data dimension.}

\item{value}{a vector of names current names should be replaced with. If
only one data element exists you can also set the name to NULL.}
}
\value{
getNames returns data names of the MAgPIE-object, whereas setNames
returns the MAgPIE object with the manipulated data names.
}
\description{
Extracts dataset names of a MAgPIE-object
}
\details{
setNames is a shortcut to use a MAgPIE object with manipulated data names.
The setNames method uses the variable names "object" and "nm" in order to be
consistent to the already existing function setNames.
}
\section{Functions}{
\itemize{
\item \code{getNames(x, dim = NULL) <- value}: set names

}}
\examples{

a <- as.magpie(1)
getNames(a)
setNames(a, "bla")

x <- new.magpie("GLO", 2000, c("a.o1", "b.o1", "a.o2"))
getNames(x, dim = 2)

getSets(x, fulldim = FALSE)[3] <- "bla.blub"
getNames(x, dim = "bla")

getSets(x)[4] <- "ble"
getNames(x, dim = "ble") <- c("Hi", "Bye")
x
}
\seealso{
\code{\link{setNames-methods}}, \code{\link{getRegions}}, \code{\link{getYears}},
\code{\link{getCPR}}, \code{\link{read.magpie}},
\code{\link{write.magpie}},\code{\link{ndata}},
\code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str.magpie.R
\name{str.magpie}
\alias{str.magpie}
\title{str}
\usage{
\method{str}{magpie}(object, ...)
}
\arguments{
\item{object}{MAgPIE object}

\item{...}{arguments to be passed to or from other methods.}
}
\description{
str method for MAgPIE objects for conventient display of the structure of a magpie object.
}
\details{
In contrast to the default str this will not show the attributes of object@.Data as these
contain only a duplicate of dimnames. Also, dim is not shown, because the information it
provides is implicitly included in dimnames.
}
\examples{
str(maxample("pop"))
}
\seealso{
\code{\link[utils]{str}}

Other Display: 
\code{\link{head.magpie}()},
\code{\link{maxample}()},
\code{\link{mplot}()},
\code{\link{print.magpie}()},
\code{\link{show-methods}}
}
\author{
Pascal Sauer
}
\concept{Display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lowpass.R
\name{lowpass}
\alias{lowpass}
\title{Lowpass Filter}
\usage{
lowpass(x, i = 1, fix = NULL, altFilter = NULL, warn = TRUE)
}
\arguments{
\item{x}{Vector of data points, that should be filtered or MAgPIE object}

\item{i}{number of iterations the filter should be applied to the data}

\item{fix}{Fixes the starting and/or ending data point. Default value is
\code{NULL} which doesn't fix any point. Available options are:
\code{"start"} for fixing the starting point, \code{"end"} for fixing the
ending point and \code{"both"} for fixing both ends of the data.}

\item{altFilter}{set special filter rule to indexes defined in this
parameter. Special filter structure: `x'(n) = (2*x(n)+x(n+1))/3`}

\item{warn}{boolean deciding whether lowpass issues a warning for
critical parameter choices or not}
}
\value{
The filtered data vector or MAgPIE object
}
\description{
Filters high frequencies out of a time series. The filter
has the structure `x'(n) = (x(n-1)+2*x(n)+x(n+1))/4`
}
\examples{

lowpass(c(1, 2, 11, 3, 4))
# to fix the starting point
lowpass(c(0, 9, 1, 5, 14, 20, 6, 11, 0), i = 2, fix = "start")
}
\seealso{
Other TemporalOperations: 
\code{\link{commonYears}()},
\code{\link{convergence}()},
\code{\link{time_interpolate}()}
}
\author{
Jan Philipp Dietrich, Misko Stevanovic
}
\concept{TemporalOperations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commonYears.R
\name{commonYears}
\alias{commonYears}
\title{commonYears}
\usage{
commonYears(..., asInteger = FALSE)
}
\arguments{
\item{...}{One or more magclass objects, or alternatively a list of magclass objects.}

\item{asInteger}{Passed on to getYears. If TRUE the output format is 1995 otherwise "y1995".}
}
\value{
The years (as returned by getYears) the input objects have in common.
}
\description{
Given (a list of) magclass objects, return the years (getYears) they have in common,
ignoring magclass objects with 0 or 1 years.
}
\seealso{
Other TemporalOperations: 
\code{\link{convergence}()},
\code{\link{lowpass}()},
\code{\link{time_interpolate}()}
}
\author{
Pascal Sauer
}
\concept{TemporalOperations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matchDim.R
\name{matchDim}
\alias{matchDim}
\title{Match dimensions of a magpie object to those of a reference object}
\usage{
matchDim(x, ref, dim = 1:3, fill = NA)
}
\arguments{
\item{x}{A \code{magpie} object to be modified.}

\item{ref}{A \code{magpie} object used as a reference for the modification.
Returns \code{x} if \code{ref} is \code{NULL}.}

\item{dim}{Subset of dimensions for which the matching should be done.  Can
be either a number between 1 and 3 or a vector of these.  Defaults to all
dimensions (i.e. \code{1:3}).}

\item{fill}{Value to be set in new dimensions.}
}
\value{
The modified \code{magpie} object.
}
\description{
A helper that restricts and expands a magpie object \code{x} to the size of a
magpie object \code{ref}.  Dimension names not present in \code{x} are added and set to
the value provided by \code{fill}.  Dimension names not present in \code{ref} are
cropped.
}
\seealso{
Other DataBinding: 
\code{\link{cbind.magpie}()},
\code{\link{extend}()},
\code{\link{magpie_expand}()},
\code{\link{magpiesort}()},
\code{\link{mbind}()}
}
\author{
Falk Benke
}
\concept{DataBinding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcalc.R
\name{mcalc}
\alias{mcalc}
\alias{mcalc<-}
\title{mcalc}
\usage{
mcalc(x, f, dim = NULL, append = FALSE)
}
\arguments{
\item{x}{MAgPIE object}

\item{f}{A formula describing the calculation that should be performed}

\item{dim}{The dimension in which the manipulation should take place. If set
to NULL function tries to detect the dimension automatically.}

\item{append}{If set to TRUE the result will be appended to x, otherwise the
result will be returned.}
}
\value{
The calculated MAgPIE object in the case that append is set to
FALSE. Otherwise nothing is returned (as x is appended in place)
}
\description{
Select values from a MAgPIE-object
}
\details{
This functions only work for MAgPIE objects with named dimensions as the
dimension name (set_name) has to be used to indicate in which dimension the
entries should be searched for!
}
\examples{

pop <- maxample("pop")
pop
mcalc(pop, X12 ~ A2 * B1, append = TRUE)
pop
mcalc(pop, `Nearly B1` ~ 0.5 * A2 + 99.5 * B1)
}
\seealso{
\code{\link{mselect}}

Other SelectionCalculation: 
\code{\link{magpply}()},
\code{\link{mselect}()},
\code{\link{where}()}
}
\author{
Jan Philipp Dietrich
}
\concept{SelectionCalculation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_non_finite.R
\name{replace_non_finite}
\alias{replace_non_finite}
\title{Replace Non-Finite Data}
\usage{
replace_non_finite(x, replace = 0)
}
\arguments{
\item{x}{A vector or \code{\link[=magclass]{magpie}} object.}

\item{replace}{A value to replace non-finite data with.}
}
\value{
A vector or \code{\link[=magclass]{magpie}} object, same as \code{x}.
}
\description{
Replaces all instances of non-finite data (\code{NA}, \code{NaN}, \code{Inf}, and \code{-Inf}).
}
\examples{
part  <- new.magpie(letters[1:3], years = "y1995", names = "foo")
total <- new.magpie(letters[1:3], years = "y1995", names = "foo")

part[, , ]  <- c(0, 1, 2)
total[, , ] <- c(0, 10, 10)

part / total

replace_non_finite(part / total)

}
\seealso{
Other Utility: 
\code{\link{unitjoin}()},
\code{\link{unitsplit}()}
}
\author{
Michaja Pehl
}
\concept{Utility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete_magpie.R
\name{complete_magpie}
\alias{complete_magpie}
\title{complete_magpie}
\usage{
complete_magpie(x, fill = NA, dim = 3)
}
\arguments{
\item{x}{MAgPIE object which should be completed.}

\item{fill}{Value that shall be written into the missing entries}

\item{dim}{dimensions in which the completion should take place (1, 2 and/or 3). For
full completion use \code{1:3}}
}
\value{
The completed MAgPIE object
}
\description{
MAgPIE objects can be incomplete to reduce memory. This function blows up a
magpie object to its real dimensions, so you can apply unwrap.
}
\examples{

pop <- maxample("pop")
complete_magpie(pop)

ani <- maxample("animal")
complete_magpie(ani)
}
\seealso{
\code{\link{addDim}},\code{\link{clean_magpie}}

Other ObjectCreation: 
\code{\link{clean_magpie}()},
\code{\link{copy.attributes}()},
\code{\link{new.magpie}()}
}
\author{
Jan Philipp Dietrich, Benjamin Bodirsky
}
\concept{ObjectCreation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round-method.R
\docType{methods}
\name{round-methods}
\alias{round-methods}
\alias{round,magpie-method}
\title{Round-method for MAgPIE objects}
\usage{
\S4method{round}{magpie}(x, digits = 0)
}
\arguments{
\item{x}{a magpie object}

\item{digits}{integer indicating the number of decimal places (round) or significant
digits (signif) to be used. Negative values are allowed.}
}
\description{
Round-method for MAgPIE-objects respectively. Works exactly as for arrays.
}
\section{Methods}{
 \describe{

\item{x = "magpie"}{ works as round(x) for arrays. }

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.report.R
\name{read.report}
\alias{read.report}
\title{Read file in report format}
\usage{
read.report(file, as.list = TRUE, showSeparatorWarning = TRUE)
}
\arguments{
\item{file}{file name the object should be read from.}

\item{as.list}{if TRUE a list is returned (default), if FALSE it is tried to
merge all information in one MAgPIE object (still under development and
works currently only if the entries for the different models and scenarios
have exactly the same regions and years).}

\item{showSeparatorWarning}{Boolean (default value TRUE) that decides whether the warning
about the replacement of dots in variable names is displayed (default value) or not.}
}
\description{
This function reads the content of a reporting file (a file in the model
intercomparison file format *.mif) into a list of MAgPIE objects or a single
MAgPIE object.
}
\details{
The \strong{Model Intercomparison File Format (MIF)} is the default file
format for data produced by Integrated Assessment Models. It is based on the
common format used for Model Intercomparison Projects such as EMF and SSP
with some slight changes/clarifications in its definition. For interactions
between models this format should be used. For everything else it is at least
recommended to use this format, too.

Aim of this standardization is to achieve a more flexible and smooth
communication between models and to facilitate the creation of aggregated
outputs from integrated assessment scenario runs which then can easily be
uploaded to external databases such as the EMF or SSP database. By using this
standard most of the required decisions for a working input output interface
between models have already been specified which significantly reduces the
required work to get a new interaction running.

\strong{Definition}

The format is characterized by the following features:

\itemize{
  \item The file ending is ".mif"
  \item The file is written in ASCII format
  \item Entries are separated with ";", every line ends with a ";"
  \item The file always contains a header
  \item The format of the header is: \code{Model;Scenario;Region;Variable;Unit;<ADDITIONAL_COLUMNS>;<YEARS>;}
}

The first 5 entries always have to exist, <ADDITIONAL_COLUMNS> is additional
information which can be added optionally (e.g. "Description") and <YEARS>
are the years for which data is delivered. <YEARS> are always written as 4
digit numbers. In the (very unlikely) case that a year before 1000 is used
the number has to start with a 0, e.g. 0950. <ADDITIONAL_COLUMNS> can be
anything, there are no further rules at the moment what it can contain.
However, there are strict rules for naming these columns. Allowed are single
names starting with a capital letter without special characters in it except
"_" which is allowed. Examples: "Description" allowed, "More Description" not
allowed, "More_Description" allowed, "123Description" not allowed,
"Description123" allowed. Scripts using this format must be able to ignore
additional columns. For years there are no specific limitations/requirements
which years should be reported. Scripts dealing with this data must be able
to work with different temporal resolutions. For variables basically
everything can be reported here. Missing values have to be marked with "N/A".
}
\examples{
\dontrun{
read.report("report.csv")
}

}
\seealso{
\code{\link{write.report}}

Other FileIO: 
\code{\link{copy.magpie}()},
\code{\link{read.magpie}()},
\code{\link{write.magpie}()},
\code{\link{write.report}()},
\code{\link{write.report2}()}
}
\author{
Jan Philipp Dietrich
}
\concept{FileIO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_tibble.R
\name{as_tibble.magpie}
\alias{as_tibble.magpie}
\title{magpie method for tibble::as_tibble}
\usage{
\method{as_tibble}{magpie}(
  x,
  ...,
  .rows = NULL,
  .name_repair = c("check_unique", "unique", "universal", "minimal"),
  rownames = pkgconfig::get_config("tibble::rownames", NULL)
)
}
\arguments{
\item{x}{A magpie object}

\item{...}{Unused, for extensibility.}

\item{.rows}{The number of rows, useful to create a 0-column tibble or
just as an additional check.}

\item{.name_repair}{Treatment of problematic column names:
\itemize{
\item \code{"minimal"}: No name repair or checks, beyond basic existence,
\item \code{"unique"}: Make sure names are unique and not empty,
\item \code{"check_unique"}: (default value), no name repair, but check they are
\code{unique},
\item \code{"universal"}: Make the names \code{unique} and syntactic
\item \code{"unique_quiet"}: Same as \code{"unique"}, but "quiet"
\item \code{"universal_quiet"}: Same as \code{"universal"}, but "quiet"
\item a function: apply custom name repair (e.g., \code{.name_repair = make.names}
for names in the style of base R).
\item A purrr-style anonymous function, see \code{\link[rlang:as_function]{rlang::as_function()}}
}

This argument is passed on as \code{repair} to \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}}.
See there for more details on these terms and the strategies used
to enforce them.}

\item{rownames}{How to treat existing row names of a data frame or matrix:
\itemize{
\item \code{NULL}: remove row names. This is the default.
\item \code{NA}: keep row names.
\item A string: the name of a new column. Existing rownames are transferred
into this column and the \code{row.names} attribute is deleted.
No name repair is applied to the new column name, even if \code{x} already contains
a column of that name.
Use \code{as_tibble(rownames_to_column(...))} to safeguard against this case.
}

Read more in \link[tibble]{rownames}.}
}
\value{
A tibble object
}
\description{
magpie method for tibble::as_tibble
}
\seealso{
Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as.data.frame-methods}},
\code{\link{unwrap}()},
\code{\link{wrap}()}
}
\concept{MAgPIE-Conversions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrap.R
\name{wrap}
\alias{wrap}
\title{Wrap}
\usage{
wrap(x, map = list(NA), sep = ".")
}
\arguments{
\item{x}{An array}

\item{map}{A list of length equal to the number of dimensions in the
reshaped array. Each element should be an integer vectors specifying the
dimensions to be joined in corresponding new dimension. One element may
equal NA to indicate that that dimension should be a join of all
non-specified (remaining) dimensions. Default is to wrap everything into a
vector.}

\item{sep}{A character separating joined dimension names}
}
\description{
Reshape an array or a matrix by permuting and/or joining dimensions.
}
\note{
This function is extracted from the R.utils library which is licensed
under LGPL>=2.1 and written by Henrik Bengtsson.
}
\seealso{
\code{\link{unwrap}},\code{\link{fulldim}}

Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as.data.frame-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{unwrap}()}
}
\author{
Henrik Bengtsson, Jan Philipp Dietrich
}
\concept{MAgPIE-Conversions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.temporal.R
\name{is.temporal}
\alias{is.temporal}
\alias{is.spatial}
\title{is.temporal, is.spatial}
\usage{
is.temporal(x)
}
\arguments{
\item{x}{A vector}
}
\value{
Returns TRUE or FALSE
}
\description{
Functions to find out whether a vector consists of strings consistent with
the definition for auto-detection of temporal or spatial data.
}
\examples{

is.temporal(1991:1993)
is.spatial(c("GLO", "AFR"))

}
\seealso{
Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{isYear}()},
\code{\link{ncells}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/head.magpie.R
\name{head.magpie}
\alias{head.magpie}
\alias{tail.magpie}
\title{head/tail}
\usage{
\method{head}{magpie}(x, n1 = 3L, n2 = 6L, n3 = 2L, ...)
}
\arguments{
\item{x}{MAgPIE object}

\item{n1, n2, n3}{number of lines in first, second and third dimension that
should be returned. If the given number is higher than the length of the
dimension all entries in this dimension will be returned.}

\item{...}{arguments to be passed to or from other methods.}
}
\value{
head returns the first n1 x n2 x n3 entries, tail returns the last
n1 x n2 x n3 entries.
}
\description{
head and tail methods for MAgPIE objects to extract the head or tail of an
object
}
\examples{

pop <- maxample("pop")
head(pop)
 tail(pop, 2, 4, 1)

}
\seealso{
\code{\link[utils]{head}}, \code{\link[utils]{tail}}

Other Display: 
\code{\link{maxample}()},
\code{\link{mplot}()},
\code{\link{print.magpie}()},
\code{\link{show-methods}},
\code{\link{str.magpie}()}
}
\author{
Jan Philipp Dietrich
}
\concept{Display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRegions.R
\name{getRegions}
\alias{getRegions}
\alias{getRegions<-}
\title{Get regions}
\usage{
getRegions(x)

getRegions(x) <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{value}{Vector containing the new region names of the MAgPIE objects.
If you also want to change the mapping of regions to cell please use
\code{\link{getRegionList}} instead.}
}
\value{
Regions of the MAgPIE-object
}
\description{
Extracts regions of a MAgPIE-object
}
\section{Functions}{
\itemize{
\item \code{getRegions(x) <- value}: overwrite region names

}}
\examples{

# a <- read.magpie("example.mz")
# getRegions(a)
}
\seealso{
\code{\link{getYears}}, \code{\link{getNames}},
\code{\link{getCPR}}, \code{\link{read.magpie}}, \code{\link{write.magpie}},
\code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unitjoin.R
\name{unitjoin}
\alias{unitjoin}
\title{joins a data.frame or vector of strings with variable and unit separated into
a data.frame with variable and unit joined as 'variable (unit)'.
Use magclass::unitsplit to split them again}
\usage{
unitjoin(x, unit = NULL, col = "variable")
}
\arguments{
\item{x}{data.frame or vector of strings}

\item{unit}{vector of strings. If NULL, col 'unit' in x is used}

\item{col}{column name. Default: variable}
}
\value{
data.frame or vector of strings, dependent on x
}
\description{
joins a data.frame or vector of strings with variable and unit separated into
a data.frame with variable and unit joined as 'variable (unit)'.
Use magclass::unitsplit to split them again
}
\seealso{
Other Utility: 
\code{\link{replace_non_finite}()},
\code{\link{unitsplit}()}
}
\concept{Utility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimSums.R
\name{dimSums}
\alias{dimSums}
\title{Summation over dimensions}
\usage{
dimSums(x, dim = 3, na.rm = FALSE)
}
\arguments{
\item{x}{A MAgPIE-object}

\item{dim}{The dimensions(s) to sum over. A vector of dimension codes or dimension names.
See \code{\link{dimCode}} for more information}

\item{na.rm}{logical. Should missing values (including NaN) be omitted from
the calculations?}
}
\value{
A MAgPIE object with values summed over the specified dimensions
}
\description{
This function sums over any (sub-)dimension of a magpie object
}
\examples{
a <- maxample("animal")
dimSums(a, dim = c(1, 2, 3.2))
dimSums(a, dim = c("x", "y", "cell", "month"))
}
\seealso{
\code{\link{rowSums}}, \code{\link{getItems}}, \code{\link{dimCode}}

Other Aggregation: 
\code{\link{colSums-methods}},
\code{\link{rowSums-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{Aggregation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log-method.R
\docType{methods}
\name{log-methods}
\alias{log-methods}
\alias{log,magpie-method}
\alias{logb-methods}
\alias{logb,magpie-method}
\title{log-method for MAgPIE objects}
\usage{
\S4method{log}{magpie}(x, base = exp(1))
}
\arguments{
\item{x}{a magpie object}

\item{base}{ia positive or complex number: the base with respect to which
logarithms are computed. Defaults to e=exp(1).}
}
\description{
log-method for MAgPIE-objects respectively. Works exactly as for arrays.
}
\section{Methods}{
 \describe{

\item{x = "magpie"}{ works as log(x) for arrays. }

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind.magpie.R
\name{cbind.magpie}
\alias{cbind.magpie}
\title{cbind method for MAgPIE objects}
\usage{
\method{cbind}{magpie}(..., deparse.level = 1)
}
\arguments{
\item{...}{see \link[base]{cbind}}

\item{deparse.level}{see \link[base]{cbind}}
}
\value{
A matrix as if cbind was applied to the internal vector of the MAgPIE objects
}
\description{
cbind method for MAgPIE-objects. Issues a warning that mbind should be used
if all objects are magpie objects and continues afterwards.
}
\seealso{
Other DataBinding: 
\code{\link{extend}()},
\code{\link{magpie_expand}()},
\code{\link{magpiesort}()},
\code{\link{matchDim}()},
\code{\link{mbind}()}
}
\author{
Patrick Rein
}
\concept{DataBinding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.data.frame.R
\docType{methods}
\name{as.data.frame-methods}
\alias{as.data.frame-methods}
\alias{as.data.frame}
\alias{as.data.frame,ANY-method}
\alias{as.data.frame,magpie-method}
\title{~~ Methods for Function as.data.frame ~~}
\usage{
\S4method{as.data.frame}{magpie}(x, rev = 1, raw = FALSE)
}
\arguments{
\item{x}{A MAgPIE-object}

\item{rev}{The revision of the algorithm that should be used for conversion.
rev=1 creates columns with the predefined names Cell, Region, Year, Data1,
Data2,... and Value, rev=2 uses the set names of the MAgPIE object for
naming and adds an attribute "dimtype" to the data.frame which contains
information about the types of the different columns (spatial, temporal,
data or value), rev=3 is identical to rev=2 except that characters are
not being converted to factors (stringsAsFactors = FALSE).}

\item{raw}{Logical to control whether years beginning with "y" should be
converted to integers (without "y") and coordinates should be converted to
numerics. If set to raw columns are returned as they are in the initial
object.}
}
\description{
~~ Methods for function \code{as.data.frame} ~~
}
\section{Methods}{
 \describe{

\item{list("signature(x = \"magpie\")")}{ Conversion creates columns for
Cell, Region, Year, Data1, Data2,... and Value } }
}

\examples{

pop <- maxample("pop")
head(as.data.frame(pop))
head(as.data.frame(pop, rev = 2))

a <- maxample("animal")
head(as.data.frame(a, rev = 3))
head(as.data.frame(a, rev = 3, raw = TRUE))
attr(as.data.frame(a, rev = 3), "dimtype")

}
\seealso{
Other MAgPIE-Conversions: 
\code{\link{as.RasterBrick}()},
\code{\link{as.SpatRaster}()},
\code{\link{as.SpatRasterDataset}()},
\code{\link{as.SpatVector}()},
\code{\link{as.array-methods}},
\code{\link{as_tibble.magpie}()},
\code{\link{unwrap}()},
\code{\link{wrap}()}
}
\concept{MAgPIE-Conversions}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCoords.R
\name{getCoords}
\alias{getCoords}
\alias{getCoords<-}
\title{Get Coordinates}
\usage{
getCoords(x, xlab = "x", ylab = "y")

getCoords(x, xlab = "x", ylab = "y") <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{xlab}{label of x-dimension}

\item{ylab}{label of y-dimension}

\item{value}{coordinates as two column data.frame the data should be set to
(first column = x, second column = y).}
}
\value{
coordinates of the MAgPIE-object
}
\description{
Extracts spatial coordinates of a MAgPIE-object
}
\section{Functions}{
\itemize{
\item \code{getCoords(x, xlab = "x", ylab = "y") <- value}: set coordinates

}}
\examples{
a <- maxample("animal")
getCoords(a)
}
\seealso{
\code{\link{as.RasterBrick}}, \code{\link{getItems}}, \code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time_interpolate.R
\name{time_interpolate}
\alias{time_interpolate}
\title{time_interpolate}
\usage{
time_interpolate(
  dataset,
  interpolated_year,
  integrate_interpolated_years = FALSE,
  extrapolation_type = "linear"
)
}
\arguments{
\item{dataset}{An MAgPIE object}

\item{interpolated_year}{Vector of years, of which values are required. Can
be in the formats 1999 or y1999.}

\item{integrate_interpolated_years}{FALSE returns only the dataset of the
interpolated year, TRUE returns the whole dataset, including all years of
data and the itnerpolated year}

\item{extrapolation_type}{Determines what happens if extrapolation is
required, i.e. if a requested year lies outside the range of years in
\code{dataset}. Specify "linear" for a linear extrapolation. "constant" uses
the value from dataset closest in time to the requested year.}
}
\value{
Uses linear extrapolation to estimate the values of the interpolated
year, using the values of the two surrounding years. If the value is before
or after the years in data, the two closest neighbours are used for
extrapolation.
}
\description{
Function to extrapolate missing years in MAgPIE objects.
}
\examples{

p <- maxample("pop")
time_interpolate(p, "y2000", integrate = TRUE)
time_interpolate(p, c("y1980", "y2000"), integrate = TRUE, extrapolation_type = "constant")
}
\seealso{
\code{\link{convergence}}

Other TemporalOperations: 
\code{\link{commonYears}()},
\code{\link{convergence}()},
\code{\link{lowpass}()}
}
\author{
Benjamin Bodirsky, Jan Philipp Dietrich
}
\concept{TemporalOperations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getDim.R
\name{getDim}
\alias{getDim}
\title{getDim}
\usage{
getDim(elems, x, fullmatch = FALSE, dimCode = TRUE)
}
\arguments{
\item{elems}{A vector of characters containing the elements that should be
found in the MAgPIE object}

\item{x}{MAgPIE object in which elems should be searched for.}

\item{fullmatch}{If enabled, only dimensions which match exactly the elements
provided will be returned. Otherwise, it is sufficient if elems contains a subset
of the dimension.}

\item{dimCode}{If enabled, the dimCode will be returned, otherwise the name
of the dimension.}
}
\value{
The name or dimCode of the dimensions in which elems were found.
}
\description{
Function which tries to detect the dimension to which the given elems belong
}
\examples{

pop <- maxample("pop")
getDim(c("AFR", "CPA"), pop)
 getDim(c("AFR", "CPA"), pop, fullmatch = TRUE)
 getDim(c("AFR", "CPA"), pop, dimCode = FALSE)

}
\seealso{
\code{\link{mcalc}},\code{\link{dimCode}}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.magpie.R
\name{read.magpie}
\alias{read.magpie}
\title{Read MAgPIE-object from file}
\usage{
read.magpie(
  file_name,
  file_folder = "",
  file_type = NULL,
  as.array = FALSE,
  comment.char = "*",
  check.names = FALSE,
  ...
)
}
\arguments{
\item{file_name}{file name including file ending (wildcards are supported).
Optionally also the full path can be specified here (instead of splitting it
to file_name and file_folder)}

\item{file_folder}{folder the file is located in (alternatively you can also
specify the full path in file_name - wildcards are supported)}

\item{file_type}{format the data is stored in. If file_type=NULL the file ending
of the file_name is used as format. If format is different to the formats
mentioned standard MAgPIE format is assumed. See \code{\link{write.magpie}}
for a list of supported file formats.}

\item{as.array}{Should the input be transformed to an array? This can be
useful for regional or global inputs, but all advantages of the magpie-class
are lost.}

\item{comment.char}{character: a character vector of length one containing a
single character or an empty string. Use "" to turn off the interpretation
of comments altogether. If a comment is found it will be stored in
attr(,"comment"). In text files the comment has to be at the beginning of
the file in order to be recognized by read.magpie.}

\item{check.names}{logical. If TRUE then the names of the variables in the
data frame are checked to ensure that they are syntactically valid variable
names. Same functionality as in read.table.}

\item{...}{additional arguments passed to specific read functions (e.g.
\code{varname} for specifying the variable to be read in from a multi-variable
NCDF file.)}
}
\value{
\item{x}{MAgPIE-object}
}
\description{
Reads a MAgPIE-file and converts it to a 3D array of the structure
(cells,years,datacolumn)
}
\details{
See \code{\link{write.magpie}} for a list of supported file formats.
}
\note{
See \code{\link{write.magpie}} for the detailed structure of binary MAgPIE formats .m and .mz.
}
\seealso{
\code{"\linkS4class{magpie}"}, \code{\link{write.magpie}}

Other FileIO: 
\code{\link{copy.magpie}()},
\code{\link{read.report}()},
\code{\link{write.magpie}()},
\code{\link{write.report}()},
\code{\link{write.report2}()}
}
\author{
Jan Philipp Dietrich, Stephen Bi, Florian Humpenoeder, Pascal Sauer
}
\concept{FileIO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/magpie_expand_dim.R
\name{magpie_expand_dim}
\alias{magpie_expand_dim}
\title{magpie_expand_dim}
\usage{
magpie_expand_dim(x, ref, dim = 1)
}
\arguments{
\item{x}{MAgPIE object that should be expanded}

\item{ref}{MAgPIE object that serves as a reference}

\item{dim}{dimension that should be expanded}
}
\value{
An expanded version of x.
}
\description{
Expands a single MAgPIE object dimension
}
\details{
Expansion means here that the dimensions of x are expanded acordingly to
ref. Please note that this is really only about expansion. In the case that
one dimension of ref is smaller than of x nothing happens with this
dimension. At the moment magpie_expand is only internally available in the
magclass library

In contrast to \code{\link{magpie_expand}} this function is expanding only a single
dimension. It is meant as a support function for \code{\link{magpie_expand}} itself.
}
\examples{
d <- new.magpie(c("AFR.BLUB.1", "AFR.BLUB.2", "EUR.BLUB.1",
                    "AFR.BLA.1", "AFR.BLA.2", "EUR.BLA.1"), fill = 1)
 getSets(d)[1:3] <- c("reg", "b", "i")
 e <- new.magpie(c("BLA.AFR.A", "BLA.EUR.A", "BLUB.AFR.A", "BLUB.EUR.A",
                    "BLA.AFR.B", "BLA.EUR.B", "BLUB.AFR.B", "BLUB.EUR.B"), fill = 2)
 getSets(e)[1:3] <- c("b", "reg", "a")
 magclass:::magpie_expand_dim(d, e, dim = 1)
}
\seealso{
\code{\link{as.magpie}}, \code{\link[base]{options}}
}
\author{
Jan Philipp Dietrich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncells.R, R/ndata.R, R/nregions.R, R/nyears.R
\name{ncells}
\alias{ncells}
\alias{nyears}
\alias{ndata}
\alias{nregions}
\title{Count elements}
\usage{
ncells(x)

ndata(x)

nregions(x)

nyears(x)
}
\arguments{
\item{x}{A MAgPIE-object}
}
\value{
\item{value}{The number of cells/years/datasets/regions of \code{x}}
}
\description{
Functions to count the number of cells/years/datasets/regions of an
MAgPIE-object
}
\section{Functions}{
\itemize{
\item \code{ndata()}: count datasets

\item \code{nregions()}: count regions

\item \code{nyears()}: count years

}}
\examples{

a <- is.magpie(NULL)
ncells(a)
nyears(a)
ndata(a)
nregions(a)
}
\seealso{
Other ObjectInfo: 
\code{\link{dimExists}()},
\code{\link{fulldim}()},
\code{\link{hasCoords}()},
\code{\link{hasSets}()},
\code{\link{is.temporal}()},
\code{\link{isYear}()},
\code{\link{ndim}()},
\code{\link{sameDims}()}
}
\author{
Jan Philipp Dietrich
}
\concept{ObjectInfo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.magpie.R
\name{write.magpie}
\alias{write.magpie}
\title{Write MAgPIE-object to file}
\usage{
write.magpie(
  x,
  file_name,
  file_folder = "",
  file_type = NULL,
  append = FALSE,
  comment = NULL,
  comment.char = "*",
  mode = NULL,
  zname = "time",
  ...
)
}
\arguments{
\item{x}{a magclass object. An exception is that formats written via the raster package
(currently "asc", "grd" and "tif") also accept RasterBrick objects which have
been previously created from a magclass object via as.RasterBrick)}

\item{file_name}{file name including file ending (wildcards are supported).
Optionally also the full path can be specified here (instead of splitting it
to file_name and file_folder)}

\item{file_folder}{folder the file should be written to (alternatively you
can also specify the full path in file_name - wildcards are supported)}

\item{file_type}{Format the data should be stored as. If file_type=NULL
the file ending of the file_name is used as format. See detailed
description for a list of available file types. Please be aware that the
file_name is independent of the file_type you choose here, so no additional
file ending will be added!}

\item{append}{Decides whether an existing file should be overwritten (FALSE)
or the data should be added to it (TRUE). Append = TRUE only works if the
existing data can be combined with the new data using the mbind function}

\item{comment}{Vector of strings: Optional comment giving additional
information about the data. If different to NULL this will overwrite the
content of attr(x,"comment"). For nc files the unit can be passed
with e.g. 'comment = "unit: kg"'.}

\item{comment.char}{character: a character vector of length one containing a
single character or an empty string. Use "" to turn off the interpretation
of comments altogether.}

\item{mode}{File permissions the file should be written with as 3-digit
number (e.g. "777" means full access for user, group and all, "750" means
full access for user, read access for group and no acess for anybody else).
Set to NULL system defaults will be used. Access codes are identical to the
codes used in unix function chmod.}

\item{zname}{name of the time variable for raster files like nc, asc, grd and tif}

\item{...}{additional arguments passed to specific write functions.
Check ?magclass::writeNC for available arguments when writing nc files.}
}
\description{
Writes a magpie object to a file. The file type is determined by the filename extension.
The written file can be read again using \code{\link{read.magpie}}.
}
\details{
This function supports writing the following file types:
\itemize{
\item "cs2" is the new standard format for cellular data with or without header
and the first columns (year,regiospatial) or only (regiospatial)
\item "cs2b" is identical to "cs2" except that it will suppress the data name
if it has only 1 element in the data dimension.
\item "csv" is the standard format for regional data with or without header and
the first columns (year,region,cellnumber) or only (region,cellnumber)
\item "cs3" is another csv format which is
specifically designed for multidimensional data for usage in GAMS.
\item "cs4" alternative multidimensional format compatible to GAMS, in contrast
to cs3 it can also handle sparse data
\item "csvr", "cs2r", "cs3r", "cs4r" which are the same formats as the ones
previously explained with the only difference that they have a REMIND
compatible format
\item "cs5" a more generalized version of cs4
\item "rds" is an R-default format for storing R objects
\item "m" (magpie) and "mz" (magpie zipped) are new formats developed to allow
a less storage intensive management of MAgPIE-data. The only difference
between both formats is that .mz is gzipped whereas .m is not compressed. So
.mz needs less memory, whereas .m might have a higher compatibility to other
languages
\item "asc" is the ASCII grid format. It can only be applied for gridded data
and writes one file per year per data column. In the case that more than
one year and data column is supplied several files are written with the
structure filename_year_datacolumn.asc
\item "tif" is the  GEOtiff format for gridded data.
\item "grd" is the native raster format for gridded data.
\item "nc" is the netCDF format for gridded data. Check ?magclass::writeNC
for more details on how nc files are written.
}
}
\note{
The binary MAgPIE formats .m and .mz have the following content/structure
(you only have to care for that if you want to implement
read.magpie/write.magpie functions in other languages): \cr \cr
[ FileFormatVersion | Current file format version number (currently 6) | integer | 2 Byte ] \cr
[ nchar_comment | Number of character bytes of the file comment | integer | 4 Byte ] \cr
[ nbyte_metadata | Number of bytes of the serialized metadata (currently = 0) | integer | 4 Byte ] \cr
[ nchar_sets | Number of characters bytes of all regionnames + 2 delimiter | integer | 2 Byte] \cr
[ nyears | Number of years | integer | 2 Byte ]\cr
[ yearList | All years of the dataset (0, if year is not present) | integer | 2*nyears Byte ] \cr
[ ncells | Number of cells | integer | 4 Byte ]\cr
[ nchar_cell | Number of characters bytes of all regionnames + (nreg-1) for delimiters | integer | 4 Byte ] \cr
[ cells | Cell names saved as cell1\\cell2 (\\n is the delimiter) | character | 1*nchar_cell Byte ] \cr
[ nelem | Total number of data elements | integer | 4 Byte ] \cr
[ nchar_data | Number of char. bytes of all datanames + (ndata - 1) for delimiters | integer | 4 Byte ] \cr
[ datanames | Names saved in the format data1\\ndata2 (\\n as del.) | character | 1*nchar_data Byte ] \cr
[ data | Data of the MAgPIE array in vectorized form | numeric | 4*nelem Byte ] \cr
[ comment | Comment with additional information about the data | character | 1*nchar_comment Byte ] \cr
[ sets | Set names with \\n as delimiter | character | 1*nchar_sets Byte] \cr
[ metadata | serialized metadata information (currently not in use) | bytes | 1*nbyte_metadata Byte] \cr
}
\examples{
pop <- maxample("pop")
path <- tempfile(fileext = ".mz")
write.magpie(pop, path)
pop2 <- read.magpie(path)
}
\seealso{
\code{"\linkS4class{magpie}"},
\code{\link{read.magpie}},\code{\link{mbind}}

Other FileIO: 
\code{\link{copy.magpie}()},
\code{\link{read.magpie}()},
\code{\link{read.report}()},
\code{\link{write.report}()},
\code{\link{write.report2}()}
}
\author{
Jan Philipp Dietrich, Stephen Bi, Florian Humpenoeder, Pascal Sauer
}
\concept{FileIO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/magpie_expand.R
\name{magpie_expand}
\alias{magpie_expand}
\title{magpie_expand}
\usage{
magpie_expand(x, ref)
}
\arguments{
\item{x}{MAgPIE object that should be expanded}

\item{ref}{MAgPIE object that serves as a reference}
}
\value{
An expanded version of x.
}
\description{
Expands a MAgPIE object based on a reference
}
\details{
Expansion means here that the dimensions of x are expanded accordingly to
ref. Please note that this is really only about expansion. In the case that
one dimension of ref is smaller than of x nothing happens with this
dimension.

You can influence the verbosity of this function by setting the option
"magclass.verbosity". By default verbosity is set to 1 which means that
only warnings are returned. Setting verbosity to 2 means that warnings
as well as additional notes are returned. This is done by
options(verbosity.level=2)

With version 5 of the package magpie_expand has been updated to a newer version
(currently 2.1) and since version 6 this is the only currently supported
version. To switch to the old setup you have to install magclass  in a version < 6
and set \code{options(magclass_expand_version=1)}.

By default expansion is based on the elements in a dimension ignoring the set name of
the dimension. To expand based on set names instead of contents (recommended) you can
switch \code{options(magclass_setMatching=TRUE)}. Please be careful with this setting
as it alters the behavior of magclass objects quite significantly! For more information
have a look at \code{vignette("magclass-expansion")}.
}
\examples{

a <- new.magpie(c("AFR", "CPA"), "y1995", c("m", "n"))
b <- new.magpie("GLO", "y1995", c("bla", "blub"))
magpie_expand(b, a)
options(magclass.verbosity = 2)
magpie_expand(b, a)
}
\seealso{
\code{\link{as.magpie}}, \code{\link[base]{options}}

Other DataBinding: 
\code{\link{cbind.magpie}()},
\code{\link{extend}()},
\code{\link{magpiesort}()},
\code{\link{matchDim}()},
\code{\link{mbind}()}
}
\author{
Jan Philipp Dietrich
}
\concept{DataBinding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimCode.R
\name{dimCode}
\alias{dimCode}
\title{dimCode}
\usage{
dimCode(dim, x, missing = 0, strict = FALSE, sep = ".")
}
\arguments{
\item{dim}{A vector of dimension numbers or dimension names which should be
translated}

\item{x}{MAgPIE object in which the dimensions should be searched for.}

\item{missing}{Either a value to which a dimension should be set in case
that it is not found (default is 0), or "stop" indicating that the function
should throw an error in these cases.}

\item{strict}{if set to TRUE also properly set dimension names which refer
to non-existing subdimensions will be treated as missing, otherwise these
dimension codes will be returned, even if the subdimension does not exist}

\item{sep}{A character separating joined dimension names}
}
\value{
A dimension code identifying the dimension. Either a integer which
represents the main dimensions (1=spatial, 2=temporal, 3=data) or a numeric,
representing the subdimensions of a dimension (e.g. 3.2 for the second data
dimension).
}
\description{
Function converts a dimension name or number to a dimension Code used for
MAgPIE objects
}
\examples{

pop <- maxample("pop")
dimCode(c("t", "scenario", "blablub"), pop)
}
\seealso{
\code{\link{mselect}}, \code{\link{getDim}}

Other DimensionManipulation: 
\code{\link{addDim}()},
\code{\link{add_columns}()},
\code{\link{collapseDim}()},
\code{\link{collapseNames}()},
\code{\link{dimOrder}()},
\code{\link{dimReduce}()}
}
\author{
Jan Philipp Dietrich, Kristine Karstens
}
\concept{DimensionManipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_columns.R
\name{add_columns}
\alias{add_columns}
\title{add_columns}
\usage{
add_columns(x, addnm = "new", dim = 3.1, fill = NA)
}
\arguments{
\item{x}{MAgPIE object which should be extended.}

\item{addnm}{The new elements that should be added to the (sub)dimension}

\item{dim}{The (sub)dimension to be filled either identified via
name or dimension code (see \code{\link{dimCode}} for more information)}

\item{fill}{fill value of length 1 for the newly added columns (NA by default)}
}
\value{
The extended MAgPIE object
}
\description{
Function adds new columns to the existing magpie object.
}
\examples{
a <- maxample("animal")
a2 <- add_columns(a, addnm = c("horse", "magpie"), dim = "species", fill = 42)
getItems(a2, dim = 3)
getItems(a2, dim = 3, split = TRUE)
head(a2[, , "magpie"])
}
\seealso{
\code{\link{addDim}},\code{\link{dimCode}}

Other DimensionManipulation: 
\code{\link{addDim}()},
\code{\link{collapseDim}()},
\code{\link{collapseNames}()},
\code{\link{dimCode}()},
\code{\link{dimOrder}()},
\code{\link{dimReduce}()}
}
\author{
Jan Philipp Dietrich, Benjamin Bodirsky
}
\concept{DimensionManipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapseDim.R
\name{collapseDim}
\alias{collapseDim}
\title{Collapse dataset dimensions}
\usage{
collapseDim(x, dim = NULL, keepdim = NULL)
}
\arguments{
\item{x}{MAgPIE object}

\item{dim}{Either NULL, dimension code or
name of dimension or a vector of these. If set to NULL all single entry subdimensions will
be removed as they are irrelevant to uniquely identfy a data element. If specified, only the specified
subdimensions will be removed (See \code{\link{dimCode}} for more details how to specify a subdimension).
CAUTION: The function also allows to specify subdimensions which are otherwise needed to clearly identify
an entry. By removing these subdimensions duplicates in the data will be created potentially causing
problems in the further use of the data set. Be careful in removing subdimensions.}

\item{keepdim}{(only considered if \code{dim} is not specified) Can be used to converse
single element subdimension which otherwise would get deleted. If \code{dim} is specified
this setting will not have any effect.}
}
\value{
The provided MAgPIE object with collapsed dimensions
}
\description{
This function will remove names in the data dimension which are the same for
each element (meaning that this data dimension contains exactly one element)
or, if forced, remove any other subdimension. It is a generalized version
of the function \code{\link{collapseNames}}
}
\note{
This function has some similarities to \code{\link{dimReduce}}, but
serves a different purpose. While \code{\link{collapseDim}} only removes
dimensions which contain only a single element or which it is
specifically told to remove, \code{\link{dimReduce}} looks whether the
entries of a multi-entry dimension are all the same and removes dimensions
for which this is the case. In some cases both will lead to the same result
but in many other cases the results will differ.
}
\examples{

x <- new.magpie(c("GLO.1", "GLO.2"), 2000, c("bla.a", "bla.b"))
collapseDim(x)
collapseDim(x, keepdim = 1:2)
collapseDim(x, dim = 1.1)
collapseDim(x, dim = 3.2)
}
\seealso{
\code{\link{getItems}}
\code{"\linkS4class{magpie}"}

Other DimensionManipulation: 
\code{\link{addDim}()},
\code{\link{add_columns}()},
\code{\link{collapseNames}()},
\code{\link{dimCode}()},
\code{\link{dimOrder}()},
\code{\link{dimReduce}()}
}
\author{
Jan Philipp Dietrich
}
\concept{DimensionManipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mplot.R
\name{mplot}
\alias{mplot}
\title{mplot}
\usage{
mplot(px, global = TRUE, total = FALSE)
}
\arguments{
\item{px}{The magpie object to be visualized}

\item{global}{Whether data should be aggregated over regions to global values.}

\item{total}{Whether the total of all data values should also be visualized.}
}
\description{
Get a quick visualization of the content of most magpie objects.
}
\seealso{
Other Display: 
\code{\link{head.magpie}()},
\code{\link{maxample}()},
\code{\link{print.magpie}()},
\code{\link{show-methods}},
\code{\link{str.magpie}()}
}
\author{
Pascal Sauer, Patrick Rein
}
\concept{Display}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCells.R, R/setCells.R
\name{getCells}
\alias{getCells}
\alias{getCells<-}
\alias{setCells}
\title{Get Cells}
\usage{
getCells(x)

getCells(x) <- value

setCells(object, nm = "GLO")
}
\arguments{
\item{x, object}{MAgPIE object}

\item{value, nm}{cell names the data should be set to.}
}
\value{
getCells returns cell names of the MAgPIE-object, whereas setCells
returns the MAgPIE object with the manipulated cell names.
}
\description{
Extracts cell names of a MAgPIE-object
}
\details{
setCells is a shortcut to use a MAgPIE object with manipulated cell names.
setCells uses the variable names "object" and "nm" in order to be consistent
to the already existing function setNames.
}
\section{Functions}{
\itemize{
\item \code{getCells(x) <- value}: set cell names

\item \code{setCells()}: set cell names

}}
\examples{

a <- as.magpie(1)
 getCells(a)
 setCells(a, "AFR")
}
\seealso{
\code{\link{getRegions}}, \code{\link{getNames}},
\code{\link{setNames}}, \code{\link{getCPR}}, \code{\link{read.magpie}},
\code{\link{write.magpie}}, \code{"\linkS4class{magpie}"}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setNames_method.R
\docType{methods}
\name{setNames-methods}
\alias{setNames-methods}
\alias{setNames}
\alias{setNames,magpie-method}
\alias{setNames,NULL-method}
\title{Get dataset names}
\usage{
\S4method{setNames}{magpie}(object = nm, nm)
}
\arguments{
\item{object}{MAgPIE object}

\item{nm}{a vector of names current names should be replaced with. If
only one data element exists you can also set the name to NULL.}
}
\description{
Extracts dataset names of a MAgPIE-object
}
\details{
setNames is a shortcut to use a MAgPIE object with manipulated data names.
The setNames method uses the variable names "object" and "nm" in order to be
consistent to the already existing function setNames.
}
\section{Methods}{
 \describe{

\item{list("signature(object = \"ANY\")")}{ normal setNames method }

\item{list("signature(object = \"magpie\")")}{ setNames for MAgPIE objects} }
}

\seealso{
\code{\link{getNames}},

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getItems}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()}
}
\concept{GetterSetter}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sizeCheck.R
\name{sizeCheck}
\alias{sizeCheck}
\title{sizeCheck}
\usage{
sizeCheck(dim)
}
\arguments{
\item{dim}{dimensions of the current object as returned by function \code{dim}}
}
\description{
Calculates expected magclass object length and checks that it stays below
the limit defined with magclass_sizeLimit (default = 10^9). This is useful
to prevent out of memory errors in case of unwanted object expansions
Ignored if \code{getOption("magclass_sizeLimit")} is negative.
}
\examples{
pop <- maxample("pop")
magclass:::sizeCheck(dim(pop))

\dontrun{
magclass:::sizeCheck(c(6765L, 10946L, 17711L))
}
}
\author{
Jan Philipp Dietrich
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mselect.R
\name{mselect}
\alias{mselect}
\alias{mselect<-}
\title{MSelect}
\usage{
mselect(x, ..., collapseNames = FALSE)

mselect(x, ...) <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{...}{entry selections of the form
\code{set_name=c(set_elem1,set_elem2)}. Alternatively a single list element
containing these selections can be provided.}

\item{collapseNames}{Boolean which decides whether names should be collapsed
or not.}

\item{value}{values on which the selected magpie entries should be set.}
}
\value{
The reduced MAgPIE object containing only the selected entries or
the full MAgPIE object in which a selection of entries was manipulated.
}
\description{
Select values from a MAgPIE-object
}
\details{
This functions only work for MAgPIE objects with named dimensions as the
dimension name (set_name) has to be used to indicate in which dimension the
entries should be searched for!
}
\section{Functions}{
\itemize{
\item \code{mselect(x, ...) <- value}: replace values in magpie object

}}
\examples{

pop <- maxample("pop")
mselect(pop, i = c("AFR", "EUR"), scenario = "A2", t = "y2035")
}
\seealso{
\code{\link{collapseNames}}, \code{"\linkS4class{magpie}"}

Other SelectionCalculation: 
\code{\link{magpply}()},
\code{\link{mcalc}()},
\code{\link{where}()}
}
\author{
Jan Philipp Dietrich
}
\concept{SelectionCalculation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{convergence}
\alias{convergence}
\title{convergence}
\usage{
convergence(
  origin,
  aim,
  start_year = NULL,
  end_year = NULL,
  direction = NULL,
  type = "smooth",
  par = 1.5
)
}
\arguments{
\item{origin}{an object with one name-column}

\item{aim}{Can be twofold: An magpie object or a numeric value.}

\item{start_year}{year in which the convergence from origin to aim starts.
If set to NULL the the first year of aim is used as start_year}

\item{end_year}{year in which the convergence from origin to aim shall be
(nearly) reached. If set to NULL the the last year of aim is used as
end_year.}

\item{direction}{NULL, "up" or "down". NULL means normal convergence in both
directions, "up" is only a convergence if origin<aim, "down" means only a
convergence if origin>aim}

\item{type}{"smooth", "s", "linear" or "decay". Describes the function used
for convergence (x starts at 0 and ends at 1):
"linear" means linear interpolation.
"s" is an s-curve which starts from origin in start_year and reaches aim
precisely in end_year. After 50\% of the convergence time, it reaches about
the middle of the two values. It's using the function x^4/(0.07+x^4)*1.07).
"smooth" is a conversion based on the function x^3/(0.1+x^3). With "smooth"
only 90\% of convergence will be reached in the end year, full
convergence is approached as x approaches infinity.
"decay" is a conversion based on the function x / (par + x) * (par + 1).}

\item{par}{parameter value for convergence function; currently only used for
type="decay"}
}
\value{
returns a time-series with the same timesteps as origin, which
lineary fades into the values of the aim object
}
\description{
Cross-Fades the values of one MAGPIE object into the values of another over
a certain time
}
\examples{

pop <- maxample("pop")
population <- add_columns(pop, "MIX")
population[, , "MIX"] <- convergence(population[, , "A2"], population[, , "B1"])
}
\seealso{
Other TemporalOperations: 
\code{\link{commonYears}()},
\code{\link{lowpass}()},
\code{\link{time_interpolate}()}
}
\author{
Benjamin Bodirsky, Jan Philipp Dietrich
}
\concept{TemporalOperations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatRasterToDataset.R
\name{spatRasterToDataset}
\alias{spatRasterToDataset}
\title{spatRasterToDataset}
\usage{
spatRasterToDataset(x)
}
\arguments{
\item{x}{SpatRaster with names of the form "y[0-9]+..[varname]"}
}
\value{
SpatRasterDataset
}
\description{
Convert a SpatRaster to a SpatRasterDataset.
}
\seealso{
Other SpatialOperations: 
\code{\link{guessResolution}()}
}
\author{
Pascal Sauer
}
\concept{SpatialOperations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getItems.R
\name{getItems}
\alias{getItems}
\alias{getItems<-}
\title{Get Items}
\usage{
getItems(x, dim = NULL, split = FALSE, full = FALSE)

getItems(x, dim, full = NULL, maindim = NULL, raw = FALSE) <- value
}
\arguments{
\item{x}{MAgPIE object}

\item{dim}{Dimension for which the items should be returned. Either number or
name of dimension or a vector of these. See \code{\link{dimCode}} for more details.}

\item{split}{Boolean which determines whether a main dimension should be split in subdimensions.
Only applicable to main dimensions (1,2,3) and ignored for all other.}

\item{full}{if TRUE dimension names are returned as they are (including repetitions), if FALSE only
the dimension elements (unique list of entries) are returned.}

\item{maindim}{main dimension the data should be added to (does not need to be set if \code{dim} exists
in the data. Should be set if \code{dim} might not exist, or if \code{dim} might potentially exist
in a different main dimension than the one anticipated).}

\item{raw}{if set to FALSE inputs will be corrected (e.g. dots replaced by the letter "p") if necessary. If
TRUE data will be written as is (risking the creation of inconsistent objects).}

\item{value}{a vector with the length of the main dimension the dimnames should be replaced in / added to.
If set to NULL the corresponding dimension will be removed.}
}
\value{
items of the requested dimension in the MAgPIE-object. If split=TRUE and applied to a
main dimension (1,2,3) a list of items for each sub-dimension.
}
\description{
Extract items of a given (sub-)dimension of a MAgPIE-object
}
\section{Functions}{
\itemize{
\item \code{getItems(x, dim, full = NULL, maindim = NULL, raw = FALSE) <- value}: set dimension names

}}
\examples{
x <- maxample("pop")
getItems(x, "scenario")
getItems(x, 3.1)
getItems(x, "i") <- paste0("REG", seq_len(ncells(x)))
getItems(x, "i")
y <- x[, 1, ]
getItems(y, "t") <- NULL
}
\seealso{
\code{\link{dimCode}}

Other GetterSetter: 
\code{\link{getCPR}()},
\code{\link{getCells}()},
\code{\link{getComment}()},
\code{\link{getCoords}()},
\code{\link{getDim}()},
\code{\link{getNames}()},
\code{\link{getRegionList}()},
\code{\link{getRegions}()},
\code{\link{getSets}()},
\code{\link{getYears}()},
\code{\link{setItems}()},
\code{\link{setNames-methods}}
}
\author{
Jan Philipp Dietrich
}
\concept{GetterSetter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimOrder.R
\name{dimOrder}
\alias{dimOrder}
\title{dimOrder}
\usage{
dimOrder(x, perm, dim = 3)
}
\arguments{
\item{x}{magpie object}

\item{perm}{vector with the new order of the sub-dimension. Missing
sub-dimensions will added automatically at the end}

\item{dim}{main dimension in which the order of sub-dimensions should
be changed (1, 2 or 3)}
}
\value{
magpie object
}
\description{
Changes the order of the sub-dimension in a magpie object
similar to unwrapping and applying the aperm command, but more efficient.
}
\examples{
a <- maxample("animal")
head(a)
head(dimOrder(a, perm = 3:1, dim = 1))
head(dimOrder(a, perm = c(2,1,3), dim = 3))
}
\seealso{
Other DimensionManipulation: 
\code{\link{addDim}()},
\code{\link{add_columns}()},
\code{\link{collapseDim}()},
\code{\link{collapseNames}()},
\code{\link{dimCode}()},
\code{\link{dimReduce}()}
}
\author{
Jan Philipp Dietrich, Benjamin Leon Bodirsky
}
\concept{DimensionManipulation}
